/*
 * Copyright (c) 2017-2019 Cavium, Inc.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include "bcm_osal.h"
#include "ecore.h"
#include "ecore_hsi_init_tool.h"
#include "ecore_user_dbg_fw_funcs.h"
#include "qlnx_ioctl.h"


const char *prog_name = "e4dump";
const char *version = "v1.1.0_STORMFW_8_42_3_0_ECORE_8_42_9_0";

#define FWDUMP_FILE "grcdump"

static void
create_dump_file(qlnx_grcdump_t *fw_dump, int hwfn_index)
{
        int dump_fd;
        char dump_file_name[64];
        time_t t;
        struct tm* lt;
	int fw_dump_size = 0;

        t = time(NULL);
        lt = localtime (&t);

        bzero(dump_file_name, sizeof(dump_file_name));

	fw_dump_size = (fw_dump->grcdump_dwords[hwfn_index] * sizeof(uint32_t));

	if (fw_dump_size == 0)
		return;

        sprintf(dump_file_name, "%s_hwfn%d_%d_%d_%d_%d_%d_%d.bin",
                FWDUMP_FILE, hwfn_index, 
		(lt->tm_year + 1900), (lt->tm_mon + 1), lt->tm_mday,
                lt->tm_hour, lt->tm_min, lt->tm_sec);

        dump_fd = open(dump_file_name, (O_RDWR | O_CREAT));

        if (lseek(dump_fd, 0, SEEK_SET) == -1) {
                printf("%s: lseek  0 SEEK_SET failed [%s]\n",
                        prog_name, FWDUMP_FILE);
                close(dump_fd);
                return;
        }

        if (write(dump_fd, fw_dump->grcdump[hwfn_index], fw_dump_size) == -1) {
                printf("%s: write [%d] bytes failed [%s]\n",
                        prog_name, fw_dump_size, dump_file_name);
        }
        close(dump_fd);
        return;
}

void
free_grcdump(qlnx_grcdump_t *dump)
{
	int i;

	for (i = 0; i < QLNX_MAX_HW_FUNCS; i++) {
		if (dump->grcdump[i] != NULL) 
			free(dump->grcdump[i]);
	}
	return;
}

static void
get_grcdump(int fd, char *dev_path)
{
	int ret;
	qlnx_grcdump_t dump;
	int i;


	bzero(&dump, sizeof(qlnx_grcdump_t));

	ret = ioctl(fd, QLNX_GRC_DUMP_SIZE, &dump);

	if (ret == -1) {
                printf("%s: failed to get grcdump size [%s]\n",
			prog_name, dev_path);
                exit (2);
	}

	for (i = 0; i < QLNX_MAX_HW_FUNCS; i++) {

		if (dump.grcdump_size[i] == 0)
			break;

		dump.grcdump[i] = malloc(dump.grcdump_size[i]);

		if (dump.grcdump[i] == NULL) {
			printf("%s: failed to allocate grcdump buffer [%s] [0x%x]\n",
				prog_name, dev_path, dump.grcdump_size[i]);
			free_grcdump(&dump);
			exit (3);
		}
		bzero(dump.grcdump[i], dump.grcdump_size[i]);
	}
	
	ret = ioctl(fd, QLNX_GRC_DUMP, &dump);
	if (ret == -1) {
                printf("%s: failed to get grcdump [%s]\n",
			prog_name, dev_path);
		free_grcdump(&dump);
                exit (4);
	}

	for (i = 0; i < QLNX_MAX_HW_FUNCS; i++)
		create_dump_file(&dump, i);

	free_grcdump(&dump);

	return;
}


#define IDLE_CHK_FILE "idle_chk"

static void
create_idle_chk_file(qlnx_idle_chk_t *idle_chk, int hwfn_index)
{
        int idle_chk_fd;
        char idle_chk_file_name[64];
        time_t t;
        struct tm* lt;
	int idle_chk_size = 0;
        unsigned int idle_chk_parsed_size;
        u32 num_errors, num_warnings;
        char *parsed_idle_chk;

        t = time(NULL);
        lt = localtime (&t);

        bzero(idle_chk_file_name, sizeof(idle_chk_file_name));

	idle_chk_size =
		(idle_chk->idle_chk_dwords[hwfn_index] * sizeof(uint32_t));

	if (idle_chk_size == 0)
		return;

       if (ecore_get_idle_chk_results_buf_size(NULL,
		idle_chk->idle_chk[hwfn_index],
		idle_chk->idle_chk_dwords[hwfn_index],
                &idle_chk_parsed_size) != DBG_STATUS_OK) {
                printf("%s get result buf size failed\n", prog_name);
                return;
        }
 
        parsed_idle_chk = malloc(idle_chk_parsed_size);
	if(parsed_idle_chk == NULL)
		return;

        bzero(parsed_idle_chk, idle_chk_parsed_size);

        if (ecore_print_idle_chk_results(NULL, 
		idle_chk->idle_chk[hwfn_index],
		idle_chk->idle_chk_dwords[hwfn_index],
                parsed_idle_chk, &num_errors, &num_warnings) != DBG_STATUS_OK) {
                printf("%s parse result failed\n", prog_name);
                free(parsed_idle_chk);
                return;
        }

        sprintf(idle_chk_file_name, "%s_hwfn%d_%d_%d_%d_%d_%d_%d.txt",
                IDLE_CHK_FILE, hwfn_index, 
		(lt->tm_year + 1900), (lt->tm_mon + 1), lt->tm_mday,
                lt->tm_hour, lt->tm_min, lt->tm_sec);

        idle_chk_fd = open(idle_chk_file_name, (O_RDWR | O_CREAT));

        if (lseek(idle_chk_fd, 0, SEEK_SET) == -1) {
                printf("%s: lseek  0 SEEK_SET failed [%s]\n",
                        prog_name, IDLE_CHK_FILE);
		free(parsed_idle_chk);
                close(idle_chk_fd);
                return;
        }

        if (write(idle_chk_fd, parsed_idle_chk, idle_chk_parsed_size) == -1) {
                printf("%s: write [%d] bytes failed [%s]\n",
                        prog_name, idle_chk_parsed_size, IDLE_CHK_FILE);
        }
        free(parsed_idle_chk);
        close(idle_chk_fd);

        return;
}

void
free_idle_chk(qlnx_idle_chk_t *idle_chk)
{
	int i;

	for (i = 0; i < QLNX_MAX_HW_FUNCS; i++) {
		if (idle_chk->idle_chk[i] != NULL) 
			free(idle_chk->idle_chk[i]);
	}
	return;
}

static void
get_idle_chk(int fd, char *dev_path)
{
	int i;
        int ret;
        qlnx_idle_chk_t idle_chk;

        bzero(&idle_chk, sizeof(qlnx_idle_chk_t));

        ret = ioctl(fd, QLNX_IDLE_CHK_SIZE, &idle_chk);

        if (ret == -1) {
                printf("%s: failed to get idle_chk size [%s]\n",
                        prog_name, dev_path);
                exit (2);
        }
	for (i = 0; i < QLNX_MAX_HW_FUNCS; i++) {
                printf("%s: idle_chk size[%s : %d]  [0x%x]\n",
                        prog_name, dev_path, i, idle_chk.idle_chk_size[i]);
	}

	for (i = 0; i < QLNX_MAX_HW_FUNCS; i++) {

		if (idle_chk.idle_chk_size[i] == 0)
			break;

		idle_chk.idle_chk[i] = malloc(idle_chk.idle_chk_size[i]);

		if (idle_chk.idle_chk[i] == NULL) {
			printf("%s: failed to allocate idle_chk buffer [%s] [0x%x]\n",
				prog_name, dev_path, idle_chk.idle_chk_size[i]);
			exit (3);
		}
		bzero(idle_chk.idle_chk[i], idle_chk.idle_chk_size[i]);
	}


        ret = ioctl(fd, QLNX_IDLE_CHK, &idle_chk);
        if (ret == -1) {
                printf("%s: failed to get idle_chk [%s]\n",
                        prog_name, dev_path);
                exit (4);
        }

	for (i = 0; i < QLNX_MAX_HW_FUNCS; i++)
        	create_idle_chk_file(&idle_chk, i);

        free_idle_chk(&idle_chk);

        return;
}

char *trace_cmd_str[] = {
	"iilegal_cmd",
	"mcp_trace",
	"reg_fifo",
	"igu_fifo",
	"protection_override",
	"fw_asserts"
};

static void
create_trace_file(qlnx_trace_t *trace, uint16_t cmd, int hwfn_index)
{
        int trace_fd;
        char trace_file_name[64];
        time_t t;
        struct tm* lt;
	int trace_size = 0;
        unsigned int parsed_trace_size;
        char *parsed_trace;

        t = time(NULL);
        lt = localtime (&t);

        bzero(trace_file_name, sizeof(trace_file_name));

	trace_size =
		(trace->dwords[hwfn_index] * sizeof(uint32_t));

	if (trace_size == 0)
		return;

	switch (cmd) {
	case QLNX_MCP_TRACE:
		if (ecore_get_mcp_trace_results_buf_size(NULL,
			trace->buffer[hwfn_index],
			trace->dwords[hwfn_index],
			&parsed_trace_size) != DBG_STATUS_OK) {

			printf("%s: %s get result buf size failed\n",
				prog_name, trace_cmd_str[cmd]);
			return;
		}
 
		parsed_trace = malloc(parsed_trace_size);

		if (parsed_trace == NULL) {

			printf("%s: %s: parsed buffer alloc  failed\n",
				prog_name, trace_cmd_str[cmd]);
			return;
		}
		bzero(parsed_trace, parsed_trace_size);

		if (ecore_print_mcp_trace_results(NULL, 
			trace->buffer[hwfn_index],
			trace->dwords[hwfn_index],
			parsed_trace) != DBG_STATUS_OK) {

			printf("%s: %s parse result failed\n",
				prog_name, trace_cmd_str[cmd]);
        		free(parsed_trace);
			return;
		}
		break;

	case QLNX_REG_FIFO:
		if (ecore_get_reg_fifo_results_buf_size(NULL,
			trace->buffer[hwfn_index],
			trace->dwords[hwfn_index],
			&parsed_trace_size) != DBG_STATUS_OK) {

			printf("%s: %s get result buf size failed\n",
				prog_name, trace_cmd_str[cmd]);
			return;
		}
 
		parsed_trace = malloc(parsed_trace_size);

		if (parsed_trace == NULL) {

			printf("%s: %s: parsed buffer alloc  failed\n",
				prog_name, trace_cmd_str[cmd]);
			return;
		}
		bzero(parsed_trace, parsed_trace_size);

		if (ecore_print_reg_fifo_results(NULL, 
			trace->buffer[hwfn_index],
			trace->dwords[hwfn_index],
			parsed_trace) != DBG_STATUS_OK) {

			printf("%s: %s parse result failed\n",
				prog_name, trace_cmd_str[cmd]);
        		free(parsed_trace);
		}
		break;

	case QLNX_IGU_FIFO:
		if (ecore_get_igu_fifo_results_buf_size(NULL,
			trace->buffer[hwfn_index],
			trace->dwords[hwfn_index],
			&parsed_trace_size) != DBG_STATUS_OK) {

			printf("%s: %s get result buf size failed\n",
				prog_name, trace_cmd_str[cmd]);
			return;
		}
 
		parsed_trace = malloc(parsed_trace_size);

		if (parsed_trace == NULL) {

			printf("%s: %s: parsed buffer alloc  failed\n",
				prog_name, trace_cmd_str[cmd]);
			return;
		}
		bzero(parsed_trace, parsed_trace_size);

		if (ecore_print_igu_fifo_results(NULL, 
			trace->buffer[hwfn_index],
			trace->dwords[hwfn_index],
			parsed_trace) != DBG_STATUS_OK) {

			printf("%s: %s parse result failed\n",
				prog_name, trace_cmd_str[cmd]);
        		free(parsed_trace);
		}
		break;

	case QLNX_PROTECTION_OVERRIDE:
		if (ecore_get_protection_override_results_buf_size(NULL,
			trace->buffer[hwfn_index],
			trace->dwords[hwfn_index],
			&parsed_trace_size) != DBG_STATUS_OK) {

			printf("%s: %s get result buf size failed\n",
				prog_name, trace_cmd_str[cmd]);
			return;
		}
 
		parsed_trace = malloc(parsed_trace_size);

		if (parsed_trace == NULL) {

			printf("%s: %s: parsed buffer alloc  failed\n",
				prog_name, trace_cmd_str[cmd]);
			return;
		}
		bzero(parsed_trace, parsed_trace_size);

		if (ecore_print_protection_override_results(NULL, 
			trace->buffer[hwfn_index],
			trace->dwords[hwfn_index],
			parsed_trace) != DBG_STATUS_OK) {

			printf("%s: %s parse result failed\n",
				prog_name, trace_cmd_str[cmd]);
        		free(parsed_trace);
		}
		break;

	case QLNX_FW_ASSERTS:
		if (ecore_get_fw_asserts_results_buf_size(NULL,
			trace->buffer[hwfn_index],
			trace->dwords[hwfn_index],
			&parsed_trace_size) != DBG_STATUS_OK) {

			printf("%s: %s get result buf size failed\n",
				prog_name, trace_cmd_str[cmd]);
			return;
		}
 
		parsed_trace = malloc(parsed_trace_size);

		if (parsed_trace == NULL) {

			printf("%s: %s: parsed buffer alloc  failed\n",
				prog_name, trace_cmd_str[cmd]);
			return;
		}
		bzero(parsed_trace, parsed_trace_size);

		if (ecore_print_fw_asserts_results(NULL, 
			trace->buffer[hwfn_index],
			trace->dwords[hwfn_index],
			parsed_trace) != DBG_STATUS_OK) {

			printf("%s: %s parse result failed\n",
				prog_name, trace_cmd_str[cmd]);
        		free(parsed_trace);
		}
		break;
	}

        sprintf(trace_file_name, "%s_hwfn%d_%d_%d_%d_%d_%d_%d.txt",
                trace_cmd_str[cmd], hwfn_index, 
		(lt->tm_year + 1900), (lt->tm_mon + 1), lt->tm_mday,
                lt->tm_hour, lt->tm_min, lt->tm_sec);

        trace_fd = open(trace_file_name, (O_RDWR | O_CREAT));

        if (lseek(trace_fd, 0, SEEK_SET) == -1) {
                printf("%s: lseek  0 SEEK_SET failed [%s]\n",
                        prog_name, trace_cmd_str[cmd]);
		free(parsed_trace);
                close(trace_fd);
                return;
        }

        if (write(trace_fd, parsed_trace, parsed_trace_size) == -1) {
                printf("%s: write [%d] bytes failed [%s]\n",
                        prog_name, parsed_trace_size, trace_cmd_str[cmd]);
        }
        free(parsed_trace);
        close(trace_fd);

        return;
}

void
free_trace_buffer(qlnx_trace_t *trace)
{
	int i;

	for (i = 0; i < QLNX_MAX_HW_FUNCS; i++) {
		if (trace->buffer[i] != NULL) 
			free(trace->buffer[i]);
	}
	return;
}

static void
get_trace(int fd, uint16_t cmd, char *dev_path)
{
	int i;
        int ret;
        qlnx_trace_t trace;

        bzero(&trace, sizeof(qlnx_trace_t));

	trace.cmd = cmd;
        ret = ioctl(fd, QLNX_TRACE_SIZE, &trace);

        if (ret == -1) {
                printf("%s: failed to get [%s] trace size [%s]\n",
                        prog_name, trace_cmd_str[cmd], dev_path);
                exit (2);
        }
	for (i = 0; i < QLNX_MAX_HW_FUNCS; i++) {
                printf("%s: %s trace size[%s : %d]  [0x%x]\n",
                        prog_name, trace_cmd_str[cmd],
			dev_path, i, trace.size[i]);
	}

	for (i = 0; i < QLNX_MAX_HW_FUNCS; i++) {

		if (trace.size[i] == 0)
			break;

		trace.buffer[i] = malloc(trace.size[i]);

		if (trace.buffer[i] == NULL) {
			printf("%s: failed to allocate [%s] trace buffer"
				" [%s] [0x%x]\n", prog_name, trace_cmd_str[cmd],
				dev_path, trace.size[i]);
			exit (3);
		}
		bzero(trace.buffer[i], trace.size[i]);
	}


        ret = ioctl(fd, QLNX_TRACE, &trace);
        if (ret == -1) {
                printf("%s: failed to get [%s] trace [%s]\n",
                        prog_name, trace_cmd_str[cmd], dev_path);
                exit (4);
        }

	for (i = 0; i < QLNX_MAX_HW_FUNCS; i++)
        	create_trace_file(&trace, cmd, i);

        free_trace_buffer(&trace);

        return;
}

static void
get_storm_stats(int fd, char *dev_path)
{
	int i, j;
        int ret;
        qlnx_storm_stats_dump_t storm_stats;
	qlnx_storm_stats_t *stats;

	bzero(&storm_stats, sizeof (qlnx_storm_stats_dump_t));

	ret = ioctl(fd, QLNX_STORM_STATS, &storm_stats);
        if (ret == -1) {
                printf("%s [%s]: failed to get stats\n", prog_name, dev_path);
                exit (-1);
        }

	for (i = 0; i < storm_stats.num_hwfns; i++) {
		storm_stats.buffer[i] = malloc(QLNX_STORM_STATS_BYTES_PER_HWFN);

		if (storm_stats.buffer[i] == NULL) {
			printf("%s [%s]: failed to allocate stats buffer[%d]\n",
				 prog_name, dev_path, i);
			goto get_storm_stats_exit;
		}
		bzero(storm_stats.buffer[i], QLNX_STORM_STATS_BYTES_PER_HWFN);
	}

	ret = ioctl(fd, QLNX_STORM_STATS, &storm_stats);
        if (ret == -1) {
                printf("%s [%s]: failed to get stats\n", prog_name, dev_path);
		goto get_storm_stats_exit;
        } else {
		printf("%s [%s]: number of samples = %d\n",
			prog_name, dev_path, storm_stats.num_samples);

		for (i = 0; i < storm_stats.num_hwfns; i++) {
			printf("%s [%s]: storm stats for hw engine [%d] in this order\n",
				prog_name, dev_path, i);

			printf("\n");
			printf("[xstorm_active_cycles, xstorm_stall_cycles, "
				"xstorm_sleeping_cycles, xstorm_inactive_cycles]\n");
			printf("[ystorm_active_cycles, ystorm_stall_cycles, "
				"ystorm_sleeping_cycles, ystorm_inactive_cycles]\n");
			printf("[pstorm_active_cycles, pstorm_stall_cycles, "
				"pstorm_sleeping_cycles, pstorm_inactive_cycles]\n");
			printf("[tstorm_active_cycles, tstorm_stall_cycles, "
				"tstorm_sleeping_cycles, tstorm_inactive_cycles]\n");
			printf("[mstorm_active_cycles, mstorm_stall_cycles, "
				"mstorm_sleeping_cycles, mstorm_inactive_cycles]\n");
			printf("[ustorm_active_cycles, ustorm_stall_cycles, "
				"ustorm_sleeping_cycles, ustorm_inactive_cycles]\n");
			printf("\n");

			stats = (qlnx_storm_stats_t *)storm_stats.buffer[i];

			for (j = 0; j < QLNX_STORM_STATS_SAMPLES_PER_HWFN; j++) {

				printf("\n");
				printf("0x%04x : 0x%08x 0x%08x 0x%08x 0x%08x\n", j,
					stats->xstorm_active_cycles,
					stats->xstorm_stall_cycles,
					stats->xstorm_sleeping_cycles,
					stats->xstorm_inactive_cycles);
				printf("0x%04x : 0x%08x 0x%08x 0x%08x 0x%08x\n", j,
					stats->ystorm_active_cycles,
					stats->ystorm_stall_cycles,
					stats->ystorm_sleeping_cycles,
					stats->ystorm_inactive_cycles);
				printf("0x%04x : 0x%08x 0x%08x 0x%08x 0x%08x\n", j,
					stats->pstorm_active_cycles,
					stats->pstorm_stall_cycles,
					stats->pstorm_sleeping_cycles,
					stats->pstorm_inactive_cycles);
				printf("0x%04x : 0x%08x 0x%08x 0x%08x 0x%08x\n", j,
					stats->tstorm_active_cycles,
					stats->tstorm_stall_cycles,
					stats->tstorm_sleeping_cycles,
					stats->tstorm_inactive_cycles);
				printf("0x%04x : 0x%08x 0x%08x 0x%08x 0x%08x\n", j,
					stats->mstorm_active_cycles,
					stats->mstorm_stall_cycles,
					stats->mstorm_sleeping_cycles,
					stats->mstorm_inactive_cycles);
				printf("0x%04x : 0x%08x 0x%08x 0x%08x 0x%08x\n", j,
					stats->ustorm_active_cycles,
					stats->ustorm_stall_cycles,
					stats->ustorm_sleeping_cycles,
					stats->ustorm_inactive_cycles);
				printf("\n");

				stats++;
			}
		}
	}

get_storm_stats_exit:
	for (i = 0; i < storm_stats.num_hwfns; i++) {
		if (storm_stats.buffer[i] != NULL)
			free(storm_stats.buffer[i]);
	}
	return;
}

static void
usage(void)
{
        printf("usage: %s -v ; get version\n", prog_name);
        printf("       %s -g [device path] ; get grcdump\n", prog_name);
        printf("       %s -i [device path] ; get idle_chk\n", prog_name);
        printf("       %s -s [device path] ; get storm stats; requires sysctl dev.ql.#.storm_stats_enable=1\n", prog_name);
        printf("       %s -m [device path] ; get mcp trace\n", prog_name);
        printf("       %s -r [device path] ; get reg_fifo trace\n", prog_name);
        printf("       %s -u [device path] ; get IGU trace\n", prog_name);
        printf("       %s -p [device path] ; get protection override trace\n",
		prog_name);
        printf("       %s -f [device path] ; get firmware asserts trace\n", prog_name);

        return;
}


int
main(int argc, char **argv)

{
        int fd;
        char *dev_path;
        char opt;

        if (argc < 3) {
		if (argc == 2) {
        		opt  = *(argv[1] + 1);
			if (opt == 'v')
                		printf("%s: version [%s]\n", prog_name, version);
			else
				usage();
		} else
			usage();
                exit(0);
        }

        opt  = *(argv[1] + 1);
        dev_path = argv[2];

        if ((fd = open(dev_path, O_RDWR)) == -1) {
                printf("%s: failed to open [%s]\n", prog_name, dev_path);
                exit (1);
        }

        switch (opt) {

        case 'g':
                get_grcdump(fd, dev_path);
                break;

        case 'i':
                get_idle_chk(fd, dev_path);
                break;

        case 's':
                get_storm_stats(fd, dev_path);
                break;

        case 'm':
                get_trace(fd, QLNX_MCP_TRACE, dev_path);
                break;

        case 'r':
                get_trace(fd, QLNX_REG_FIFO, dev_path);
                break;

        case 'u':
                get_trace(fd, QLNX_IGU_FIFO, dev_path);
                break;

        case 'p':
                get_trace(fd, QLNX_PROTECTION_OVERRIDE, dev_path);
                break;

        case 'f':
                get_trace(fd, QLNX_FW_ASSERTS, dev_path);
                break;

        default:
                usage();
                break;
        }

        exit (0);
}

