/*
 * Copyright (c) 2017-2019 Cavium, Inc.
 * All rights reserved.                
 *                                     
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:                                                          
 *                                                                    
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer. 
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *                                                                         
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE   
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 *  POSSIBILITY OF SUCH DAMAGE.                                                
 */    
#ifndef _USER_DBG_FW_FUNCS_H
#define _USER_DBG_FW_FUNCS_H
#include "ecore_hsi_func_common.h"
/******************************* Data Types **********************************/

/* MCP Trace  structure, generated by a perl script during MFW build.
 * Therefore, this struct is duplicated in the perl script.
 */
struct mcp_trace_format {
	u32 data;
#define MCP_TRACE_FORMAT_MODULE_MASK	0x0000ffff
#define MCP_TRACE_FORMAT_MODULE_OFFSET	0
#define MCP_TRACE_FORMAT_LEVEL_MASK	0x00030000
#define MCP_TRACE_FORMAT_LEVEL_OFFSET	16
#define MCP_TRACE_FORMAT_P1_SIZE_MASK	0x000c0000
#define MCP_TRACE_FORMAT_P1_SIZE_OFFSET	18
#define MCP_TRACE_FORMAT_P2_SIZE_MASK	0x00300000
#define MCP_TRACE_FORMAT_P2_SIZE_OFFSET	20
#define MCP_TRACE_FORMAT_P3_SIZE_MASK	0x00c00000
#define MCP_TRACE_FORMAT_P3_SIZE_OFFSET	22
#define MCP_TRACE_FORMAT_LEN_MASK	0xff000000
#define MCP_TRACE_FORMAT_LEN_OFFSET	24

	char *format_str;
};

/* MCP Trace Meta data structure, generated by a perl script during MFW build.
 * Therefore, this struct is duplicated in the perl script.
 */
struct mcp_trace_meta {
	u32 modules_num;
	char **modules;
	u32 formats_num;
	struct mcp_trace_format	*formats;
	bool is_allocated;
};

/* Debug Tools user data */
struct dbg_tools_user_data {
	struct mcp_trace_meta mcp_trace_meta;
	const u32 *mcp_trace_user_meta_buf;
};

/******************************** Constants **********************************/

#define MAX_NAME_LEN	16

/***************************** Public Functions *******************************/

/**
 * @brief ecore_dbg_user_set_bin_ptr - Sets a pointer to the binary data with
 * debug arrays.
 *
 * @param p_hwfn -	    HW device data
 * @param bin_ptr - a pointer to the binary data with debug arrays.
 */
enum dbg_status ecore_dbg_user_set_bin_ptr(struct ecore_hwfn *p_hwfn,
					   const u8 *const bin_ptr);

/**
* @brief ecore_dbg_alloc_user_data - Allocates user debug data.
*
* @param p_hwfn -		 HW device data
* @param user_data_ptr - OUT: a pointer to the allocated memory.
*/
enum dbg_status ecore_dbg_alloc_user_data(struct ecore_hwfn *p_hwfn,
					  void **user_data_ptr);

/**
 * @brief ecore_dbg_get_storm_id - Returns an ID for the specified storm name.
 *
 * @param storm_name - Storm name.
 *
 * @return an ID for the specified storm name, or NUM_OF_STORMS if not found.
 */
enum dbg_storms ecore_dbg_get_storm_id(const char *storm_name);

/**
 * @brief ecore_dbg_get_block_id - Returns an ID for the specified block name.
 *
 * @param p_hwfn -	       HW device data
 * @param block_name - Block name.
 *
 * @return an ID for the specified block name, or NUM_OF_BLOCKS if not found.
 */
enum block_id ecore_dbg_get_block_id(struct ecore_hwfn *p_hwfn,
				     const char *block_name);

/**
 * @brief ecore_dbg_get_storm_mode_id - Returns an ID for the specified Storm
 * mode name.
 *
 * @param storm_mode_name - Storm mode name.
 *
 * @return an ID for the specified Storm mode name, or MAX_DBG_BUS_STORM_MODES
 * if not found.
 */
enum dbg_bus_storm_modes ecore_dbg_get_storm_mode_id(const char *storm_mode_name);

/**
 * @brief ecore_dbg_get_constraint_op_id - Returns an ID for the specified
 * constraint operation name.
 *
 * @param op_name - operation name.
 *
 * @return an ID for the specified constraint operation name, or
 * MAX_DBG_BUS_CONSTRAINT_OPS if not found.
 */
enum dbg_bus_constraint_ops ecore_dbg_get_constraint_op_id(const char *op_name);

/**
 * @brief ecore_dbg_get_status_str - Returns a string for the specified status.
 *
 * @param status - a debug status code.
 *
 * @return a string for the specified status
 */
const char* ecore_dbg_get_status_str(enum dbg_status status);

/**
 * @brief ecore_dbg_get_grc_param_id - Returns an ID for the specified GRC
 * param name.
 *
 * @param param_name - GRC param name.
 *
 * @return an ID for the specified GRC param name, or NUM_OF_GRC_PARAMS if not
 * found.
 */
enum dbg_grc_params ecore_dbg_get_grc_param_id(const char *param_name);

/**
 * @brief ecore_dbg_get_dbg_bus_line - Returns an ID for the specified Debug Bus
 * line.
 *
 * @param p_hwfn -	     HW device data
 * @param block_id - block ID
 * @param chip_id -  chip ID
 * @param line -     a string containing a debug line name that belongs to the
 *		     specified block/chip, or an 8-bit debug line number.
 *
 * @return an ID for the specified Debug Bus line name, or -1 if not found.
 */
int ecore_dbg_get_dbg_bus_line(struct ecore_hwfn *p_hwfn,
			       enum block_id block_id,
			       enum chip_ids chip_id,
			       const char *line);

/**
 * @brief ecore_get_idle_chk_results_buf_size - Returns the required buffer
 * size for idle check results (in bytes).
 *
 * @param p_hwfn -		      HW device data
 * @param dump_buf -	      idle check dump buffer.
 * @param num_dumped_dwords - number of dwords that were dumped.
 * @param results_buf_size -  OUT: required buffer size (in bytes) for the
 *			      parsed results.
 *
 * @return error if the parsing fails, ok otherwise.
 */
enum dbg_status ecore_get_idle_chk_results_buf_size(struct ecore_hwfn *p_hwfn,
						    u32 *dump_buf,
						    u32 num_dumped_dwords,
						    u32 *results_buf_size);

/**
 * @brief ecore_print_idle_chk_results - Prints idle check results
 *
 * @param p_hwfn -			HW device data
 * @param dump_buf -		idle check dump buffer.
 * @param num_dumped_dwords -	number of dwords that were dumped.
 * @param results_buf -		buffer for printing the idle check results.
 * @param num_errors -		OUT: number of errors found in idle check.
 * @param num_warnings -	OUT: number of warnings found in idle check.
 *
 * @return error if the parsing fails, ok otherwise.
 */
enum dbg_status ecore_print_idle_chk_results(struct ecore_hwfn *p_hwfn,
					     u32 *dump_buf,
					     u32 num_dumped_dwords,
					     char *results_buf,
					     u32 *num_errors,
					     u32 *num_warnings);

/**
 * @brief ecore_dbg_mcp_trace_set_meta_data - Sets the MCP Trace meta data.
 *
 * Needed in case the MCP Trace dump doesn't contain the meta data (e.g. due to
 * no NVRAM access).
 *
 * @param p_hwfn -	     HW device data
 * @param meta_buf - pointer to MCP Trace meta data
 */
void ecore_dbg_mcp_trace_set_meta_data(struct ecore_hwfn *p_hwfn,
				       const u32 *meta_buf);

/**
 * @brief ecore_get_mcp_trace_results_buf_size - Returns the required buffer
 * size for MCP Trace results (in bytes).
 *
 * @param p_hwfn -		      HW device data
 * @param dump_buf -	      MCP Trace dump buffer.
 * @param num_dumped_dwords - number of dwords that were dumped.
 * @param results_buf_size -  OUT: required buffer size (in bytes) for the
 *			      parsed results.
 *
 * @return error if the parsing fails, ok otherwise.
 */
enum dbg_status ecore_get_mcp_trace_results_buf_size(struct ecore_hwfn *p_hwfn,
						     u32 *dump_buf,
						     u32 num_dumped_dwords,
						     u32 *results_buf_size);

/**
 * @brief ecore_print_mcp_trace_results - Prints MCP Trace results
 *
 * @param p_hwfn -		      HW device data
 * @param dump_buf -	      mcp trace dump buffer, starting from the header.
 * @param num_dumped_dwords - number of dwords that were dumped.
 * @param results_buf -	      buffer for printing the mcp trace results.
 *
 * @return error if the parsing fails, ok otherwise.
 */
enum dbg_status ecore_print_mcp_trace_results(struct ecore_hwfn *p_hwfn,
					      u32 *dump_buf,
					      u32 num_dumped_dwords,
					      char *results_buf);

/**
 * @brief ecore_print_mcp_trace_results_cont - Prints MCP Trace results, and
 * keeps the MCP trace meta data allocated, to support continuous MCP Trace
 * parsing. After the continuous parsing ends, mcp_trace_free_meta_data should
 * be called to free the meta data.
 *
 * @param p_hwfn -		      HW device data
 * @param dump_buf -	      mcp trace dump buffer, starting from the header.
 * @param num_dumped_dwords - number of dwords that were dumped.
 * @param results_buf -	      buffer for printing the mcp trace results.
 *
 * @return error if the parsing fails, ok otherwise.
 */
enum dbg_status ecore_print_mcp_trace_results_cont(struct ecore_hwfn *p_hwfn,
						   u32 *dump_buf,
						   u32 num_dumped_dwords,
						   char *results_buf);

/**
 * @brief print_mcp_trace_line - Prints MCP Trace results for a single line
 *
 * @param p_hwfn -		      HW device data
 * @param dump_buf -	      mcp trace dump buffer, starting from the header.
 * @param num_dumped_bytes -  number of bytes that were dumped.
 * @param results_buf -	      buffer for printing the mcp trace results.
 *
 * @return error if the parsing fails, ok otherwise.
 */
enum dbg_status ecore_print_mcp_trace_line(struct ecore_hwfn *p_hwfn,
					   u8 *dump_buf,
					   u32 num_dumped_bytes,
					   char *results_buf);

/**
* @brief mcp_trace_free_meta_data - Frees the MCP Trace meta data.
* Should be called after continuous MCP Trace parsing.
*
* @param p_hwfn - HW device data
*/
void ecore_mcp_trace_free_meta_data(struct ecore_hwfn *p_hwfn);

/**
 * @brief ecore_get_reg_fifo_results_buf_size - Returns the required buffer
 * size for reg_fifo results (in bytes).
 *
 * @param p_hwfn -		      HW device data
 * @param dump_buf -	      reg fifo dump buffer.
 * @param num_dumped_dwords - number of dwords that were dumped.
 * @param results_buf_size -  OUT: required buffer size (in bytes) for the
 *			      parsed results.
 *
 * @return error if the parsing fails, ok otherwise.
 */
enum dbg_status ecore_get_reg_fifo_results_buf_size(struct ecore_hwfn *p_hwfn,
						    u32 *dump_buf,
						    u32 num_dumped_dwords,
						    u32 *results_buf_size);

/**
 * @brief ecore_print_reg_fifo_results - Prints reg fifo results
 *
 * @param p_hwfn -			HW device data
 * @param dump_buf -		reg fifo dump buffer, starting from the header.
 * @param num_dumped_dwords -	number of dwords that were dumped.
 * @param results_buf -		buffer for printing the reg fifo results.
 *
 * @return error if the parsing fails, ok otherwise.
 */
enum dbg_status ecore_print_reg_fifo_results(struct ecore_hwfn *p_hwfn,
					     u32 *dump_buf,
					     u32 num_dumped_dwords,
					     char *results_buf);

/**
 * @brief ecore_get_igu_fifo_results_buf_size - Returns the required buffer size
 * for igu_fifo results (in bytes).
 *
 * @param p_hwfn -		      HW device data
 * @param dump_buf -	      IGU fifo dump buffer.
 * @param num_dumped_dwords - number of dwords that were dumped.
 * @param results_buf_size -  OUT: required buffer size (in bytes) for the
 *			      parsed results.
 *
 * @return error if the parsing fails, ok otherwise.
 */
enum dbg_status ecore_get_igu_fifo_results_buf_size(struct ecore_hwfn *p_hwfn,
						    u32 *dump_buf,
						    u32 num_dumped_dwords,
						    u32 *results_buf_size);

/**
* @brief ecore_print_igu_fifo_results - Prints IGU fifo results
*
* @param p_hwfn -		     HW device data
* @param dump_buf -	     IGU fifo dump buffer, starting from the header.
* @param num_dumped_dwords - number of dwords that were dumped.
* @param results_buf -	     buffer for printing the IGU fifo results.
*
* @return error if the parsing fails, ok otherwise.
*/
enum dbg_status ecore_print_igu_fifo_results(struct ecore_hwfn *p_hwfn,
					     u32 *dump_buf,
					     u32 num_dumped_dwords,
					     char *results_buf);

/**
 * @brief ecore_get_protection_override_results_buf_size - Returns the required
 * buffer size for protection override results (in bytes).
 *
 * @param p_hwfn -		      HW device data
 * @param dump_buf -	      protection override dump buffer.
 * @param num_dumped_dwords - number of dwords that were dumped.
 * @param results_buf_size -  OUT: required buffer size (in bytes) for the parsed results.
 *
 * @return error if the parsing fails, ok otherwise.
 */
enum dbg_status ecore_get_protection_override_results_buf_size(struct ecore_hwfn *p_hwfn,
							       u32 *dump_buf,
							       u32 num_dumped_dwords,
							       u32 *results_buf_size);

/*
 * @brief ecore_print_protection_override_results - Prints protection override
 * results.
 *
 * @param p_hwfn -               HW device data
 * @param dump_buf -          protection override dump buffer, starting from
 *			      the header.
 * @param num_dumped_dwords - number of dwords that were dumped.
 * @param results_buf -       buffer for printing the reg fifo results.
 *
 * @return error if the parsing fails, ok otherwise.
 */
enum dbg_status ecore_print_protection_override_results(struct ecore_hwfn *p_hwfn,
							u32 *dump_buf,
							u32 num_dumped_dwords,
							char *results_buf);

/**
 * @brief ecore_get_fw_asserts_results_buf_size - Returns the required buffer
 * size for FW Asserts results (in bytes).
 *
 * @param p_hwfn -		      HW device data
 * @param dump_buf -	      FW Asserts dump buffer.
 * @param num_dumped_dwords - number of dwords that were dumped.
 * @param results_buf_size -  OUT: required buffer size (in bytes) for the
 *			      parsed results.
 *
 * @return error if the parsing fails, ok otherwise.
*/
enum dbg_status ecore_get_fw_asserts_results_buf_size(struct ecore_hwfn *p_hwfn,
						      u32 *dump_buf,
						      u32 num_dumped_dwords,
						      u32 *results_buf_size);

/**
* @brief ecore_print_fw_asserts_results - Prints FW Asserts results
*
* @param p_hwfn -		     HW device data
* @param dump_buf -	     FW Asserts dump buffer, starting from the header.
* @param num_dumped_dwords - number of dwords that were dumped.
* @param results_buf -	     buffer for printing the FW Asserts results.
*
* @return error if the parsing fails, ok otherwise.
*/
enum dbg_status ecore_print_fw_asserts_results(struct ecore_hwfn *p_hwfn,
					       u32 *dump_buf,
					       u32 num_dumped_dwords,
					       char *results_buf);

/**
 * @brief ecore_dbg_parse_attn - Parses and prints attention registers values in
 * the specified results struct.
 *
 * @param p_hwfn -	    HW device data
 * @param results - Pointer to the attention read results
 *
 * @return error if one of the following holds:
 *	- the version wasn't set
 * Otherwise, returns ok.
 */
enum dbg_status ecore_dbg_parse_attn(struct ecore_hwfn *p_hwfn,
				     struct dbg_attn_block_result *results);

#endif
