/****************************************************************************
 * Name:        append.h
 *
 * Description:
 *      This is a utility to append firmware and other images into a
 *      single file. The primary use of this is to combine two phases
 *      of the bootcode into a single file. It appends some header
 *      information (used for parsing the file) and the input image to
 *      the output file. (If the output file does not yet exist, it'll
 *      create one.)
 *      This header file defines the image header information.
 * 
 * Bundle image layout:
 *      ========================================================
 *      = Bundle Header <bundle_header>
 *      =                             <magic     > - 0xbdbdbdbd
 *	=                             <version   > - Currently 1
 *      =                             <num_images> - Of the bundle
 *      =                             <total_size> - Of the bundle
 *      ========================================================
 *      = Img1 Hdr      <image_header>
 *      =                             <magic     > - 0x669955aa
 *      =                             <version   > - Currently 2
 *      =                             <type      > 
 *      =                             <image_info>
 *      =                             <start_addr>
 *      =                             <run_addr  >
 *      =                             <byte_cnt  >
 *      ========================================================
 *      =     Img1 data 
 *      ========================================================
 *      ========================================================
 *      =     ImgN Hdr     <image_header>
 *      ========================================================
 *      =     ImgN data   
 *      ========================================================
 *
 ****************************************************************************/

#ifndef APPEND_H
#define APPEND_H

#define SIGNATURE_MAX_DER_SIZE 128
#define SIGNATURE_MIN_DER_SIZE 64

struct image_header {
#pragma pack(push, 1)
	u32 magic;
#define FILE_MAGIC                       0x669955aa
	u32 version;
#define FORMAT_VERSION_1                 0x1
#define FORMAT_VERSION_2	         0x2
#define FORMAT_VERSION_3		 0x3
#define LATEST_FORMAT_VERSION            FORMAT_VERSION_3
	u32 type;
	u32 image_info;
	/* MAX_MEM base value is 8K, means if MAX_MEM value is 0,
	 * the size is 8K. */
#define IMAGE_INFO_MAX_MEM_BASE                  8
	/* Runtime mem size required in k, Encoded with value +
	 * IMAGE_INFO_MAX_MEM_BASE */
#define IMAGE_INFO_MAX_MEM_MASK         0x0000001f

	/* bit 23:16 reserved for bit define that device it can support.
	 * These are bit fields. */
#define IMAGE_INFO_CHIP_MASK            0x00ff0000
#define IMAGE_INFO_CHIP_57940		0x00200000
#define IMAGE_INFO_CHIP_579XX		0x00400000
#define IMAGE_INFO_CHIP_579XX_B0_ONLY	0x00500000 /* For PCIE2 */
#define IMAGE_INFO_CHIP_51000		0x00510000 /* E5 */


	u32 start_addr;
	u32 run_addr;
	u32 byte_cnt;
	u32 image[1];        /* Unbounded */
#pragma pack(pop)
};

#define IMG_HDR_LEN (sizeof(struct image_header))

struct bundle_header {
	u32 magic;
#define BUNDLE_MAGIC 0xbdbdbdbd
	u32 version;
#define BUNDLE_IMAGE_VER 1
	u32 num_images;
	u32 total_size;
};

#define MAX_IMAGES_IN_BUNDLE 16
struct hdr_type_entry {
	u32 type;
	u32 max_img_size;
	int is_gen_fw_header;
	u32 sig_idx; /* If applicable */
};
struct bundle_info {
	char *name;
	u32 num_images;
	struct hdr_type_entry entries[MAX_IMAGES_IN_BUNDLE];
};
int generate_mfw(int argc, char * argv [], int tim_pk_idx, int mim_pk_idx);
#endif				/*APPEND_H */
