#include "bcm_types.h"
#include "reg_addr.h"
#include "append.h"
#include "nvm_map.h"
#include "mcp_public.h"
#include "nvm_cfg.h"
#include "misc_bits.h"
#include "mono_image.h"
#include "checksum.h"
#include "Decompress.h"
#include "spad_layout.h"
#include "crctab.h"
#include "common_nvm.h"
#include "drv_nvm_funcs_al.h"
#include "nvmcfg_meta.h"
#include "bdn.h"
#include "mcp_debug.h"

#ifdef __cplusplus
extern "C" {
#endif  /* __cplusplus */

#define PCI_CODE_X86		0x00
#define PCI_IND_LAST_IMAGE	0x80
#define DONT_CARE_BUNDLE_ID 0xFF
#define READ_FROM_NVM 		0
#define READ_FROM_IMAGE 	1
#define MAX_VER_STR_LEN		70
#define MAX_MSG_LEN 		120
#define DIR_MAX_NAME 		64
#define NVRAM_TIMEOUT_COUNT 10000

#define OEM1_SVID 			0x1590
#define OEM2_SVID 			0x1137

#define SDID_FALCON1 		0x0228
#define SDID_FALCON2 		0x027B


typedef struct _mba_versions_t {
	u32 file_vers[ROMIMG_NUM_MAX];
	u32 nvm_vers[ROMIMG_NUM_MAX];
	char *type_str[ROMIMG_NUM_MAX];
	u32 efi_version;
} mba_versions;

enum upgrade_type_e {
	NVM_UPGRADE_MFW,
	NVM_UPGRADE_IMAGE,
	EXTPHY_FW_UPGRADE,
	PK_UPGRADE,
	NVM_CFG_UPGRADE,
	NVM_META_UPGRADE,
	NVM_UPGRADE_MBI
};

#if !(defined htonl) && !(defined WINEDIAG)
u16 htons (u16 data);
u16 ntohs (u16 data);
u32 htonl (u32 data);
u32 ntohl (u32 data);
#endif

typedef enum {
	PORT_MODE_1_PORTS,
	PORT_MODE_2_PORTS,
	PORT_MODE_4_PORTS,
	PORT_MODE_INVALID
} port_mode_t;

typedef enum {
	CHIP_TYPE_BB,
	CHIP_TYPE_AH,
	CHIP_TYPE_INVALID
} chip_type_t;

typedef struct {
	chip_type_t chip_type;
	int nvm_size;
	FILE *fd;
	u8 *nvm_buff;
} virt_nvm_t;

typedef int (*nvm_read_func)(u32 addr,
							 u32 byte_cnt,
							 u32 *ret_buf,
							 u8 show_progress);

typedef int (*nvm_file_delete_func)(u32 image_type);

typedef int (*nvm_write_func)(u32 addr,
							  u32 byte_cnt,
							  u8 *buf,
							  u8 show_progress);

typedef int (*nvm_write_data_func)(u32 command,
								   u32 addr,
								   u8 *buf,
								   u32 byte_cnt,
								   bool show_progress,
								   bool *is_put_file_complete);

typedef int (*nvm_upgrade_func)(struct image_header *p_image_header, /* Just for image upgrade. */
								u8 *p_image_data,
								u32 file_size,
								enum upgrade_type_e upgrade_type,
								bool b_validate_version);

typedef int (*nvm_dir_func)(u32 nvm_dir_id,
						  struct nvm_dir *p_dir);

typedef int (*get_file_func)(u32 image_type,
							 u32 file_size,
							 u32 *buf);

typedef int (*mcp_reset_func)(void);

typedef int (*running_mfw_func) (u32 *mcp_ver,
								 u32 *running_bundle_id);

struct dispatch_nvm_func {
	nvm_read_func read;
	nvm_write_func write;
	nvm_write_data_func write_data;
	nvm_upgrade_func upgrade;
	nvm_dir_func dir;
	nvm_file_delete_func delete;
	mcp_reset_func mcp_reset;
	running_mfw_func running_mfw; 

};


/********************************************************************* 
 *  
 * 						Secure NVM functions 
 *  
 *********************************************************************/
int secure_nvm_read(u32 addr,
				    u32 byte_cnt,
					u32 *ret_buf,
					u8 show_progress);

int secure_nvm_write(u32 address,
					 u32 byte_cnt,
					 u8 *buf,
					 u8 show_progress);

int secure_nvm_write_data(u32 command,
						  u32 addr,
						  u8 *buf,
						  u32 byte_cnt,
						  bool show_progress,
						  bool *is_put_file_complete);

int secure_nvm_upgrade(struct image_header *p_image_header,
					   u8 *p_buf,
					   u32 file_size,
					   enum upgrade_type_e upgrade_type,
					   bool b_validate_version);

int secure_nvm_dir(u32 dir_id, struct nvm_dir *p_dir);

int secure_nvm_file_del(u32 image_type);

int secure_mcp_reset(void);

int drv_dbg_secure_nvm_read(u32 addr,
							u32 byte_cnt,
							u32 *ret_buf,
							u8 show_progress);

int drv_dbg_secure_nvm_write(u32 address,
							 u32 byte_cnt,
							 u8 *buf, 
							 u8 show_progress);

int drv_dbg_secure_nvm_write_data(u32 command,
								  u32 addr,
								  u8 *buf,
								  u32 byte_cnt,
								  bool show_progress,
								  bool *is_put_file_complete);

/********************************************************************* 
 *  
 * 					Non-Secure (Legacy) NVM functions 
 *  
 *********************************************************************/

int acquire_nvram_lock();

int release_nvram_lock();

int legacy_nvm_read(u32 addr,
					u32 byte_cnt,
					u32 *ret_buf,
					u8 show_progress);

int legacy_nvm_write(u32 nvm_addr,
					 u32 byte_cnt,
					 u8 *buf,
					 u8 show_progress);

int legacy_nvm_upgrade(struct image_header *p_image_header,
					   u8 *p_buf,
					   u32 file_size,
					   enum upgrade_type_e upgrade_type,
					   bool b_validate_version);

int legacy_nvm_dir(u32 dir_id, 
				   struct nvm_dir *p_dir);

int legacy_nvm_delete(u32 image_type);

int legacy_mcp_reset(void);

int mcp_debug_mb(u32 *p_debug_mb_offset, u32 cmd, u32 offset, u32 val);

int serial_nvm_read(u32 addr,
                    u32 byte_cnt,
                    u32 *ret_buf,
                    u8 show_progress);

int serial_nvm_write(u32 addr,
                     u32 byte_cnt,
                     u8 *buf,
                     u8 show_progress);

int don_nvm_read(u32 addr,
				u32 byte_cnt,
				u32 *ret_buf,
				u8 show_progress);

int don_nvm_write(u32 nvm_addr,
				 u32 byte_cnt,
				 u8 *buf,
				 u8 show_progress);
#ifndef TEDIBEAR
int virt_nvm_read(u32 nvm_addr,
		 		 u32 byte_cnt,
				 u32 *ret_buf,
				 u8 show_progress);

int virt_nvm_write(u32 nvm_addr,
				 u32 byte_cnt,
				 u8 *buf,
				 u8 show_progress);

int virt_running_mfw(u32 *mcp_ver,
					u32 *running_bundle_id);

int virt_mcp_reset(void);
#endif


/***************************************************************************
 * 
 * 					Utilities 
 *  
 ***************************************************************************/
void nvm_func_init(u32 is_mcp_running);
int nvm_read(u32 nvm_addr, 
			 u32 n_bytes, 
			 u32 *read_buf);

int nvm_write(u32_t nvm_addr,
			   u32_t byte_cnt, 
			  u32_t *buf);

int get_nvm_secure_mode(void);
int analyze_nvm_rsp(u32 rsp_cmd);
int find_image_name_by_type(u32 type,
						  char *name);

int get_running_mfw(u32 *mcp_ver, 
					u32 *running_bundle_id);

int alloc_and_get_nvm_dir(struct nvm_dir **pp_dir, 
						  u32 *dir_id_o, 
						  u32 *next_mfw_o);

int nvm_find_image(u32 image_type,
			      u32 *nvm_offset,
			      u32 *length,
			      struct nvm_dir *p_dir);

int find_image_in_table(char *option,
						u32 *image_idx);

int image_name_2_type(char *name,
					u32 *o_type);


int find_image_name_by_type(u32 type,
						  char *name);

int validate_image_version(u32 image_type,
						   char *old_version,
						   char *new_version,
						   u8    *p_is_same_ver);

u32 get_image_table_size();
char *get_image_option_by_index(u32 idx);

void nvm_write_progress_cb(u32 byte_cnt,
						   u32 orig_byte_cnt);

int nvm_compute_crc(u32 nvram_start_addr,
 			        u32 byte_cnt,
				    u32 *crc_ptr);

void read_fw_version(u8 source,
 					 u32 type,
					 struct nvm_dir *p_dir,
					 u32 nvm_addr,
					 u8 *p_image,
					 char *p_str);

int update_mba_hdr(u8 *buf);

int update_image_in_nvm_by_type(u8 *buf, 
								u32 byte_cnt, 
								u32 image_type, 
								struct nvm_dir *p_dir);

int upgrade_nvm_image(struct image_header *p_image_header,
					  u8 *buf,
					  struct nvm_dir *p_dir,
					  bool b_validate_version);

struct image_header *create_image_header(u8 *buf,
										 u32 bytes_read,
										 u32 image_idx,
										 enum upgrade_type_e upgrade_type,
										 u8 **p_image_data,
										 struct image_header *tmp_image_header);

int nvm_read_svid(u16 *p_svid);
int nvm_read_sdid(u16 *p_sdid);
int is_selected_sdid(u16 sdid);

int read_num_ports_from_nvm(u32 *p_num_ports);

int read_mac_addr_from_nvm(u32 func_idx, char *p_mac_addr);
int read_fcoe_node_mac_addr_from_nvm(u32 func_idx, char *p_mac_addr);
int read_fcoe_port_mac_addr_from_nvm(u32 func_idx, char *p_mac_addr);
int read_lldp_mac_addr_from_nvm(u32 func_idx, char *p_mac_addr);

int read_serial_no_from_nvm(char *p_serial_no);

int read_personality_from_nvm(u32 port_idx, u32 device_id, char *p_personality);

int read_num_provisioned_mac_from_nvm(u16 *p_num_provisioned_mac, 
                                      u8 *p_num_provisioned_vf_mac, 
                                      u8 *p_is_bmc_mac);

int read_trace_modules_from_nvm(u32 *p_trace_modules);

int write_trace_modules_in_nvm(u32 trace_modules);

int set_mfw_version_in_vpd(Tcl_Interp *interp);

int set_efi_version_in_vpd(Tcl_Interp *interp);

ediag_status_t clear_efi_version_in_vpd(Tcl_Interp *interp);

int get_mbi_version(struct nvm_dir *p_dir, u32 *mbi_version, u32 *mbi_date);

int create_default_cfg_image(void);

int restore_factory_default_cfg(void);


int dump_meta_text_file(char *filename, u8 *buf, u32 size);

int nvmcfg_update(u8 *p_incoming_nvm_cfg_buffer,
				  u32 incoming_nvm_cfg_length,
				  u8 *p_incoming_nvm_meta_buffer,
				  u32 incoming_nvm_meta_length);

int nvmcfg_update_cfg(u8 *onboard_default_cfg_buffer,u32 onboard_default_cfg_length,
					  u8 *onboard_nvm_cfg_buffer,u32 onboard_nvm_cfg_length,
					  u8 *onboard_nvm_meta_buffer,u32 onboard_nvm_meta_length,
					  u8 *incoming_nvm_cfg_buffer, u32 incoming_nvm_cfg_length,
					  u8 *incoming_nvm_meta_buffer, u32 incoming_nvm_meta_length,
					  u8  *outgoing_nvm_cfg_buffer, u32 *outgoing_nvm_cfg_length,
					  u8  *outgoing_default_cfg_buffer, u32 *outgoing_default_cfg_length);

int add_bdn_image(void);

int extphy_fw_upgrade(u8 *extphy_buf_p, u32 file_size);
int extphy_fw_upgrade_verify(void);

int nvm_read_via_debug_mb(u32 addr,
                          u32 byte_cnt,
                          u32 *ret_buf);

#ifdef __cplusplus
}
#endif  /* __cplusplus */