#ifndef __ECORE_LL2_H__
#define __ECORE_LL2_H__

#include "ecore.h"
#include "ecore_hsi_eth.h"
#include "ecore_chain.h"
#include "ecore_hsi_common.h"
#include "ecore_ll2_api.h"
#include "ecore_sp_api.h"

/* ECORE LL2: internal structures and functions*/

/* LL2 queues handles will be split as follows:
 * first will be legacy queues, and then the ctx based queues.
 * for VF it will always start from 0 and will have its own allowed connections
 */
#define ECORE_MAX_NUM_OF_LL2_CONNS_PF	       (4)
#define ECORE_MAX_NUM_OF_LEGACY_LL2_CONNS_PF   (3)

#define ECORE_MAX_NUM_OF_CTX_LL2_CONNS_PF \
	(ECORE_MAX_NUM_OF_LL2_CONNS_PF - ECORE_MAX_NUM_OF_LEGACY_LL2_CONNS_PF)

#define ECORE_LL2_LEGACY_CONN_BASE_PF	0
#define ECORE_LL2_CTX_CONN_BASE_PF	ECORE_MAX_NUM_OF_LEGACY_LL2_CONNS_PF
#define ECORE_LL2_CONN_BASE_VF		ECORE_MAX_NUM_OF_CTX_LL2_CONNS_PF

#define ECORE_MAX_NUM_OF_LL2_CONNS(p_hwfn) \
	(IS_PF(p_hwfn->p_dev) ? ECORE_MAX_NUM_OF_LL2_CONNS_PF : \
			       ECORE_MAX_NUM_OF_LL2_CONNS_VF)

#define ECORE_LL2_RX_REGISTERED(ll2)	((ll2)->rx_queue.b_cb_registered)
#define ECORE_LL2_TX_REGISTERED(ll2)	((ll2)->tx_queue.b_cb_registered)

static OSAL_INLINE u8 ecore_ll2_handle_to_queue_id(struct ecore_hwfn *p_hwfn,
					      u8 handle, u8 ll2_queue_type,
					      u8 vf_id)
{
	u8 qid;

	if (ll2_queue_type == ECORE_LL2_RX_TYPE_LEGACY) {
		return p_hwfn->hw_info.resc_start[ECORE_LL2_RAM_QUEUE] + handle;
	} else {
		/* ECORE_LL2_RX_TYPE_CTX
		 * FW distinguish between the legacy queus (ram based) and the
		 * ctx based queues by the queue_id.
		 * The first MAX_NUM_LL2_RX_RAM_QUEUES queues are legacy
		 * and the queue ids above that are ctx base.
		 */
		qid = p_hwfn->hw_info.resc_start[ECORE_LL2_CTX_QUEUE] +
		      MAX_NUM_LL2_RX_RAM_QUEUES;

		/* See comment on th acquire conneciton for how the ll2
		 * queues handles are divided.
		 */
		if (vf_id == ECORE_CXT_PF_CID)
			qid += (handle - ECORE_MAX_NUM_OF_LEGACY_LL2_CONNS_PF);
		else
			qid += ECORE_LL2_CONN_BASE_VF + vf_id;
	}

	return qid;
}

struct ecore_ll2_rx_packet
{
	osal_list_entry_t   list_entry;
	struct core_rx_bd_with_buff_len   *rxq_bd;
	dma_addr_t          rx_buf_addr;
	u16                 buf_length;
	void                *cookie;
	u8                  placement_offset;
	u16                 parse_flags;
	u16                 packet_length;
	u16                 vlan;
	u32                 opaque_data[2];
};

struct ecore_ll2_tx_packet
{
	osal_list_entry_t       list_entry;
	u16                     bd_used;
	bool                    notify_fw;
	void                    *cookie;
	struct {
		struct core_tx_bd       *txq_bd;
		dma_addr_t              tx_frag;
		u16                     frag_len;
	}   bds_set[1];
	/* Flexible Array of bds_set determined by max_bds_per_packet */
};

struct ecore_ll2_rx_queue {
	osal_spinlock_t			lock;
	struct ecore_chain		rxq_chain;
	struct ecore_chain		rcq_chain;
	u8				rx_sb_index;
	u8				ctx_based;
	bool				b_cb_registered;
	bool				b_started;
	__le16				*p_fw_cons;
	osal_list_t			active_descq;
	osal_list_t			free_descq;
	osal_list_t			posting_descq;
	struct ecore_ll2_rx_packet	*descq_array;

	/* Either RAM addr or doorbell (depends on qid) */
	void OSAL_IOMEM		*set_prod_addr;
	struct core_pwm_prod_update_data db_data;
};

struct ecore_ll2_tx_queue {
	osal_spinlock_t			lock;
	struct ecore_chain		txq_chain;
	u8				tx_sb_index;
	bool				b_cb_registered;
	bool				b_started;
	__le16				*p_fw_cons;
	osal_list_t			active_descq;
	osal_list_t			free_descq;
	osal_list_t			sending_descq;
	struct ecore_ll2_tx_packet	*descq_array;
	struct ecore_ll2_tx_packet	*cur_send_packet;
	struct ecore_ll2_tx_packet	cur_completing_packet;
	u16				cur_completing_bd_idx;
	void OSAL_IOMEM			*doorbell_addr;
	struct core_db_data		db_msg;
	u16				bds_idx;
	u16				cur_send_frag_num;
	u16				cur_completing_frag_num;
	bool				b_completing_packet;
};

struct ecore_ll2_info {
	osal_mutex_t			mutex;

	struct ecore_ll2_acquire_data_inputs input;
	u32				cid;
	u8				my_id;
	u8				queue_id;
	u8				tx_stats_id;
	bool				b_active;
	enum core_tx_dest		tx_dest;
	u8				tx_stats_en;
	u8				main_func_queue;
	struct ecore_ll2_rx_queue	rx_queue;
	struct ecore_ll2_tx_queue	tx_queue;
	struct ecore_ll2_cbs		cbs;
};

struct ecore_sp_ll2_rx_queue_start_params {
	u64 bd_base_addr;
	u64 cqe_pbl_addr;
	u32 cid;
	u16 opaque_fid;
	u16 sb_id;
	u16 mtu;
	u16 num_of_pbl_pages;
	u8 sb_index;
	u8 drop_ttl0_flg;
	u8 inner_vlan_stripping_en;
	u8 report_outer_vlan;
	u8 gsi_enable;
	u8 queue_id;
	u8 main_func_queue;
	u8 mf_si_bcast_accept_all;
	u8 mf_si_mcast_accept_all;
	u8 vport_id_valid;
	u8 vport_id;
	enum ecore_ll2_error_handle err_packet_too_big;
	enum ecore_ll2_error_handle err_no_buf;
};

struct ecore_sp_ll2_tx_queue_start_params {
	u64 pbl_addr;
	u32 cid;
	u16 pbl_size;
	u16 pq_id;
	u16 opaque_fid;
	u16 sb_id;
	u16 mtu;
	u8 sb_index;
	u8 stats_en;
	u8 stats_id;
	u8 conn_type;
	u8 gsi_enable;
	u8 vport_id_valid;
	u8 vport_id;
};

struct ecore_sp_ll2_rx_queue_stop_params {
	u32 cid;
	u16 opaque_fid;
	u8 queue_id;
};

struct ecore_sp_ll2_tx_queue_stop_params {
	u32 cid;
	u16 opaque_fid;
};

/**
* @brief ecore_ll2_alloc - Allocates LL2 connections set
*
* @param p_hwfn
*
* @return enum _ecore_status_t
*/
enum _ecore_status_t ecore_ll2_alloc(struct ecore_hwfn *p_hwfn);

/**
* @brief ecore_ll2_setup - Inits LL2 connections set
*
* @param p_hwfn
*
*/
void ecore_ll2_setup(struct ecore_hwfn *p_hwfn);

/**
* @brief ecore_ll2_free - Releases LL2 connections set
*
* @param p_hwfn
*
*/
void ecore_ll2_free(struct ecore_hwfn *p_hwfn);

/**
 * @brief ecore_ll2_handle_sanity_lock - performs sanity on the connection
 *					 handle and return the corresponding
 *					 ll2_info
 *
 * @param p_hwfn
 * @param connection_handle    LL2 connection's handle
 *                              obtained from
 *                              ecore_ll2_require_connection
 *
 * @return struct ecore_ll2_info*
 */
struct ecore_ll2_info *
ecore_ll2_handle_sanity_lock(struct ecore_hwfn *p_hwfn,
			     u8 connection_handle);

/**
 * @brief ecore_sp_ll2_rx_queue_start - Sends RX_QUEUE_START ramrod to the FW
 *
 * @param p_hwfn
 * @param params    ll2_rx_queue_start_params
 *
 * @return enum _ecore_status_t
 */
enum _ecore_status_t
ecore_sp_ll2_rx_queue_start(struct ecore_hwfn *p_hwfn,
			    struct ecore_sp_ll2_rx_queue_start_params *params);

/**
 * @brief ecore_sp_ll2_tx_queue_start - Sends TX_QUEUE_START ramrod to the FW
 *
 * @param p_hwfn
 * @param params    ll2_tx_queue_start_params
 *
 * @return enum _ecore_status_t
 */
enum _ecore_status_t
ecore_sp_ll2_tx_queue_start(struct ecore_hwfn *p_hwfn,
			    struct ecore_sp_ll2_tx_queue_start_params *params);

/**
 * @brief ecore_sp_ll2_tx_queue_stop - Sends TX_QUEUE_STOP ramrod to the FW
 *
 * @param p_hwfn
 * @param params    ll2_tx_queue_stop_params
 *
 * @return enum _ecore_status_t
 */
enum _ecore_status_t
ecore_sp_ll2_tx_queue_stop(struct ecore_hwfn *p_hwfn,
			   struct ecore_sp_ll2_tx_queue_stop_params *params);

/**
 * @brief ecore_sp_ll2_rx_queue_stop - Sends RX_QUEUE_STOP ramrod to the FW
 *
 * @param p_hwfn
 * @param params    ll2_rx_queue_stop_params
 *
 * @return enum _ecore_status_t
 */
enum _ecore_status_t
ecore_sp_ll2_rx_queue_stop(struct ecore_hwfn *p_hwfn,
			   struct ecore_sp_ll2_rx_queue_stop_params *params);

#ifndef LINUX_REMOVE
/**
 * @brief ecore_ll2_get_fragment_of_tx_packet
 *
 * @param p_hwfn
 * @param connection_handle    LL2 connection's handle
 *                              obtained from
 *                              ecore_ll2_require_connection
 * @param addr
 * @param last_fragment)
 *
 * @return enum _ecore_status_t
 */
enum _ecore_status_t
ecore_ll2_get_fragment_of_tx_packet(struct ecore_hwfn *p_hwfn,
				    u8 connection_handle,
				    dma_addr_t *addr,
				    bool *last_fragment);
#endif

#endif /*__ECORE_LL2_H__*/
