/*
 * Copyright (c) 2017-2019 Cavium, Inc.
 * All rights reserved.                
 *                                     
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:                                                          
 *                                                                    
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer. 
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *                                                                         
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE   
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 *  POSSIBILITY OF SUCH DAMAGE.                                                
 */    
#include "bcm_osal.h"

#include "ecore.h"
#include "ecore_status.h"
#include "ecore_ll2.h"
#include "ecore_iscsi.h"
#include "ecore_iro.h"
#include "reg_addr.h"
#include "ecore_int.h"
#include "ecore_cxt.h"
#include "ecore_sp_commands.h"
#include "ecore_hw.h"
#include "ecore_gtt_reg_addr.h"
#include "ecore_dev_api.h"

#ifdef _NTDDK_
#pragma warning(push)
#pragma warning(disable : 28167)
#pragma warning(disable : 28123)
#pragma warning(disable : 6011)
#endif

static enum _ecore_status_t
ecore_iscsi_async_event(struct ecore_hwfn *p_hwfn,
			u8 fw_event_code,
			u16 OSAL_UNUSED echo,
			union event_ring_data *data,
			u8 OSAL_UNUSED fw_return_code,
			u8 OSAL_UNUSED vf_id)
{
	if (p_hwfn->p_iscsi_info->event_cb != OSAL_NULL) {
		struct ecore_iscsi_info *p_iscsi = p_hwfn->p_iscsi_info;

		return p_iscsi->event_cb(p_iscsi->event_context,
					 fw_event_code, data);
	} else {
		DP_NOTICE(p_hwfn,
			  false, "iSCSI async completion is not set\n");
		return ECORE_INVAL;
	}
}

LNX_STATIC enum _ecore_status_t
ecore_sp_iscsi_func_start(struct ecore_hwfn *p_hwfn,
			  enum spq_mode comp_mode,
			  struct ecore_spq_comp_cb *p_comp_addr,
			  void *event_context,
			  iscsi_event_cb_t async_event_cb)
{
	struct iscsi_init_ramrod_params	*p_ramrod = OSAL_NULL;
	struct scsi_init_func_queues *p_queue = OSAL_NULL;
	struct ecore_iscsi_pf_params *p_params = OSAL_NULL;
	struct iscsi_spe_func_init *p_init = OSAL_NULL;
	struct ecore_spq_entry *p_ent = OSAL_NULL;
	struct ecore_sp_init_data init_data;
	u8 i, num_sbs;
	enum _ecore_status_t rc = ECORE_SUCCESS;
	u32 dval;
	u16 val;

	/* Get SPQ entry */
	OSAL_MEMSET(&init_data, 0, sizeof(init_data));
	init_data.cid = ecore_spq_get_cid(p_hwfn);
	init_data.opaque_fid = p_hwfn->hw_info.opaque_fid;
	init_data.comp_mode = comp_mode;
	init_data.p_comp_data = p_comp_addr;

	rc = ecore_sp_init_request(p_hwfn, &p_ent,
				   ISCSI_RAMROD_CMD_ID_INIT_FUNC,
				   PROTOCOLID_ISCSI, &init_data);
	if (rc != ECORE_SUCCESS)
		return rc;

	p_ramrod = &p_ent->ramrod.iscsi_init;
	p_init = &p_ramrod->iscsi_init_spe;
	p_params = &p_hwfn->pf_params.iscsi_pf_params;
	p_queue = &p_init->q_params;

	/* Sanity */
	if (p_params->num_queues > p_hwfn->hw_info.feat_num[ECORE_ISCSI_CQ]) {
		DP_ERR(p_hwfn,
		       "Cannot satisfy CQ amount. Queues requested %d, CQs available %d. Aborting function start\n",
		       p_params->num_queues,
		       p_hwfn->hw_info.feat_num[ECORE_ISCSI_CQ]);
		ecore_sp_destroy_request(p_hwfn, p_ent);
		return ECORE_INVAL;
	}

	val = p_params->half_way_close_timeout;
	p_init->half_way_close_timeout = OSAL_CPU_TO_LE16(val);
	p_init->num_sq_pages_in_ring = p_params->num_sq_pages_in_ring;
	p_init->num_r2tq_pages_in_ring = p_params->num_sq_pages_in_ring;
	p_init->num_uhq_pages_in_ring = p_params->num_uhq_pages_in_ring;
	p_init->ll2_rx_queue_id =
		p_hwfn->hw_info.resc_start[ECORE_LL2_RAM_QUEUE] +
		p_params->ll2_ooo_queue_id;

	if (p_params->disable_stats_collection) {
		SET_FIELD(p_init->flags, ISCSI_SPE_FUNC_INIT_COUNTERS_EN, 0);
	} else {
		SET_FIELD(p_init->flags, ISCSI_SPE_FUNC_INIT_COUNTERS_EN, 1);
	}

	p_init->func_params.log_page_size = p_params->log_page_size;
	p_init->func_params.log_page_size_conn = p_params->log_page_size_conn;
	val = p_params->num_tasks;
	p_init->func_params.num_tasks = OSAL_CPU_TO_LE16(val);
	p_init->debug_mode.flags = p_params->debug_mode;

	DMA_REGPAIR_LE(p_queue->glbl_q_params_addr,
		       p_params->glbl_q_params_addr);

	val = p_params->cq_num_entries;
	p_queue->cq_num_entries = OSAL_CPU_TO_LE16(val);
	val = p_params->cmdq_num_entries;
	p_queue->cmdq_num_entries = OSAL_CPU_TO_LE16(val);
	p_queue->num_queues = p_params->num_queues;
	dval = (u8)p_hwfn->hw_info.resc_start[ECORE_CMDQS_CQS];
	p_queue->queue_relative_offset = (u8)dval;
	p_queue->cq_sb_pi = p_params->gl_rq_pi;
	p_queue->cmdq_sb_pi = p_params->gl_cmd_pi;

	/* In storage target each CQ/CmdQ pair uses a single SB */
	num_sbs = p_params->num_queues;
	if (p_params->is_target)
		num_sbs /= 2;
	for (i = 0; i < num_sbs; i++) {
		val = ecore_get_igu_sb_id(p_hwfn, i);
		p_queue->cq_cmdq_sb_num_arr[i] = OSAL_CPU_TO_LE16(val);
	}

	p_queue->storage_func_id = (u8)RESC_START(p_hwfn, ECORE_BDQ);

	DMA_REGPAIR_LE(p_queue->bdq_pbl_base_address[BDQ_ID_RQ],
		       p_params->bdq_pbl_base_addr[BDQ_ID_RQ]);
	p_queue->bdq_pbl_num_entries[BDQ_ID_RQ] =
		p_params->bdq_pbl_num_entries[BDQ_ID_RQ];
	val = p_params->bdq_xoff_threshold[BDQ_ID_RQ];
	p_queue->bdq_xoff_threshold[BDQ_ID_RQ] = OSAL_CPU_TO_LE16(val);
	val = p_params->bdq_xon_threshold[BDQ_ID_RQ];
	p_queue->bdq_xon_threshold[BDQ_ID_RQ] = OSAL_CPU_TO_LE16(val);

	DMA_REGPAIR_LE(p_queue->bdq_pbl_base_address[BDQ_ID_IMM_DATA],
		       p_params->bdq_pbl_base_addr[BDQ_ID_IMM_DATA]);
	p_queue->bdq_pbl_num_entries[BDQ_ID_IMM_DATA] =
		p_params->bdq_pbl_num_entries[BDQ_ID_IMM_DATA];
	val = p_params->bdq_xoff_threshold[BDQ_ID_IMM_DATA];
	p_queue->bdq_xoff_threshold[BDQ_ID_IMM_DATA] = OSAL_CPU_TO_LE16(val);
	val = p_params->bdq_xon_threshold[BDQ_ID_IMM_DATA];
	p_queue->bdq_xon_threshold[BDQ_ID_IMM_DATA] = OSAL_CPU_TO_LE16(val);

	DMA_REGPAIR_LE(p_queue->bdq_pbl_base_address[BDQ_ID_TQ],
		       p_params->bdq_pbl_base_addr[BDQ_ID_TQ]);
	p_queue->bdq_pbl_num_entries[BDQ_ID_TQ] =
		p_params->bdq_pbl_num_entries[BDQ_ID_TQ];
	val = p_params->bdq_xoff_threshold[BDQ_ID_TQ];
	p_queue->bdq_xoff_threshold[BDQ_ID_TQ] = OSAL_CPU_TO_LE16(val);
	val = p_params->bdq_xon_threshold[BDQ_ID_TQ];
	p_queue->bdq_xon_threshold[BDQ_ID_TQ] = OSAL_CPU_TO_LE16(val);

	val = p_params->rq_buffer_size;
	p_queue->rq_buffer_size = OSAL_CPU_TO_LE16(val);
	if (p_params->is_target) {
		SET_FIELD(p_queue->q_validity,
			  SCSI_INIT_FUNC_QUEUES_RQ_VALID, 1);
		if (p_queue->bdq_pbl_num_entries[BDQ_ID_IMM_DATA]) {
			SET_FIELD(p_queue->q_validity,
				  SCSI_INIT_FUNC_QUEUES_IMM_DATA_VALID, 1);
		}
		if (p_queue->bdq_pbl_num_entries[BDQ_ID_TQ]) {
			SET_FIELD(p_queue->q_validity,
				  SCSI_INIT_FUNC_QUEUES_TQ_VALID, 1);
			if (p_params->is_soc_en) {
				SET_FIELD(p_queue->q_validity,
					  SCSI_INIT_FUNC_QUEUES_SOC_EN, 1);
				SET_FIELD(p_queue->q_validity,
					  SCSI_INIT_FUNC_QUEUES_SOC_NUM_OF_BLOCKS_LOG,
					  p_params->soc_num_of_blocks_log);
			}
		}
		SET_FIELD(p_queue->q_validity,
			  SCSI_INIT_FUNC_QUEUES_CMD_VALID, 1);
	} else {
		SET_FIELD(p_queue->q_validity,
			  SCSI_INIT_FUNC_QUEUES_RQ_VALID, 1);
	}
	p_ramrod->tcp_init.two_msl_timer =
		OSAL_CPU_TO_LE32(p_params->two_msl_timer);
	val = p_params->tx_sws_timer;
	p_ramrod->tcp_init.tx_sws_timer = OSAL_CPU_TO_LE16(val);
	p_ramrod->tcp_init.max_fin_rt = p_params->max_fin_rt;
	SET_FIELD(p_ramrod->iscsi_init_spe.params,
		  ISCSI_SPE_FUNC_INIT_MAX_SYN_RT, p_params->max_syn_rt);

	p_hwfn->p_iscsi_info->event_context = event_context;
	p_hwfn->p_iscsi_info->event_cb = async_event_cb;

	ecore_spq_register_async_cb(p_hwfn, PROTOCOLID_ISCSI,
				    ecore_iscsi_async_event);

	return ecore_spq_post(p_hwfn, p_ent, OSAL_NULL);
}

LNX_STATIC enum _ecore_status_t
ecore_sp_iscsi_conn_offload(struct ecore_hwfn *p_hwfn,
			    struct ecore_iscsi_conn *p_conn,
			    enum spq_mode comp_mode,
			    struct ecore_spq_comp_cb *p_comp_addr)
{
	struct iscsi_spe_conn_offload *p_ramrod = OSAL_NULL;
	struct tcp_offload_params_opt2 *p_tcp2 = OSAL_NULL;
	struct tcp_offload_params *p_tcp = OSAL_NULL;
	struct ecore_spq_entry *p_ent = OSAL_NULL;
	struct ecore_sp_init_data init_data;
	dma_addr_t r2tq_pbl_addr;
	dma_addr_t xhq_pbl_addr;
	dma_addr_t uhq_pbl_addr;
	u16 physical_q;
	enum _ecore_status_t rc = ECORE_SUCCESS;
	u32 dval;
	u16 wval;
	u8 ucval;
	u8 i;

	/* Get SPQ entry */
	OSAL_MEMSET(&init_data, 0, sizeof(init_data));
	init_data.cid = p_conn->icid;
	init_data.opaque_fid = p_hwfn->hw_info.opaque_fid;
	init_data.comp_mode = comp_mode;
	init_data.p_comp_data = p_comp_addr;

	rc = ecore_sp_init_request(p_hwfn, &p_ent,
				   ISCSI_RAMROD_CMD_ID_OFFLOAD_CONN,
				   PROTOCOLID_ISCSI, &init_data);
	if (rc != ECORE_SUCCESS)
		return rc;

	p_ramrod = &p_ent->ramrod.iscsi_conn_offload;

	/* Transmission PQ is the first of the PF */
	physical_q = ecore_get_cm_pq_idx(p_hwfn, PQ_FLAGS_OFLD);
	p_conn->physical_q0 = OSAL_CPU_TO_LE16(physical_q);
	p_ramrod->iscsi.physical_q0 = OSAL_CPU_TO_LE16(physical_q);

	/* iSCSI Pure-ACK PQ */
	physical_q = ecore_get_cm_pq_idx(p_hwfn, PQ_FLAGS_ACK);
	p_conn->physical_q1 = OSAL_CPU_TO_LE16(physical_q);
	p_ramrod->iscsi.physical_q1 = OSAL_CPU_TO_LE16(physical_q);

	p_ramrod->conn_id = OSAL_CPU_TO_LE16(p_conn->conn_id);

	DMA_REGPAIR_LE(p_ramrod->iscsi.sq_pbl_addr, p_conn->sq_pbl_addr);

	r2tq_pbl_addr = ecore_chain_get_pbl_phys(&p_conn->r2tq);
	DMA_REGPAIR_LE(p_ramrod->iscsi.r2tq_pbl_addr, r2tq_pbl_addr);

	xhq_pbl_addr = ecore_chain_get_pbl_phys(&p_conn->xhq);
	DMA_REGPAIR_LE(p_ramrod->iscsi.xhq_pbl_addr, xhq_pbl_addr);

	uhq_pbl_addr = ecore_chain_get_pbl_phys(&p_conn->uhq);
	DMA_REGPAIR_LE(p_ramrod->iscsi.uhq_pbl_addr, uhq_pbl_addr);

	p_ramrod->iscsi.initial_ack = OSAL_CPU_TO_LE32(p_conn->initial_ack);
	p_ramrod->iscsi.flags = p_conn->offl_flags;
	p_ramrod->iscsi.default_cq = p_conn->default_cq;
	p_ramrod->iscsi.stat_sn = OSAL_CPU_TO_LE32(p_conn->stat_sn);

	if (!GET_FIELD(p_ramrod->iscsi.flags,
			ISCSI_CONN_OFFLOAD_PARAMS_TCP_ON_CHIP_1B)) {
		p_tcp = &p_ramrod->tcp;
		ucval = p_conn->local_mac[1];
		((u8 *)(&p_tcp->local_mac_addr_hi))[0] = ucval;
		ucval = p_conn->local_mac[0];
		((u8 *)(&p_tcp->local_mac_addr_hi))[1] = ucval;
		ucval = p_conn->local_mac[3];
		((u8 *)(&p_tcp->local_mac_addr_mid))[0] = ucval;
		ucval = p_conn->local_mac[2];
		((u8 *)(&p_tcp->local_mac_addr_mid))[1] = ucval;
		ucval = p_conn->local_mac[5];
		((u8 *)(&p_tcp->local_mac_addr_lo))[0] = ucval;
		ucval = p_conn->local_mac[4];
		((u8 *)(&p_tcp->local_mac_addr_lo))[1] = ucval;
		ucval = p_conn->remote_mac[1];
		((u8 *)(&p_tcp->remote_mac_addr_hi))[0] = ucval;
		ucval = p_conn->remote_mac[0];
		((u8 *)(&p_tcp->remote_mac_addr_hi))[1] = ucval;
		ucval = p_conn->remote_mac[3];
		((u8 *)(&p_tcp->remote_mac_addr_mid))[0] = ucval;
		ucval = p_conn->remote_mac[2];
		((u8 *)(&p_tcp->remote_mac_addr_mid))[1] = ucval;
		ucval = p_conn->remote_mac[5];
		((u8 *)(&p_tcp->remote_mac_addr_lo))[0] = ucval;
		ucval = p_conn->remote_mac[4];
		((u8 *)(&p_tcp->remote_mac_addr_lo))[1] = ucval;

		p_tcp->vlan_id = OSAL_CPU_TO_LE16(p_conn->vlan_id);

		p_tcp->flags = OSAL_CPU_TO_LE16(p_conn->tcp_flags);
		p_tcp->ip_version = p_conn->ip_version;
		for (i = 0; i < 4; i++) {
			dval = p_conn->remote_ip[i];
			p_tcp->remote_ip[i] = OSAL_CPU_TO_LE32(dval);
			dval = p_conn->local_ip[i];
			p_tcp->local_ip[i] = OSAL_CPU_TO_LE32(dval);
		}
		p_tcp->ka_max_probe_cnt = p_conn->ka_max_probe_cnt;
		p_tcp->dup_ack_theshold = p_conn->dup_ack_theshold;

		p_tcp->rcv_next = OSAL_CPU_TO_LE32(p_conn->rcv_next);
		p_tcp->snd_una = OSAL_CPU_TO_LE32(p_conn->snd_una);
		p_tcp->snd_next = OSAL_CPU_TO_LE32(p_conn->snd_next);
		p_tcp->snd_max = OSAL_CPU_TO_LE32(p_conn->snd_max);
		p_tcp->snd_wnd = OSAL_CPU_TO_LE32(p_conn->snd_wnd);
		p_tcp->rcv_wnd = OSAL_CPU_TO_LE32(p_conn->rcv_wnd);
		p_tcp->snd_wl1 = OSAL_CPU_TO_LE32(p_conn->snd_wl1);
		p_tcp->cwnd = OSAL_CPU_TO_LE32(p_conn->cwnd);
		p_tcp->ss_thresh = OSAL_CPU_TO_LE32(p_conn->ss_thresh);
		p_tcp->srtt = OSAL_CPU_TO_LE16(p_conn->srtt);
		p_tcp->rtt_var = OSAL_CPU_TO_LE16(p_conn->rtt_var);
		p_tcp->ts_recent = OSAL_CPU_TO_LE32(p_conn->ts_recent);
		p_tcp->ts_recent_age = OSAL_CPU_TO_LE32(p_conn->ts_recent_age);
		p_tcp->total_rt = OSAL_CPU_TO_LE32(p_conn->total_rt);
		dval = p_conn->ka_timeout_delta;
		p_tcp->ka_timeout_delta = OSAL_CPU_TO_LE32(dval);
		dval = p_conn->rt_timeout_delta;
		p_tcp->rt_timeout_delta = OSAL_CPU_TO_LE32(dval);
		p_tcp->dup_ack_cnt = p_conn->dup_ack_cnt;
		p_tcp->snd_wnd_probe_cnt = p_conn->snd_wnd_probe_cnt;
		p_tcp->ka_probe_cnt = p_conn->ka_probe_cnt;
		p_tcp->rt_cnt = p_conn->rt_cnt;
		p_tcp->flow_label = OSAL_CPU_TO_LE32(p_conn->flow_label);
		p_tcp->ka_timeout = OSAL_CPU_TO_LE32(p_conn->ka_timeout);
		p_tcp->ka_interval = OSAL_CPU_TO_LE32(p_conn->ka_interval);
		p_tcp->max_rt_time = OSAL_CPU_TO_LE32(p_conn->max_rt_time);
		dval = p_conn->initial_rcv_wnd;
		p_tcp->initial_rcv_wnd = OSAL_CPU_TO_LE32(dval);
		p_tcp->ttl = p_conn->ttl;
		p_tcp->tos_or_tc = p_conn->tos_or_tc;
		p_tcp->remote_port = OSAL_CPU_TO_LE16(p_conn->remote_port);
		p_tcp->local_port = OSAL_CPU_TO_LE16(p_conn->local_port);
		p_tcp->mss = OSAL_CPU_TO_LE16(p_conn->mss);
		p_tcp->snd_wnd_scale = p_conn->snd_wnd_scale;
		p_tcp->rcv_wnd_scale = p_conn->rcv_wnd_scale;
		wval = p_conn->da_timeout_value;
		p_tcp->da_timeout_value = OSAL_CPU_TO_LE16(wval);
		p_tcp->ack_frequency = p_conn->ack_frequency;
		p_tcp->connect_mode = p_conn->connect_mode;
	} else {
		p_tcp2 = &((struct iscsi_spe_conn_offload_option2 *)p_ramrod)->tcp;
		ucval = p_conn->local_mac[1];
		((u8 *)(&p_tcp2->local_mac_addr_hi))[0] = ucval;
		ucval = p_conn->local_mac[0];
		((u8 *)(&p_tcp2->local_mac_addr_hi))[1] = ucval;
		ucval = p_conn->local_mac[3];
		((u8 *)(&p_tcp2->local_mac_addr_mid))[0] = ucval;
		ucval = p_conn->local_mac[2];
		((u8 *)(&p_tcp2->local_mac_addr_mid))[1] = ucval;
		ucval = p_conn->local_mac[5];
		((u8 *)(&p_tcp2->local_mac_addr_lo))[0] = ucval;
		ucval = p_conn->local_mac[4];
		((u8 *)(&p_tcp2->local_mac_addr_lo))[1] = ucval;

		ucval = p_conn->remote_mac[1];
		((u8 *)(&p_tcp2->remote_mac_addr_hi))[0] = ucval;
		ucval = p_conn->remote_mac[0];
		((u8 *)(&p_tcp2->remote_mac_addr_hi))[1] = ucval;
		ucval = p_conn->remote_mac[3];
		((u8 *)(&p_tcp2->remote_mac_addr_mid))[0] = ucval;
		ucval = p_conn->remote_mac[2];
		((u8 *)(&p_tcp2->remote_mac_addr_mid))[1] = ucval;
		ucval = p_conn->remote_mac[5];
		((u8 *)(&p_tcp2->remote_mac_addr_lo))[0] = ucval;
		ucval = p_conn->remote_mac[4];
		((u8 *)(&p_tcp2->remote_mac_addr_lo))[1] = ucval;

		p_tcp2->vlan_id = OSAL_CPU_TO_LE16(p_conn->vlan_id);
		p_tcp2->flags = OSAL_CPU_TO_LE16(p_conn->tcp_flags);

		p_tcp2->ip_version = p_conn->ip_version;
		for (i = 0; i < 4; i++) {
			dval = p_conn->remote_ip[i];
			p_tcp2->remote_ip[i] = OSAL_CPU_TO_LE32(dval);
			dval = p_conn->local_ip[i];
			p_tcp2->local_ip[i] = OSAL_CPU_TO_LE32(dval);
		}

		p_tcp2->flow_label = OSAL_CPU_TO_LE32(p_conn->flow_label);
		p_tcp2->ttl = p_conn->ttl;
		p_tcp2->tos_or_tc = p_conn->tos_or_tc;
		p_tcp2->remote_port = OSAL_CPU_TO_LE16(p_conn->remote_port);
		p_tcp2->local_port = OSAL_CPU_TO_LE16(p_conn->local_port);
		p_tcp2->mss = OSAL_CPU_TO_LE16(p_conn->mss);
		p_tcp2->rcv_wnd_scale = p_conn->rcv_wnd_scale;
		p_tcp2->connect_mode = p_conn->connect_mode;
		wval = p_conn->syn_ip_payload_length;
		p_tcp2->syn_ip_payload_length = OSAL_CPU_TO_LE16(wval);
		p_tcp2->syn_phy_addr_lo = DMA_LO_LE(p_conn->syn_phy_addr);
		p_tcp2->syn_phy_addr_hi = DMA_HI_LE(p_conn->syn_phy_addr);
		p_tcp2->cwnd = OSAL_CPU_TO_LE32(p_conn->cwnd);
		p_tcp2->ka_max_probe_cnt = p_conn->ka_probe_cnt;
		p_tcp2->ka_timeout = OSAL_CPU_TO_LE32(p_conn->ka_timeout);
		p_tcp2->max_rt_time = OSAL_CPU_TO_LE32(p_conn->max_rt_time);
		p_tcp2->ka_interval = OSAL_CPU_TO_LE32(p_conn->ka_interval);
	}

	return ecore_spq_post(p_hwfn, p_ent, OSAL_NULL);
}

LNX_STATIC enum _ecore_status_t
ecore_sp_iscsi_conn_update(struct ecore_hwfn *p_hwfn,
			   struct ecore_iscsi_conn *p_conn,
			   enum spq_mode comp_mode,
			   struct ecore_spq_comp_cb *p_comp_addr)
{
	struct iscsi_conn_update_ramrod_params *p_ramrod = OSAL_NULL;
	struct dif_on_immediate_params *p_def_dif_conf = OSAL_NULL;
	struct ecore_spq_entry *p_ent = OSAL_NULL;
	struct ecore_sp_init_data init_data;
	enum _ecore_status_t rc = ECORE_NOTIMPL;
	u32 dval;
	u16 val;

	/* Get SPQ entry */
	OSAL_MEMSET(&init_data, 0, sizeof(init_data));
	init_data.cid = p_conn->icid;
	init_data.opaque_fid = p_hwfn->hw_info.opaque_fid;
	init_data.comp_mode = comp_mode;
	init_data.p_comp_data = p_comp_addr;

	rc = ecore_sp_init_request(p_hwfn, &p_ent,
				   ISCSI_RAMROD_CMD_ID_UPDATE_CONN,
				   PROTOCOLID_ISCSI, &init_data);
	if (rc != ECORE_SUCCESS)
		return rc;

	p_ramrod = &p_ent->ramrod.iscsi_conn_update;

	p_ramrod->conn_id = OSAL_CPU_TO_LE16(p_conn->conn_id);
	p_ramrod->flags = p_conn->update_flag;
	p_ramrod->max_seq_size = OSAL_CPU_TO_LE32(p_conn->max_seq_size);
	dval = p_conn->max_recv_pdu_length;
	p_ramrod->max_recv_pdu_length = OSAL_CPU_TO_LE32(dval);
	dval = p_conn->max_send_pdu_length;
	p_ramrod->max_send_pdu_length = OSAL_CPU_TO_LE32(dval);
	dval = p_conn->first_seq_length;
	p_ramrod->first_seq_length = OSAL_CPU_TO_LE32(dval);
	p_ramrod->exp_stat_sn = OSAL_CPU_TO_LE32(p_conn->exp_stat_sn);

	if (p_conn->dif_on_immediate == ECORE_ISCSI_CONN_DIF_ON_IMM_DEFAULT) {
		SET_FIELD(p_ramrod->flags,
			  ISCSI_CONN_UPDATE_RAMROD_PARAMS_DIF_ON_IMM_EN, 1);
		SET_FIELD(p_ramrod->flags,
			  ISCSI_CONN_UPDATE_RAMROD_PARAMS_LUN_MAPPER_EN, 0);
		p_def_dif_conf = &p_ramrod->dif_on_imme_params.def_dif_conf;
		dval = p_conn->initial_ref_tag;
		p_def_dif_conf->initial_ref_tag = OSAL_CPU_TO_LE32(dval);
		val = p_conn->application_tag;
		p_def_dif_conf->application_tag = OSAL_CPU_TO_LE16(val);
		val = p_conn->application_tag_mask;
		p_def_dif_conf->application_tag_mask = OSAL_CPU_TO_LE16(val);

		val = 0;
		SET_FIELD(val, DIF_ON_IMMEDIATE_PARAMS_VALIDATE_GUARD,
			  p_conn->validate_guard);
		SET_FIELD(val, DIF_ON_IMMEDIATE_PARAMS_VALIDATE_APP_TAG,
			  p_conn->validate_app_tag);
		SET_FIELD(val, DIF_ON_IMMEDIATE_PARAMS_VALIDATE_REF_TAG,
			  p_conn->validate_ref_tag);
		SET_FIELD(val, DIF_ON_IMMEDIATE_PARAMS_FORWARD_GUARD,
			  p_conn->forward_guard);
		SET_FIELD(val, DIF_ON_IMMEDIATE_PARAMS_FORWARD_APP_TAG,
			  p_conn->forward_app_tag);
		SET_FIELD(val, DIF_ON_IMMEDIATE_PARAMS_FORWARD_REF_TAG,
			  p_conn->forward_ref_tag);
		SET_FIELD(val, DIF_ON_IMMEDIATE_PARAMS_INTERVAL_SIZE,
			  p_conn->interval_size);
		SET_FIELD(val, DIF_ON_IMMEDIATE_PARAMS_NETWORK_INTERFACE,
			  p_conn->network_interface);
		SET_FIELD(val, DIF_ON_IMMEDIATE_PARAMS_HOST_INTERFACE,
			  p_conn->host_interface);
		SET_FIELD(val, DIF_ON_IMMEDIATE_PARAMS_REF_TAG_MASK,
			  p_conn->ref_tag_mask);
		SET_FIELD(val,
			  DIF_ON_IMMEDIATE_PARAMS_FORWARD_APP_TAG_WITH_MASK,
			  p_conn->forward_app_tag_with_mask);
		SET_FIELD(val,
			  DIF_ON_IMMEDIATE_PARAMS_FORWARD_REF_TAG_WITH_MASK,
			  p_conn->forward_ref_tag_with_mask);
		p_def_dif_conf->flags1 = OSAL_CPU_TO_LE16(val);

		SET_FIELD(p_def_dif_conf->flags0,
			  DIF_ON_IMMEDIATE_PARAMS_IGNORE_APP_TAG,
			  p_conn->ignore_app_tag);
		SET_FIELD(p_def_dif_conf->flags0,
			  DIF_ON_IMMEDIATE_PARAMS_INITIAL_REF_TAG_IS_VALID,
			  p_conn->initial_ref_tag_is_valid);
		SET_FIELD(p_def_dif_conf->flags0,
			  DIF_ON_IMMEDIATE_PARAMS_HOST_GUARD_TYPE,
			  p_conn->host_guard_type);
		SET_FIELD(p_def_dif_conf->flags0,
			  DIF_ON_IMMEDIATE_PARAMS_PROTECTION_TYPE,
			  p_conn->protection_type);
		SET_FIELD(p_def_dif_conf->flags0,
			  DIF_ON_IMMEDIATE_PARAMS_CRC_SEED,
			  p_conn->crc_seed);
		SET_FIELD(p_def_dif_conf->flags0,
			  DIF_ON_IMMEDIATE_PARAMS_KEEP_REF_TAG_CONST,
			  p_conn->keep_ref_tag_const);

	} else if (p_conn->dif_on_immediate ==
			ECORE_ISCSI_CONN_DIF_ON_IMM_LUN_MAPPER) {
		SET_FIELD(p_ramrod->flags,
			  ISCSI_CONN_UPDATE_RAMROD_PARAMS_DIF_ON_IMM_EN, 1);
		SET_FIELD(p_ramrod->flags,
			  ISCSI_CONN_UPDATE_RAMROD_PARAMS_LUN_MAPPER_EN, 1);
		DMA_REGPAIR_LE(p_ramrod->dif_on_imme_params.lun_mapper_address.lun_mapper_addr,
			       p_conn->lun_mapper_phys_addr);
	}

	return ecore_spq_post(p_hwfn, p_ent, OSAL_NULL);
}

LNX_STATIC enum _ecore_status_t
ecore_sp_iscsi_mac_update(struct ecore_hwfn *p_hwfn,
			  struct ecore_iscsi_conn *p_conn,
			  enum spq_mode comp_mode,
			  struct ecore_spq_comp_cb *p_comp_addr)
{
	struct iscsi_spe_conn_mac_update *p_ramrod = OSAL_NULL;
	struct ecore_spq_entry *p_ent = OSAL_NULL;
	enum _ecore_status_t rc = ECORE_NOTIMPL;
	struct ecore_sp_init_data init_data;
	u8 ucval;

	/* Get SPQ entry */
	OSAL_MEMSET(&init_data, 0, sizeof(init_data));
	init_data.cid = p_conn->icid;
	init_data.opaque_fid = p_hwfn->hw_info.opaque_fid;
	init_data.comp_mode = comp_mode;
	init_data.p_comp_data = p_comp_addr;

	rc = ecore_sp_init_request(p_hwfn, &p_ent,
				   ISCSI_RAMROD_CMD_ID_MAC_UPDATE,
				   PROTOCOLID_ISCSI, &init_data);
	if (rc != ECORE_SUCCESS)
		return rc;

	p_ramrod = &p_ent->ramrod.iscsi_conn_mac_update;

	p_ramrod->conn_id = OSAL_CPU_TO_LE16(p_conn->conn_id);
	ucval = p_conn->remote_mac[1];
	((u8 *)(&p_ramrod->remote_mac_addr_hi))[0] = ucval;
	ucval = p_conn->remote_mac[0];
	((u8 *)(&p_ramrod->remote_mac_addr_hi))[1] = ucval;
	ucval = p_conn->remote_mac[3];
	((u8 *)(&p_ramrod->remote_mac_addr_mid))[0] = ucval;
	ucval = p_conn->remote_mac[2];
	((u8 *)(&p_ramrod->remote_mac_addr_mid))[1] = ucval;
	ucval = p_conn->remote_mac[5];
	((u8 *)(&p_ramrod->remote_mac_addr_lo))[0] = ucval;
	ucval = p_conn->remote_mac[4];
	((u8 *)(&p_ramrod->remote_mac_addr_lo))[1] = ucval;

	rc =  ecore_spq_post(p_hwfn, p_ent, OSAL_NULL);

	return rc;
}

#ifndef __EXTRACT__LINUX__THROW__
LNX_STATIC enum _ecore_status_t
ecore_sp_iscsi_stats_tcp_update(struct ecore_hwfn *p_hwfn,
				struct ecore_iscsi_conn *p_conn,
				bool reset,
				enum spq_mode comp_mode,
				struct ecore_spq_comp_cb *p_comp_addr)
{
	struct iscsi_spe_conn_statistics *p_ramrod = OSAL_NULL;
	struct ecore_spq_entry *p_ent = OSAL_NULL;
	enum _ecore_status_t rc = ECORE_NOTIMPL;
	struct ecore_sp_init_data init_data;

	/* Get SPQ entry */
	OSAL_MEMSET(&init_data, 0, sizeof(init_data));
	init_data.cid = p_conn->icid;
	init_data.opaque_fid = p_hwfn->hw_info.opaque_fid;
	init_data.comp_mode = comp_mode;
	init_data.p_comp_data = p_comp_addr;

	rc = ecore_sp_init_request(p_hwfn, &p_ent,
				   ISCSI_RAMROD_CMD_ID_CONN_STATS,
				   PROTOCOLID_ISCSI, &init_data);
	if (rc != ECORE_SUCCESS)
		return rc;

	p_ramrod = &p_ent->ramrod.iscsi_conn_statistics;

	p_ramrod->conn_id = OSAL_CPU_TO_LE16(p_conn->conn_id);
	p_ramrod->reset_stats = reset;
	DMA_REGPAIR_LE(p_ramrod->stats_cnts_addr,
		       p_conn->conn_stats_params_phys_addr);

	rc =  ecore_spq_post(p_hwfn, p_ent, OSAL_NULL);

	return rc;
}
#endif

LNX_STATIC enum _ecore_status_t
ecore_sp_iscsi_conn_terminate(struct ecore_hwfn *p_hwfn,
			      struct ecore_iscsi_conn *p_conn,
			      enum spq_mode comp_mode,
			      struct ecore_spq_comp_cb *p_comp_addr)
{
	struct iscsi_spe_conn_termination *p_ramrod = OSAL_NULL;
	struct ecore_spq_entry *p_ent = OSAL_NULL;
	struct ecore_sp_init_data init_data;
	enum _ecore_status_t rc = ECORE_NOTIMPL;

	/* Get SPQ entry */
	OSAL_MEMSET(&init_data, 0, sizeof(init_data));
	init_data.cid = p_conn->icid;
	init_data.opaque_fid = p_hwfn->hw_info.opaque_fid;
	init_data.comp_mode = comp_mode;
	init_data.p_comp_data = p_comp_addr;

	rc = ecore_sp_init_request(p_hwfn, &p_ent,
				   ISCSI_RAMROD_CMD_ID_TERMINATION_CONN,
				   PROTOCOLID_ISCSI, &init_data);
	if (rc != ECORE_SUCCESS)
		return rc;

	p_ramrod = &p_ent->ramrod.iscsi_conn_terminate;

	p_ramrod->conn_id = OSAL_CPU_TO_LE16(p_conn->conn_id);
	p_ramrod->abortive = p_conn->abortive_dsconnect;

	DMA_REGPAIR_LE(p_ramrod->query_params_addr,
		       p_conn->tcp_upload_params_phys_addr);
	DMA_REGPAIR_LE(p_ramrod->queue_cnts_addr,
		       p_conn->queue_cnts_phys_addr);

	return ecore_spq_post(p_hwfn, p_ent, OSAL_NULL);
}

LNX_STATIC enum _ecore_status_t
ecore_sp_iscsi_conn_clear_sq(struct ecore_hwfn *p_hwfn,
			     struct ecore_iscsi_conn *p_conn,
			     enum spq_mode comp_mode,
			     struct ecore_spq_comp_cb *p_comp_addr)
{
	struct ecore_spq_entry *p_ent = OSAL_NULL;
	struct ecore_sp_init_data init_data;
	enum _ecore_status_t rc = ECORE_NOTIMPL;

	/* Get SPQ entry */
	OSAL_MEMSET(&init_data, 0, sizeof(init_data));
	init_data.cid = p_conn->icid;
	init_data.opaque_fid = p_hwfn->hw_info.opaque_fid;
	init_data.comp_mode = comp_mode;
	init_data.p_comp_data = p_comp_addr;

	rc = ecore_sp_init_request(p_hwfn, &p_ent,
				   ISCSI_RAMROD_CMD_ID_CLEAR_SQ,
				   PROTOCOLID_ISCSI, &init_data);
	if (rc != ECORE_SUCCESS)
		return rc;

	return ecore_spq_post(p_hwfn, p_ent, OSAL_NULL);
}

LNX_STATIC enum _ecore_status_t
ecore_sp_iscsi_func_stop(struct ecore_hwfn *p_hwfn,
			 enum spq_mode comp_mode,
			 struct ecore_spq_comp_cb *p_comp_addr)
{
	struct ecore_spq_entry *p_ent = OSAL_NULL;
	struct ecore_sp_init_data init_data;
	enum _ecore_status_t rc = ECORE_SUCCESS;

	/* Get SPQ entry */
	OSAL_MEMSET(&init_data, 0, sizeof(init_data));
	init_data.cid = ecore_spq_get_cid(p_hwfn);
	init_data.opaque_fid = p_hwfn->hw_info.opaque_fid;
	init_data.comp_mode = comp_mode;
	init_data.p_comp_data = p_comp_addr;

	rc = ecore_sp_init_request(p_hwfn, &p_ent,
				   ISCSI_RAMROD_CMD_ID_DESTROY_FUNC,
				   PROTOCOLID_ISCSI, &init_data);
	if (rc != ECORE_SUCCESS)
		return rc;

	rc = ecore_spq_post(p_hwfn, p_ent, OSAL_NULL);

	ecore_spq_unregister_async_cb(p_hwfn, PROTOCOLID_ISCSI);
	return rc;
}

LNX_STATIC void OSAL_IOMEM *
ecore_iscsi_get_db_addr(struct ecore_hwfn *p_hwfn, u32 cid)
{
	return (u8 OSAL_IOMEM *)p_hwfn->doorbells +
				DB_ADDR(cid, DQ_DEMS_LEGACY);
}

#ifndef __EXTRACT__LINUX__THROW__
void OSAL_IOMEM *ecore_iscsi_get_global_cmdq_cons(struct ecore_hwfn *p_hwfn,
						  u8 relative_q_id)
{
	u8 hw_q_id = (u8)p_hwfn->hw_info.resc_start[ECORE_CMDQS_CQS] +
			 relative_q_id;

	return (u8 OSAL_IOMEM *)p_hwfn->regview +
				GTT_BAR0_MAP_REG_TSDM_RAM +
				TSTORM_SCSI_CMDQ_CONS_OFFSET(hw_q_id);
}
#endif

LNX_STATIC void OSAL_IOMEM *
ecore_iscsi_get_primary_bdq_prod(struct ecore_hwfn *p_hwfn, u8 bdq_id)
{
	if (RESC_NUM(p_hwfn, ECORE_BDQ)) {
		return (u8 OSAL_IOMEM *)p_hwfn->regview +
			GTT_BAR0_MAP_REG_MSDM_RAM +
			MSTORM_SCSI_BDQ_EXT_PROD_OFFSET(
			RESC_START(p_hwfn, ECORE_BDQ), bdq_id);
	} else {
		DP_NOTICE(p_hwfn, true, "BDQ is not allocated!\n");
		return OSAL_NULL;
	}
}

LNX_STATIC void OSAL_IOMEM *
ecore_iscsi_get_secondary_bdq_prod(struct ecore_hwfn *p_hwfn, u8 bdq_id)
{
	if (RESC_NUM(p_hwfn, ECORE_BDQ)) {
		return (u8 OSAL_IOMEM *)p_hwfn->regview +
			GTT_BAR0_MAP_REG_TSDM_RAM +
			TSTORM_SCSI_BDQ_EXT_PROD_OFFSET(
			RESC_START(p_hwfn, ECORE_BDQ), bdq_id);
	} else {
		DP_NOTICE(p_hwfn, true, "BDQ is not allocated!\n");
		return OSAL_NULL;
	}
}

LNX_STATIC enum _ecore_status_t
ecore_iscsi_setup_connection(struct ecore_iscsi_conn *p_conn)
{
	if (!p_conn->queue_cnts_virt_addr)
		goto nomem;
	OSAL_MEM_ZERO(p_conn->queue_cnts_virt_addr,
		      sizeof(*p_conn->queue_cnts_virt_addr));

	if (!p_conn->tcp_upload_params_virt_addr)
		goto nomem;
	OSAL_MEM_ZERO(p_conn->tcp_upload_params_virt_addr,
		      sizeof(*p_conn->tcp_upload_params_virt_addr));

	if (!p_conn->r2tq.p_virt_addr)
		goto nomem;
	ecore_chain_pbl_zero_mem(&p_conn->r2tq);

	if (!p_conn->uhq.p_virt_addr)
		goto nomem;
	ecore_chain_pbl_zero_mem(&p_conn->uhq);

	if (!p_conn->xhq.p_virt_addr)
		goto nomem;
	ecore_chain_pbl_zero_mem(&p_conn->xhq);

	return ECORE_SUCCESS;
nomem:
	return ECORE_NOMEM;
}

static enum _ecore_status_t
ecore_iscsi_allocate_connection(struct ecore_hwfn *p_hwfn,
				struct ecore_iscsi_conn	**p_out_conn)
{
	u16 uhq_num_elements = 0, xhq_num_elements = 0, r2tq_num_elements = 0;
	struct scsi_terminate_extra_params *p_q_cnts = OSAL_NULL;
	struct ecore_iscsi_pf_params *p_params = OSAL_NULL;
	struct tcp_upload_params *p_tcp = OSAL_NULL;
	struct ecore_iscsi_conn	*p_conn = OSAL_NULL;
	struct iscsi_conn_stats_params *p_stats = OSAL_NULL;
	struct ecore_chain_params chain_params;
	enum _ecore_status_t rc = ECORE_SUCCESS;

	/* Try finding a free connection that can be used */
	OSAL_SPIN_LOCK(&p_hwfn->p_iscsi_info->lock);
	if (!OSAL_LIST_IS_EMPTY(&p_hwfn->p_iscsi_info->free_list))
		p_conn = OSAL_LIST_FIRST_ENTRY(&p_hwfn->p_iscsi_info->free_list,
					       struct ecore_iscsi_conn,
					       list_entry);
	if (p_conn) {
		OSAL_LIST_REMOVE_ENTRY(&p_conn->list_entry,
				       &p_hwfn->p_iscsi_info->free_list);
		OSAL_SPIN_UNLOCK(&p_hwfn->p_iscsi_info->lock);
		*p_out_conn = p_conn;
		return ECORE_SUCCESS;
	}
	OSAL_SPIN_UNLOCK(&p_hwfn->p_iscsi_info->lock);

	/* Need to allocate a new connection */
	p_params = &p_hwfn->pf_params.iscsi_pf_params;

	p_conn = OSAL_ZALLOC(p_hwfn->p_dev, GFP_KERNEL,
			     sizeof(*p_conn));
	if (!p_conn)
		return ECORE_NOMEM;

	p_q_cnts = OSAL_DMA_ALLOC_COHERENT(p_hwfn->p_dev,
					   &p_conn->queue_cnts_phys_addr,
					   sizeof(*p_q_cnts));
	if (!p_q_cnts)
		goto nomem_queue_cnts_param;
	p_conn->queue_cnts_virt_addr = p_q_cnts;

	p_tcp = OSAL_DMA_ALLOC_COHERENT(p_hwfn->p_dev,
					&p_conn->tcp_upload_params_phys_addr,
					sizeof(*p_tcp));
	if (!p_tcp)
		goto nomem_upload_param;
	p_conn->tcp_upload_params_virt_addr = p_tcp;

	p_stats = OSAL_DMA_ALLOC_COHERENT(p_hwfn->p_dev,
					 &p_conn->conn_stats_params_phys_addr,
					 sizeof(*p_stats));
	if (!p_stats)
		goto nomem_stats_param;
	p_conn->conn_stats_params_virt_addr = p_stats;

	r2tq_num_elements = p_params->num_sq_pages_in_ring *
			    ECORE_CHAIN_PAGE_SIZE / sizeof(struct iscsi_wqe);
	ecore_chain_params_init(&chain_params,
				ECORE_CHAIN_USE_TO_CONSUME_PRODUCE,
				ECORE_CHAIN_MODE_PBL,
				ECORE_CHAIN_CNT_TYPE_U16,
				r2tq_num_elements,
				sizeof(struct iscsi_wqe));
	rc = ecore_chain_alloc(p_hwfn->p_dev, &p_conn->r2tq, &chain_params);
	if (rc != ECORE_SUCCESS)
		goto nomem_r2tq;

	uhq_num_elements = p_params->num_uhq_pages_in_ring *
			   ECORE_CHAIN_PAGE_SIZE / sizeof(struct iscsi_uhqe);
	ecore_chain_params_init(&chain_params,
				ECORE_CHAIN_USE_TO_CONSUME_PRODUCE,
				ECORE_CHAIN_MODE_PBL,
				ECORE_CHAIN_CNT_TYPE_U16,
				uhq_num_elements,
				sizeof(struct iscsi_uhqe));
	rc = ecore_chain_alloc(p_hwfn->p_dev, &p_conn->uhq, &chain_params);
	if (rc != ECORE_SUCCESS)
		goto nomem_uhq;

	xhq_num_elements =  uhq_num_elements;
	ecore_chain_params_init(&chain_params,
				ECORE_CHAIN_USE_TO_CONSUME_PRODUCE,
				ECORE_CHAIN_MODE_PBL,
				ECORE_CHAIN_CNT_TYPE_U16,
				xhq_num_elements,
				sizeof(struct iscsi_xhqe));
	rc = ecore_chain_alloc(p_hwfn->p_dev, &p_conn->xhq, &chain_params);
	if (rc != ECORE_SUCCESS)
		goto nomem;

	p_conn->free_on_delete = true;
	*p_out_conn = p_conn;
	return ECORE_SUCCESS;

nomem:
	ecore_chain_free(p_hwfn->p_dev, &p_conn->uhq);
nomem_uhq:
	ecore_chain_free(p_hwfn->p_dev, &p_conn->r2tq);
nomem_r2tq:
	OSAL_DMA_FREE_COHERENT(p_hwfn->p_dev,
			       p_conn->conn_stats_params_virt_addr,
			       p_conn->conn_stats_params_phys_addr,
			       sizeof(struct iscsi_conn_stats_params));
nomem_stats_param:
	OSAL_DMA_FREE_COHERENT(p_hwfn->p_dev,
			       p_conn->tcp_upload_params_virt_addr,
			       p_conn->tcp_upload_params_phys_addr,
			       sizeof(struct tcp_upload_params));
nomem_upload_param:
	OSAL_DMA_FREE_COHERENT(p_hwfn->p_dev,
			       p_conn->queue_cnts_virt_addr,
			       p_conn->queue_cnts_phys_addr,
			       sizeof(struct scsi_terminate_extra_params));
nomem_queue_cnts_param:
	OSAL_FREE(p_hwfn->p_dev, p_conn);

	return ECORE_NOMEM;
}

LNX_STATIC enum _ecore_status_t
ecore_iscsi_acquire_connection(struct ecore_hwfn *p_hwfn,
			       struct ecore_iscsi_conn *p_in_conn,
			       struct ecore_iscsi_conn **p_out_conn)
{
	struct ecore_iscsi_conn *p_conn = OSAL_NULL;
	enum _ecore_status_t rc = ECORE_SUCCESS;
	u32 icid;

	OSAL_SPIN_LOCK(&p_hwfn->p_iscsi_info->lock);
	rc = ecore_cxt_acquire_cid(p_hwfn, PROTOCOLID_ISCSI, &icid);
	OSAL_SPIN_UNLOCK(&p_hwfn->p_iscsi_info->lock);
	if (rc != ECORE_SUCCESS)
		return rc;

	/* Use input connection or allocate a new one */
	if (p_in_conn)
		p_conn = p_in_conn;
	else
		rc = ecore_iscsi_allocate_connection(p_hwfn, &p_conn);

	if (rc == ECORE_SUCCESS)
		rc = ecore_iscsi_setup_connection(p_conn);

	if (rc != ECORE_SUCCESS) {
		/* TODO - Aren't we possibly leaking a connection here? */
		OSAL_SPIN_LOCK(&p_hwfn->p_iscsi_info->lock);
		ecore_cxt_release_cid(p_hwfn, icid);
		OSAL_SPIN_UNLOCK(&p_hwfn->p_iscsi_info->lock);
		return rc;
	}

	p_conn->icid = icid;
	p_conn->conn_id = (u16)icid;
	p_conn->fw_cid = (p_hwfn->hw_info.opaque_fid << 16) | icid;

	*p_out_conn = p_conn;

	return rc;
}

#ifndef __EXTRACT__LINUX__THROW__
enum _ecore_status_t
ecore_iscsi_offload_connection(struct ecore_hwfn *p_hwfn,
			       struct ecore_iscsi_conn *p_conn)
{
	return ecore_sp_iscsi_conn_offload(p_hwfn, p_conn,
					   ECORE_SPQ_MODE_EBLOCK, OSAL_NULL);
}


enum _ecore_status_t
ecore_iscsi_update_connection(struct ecore_hwfn *p_hwfn,
			      struct ecore_iscsi_conn *p_conn)
{
	return ecore_sp_iscsi_conn_update(p_hwfn, p_conn,
					  ECORE_SPQ_MODE_EBLOCK, OSAL_NULL);
}

enum _ecore_status_t
ecore_iscsi_update_remote_mac(struct ecore_hwfn *p_hwfn,
			      struct ecore_iscsi_conn *p_conn)
{
	return ecore_sp_iscsi_mac_update(p_hwfn, p_conn,
					 ECORE_SPQ_MODE_EBLOCK, OSAL_NULL);
}

enum _ecore_status_t
ecore_iscsi_get_tcp_stats(struct ecore_hwfn *p_hwfn,
			  struct ecore_iscsi_conn *p_conn,
			  struct ecore_iscsi_tcp_stats *p_stats,
			  u8 reset)
{
	struct iscsi_conn_stats_params *p_conn_stats;
	enum _ecore_status_t rc;
	u32 hi, lo;

	rc = ecore_sp_iscsi_stats_tcp_update(p_hwfn, p_conn, !!reset,
					     ECORE_SPQ_MODE_EBLOCK, OSAL_NULL);
	if (rc != ECORE_SUCCESS)
		return rc;

	if (p_stats == OSAL_NULL)
		goto out;

	p_conn_stats = p_conn->conn_stats_params_virt_addr;
	OSAL_MEM_ZERO(p_stats, sizeof(*p_stats));

	hi = OSAL_LE32_TO_CPU(p_conn_stats->iscsi_tcp_tx_packets_cnt.hi);
	lo = OSAL_LE32_TO_CPU(p_conn_stats->iscsi_tcp_tx_packets_cnt.lo);
	p_stats->iscsi_tcp_tx_packets_cnt = HILO_64(hi, lo);

	hi = OSAL_LE32_TO_CPU(p_conn_stats->iscsi_tcp_tx_bytes_cnt.hi);
	lo = OSAL_LE32_TO_CPU(p_conn_stats->iscsi_tcp_tx_bytes_cnt.lo);
	p_stats->iscsi_tcp_tx_bytes_cnt = HILO_64(hi, lo);

	hi = OSAL_LE32_TO_CPU(p_conn_stats->iscsi_tcp_tx_rxmit_cnt.hi);
	lo = OSAL_LE32_TO_CPU(p_conn_stats->iscsi_tcp_tx_rxmit_cnt.lo);
	p_stats->iscsi_tcp_tx_rxmit_cnt = HILO_64(hi, lo);

	hi = OSAL_LE32_TO_CPU(p_conn_stats->iscsi_tcp_rx_packets_cnt.hi);
	lo = OSAL_LE32_TO_CPU(p_conn_stats->iscsi_tcp_rx_packets_cnt.lo);
	p_stats->iscsi_tcp_rx_packets_cnt = HILO_64(hi, lo);

	hi = OSAL_LE32_TO_CPU(p_conn_stats->iscsi_tcp_rx_bytes_cnt.hi);
	lo = OSAL_LE32_TO_CPU(p_conn_stats->iscsi_tcp_rx_bytes_cnt.lo);
	p_stats->iscsi_tcp_rx_bytes_cnt = HILO_64(hi, lo);

	hi = OSAL_LE32_TO_CPU(p_conn_stats->iscsi_tcp_rx_dup_ack_cnt.hi);
	lo = OSAL_LE32_TO_CPU(p_conn_stats->iscsi_tcp_rx_dup_ack_cnt.lo);
	p_stats->iscsi_tcp_rx_dup_ack_cnt = HILO_64(hi, lo);

	lo = OSAL_LE32_TO_CPU(p_conn_stats->iscsi_tcp_rx_chksum_err_cnt);
	p_stats->iscsi_tcp_rx_chksum_err_cnt = lo;
out:
	return ECORE_SUCCESS;
}

enum _ecore_status_t
ecore_iscsi_clear_connection_sq(struct ecore_hwfn *p_hwfn,
				struct ecore_iscsi_conn *p_conn)
{
	return ecore_sp_iscsi_conn_clear_sq(p_hwfn, p_conn,
					    ECORE_SPQ_MODE_EBLOCK,
					    OSAL_NULL);
}

enum _ecore_status_t
ecore_iscsi_terminate_connection(struct ecore_hwfn *p_hwfn,
				 struct ecore_iscsi_conn *p_conn)
{
	return ecore_sp_iscsi_conn_terminate(p_hwfn, p_conn,
					     ECORE_SPQ_MODE_EBLOCK,
					     OSAL_NULL);
}
#endif

LNX_STATIC void
ecore_iscsi_release_connection(struct ecore_hwfn *p_hwfn,
			       struct ecore_iscsi_conn *p_conn)
{
	OSAL_SPIN_LOCK(&p_hwfn->p_iscsi_info->lock);
	OSAL_LIST_PUSH_TAIL(&p_conn->list_entry,
			    &p_hwfn->p_iscsi_info->free_list);
	ecore_cxt_release_cid(p_hwfn, p_conn->icid);
	OSAL_SPIN_UNLOCK(&p_hwfn->p_iscsi_info->lock);
}

LNX_STATIC void
ecore_iscsi_free_connection(struct ecore_hwfn *p_hwfn,
			    struct ecore_iscsi_conn *p_conn)
{
	ecore_chain_free(p_hwfn->p_dev, &p_conn->xhq);
	ecore_chain_free(p_hwfn->p_dev, &p_conn->uhq);
	ecore_chain_free(p_hwfn->p_dev, &p_conn->r2tq);
	OSAL_DMA_FREE_COHERENT(p_hwfn->p_dev,
			       p_conn->tcp_upload_params_virt_addr,
			       p_conn->tcp_upload_params_phys_addr,
			       sizeof(struct tcp_upload_params));
	OSAL_DMA_FREE_COHERENT(p_hwfn->p_dev,
			       p_conn->conn_stats_params_virt_addr,
			       p_conn->conn_stats_params_phys_addr,
			       sizeof(struct iscsi_conn_stats_params));
	OSAL_DMA_FREE_COHERENT(p_hwfn->p_dev,
			       p_conn->queue_cnts_virt_addr,
			       p_conn->queue_cnts_phys_addr,
			       sizeof(struct scsi_terminate_extra_params));
	OSAL_FREE(p_hwfn->p_dev, p_conn);
}

enum _ecore_status_t ecore_iscsi_alloc(struct ecore_hwfn *p_hwfn)
{
	struct ecore_iscsi_info *p_iscsi_info;

	p_iscsi_info = OSAL_ZALLOC(p_hwfn->p_dev, GFP_KERNEL,
				   sizeof(*p_iscsi_info));
	if (!p_iscsi_info) {
		DP_NOTICE(p_hwfn, true,
			  "Failed to allocate ecore_iscsi_info'\n");
		return ECORE_NOMEM;
	}

#ifdef CONFIG_ECORE_LOCK_ALLOC
	OSAL_SPIN_LOCK_ALLOC(p_hwfn, &p_iscsi_info->lock);
#endif
	OSAL_LIST_INIT(&p_iscsi_info->free_list);

	p_hwfn->p_iscsi_info = p_iscsi_info;
	return ECORE_SUCCESS;
}

void ecore_iscsi_setup(struct ecore_hwfn *p_hwfn)
{
	OSAL_SPIN_LOCK_INIT(&p_hwfn->p_iscsi_info->lock);
}

void ecore_iscsi_free(struct ecore_hwfn *p_hwfn)
{
	struct ecore_iscsi_conn	*p_conn = OSAL_NULL;

	if (!p_hwfn->p_iscsi_info)
		return;

	while (!OSAL_LIST_IS_EMPTY(&p_hwfn->p_iscsi_info->free_list)) {
		p_conn = OSAL_LIST_FIRST_ENTRY(&p_hwfn->p_iscsi_info->free_list,
					       struct ecore_iscsi_conn,
					       list_entry);
		if (p_conn) {
			OSAL_LIST_REMOVE_ENTRY(&p_conn->list_entry,
					       &p_hwfn->p_iscsi_info->free_list);
			ecore_iscsi_free_connection(p_hwfn, p_conn);
		}
	}

#ifdef CONFIG_ECORE_LOCK_ALLOC
	OSAL_SPIN_LOCK_DEALLOC(&p_hwfn->p_iscsi_info->lock);
#endif

	OSAL_FREE(p_hwfn->p_dev, p_hwfn->p_iscsi_info);
	p_hwfn->p_iscsi_info = OSAL_NULL;
}

static void _ecore_iscsi_get_tstats(struct ecore_hwfn *p_hwfn,
				    struct ecore_ptt *p_ptt,
				    struct ecore_iscsi_stats *p_stats)
{
	struct tstorm_iscsi_stats_drv tstats;
	u32 tstats_addr;

	OSAL_MEMSET(&tstats, 0, sizeof(tstats));
	tstats_addr = BAR0_MAP_REG_TSDM_RAM +
		TSTORM_ISCSI_RX_STATS_OFFSET(p_hwfn->rel_pf_id);
	ecore_memcpy_from(p_hwfn, p_ptt, &tstats, tstats_addr, sizeof(tstats));

	p_stats->iscsi_rx_bytes_cnt =
		HILO_64_REGPAIR(tstats.iscsi_rx_bytes_cnt);
	p_stats->iscsi_rx_packet_cnt =
		HILO_64_REGPAIR(tstats.iscsi_rx_packet_cnt);
	p_stats->iscsi_rx_new_ooo_isle_events_cnt =
		HILO_64_REGPAIR(tstats.iscsi_rx_new_ooo_isle_events_cnt);
	p_stats->iscsi_cmdq_threshold_cnt =
		OSAL_LE32_TO_CPU(tstats.iscsi_cmdq_threshold_cnt);
	p_stats->iscsi_rq_threshold_cnt =
		OSAL_LE32_TO_CPU(tstats.iscsi_rq_threshold_cnt);
	p_stats->iscsi_immq_threshold_cnt =
		OSAL_LE32_TO_CPU(tstats.iscsi_immq_threshold_cnt);
	p_stats->iscsi_rx_tcp_payload_bytes_cnt =
		HILO_64_REGPAIR(tstats.iscsi_rx_tcp_payload_bytes_cnt);
	p_stats->iscsi_rx_tcp_pkt_cnt =
		HILO_64_REGPAIR(tstats.iscsi_rx_tcp_pkt_cnt);
	p_stats->iscsi_rx_pure_ack_cnt =
		HILO_64_REGPAIR(tstats.iscsi_rx_pure_ack_cnt);
}

static void _ecore_iscsi_get_mstats(struct ecore_hwfn *p_hwfn,
				    struct ecore_ptt *p_ptt,
				    struct ecore_iscsi_stats *p_stats)
{
	struct mstorm_iscsi_stats_drv mstats;
	u32 mstats_addr;

	OSAL_MEMSET(&mstats, 0, sizeof(mstats));
	mstats_addr = BAR0_MAP_REG_MSDM_RAM +
		MSTORM_ISCSI_RX_STATS_OFFSET(p_hwfn->rel_pf_id);
	ecore_memcpy_from(p_hwfn, p_ptt, &mstats, mstats_addr, sizeof(mstats));

	p_stats->iscsi_rx_dropped_pdus_task_not_valid =
		HILO_64_REGPAIR(mstats.iscsi_rx_dropped_PDUs_task_not_valid);
	p_stats->iscsi_rx_dup_ack_cnt = 
		HILO_64_REGPAIR(mstats.iscsi_rx_dup_ack_cnt);
}

static void _ecore_iscsi_get_ustats(struct ecore_hwfn *p_hwfn,
				    struct ecore_ptt *p_ptt,
				    struct ecore_iscsi_stats *p_stats)
{
	struct ustorm_iscsi_stats_drv ustats;
	u32 ustats_addr;

	OSAL_MEMSET(&ustats, 0, sizeof(ustats));
	ustats_addr = BAR0_MAP_REG_USDM_RAM +
		USTORM_ISCSI_RX_STATS_OFFSET(p_hwfn->rel_pf_id);
	ecore_memcpy_from(p_hwfn, p_ptt, &ustats, ustats_addr, sizeof(ustats));

	p_stats->iscsi_rx_data_pdu_cnt =
		HILO_64_REGPAIR(ustats.iscsi_rx_data_pdu_cnt);
	p_stats->iscsi_rx_r2t_pdu_cnt =
		HILO_64_REGPAIR(ustats.iscsi_rx_r2t_pdu_cnt);
	p_stats->iscsi_rx_total_pdu_cnt =
		HILO_64_REGPAIR(ustats.iscsi_rx_total_pdu_cnt);
}

static void _ecore_iscsi_get_xstats(struct ecore_hwfn *p_hwfn,
				    struct ecore_ptt *p_ptt,
				    struct ecore_iscsi_stats *p_stats)
{
	struct xstorm_iscsi_stats_drv xstats;
	u32 xstats_addr;

	OSAL_MEMSET(&xstats, 0, sizeof(xstats));
	xstats_addr = BAR0_MAP_REG_XSDM_RAM +
		XSTORM_ISCSI_TX_STATS_OFFSET(p_hwfn->rel_pf_id);
	ecore_memcpy_from(p_hwfn, p_ptt, &xstats, xstats_addr, sizeof(xstats));

	p_stats->iscsi_tx_go_to_slow_start_event_cnt =
		HILO_64_REGPAIR(xstats.iscsi_tx_go_to_slow_start_event_cnt);
	p_stats->iscsi_tx_fast_retransmit_event_cnt =
		HILO_64_REGPAIR(xstats.iscsi_tx_fast_retransmit_event_cnt);
	p_stats->iscsi_tx_pure_ack_cnt =
		HILO_64_REGPAIR(xstats.iscsi_tx_pure_ack_cnt);
	p_stats->iscsi_tx_delayed_ack_cnt =
		HILO_64_REGPAIR(xstats.iscsi_tx_delayed_ack_cnt);
}

static void _ecore_iscsi_get_ystats(struct ecore_hwfn *p_hwfn,
				    struct ecore_ptt *p_ptt,
				    struct ecore_iscsi_stats *p_stats)
{
	struct ystorm_iscsi_stats_drv ystats;
	u32 ystats_addr;

	OSAL_MEMSET(&ystats, 0, sizeof(ystats));
	ystats_addr = BAR0_MAP_REG_YSDM_RAM +
		YSTORM_ISCSI_TX_STATS_OFFSET(p_hwfn->rel_pf_id);
	ecore_memcpy_from(p_hwfn, p_ptt, &ystats, ystats_addr, sizeof(ystats));

	p_stats->iscsi_tx_data_pdu_cnt =
		HILO_64_REGPAIR(ystats.iscsi_tx_data_pdu_cnt);
	p_stats->iscsi_tx_r2t_pdu_cnt =
		HILO_64_REGPAIR(ystats.iscsi_tx_r2t_pdu_cnt);
	p_stats->iscsi_tx_total_pdu_cnt =
		HILO_64_REGPAIR(ystats.iscsi_tx_total_pdu_cnt);
	p_stats->iscsi_tx_tcp_payload_bytes_cnt =
		HILO_64_REGPAIR(ystats.iscsi_tx_tcp_payload_bytes_cnt);
	p_stats->iscsi_tx_tcp_pkt_cnt =
		HILO_64_REGPAIR(ystats.iscsi_tx_tcp_pkt_cnt);
}

static void _ecore_iscsi_get_pstats(struct ecore_hwfn *p_hwfn,
				    struct ecore_ptt *p_ptt,
				    struct ecore_iscsi_stats *p_stats)
{
	struct pstorm_iscsi_stats_drv pstats;
	u32 pstats_addr;

	OSAL_MEMSET(&pstats, 0, sizeof(pstats));
	pstats_addr = BAR0_MAP_REG_PSDM_RAM +
		PSTORM_ISCSI_TX_STATS_OFFSET(p_hwfn->rel_pf_id);
	ecore_memcpy_from(p_hwfn, p_ptt, &pstats, pstats_addr, sizeof(pstats));

	p_stats->iscsi_tx_bytes_cnt =
		HILO_64_REGPAIR(pstats.iscsi_tx_bytes_cnt);
	p_stats->iscsi_tx_packet_cnt =
		HILO_64_REGPAIR(pstats.iscsi_tx_packet_cnt);
}

LNX_STATIC enum _ecore_status_t
ecore_iscsi_get_stats(struct ecore_hwfn *p_hwfn,
		      struct ecore_iscsi_stats *stats)
{
	struct ecore_ptt *p_ptt;

	OSAL_MEMSET(stats, 0, sizeof(*stats));

	p_ptt = ecore_ptt_acquire(p_hwfn);
	if (!p_ptt) {
		DP_ERR(p_hwfn, "Failed to acquire ptt\n");
		return ECORE_INVAL;
	}

	_ecore_iscsi_get_tstats(p_hwfn, p_ptt, stats);
	_ecore_iscsi_get_mstats(p_hwfn, p_ptt, stats);
	_ecore_iscsi_get_ustats(p_hwfn, p_ptt, stats);

	_ecore_iscsi_get_xstats(p_hwfn, p_ptt, stats);
	_ecore_iscsi_get_ystats(p_hwfn, p_ptt, stats);
	_ecore_iscsi_get_pstats(p_hwfn, p_ptt, stats);

	ecore_ptt_release(p_hwfn, p_ptt);

	return ECORE_SUCCESS;
}

#ifdef _NTDDK_
#pragma warning(pop)
#endif
