/*
 * Copyright (c) 2017-2019 Cavium, Inc.
 * All rights reserved.                
 *                                     
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:                                                          
 *                                                                    
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer. 
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *                                                                         
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE   
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 *  POSSIBILITY OF SUCH DAMAGE.                                                
 */    
#include "bcm_osal.h"

#include "ecore.h"
#include "ecore_l2_api.h"
#include "reg_addr.h"
#include "ecore_hw.h"
#include "ecore_dev_api.h"
#include "ecore_mcp.h"

/* 16 nano second time quantas to wait before making a Drift adjustment */
#define ECORE_DRIFT_CNTR_TIME_QUANTA_SHIFT    0
/* Nano seconds to add/subtract when making a Drift adjustment */
#define ECORE_DRIFT_CNTR_ADJUSTMENT_SHIFT     28
/* Add/subtract the Adjustment_Value when making a Drift adjustment */
#define ECORE_DRIFT_CNTR_DIRECTION_SHIFT      31
#define ECORE_TIMESTAMP_MASK		      0x10000

static enum ecore_resc_lock ecore_ptp_dev_to_resc(struct ecore_hwfn *p_hwfn)
{
	switch (MFW_PORT(p_hwfn)) {
	case 0:
		return ECORE_RESC_LOCK_PTP_PORT0;
	case 1:
		return ECORE_RESC_LOCK_PTP_PORT1;
	case 2:
		return ECORE_RESC_LOCK_PTP_PORT2;
	case 3:
		return ECORE_RESC_LOCK_PTP_PORT3;
	default:
		return ECORE_RESC_LOCK_RESC_INVALID;
	}
}

static enum _ecore_status_t ecore_ptp_res_lock(struct ecore_hwfn *p_hwfn,
					       struct ecore_ptt *p_ptt)
{
	struct ecore_resc_lock_params params;
	enum ecore_resc_lock resource;
	enum _ecore_status_t rc;

	resource = ecore_ptp_dev_to_resc(p_hwfn);
	if (resource == ECORE_RESC_LOCK_RESC_INVALID)
		return ECORE_INVAL;

	ecore_mcp_resc_lock_default_init(&params, OSAL_NULL, resource, true);

	rc = ecore_mcp_resc_lock(p_hwfn, p_ptt, &params);
	if (rc != ECORE_SUCCESS && rc != ECORE_NOTIMPL) {
		return rc;
	} else if (rc == ECORE_NOTIMPL) {
		/* MFW doesn't support resource locking, first PF on the port
		 * has lock ownership.
		 */
		if (p_hwfn->abs_pf_id < p_hwfn->p_dev->num_ports_in_engine)
			return ECORE_SUCCESS;

		DP_INFO(p_hwfn, "PF doesn't have lock ownership\n");
		return ECORE_BUSY;
	} else if (rc == ECORE_SUCCESS && !params.b_granted) {
		DP_INFO(p_hwfn, "Failed to acquire ptp resource lock\n");
		return ECORE_BUSY;
	}

	return rc;
}

static enum _ecore_status_t ecore_ptp_res_unlock(struct ecore_hwfn *p_hwfn,
						 struct ecore_ptt *p_ptt)
{
	struct ecore_resc_unlock_params params;
	enum ecore_resc_lock resource;
	enum _ecore_status_t rc;

	resource = ecore_ptp_dev_to_resc(p_hwfn);
	if (resource == ECORE_RESC_LOCK_RESC_INVALID)
		return ECORE_INVAL;

	ecore_mcp_resc_lock_default_init(OSAL_NULL, &params, resource, true);

	rc = ecore_mcp_resc_unlock(p_hwfn, p_ptt, &params);
	if (rc == ECORE_NOTIMPL) {
		/* MFW doesn't support locking, first PF has lock ownership */
		if (p_hwfn->abs_pf_id < p_hwfn->p_dev->num_ports_in_engine) {
			rc = ECORE_SUCCESS;
		} else {
			DP_INFO(p_hwfn, "PF doesn't have lock ownership\n");
			return ECORE_INVAL;
		}
	} else if (rc != ECORE_SUCCESS)
		DP_INFO(p_hwfn, "Failed to release the ptp resource lock\n");

	return rc;
}

/* Read Rx timestamp */
LNX_STATIC enum _ecore_status_t
ecore_ptp_hw_read_rx_ts(struct ecore_dev *p_dev, u64 *timestamp)
{
	struct ecore_hwfn *p_hwfn = ECORE_LEADING_HWFN(p_dev);
	struct ecore_ptt *p_ptt = p_hwfn->p_ptp_ptt;
	u32 val;

	*timestamp = 0;
	val = ecore_rd(p_hwfn, p_ptt, NIG_REG_LLH_PTP_HOST_BUF_SEQID);
	if (!(val & ECORE_TIMESTAMP_MASK)) {
		DP_INFO(p_hwfn,
			"Invalid Rx timestamp, but seqid = %08x\n", val);
		return ECORE_INVAL;
	}

	val = ecore_rd(p_hwfn, p_ptt, NIG_REG_LLH_PTP_HOST_BUF_TS_LSB);
	*timestamp = ecore_rd(p_hwfn, p_ptt, NIG_REG_LLH_PTP_HOST_BUF_TS_MSB);
	*timestamp <<= 32;
	*timestamp |= val;

	/* Reset timestamp register to allow new timestamp */
	ecore_wr(p_hwfn, p_ptt, NIG_REG_LLH_PTP_HOST_BUF_SEQID,
		 ECORE_TIMESTAMP_MASK);

	return ECORE_SUCCESS;
}

/* Read Tx timestamp */
LNX_STATIC enum _ecore_status_t
ecore_ptp_hw_read_tx_ts(struct ecore_dev *p_dev, u64 *timestamp)
{
	struct ecore_hwfn *p_hwfn = ECORE_LEADING_HWFN(p_dev);
	struct ecore_ptt *p_ptt = p_hwfn->p_ptp_ptt;
	u32 val;

	*timestamp = 0;
	val = ecore_rd(p_hwfn, p_ptt, NIG_REG_TX_LLH_PTP_BUF_SEQID);
	if (!(val & ECORE_TIMESTAMP_MASK)) {
		DP_VERBOSE(p_hwfn, ECORE_MSG_DEBUG,
			   "Invalid Tx timestamp, buf_seqid = %08x\n", val);
		return ECORE_INVAL;
	}

	val = ecore_rd(p_hwfn, p_ptt, NIG_REG_TX_LLH_PTP_BUF_TS_LSB);
	*timestamp = ecore_rd(p_hwfn, p_ptt, NIG_REG_TX_LLH_PTP_BUF_TS_MSB);
	*timestamp <<= 32;
	*timestamp |= val;

	/* Reset timestamp register to allow new timestamp */
	ecore_wr(p_hwfn, p_ptt, NIG_REG_TX_LLH_PTP_BUF_SEQID,
		 ECORE_TIMESTAMP_MASK);

	return ECORE_SUCCESS;
}

/* Read Phy Hardware Clock */
LNX_STATIC enum _ecore_status_t
ecore_ptp_hw_read_cc(struct ecore_dev *p_dev, u64 *phc_cycles)
{
	struct ecore_hwfn *p_hwfn = ECORE_LEADING_HWFN(p_dev);
	struct ecore_ptt *p_ptt = p_hwfn->p_ptp_ptt;
	u32 temp = 0;

	/* Reading of the LSB would latch the MSB until read */
	temp = ecore_rd(p_hwfn, p_ptt, NIG_REG_TSGEN_SYNC_TIME_LSB);
	*phc_cycles = ecore_rd(p_hwfn, p_ptt, NIG_REG_TSGEN_SYNC_TIME_MSB);
	*phc_cycles <<= 32;
	*phc_cycles |= temp;

	return ECORE_SUCCESS;
}

/* Filter PTP protocol packets that need to be timestamped */
LNX_STATIC enum _ecore_status_t
ecore_ptp_hw_cfg_filters(struct ecore_dev *p_dev,
			 enum ecore_ptp_filter_type rx_type,
			 enum ecore_ptp_hwtstamp_tx_type tx_type)
{
	struct ecore_hwfn *p_hwfn = ECORE_LEADING_HWFN(p_dev);
	struct ecore_ptt *p_ptt = p_hwfn->p_ptp_ptt;
	u32 rule_mask, enable_cfg = 0x0;

	switch (rx_type) {
	case ECORE_PTP_FILTER_NONE:
		enable_cfg = 0x0;
		rule_mask = 0x3FFF;
		break;
	case ECORE_PTP_FILTER_ALL:
		enable_cfg = 0x7;
		rule_mask = 0x3CAA;
		break;
	case ECORE_PTP_FILTER_V1_L4_EVENT:
		enable_cfg = 0x3;
		rule_mask = 0x3FFA;
		break;
	case ECORE_PTP_FILTER_V1_L4_GEN:
		enable_cfg = 0x3;
		rule_mask = 0x3FFE;
		break;
	case ECORE_PTP_FILTER_V2_L4_EVENT:
		enable_cfg = 0x5;
		rule_mask = 0x3FAA;
		break;
	case ECORE_PTP_FILTER_V2_L4_GEN:
		enable_cfg = 0x5;
		rule_mask = 0x3FEE;
		break;
	case ECORE_PTP_FILTER_V2_L2_EVENT:
		enable_cfg = 0x5;
		rule_mask = 0x3CFF;
		break;
	case ECORE_PTP_FILTER_V2_L2_GEN:
		enable_cfg = 0x5;
		rule_mask = 0x3EFF;
		break;
	case ECORE_PTP_FILTER_V2_EVENT:
		enable_cfg = 0x5;
		rule_mask = 0x3CAA;
		break;
	case ECORE_PTP_FILTER_V2_GEN:
		enable_cfg = 0x5;
		rule_mask = 0x3EEE;
		break;
	default:
		DP_INFO(p_hwfn, "Invalid PTP filter type %d\n", rx_type);
		return ECORE_INVAL;
	}

	ecore_wr(p_hwfn, p_ptt, NIG_REG_LLH_PTP_PARAM_MASK, 0xF);
	ecore_wr(p_hwfn, p_ptt, NIG_REG_LLH_PTP_RULE_MASK, rule_mask);
	ecore_wr(p_hwfn, p_ptt, NIG_REG_RX_PTP_EN, enable_cfg);

	if (tx_type == ECORE_PTP_HWTSTAMP_TX_OFF) {
		ecore_wr(p_hwfn, p_ptt, NIG_REG_TX_PTP_EN, 0x0);
		ecore_wr(p_hwfn, p_ptt, NIG_REG_TX_LLH_PTP_PARAM_MASK, 0x7FF);
		ecore_wr(p_hwfn, p_ptt, NIG_REG_TX_LLH_PTP_RULE_MASK, 0x3FFF);
	} else {
		ecore_wr(p_hwfn, p_ptt, NIG_REG_TX_PTP_EN, enable_cfg);
		ecore_wr(p_hwfn, p_ptt, NIG_REG_TX_LLH_PTP_PARAM_MASK, 0xF);
		ecore_wr(p_hwfn, p_ptt, NIG_REG_TX_LLH_PTP_RULE_MASK,
			 rule_mask);
	}

	/* Reset possibly old timestamps */
	ecore_wr(p_hwfn, p_ptt, NIG_REG_LLH_PTP_HOST_BUF_SEQID,
		 ECORE_TIMESTAMP_MASK);

	return ECORE_SUCCESS;
}

/* Adjust the HW clock by a rate given in parts-per-billion (ppb) units.
 * FW/HW accepts the adjustment value in terms of 3 parameters:
 *   Drift period - adjustment happens once in certain number of nano seconds.
 *   Drift value - time is adjusted by a certain value, for example by 5 ns.
 *   Drift direction - add or subtract the adjustment value.
 * The routine translates ppb into the adjustment triplet in an optimal manner.
 */
LNX_STATIC enum _ecore_status_t
ecore_ptp_hw_adjfreq(struct ecore_dev *p_dev, s32 ppb)
{
	s64 best_val = 0, val, best_period = 0, period, approx_dev, dif, dif2;
	struct ecore_hwfn *p_hwfn = ECORE_LEADING_HWFN(p_dev);
	struct ecore_ptt *p_ptt = p_hwfn->p_ptp_ptt;
	u32 drift_ctr_cfg = 0, drift_state;
	int drift_dir = 1;

	if (ppb < 0) {
		ppb = -ppb;
		drift_dir = 0;
	}

	if (ppb > 1) {
		s64 best_dif = ppb, best_approx_dev = 1;

		/* Adjustment value is up to +/-7ns, find an optimal value in
		 * this range.
		 */
		for (val = 7; val > 0; val--) {
			period = OSAL_DIV_S64(val * 1000000000, ppb);
			period -= 8;
			period >>= 4;
			if (period < 1)
				period = 1;
			if (period > 0xFFFFFFE)
				period = 0xFFFFFFE;

			/* Check both rounding ends for approximate error */
			approx_dev = period * 16 + 8;
			dif = ppb * approx_dev - val * 1000000000;
			dif2 = dif + 16 * ppb;

			if (dif < 0)
				dif = -dif;
			if (dif2 < 0)
				dif2 = -dif2;

			/* Determine which end gives better approximation */
			if (dif * (approx_dev + 16) > dif2 * approx_dev) {
				period++;
				approx_dev += 16;
				dif = dif2;
			}

			/* Track best approximation found so far */
			if (best_dif * approx_dev > dif * best_approx_dev) {
				best_dif = dif;
				best_val = val;
				best_period = period;
				best_approx_dev = approx_dev;
			}
		}
	} else if (ppb == 1) {
		/* This is a special case as its the only value which wouldn't
		 * fit in a s64 variable. In order to prevent castings simple
		 * handle it seperately.
		 */
		best_val = 4;
		best_period = 0xee6b27f;
	} else {
		best_val = 0;
		best_period = 0xFFFFFFF;
	}

	drift_ctr_cfg = (best_period << ECORE_DRIFT_CNTR_TIME_QUANTA_SHIFT) |
			(((int)best_val) << ECORE_DRIFT_CNTR_ADJUSTMENT_SHIFT) |
			(((int)drift_dir) << ECORE_DRIFT_CNTR_DIRECTION_SHIFT);

	ecore_wr(p_hwfn, p_ptt, NIG_REG_TSGEN_RST_DRIFT_CNTR, 0x1);

	drift_state = ecore_rd(p_hwfn, p_ptt, NIG_REG_TSGEN_RST_DRIFT_CNTR);
	if (drift_state & 1) {
		ecore_wr(p_hwfn, p_ptt, NIG_REG_TSGEN_DRIFT_CNTR_CONF,
			 drift_ctr_cfg);
	} else {
		DP_INFO(p_hwfn, "Drift counter is not reset\n");
		return ECORE_INVAL;
	}

	ecore_wr(p_hwfn, p_ptt, NIG_REG_TSGEN_RST_DRIFT_CNTR, 0x0);

	return ECORE_SUCCESS;
}

LNX_STATIC enum _ecore_status_t ecore_ptp_hw_enable(struct ecore_dev *p_dev)
{
	struct ecore_hwfn *p_hwfn = ECORE_LEADING_HWFN(p_dev);
	struct ecore_ptt *p_ptt;
	enum _ecore_status_t rc;

	p_ptt = ecore_ptt_acquire(p_hwfn);
	if (p_ptt == OSAL_NULL) {
		DP_NOTICE(p_hwfn, false,
			  "Failed to acquire PTT for PTP\n");
		return ECORE_BUSY;
	}
	p_hwfn->p_ptp_ptt = p_ptt;

	rc = ecore_ptp_res_lock(p_hwfn, p_ptt);
	if (rc != ECORE_SUCCESS) {
		DP_INFO(p_hwfn,
			"Couldn't acquire the resource lock, skip ptp enable for this PF\n");
		ecore_ptt_release(p_hwfn, p_ptt);
		p_hwfn->p_ptp_ptt = OSAL_NULL;
		return rc;
	}

	/* Reset PTP event detection rules - will be configured in the IOCTL */
	ecore_wr(p_hwfn, p_ptt, NIG_REG_LLH_PTP_PARAM_MASK, 0x7FF);
	ecore_wr(p_hwfn, p_ptt, NIG_REG_LLH_PTP_RULE_MASK, 0x3FFF);
	ecore_wr(p_hwfn, p_ptt, NIG_REG_TX_LLH_PTP_PARAM_MASK, 0x7FF);
	ecore_wr(p_hwfn, p_ptt, NIG_REG_TX_LLH_PTP_RULE_MASK, 0x3FFF);

	ecore_wr(p_hwfn, p_ptt, NIG_REG_TX_PTP_EN, 7);
	ecore_wr(p_hwfn, p_ptt, NIG_REG_RX_PTP_EN, 7);

	ecore_wr(p_hwfn, p_ptt, NIG_REG_TS_OUTPUT_ENABLE_PDA, 0x1);

	/* Pause free running counter */
	if (ECORE_IS_BB_B0(p_hwfn->p_dev))
		ecore_wr(p_hwfn, p_ptt, NIG_REG_TIMESYNC_GEN_REG_BB, 2);
	if (ECORE_IS_AH(p_hwfn->p_dev))
		ecore_wr(p_hwfn, p_ptt, NIG_REG_TSGEN_FREECNT_UPDATE_K2, 2);

	ecore_wr(p_hwfn, p_ptt, NIG_REG_TSGEN_FREE_CNT_VALUE_LSB, 0);
	ecore_wr(p_hwfn, p_ptt, NIG_REG_TSGEN_FREE_CNT_VALUE_MSB, 0);
	/* Resume free running counter */
	if (ECORE_IS_BB_B0(p_hwfn->p_dev))
		ecore_wr(p_hwfn, p_ptt, NIG_REG_TIMESYNC_GEN_REG_BB, 4);
	if (ECORE_IS_AH(p_hwfn->p_dev)) {
		ecore_wr(p_hwfn, p_ptt, NIG_REG_TSGEN_FREECNT_UPDATE_K2, 4);
		/* Workaround for latching the 2-step PTP packets */
		ecore_wr(p_hwfn, p_ptt, NIG_REG_PTP_LATCH_OSTS_PKT_TIME_K2, 1);
	}

	/* Disable drift register */
	ecore_wr(p_hwfn, p_ptt, NIG_REG_TSGEN_DRIFT_CNTR_CONF, 0x0);
	ecore_wr(p_hwfn, p_ptt, NIG_REG_TSGEN_RST_DRIFT_CNTR, 0x0);

	/* Reset possibly old timestamps */
	ecore_wr(p_hwfn, p_ptt, NIG_REG_LLH_PTP_HOST_BUF_SEQID,
		 ECORE_TIMESTAMP_MASK);
	ecore_wr(p_hwfn, p_ptt, NIG_REG_TX_LLH_PTP_BUF_SEQID,
		 ECORE_TIMESTAMP_MASK);

	return ECORE_SUCCESS;
}

LNX_STATIC enum _ecore_status_t ecore_ptp_hw_disable(struct ecore_dev *p_dev)
{
	struct ecore_hwfn *p_hwfn = ECORE_LEADING_HWFN(p_dev);
	struct ecore_ptt *p_ptt = p_hwfn->p_ptp_ptt;

	if (p_ptt == OSAL_NULL)
		return ECORE_SUCCESS;

	ecore_ptp_res_unlock(p_hwfn, p_ptt);

	/* Reset PTP event detection rules */
	ecore_wr(p_hwfn, p_ptt, NIG_REG_LLH_PTP_PARAM_MASK, 0x7FF);
	ecore_wr(p_hwfn, p_ptt, NIG_REG_LLH_PTP_RULE_MASK, 0x3FFF);

	ecore_wr(p_hwfn, p_ptt, NIG_REG_TX_LLH_PTP_PARAM_MASK, 0x7FF);
	ecore_wr(p_hwfn, p_ptt, NIG_REG_TX_LLH_PTP_RULE_MASK, 0x3FFF);

	/* Disable the PTP feature */
	ecore_wr(p_hwfn, p_ptt, NIG_REG_RX_PTP_EN, 0x0);
	ecore_wr(p_hwfn, p_ptt, NIG_REG_TX_PTP_EN, 0x0);

	ecore_ptt_release(p_hwfn, p_ptt);
	p_hwfn->p_ptp_ptt = OSAL_NULL;

	return ECORE_SUCCESS;
}
