/*
 * Copyright (c) 2017-2019 Cavium, Inc.
 * All rights reserved.                
 *                                     
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:                                                          
 *                                                                    
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer. 
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *                                                                         
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE   
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 *  POSSIBILITY OF SUCH DAMAGE.                                                
 */    
#include "bcm_osal.h"
#include "ecore.h"
#include "ecore_cxt.h"
#include "ecore_hw.h"
#include "ecore_mcp.h"
#include "spad_layout.h"
#include "aeu_inputs.h"
#include "reg_addr.h"
#include "ecore_hsi_common.h"
#include "ecore_hsi_debug_tools.h"
#include "mcp_public.h"
#include "nvm_map.h"
#ifndef USE_DBG_BIN_FILE
#include "ecore_dbg_values.h"
#endif
#include "ecore_dbg_fw_funcs.h"

/* Memory groups enum */
enum mem_groups {
	MEM_GROUP_PXP_MEM,
	MEM_GROUP_DMAE_MEM,
	MEM_GROUP_CM_MEM,
	MEM_GROUP_QM_MEM,
	MEM_GROUP_DORQ_MEM,
	MEM_GROUP_BRB_RAM,
	MEM_GROUP_BRB_MEM,
	MEM_GROUP_PRS_MEM,
	MEM_GROUP_SDM_MEM,
	MEM_GROUP_PBUF,
	MEM_GROUP_IOR,
	MEM_GROUP_RAM,
	MEM_GROUP_BTB_RAM,
	MEM_GROUP_RDIF_CTX,
	MEM_GROUP_TDIF_CTX,
	MEM_GROUP_CFC_MEM,
	MEM_GROUP_CONN_CFC_MEM,
	MEM_GROUP_CAU_PI,
	MEM_GROUP_CAU_MEM,
	MEM_GROUP_CAU_MEM_EXT,
	MEM_GROUP_PXP_ILT,
	MEM_GROUP_MULD_MEM,
	MEM_GROUP_BTB_MEM,
	MEM_GROUP_IGU_MEM,
	MEM_GROUP_IGU_MSIX,
	MEM_GROUP_CAU_SB,
	MEM_GROUP_BMB_RAM,
	MEM_GROUP_BMB_MEM,
	MEM_GROUP_TM_MEM,
	MEM_GROUP_TASK_CFC_MEM,
	MEM_GROUPS_NUM
};

/* Memory groups names */
static const char* s_mem_group_names[] = {
	"PXP_MEM",
	"DMAE_MEM",
	"CM_MEM",
	"QM_MEM",
	"DORQ_MEM",
	"BRB_RAM",
	"BRB_MEM",
	"PRS_MEM",
	"SDM_MEM",
	"PBUF",
	"IOR",
	"RAM",
	"BTB_RAM",
	"RDIF_CTX",
	"TDIF_CTX",
	"CFC_MEM",
	"CONN_CFC_MEM",
	"CAU_PI",
	"CAU_MEM",
	"CAU_MEM_EXT",
	"PXP_ILT",
	"MULD_MEM",
	"BTB_MEM",
	"IGU_MEM",
	"IGU_MSIX",
	"CAU_SB",
	"BMB_RAM",
	"BMB_MEM",
	"TM_MEM",
	"TASK_CFC_MEM",
};

/* Idle check conditions */

#ifndef __PREVENT_COND_ARR__

static u32 cond5(const u32 *r, const u32 *imm) {
	return (((r[0] & imm[0]) != imm[1]) && ((r[1] & imm[2]) != imm[3]));
}

static u32 cond7(const u32 *r, const u32 *imm) {
	return (((r[0] >> imm[0]) & imm[1]) != imm[2]);
}

static u32 cond6(const u32 *r, const u32 *imm) {
	return ((r[0] & imm[0]) != imm[1]);
}

static u32 cond9(const u32 *r, const u32 *imm) {
	return ((r[0] & imm[0]) >> imm[1]) != (((r[0] & imm[2]) >> imm[3]) | ((r[1] & imm[4]) << imm[5]));
}

static u32 cond10(const u32 *r, const u32 *imm) {
	return ((r[0] & imm[0]) >> imm[1]) != (r[0] & imm[2]);
}

static u32 cond4(const u32 *r, const u32 *imm) {
	return ((r[0] & ~imm[0]) != imm[1]);
}

static u32 cond0(const u32 *r, const u32 *imm) {
	return ((r[0] & ~r[1]) != imm[0]);
}

static u32 cond1(const u32 *r, const u32 *imm) {
	return (r[0] != imm[0]);
}

static u32 cond11(const u32 *r, const u32 *imm) {
	return (r[0] != r[1] && r[2] == imm[0]);
}

static u32 cond12(const u32 *r, const u32 *imm) {
	return (r[0] != r[1] && r[2] > imm[0]);
}

static u32 cond3(const u32 *r, const u32 OSAL_UNUSED *imm) {
	return (r[0] != r[1]);
}

static u32 cond13(const u32 *r, const u32 *imm) {
	return (r[0] & imm[0]);
}

static u32 cond8(const u32 *r, const u32 *imm) {
	return (r[0] < (r[1] - imm[0]));
}

static u32 cond2(const u32 *r, const u32 *imm) {
	return (r[0] > imm[0]);
}

/* Array of Idle Check conditions */
static u32 (*cond_arr[])(const u32 *r, const u32 *imm) = {
	cond0,
	cond1,
	cond2,
	cond3,
	cond4,
	cond5,
	cond6,
	cond7,
	cond8,
	cond9,
	cond10,
	cond11,
	cond12,
	cond13,
};

#endif /* __PREVENT_COND_ARR__ */

#define NUM_PHYS_BLOCKS 84

#define NUM_DBG_RESET_REGS 8


/******************************* Data Types **********************************/

enum hw_types {
	HW_TYPE_ASIC,
	HW_TYPE_EMUL_FULL,
	HW_TYPE_EMUL_REDUCED,
	HW_TYPE_FPGA,
	PLATFORM_CHIPSIM,
	MAX_HW_TYPES
};

/* CM context types */
enum cm_ctx_types {
	CM_CTX_CONN_AG,
	CM_CTX_CONN_ST,
	CM_CTX_TASK_AG,
	CM_CTX_TASK_ST,
	NUM_CM_CTX_TYPES
};

/* HW ID allocation modes */
enum hw_id_alloc_mode {
	HW_ID_MODE_SINGLE,	/* Single HW ID to all dwords */
	HW_ID_MODE_PER_BLOCK,	/* HW ID per block */
	HW_ID_MODE_PER_DWORD,	/* HW ID per dword*/
	MAX_HW_ID_MODES
};

/* Debug bus frame modes */
enum dbg_bus_frame_modes {
	DBG_BUS_FRAME_MODE_4ST = 0,	/* 4 Storm dwords (no HW) */
	DBG_BUS_FRAME_MODE_2ST_2HW = 1, /* 2 Storm dwords, 2 HW dwords */
	DBG_BUS_FRAME_MODE_1ST_3HW = 2,	/* 1 Storm dwords, 3 HW dwords */
	DBG_BUS_FRAME_MODE_4HW = 3,	/* 4 HW dwords (no Storms) */
	DBG_BUS_FRAME_MODE_8HW = 4,	/* 8 HW dwords (no Storms) */
	DBG_BUS_NUM_FRAME_MODES
};

/* Debug bus SEMI frame modes */
enum dbg_bus_semi_frame_modes {
	DBG_BUS_SEMI_FRAME_MODE_4FAST = 0,	 /* 4 fast dwords */
	DBG_BUS_SEMI_FRAME_MODE_2FAST_2SLOW = 1, /* 2 fast dwords, 2 slow dwords */
	DBG_BUS_SEMI_FRAME_MODE_1FAST_3SLOW = 2, /* 1 fast dwords, 3 slow dwords */
	DBG_BUS_SEMI_FRAME_MODE_4SLOW = 3,	 /* 4 slow dwords */
	DBG_BUS_SEMI_NUM_FRAME_MODES
};

/* Debug bus filter types */
enum dbg_bus_filter_types {
	DBG_BUS_FILTER_TYPE_OFF,	/* Filter always off */
	DBG_BUS_FILTER_TYPE_PRE,	/* Filter before trigger only */
	DBG_BUS_FILTER_TYPE_POST,	/* Filter after trigger only */
	DBG_BUS_FILTER_TYPE_ON		/* Filter always on */
};

/* Debug bus pre-trigger recording types */
enum dbg_bus_pre_trigger_types {
	DBG_BUS_PRE_TRIGGER_FROM_ZERO,	/* Record from time 0 */
	DBG_BUS_PRE_TRIGGER_NUM_CHUNKS,	/* Record some chunks before trigger */
	DBG_BUS_PRE_TRIGGER_DROP	/* Drop data before trigger */
};

/* Debug bus post-trigger recording types */
enum dbg_bus_post_trigger_types {
	DBG_BUS_POST_TRIGGER_RECORD, /* Start recording after trigger */
	DBG_BUS_POST_TRIGGER_DROP    /* Drop data after trigger */
};

/* Debug bus other engine mode */
enum dbg_bus_other_engine_modes {
	DBG_BUS_OTHER_ENGINE_MODE_NONE,
	DBG_BUS_OTHER_ENGINE_MODE_DOUBLE_BW_TX,
	DBG_BUS_OTHER_ENGINE_MODE_DOUBLE_BW_RX,
	DBG_BUS_OTHER_ENGINE_MODE_CROSS_ENGINE_TX,
	DBG_BUS_OTHER_ENGINE_MODE_CROSS_ENGINE_RX
};

/* DBG block Framing mode definitions */
struct framing_mode_defs {
	u8 id;
	u8 blocks_dword_mask;
	u8 storms_dword_mask;
	u8 semi_framing_mode_id;
	u8 full_buf_thr;
};

/* Chip constant definitions */
struct chip_defs {
	const char *name;
	u8 dwords_per_cycle;
	u8 num_framing_modes;
	u32 num_ilt_pages;
	struct framing_mode_defs *framing_modes;
};

/* HW type constant definitions */
struct hw_type_defs {
	const char *name;
	u32 delay_factor;
	u32 dmae_thresh;
	u32 log_thresh;
};

/* RBC reset definitions */
struct rbc_reset_defs {
	u32 reset_reg_addr;
	u32 reset_val[MAX_CHIP_IDS];
};

/* Storm constant definitions.
 * Addresses are in bytes, sizes are in quad-regs.
 */
struct storm_defs {
	char letter;
	enum block_id sem_block_id;
	enum dbg_bus_clients dbg_client_id[MAX_CHIP_IDS];
	bool has_vfc;
	u32 sem_fast_mem_addr;
	u32 sem_frame_mode_addr;
	u32 sem_slow_enable_addr;
	u32 sem_slow_mode_addr;
	u32 sem_slow_mode1_conf_addr;
	u32 sem_sync_dbg_empty_addr;
	u32 sem_gpre_vect_addr;
	u32 cm_ctx_wr_addr;
	u32 cm_ctx_rd_addr[NUM_CM_CTX_TYPES];
	u32 cm_ctx_lid_sizes[MAX_CHIP_IDS][NUM_CM_CTX_TYPES];
};

/* Debug Bus Constraint operation constant definitions */
struct dbg_bus_constraint_op_defs {
	u8 hw_op_val;
	bool is_cyclic;
};

/* Storm Mode definitions */
struct storm_mode_defs {
	const char *name;
	bool is_fast_dbg;
	u8 id_in_hw;
	u32 src_disable_reg_addr;
	u32 src_enable_val;
	bool exists[MAX_CHIP_IDS];
};

struct grc_param_defs {
	u32 default_val[MAX_CHIP_IDS];
	u32 min;
	u32 max;
	bool is_preset;
	bool is_persistent;
	u32 exclude_all_preset_val;
	u32 crash_preset_val[MAX_CHIP_IDS];
};

/* address is in 128b units. Width is in bits. */
struct rss_mem_defs {
	const char *mem_name;
	const char *type_name;
	u32 addr;
	u32 entry_width;
	u32 num_entries[MAX_CHIP_IDS];
};

struct vfc_ram_defs {
	const char *mem_name;
	const char *type_name;
	u32 base_row;
	u32 num_rows;
};

struct big_ram_defs {
	const char *instance_name;
	enum mem_groups mem_group_id;
	enum mem_groups ram_mem_group_id;
	enum dbg_grc_params grc_param;
	u32 addr_reg_addr;
	u32 data_reg_addr;
	u32 is_256b_reg_addr;
	u32 is_256b_bit_offset[MAX_CHIP_IDS];
	u32 ram_size[MAX_CHIP_IDS]; /* In dwords */
};

struct phy_defs {
	const char *phy_name;

	/* PHY base GRC address */
	u32 base_addr;

	/* Relative address of indirect TBUS address register (bits 0..7) */
	u32 tbus_addr_lo_addr;

	/* Relative address of indirect TBUS address register (bits 8..10) */
	u32 tbus_addr_hi_addr;

	/* Relative address of indirect TBUS data register (bits 0..7) */
	u32 tbus_data_lo_addr;

	/* Relative address of indirect TBUS data register (bits 8..11) */
	u32 tbus_data_hi_addr;
};

/* Split type definitions */
struct split_type_defs {
	const char *name;
};

/******************************** Constants **********************************/

#define BYTES_IN_DWORD			sizeof(u32)

/* In the macros below, size and offset are specified in bits */
#define CEIL_DWORDS(size)		DIV_ROUND_UP(size, 32)
#define FIELD_BIT_OFFSET(type, field)	type##_##field##_##OFFSET
#define FIELD_BIT_SIZE(type, field)	type##_##field##_##SIZE
#define FIELD_DWORD_OFFSET(type, field)		(int)(FIELD_BIT_OFFSET(type, field) / 32)
#define FIELD_DWORD_SHIFT(type, field)	(FIELD_BIT_OFFSET(type, field) % 32)
#define FIELD_BIT_MASK(type, field)		(((1 << FIELD_BIT_SIZE(type, field)) - 1) 	<< FIELD_DWORD_SHIFT(type, field))

#define SET_VAR_FIELD(var, type, field, val) 	var[FIELD_DWORD_OFFSET(type, field)] &= 		(~FIELD_BIT_MASK(type, field)); 	var[FIELD_DWORD_OFFSET(type, field)] |= 		(val) << FIELD_DWORD_SHIFT(type, field)

#define ARR_REG_WR(dev, ptt, addr, arr, arr_size) 	for (i = 0; i < (arr_size); i++) 		ecore_wr(dev, ptt, addr, (arr)[i])

#define CHECK_ARR_SIZE(arr, size) 	OSAL_BUILD_BUG_ON(!(OSAL_ARRAY_SIZE(arr) == size))

#ifndef DWORDS_TO_BYTES
#define DWORDS_TO_BYTES(dwords)		((dwords) * BYTES_IN_DWORD)
#endif
#ifndef BYTES_TO_DWORDS
#define BYTES_TO_DWORDS(bytes)		((bytes) / BYTES_IN_DWORD)
#endif

/* extra lines include a signature line + optional latency events line */
#define NUM_EXTRA_DBG_LINES(block)		(GET_FIELD(block->flags, DBG_BLOCK_CHIP_HAS_LATENCY_EVENTS) ? 2 : 1)
#define NUM_DBG_LINES(block)		(block->num_of_dbg_bus_lines + NUM_EXTRA_DBG_LINES(block))

#define USE_DMAE			true
#define PROTECT_WIDE_BUS		true

#define RAM_LINES_TO_DWORDS(lines)	((lines) * 2)
#define RAM_LINES_TO_BYTES(lines)		DWORDS_TO_BYTES(RAM_LINES_TO_DWORDS(lines))

#define REG_DUMP_LEN_SHIFT		24
#define MEM_DUMP_ENTRY_SIZE_DWORDS		BYTES_TO_DWORDS(sizeof(struct dbg_dump_mem))

#define IDLE_CHK_RULE_SIZE_DWORDS		BYTES_TO_DWORDS(sizeof(struct dbg_idle_chk_rule))

#define IDLE_CHK_RESULT_HDR_DWORDS		BYTES_TO_DWORDS(sizeof(struct dbg_idle_chk_result_hdr))

#define IDLE_CHK_RESULT_REG_HDR_DWORDS		BYTES_TO_DWORDS(sizeof(struct dbg_idle_chk_result_reg_hdr))

#define PAGE_MEM_DESC_SIZE_DWORDS		BYTES_TO_DWORDS(sizeof(struct phys_mem_desc))

#define IDLE_CHK_MAX_ENTRIES_SIZE	32

/* The sizes and offsets below are specified in bits */
#define VFC_CAM_CMD_STRUCT_SIZE		64
#define VFC_CAM_CMD_ROW_OFFSET		48
#define VFC_CAM_CMD_ROW_SIZE		9
#define VFC_CAM_ADDR_STRUCT_SIZE	16
#define VFC_CAM_ADDR_OP_OFFSET		0
#define VFC_CAM_ADDR_OP_SIZE		4
#define VFC_CAM_RESP_STRUCT_SIZE	256
#define VFC_RAM_ADDR_STRUCT_SIZE	16
#define VFC_RAM_ADDR_OP_OFFSET		0
#define VFC_RAM_ADDR_OP_SIZE		2
#define VFC_RAM_ADDR_ROW_OFFSET		2
#define VFC_RAM_ADDR_ROW_SIZE		10
#define VFC_RAM_RESP_STRUCT_SIZE	256

#define VFC_CAM_CMD_DWORDS		CEIL_DWORDS(VFC_CAM_CMD_STRUCT_SIZE)
#define VFC_CAM_ADDR_DWORDS		CEIL_DWORDS(VFC_CAM_ADDR_STRUCT_SIZE)
#define VFC_CAM_RESP_DWORDS		CEIL_DWORDS(VFC_CAM_RESP_STRUCT_SIZE)
#define VFC_RAM_CMD_DWORDS		VFC_CAM_CMD_DWORDS
#define VFC_RAM_ADDR_DWORDS		CEIL_DWORDS(VFC_RAM_ADDR_STRUCT_SIZE)
#define VFC_RAM_RESP_DWORDS		CEIL_DWORDS(VFC_RAM_RESP_STRUCT_SIZE)

#define NUM_VFC_RAM_TYPES		4

#define VFC_CAM_NUM_ROWS		512

#define VFC_OPCODE_CAM_RD		14
#define VFC_OPCODE_RAM_RD		0

#define NUM_RSS_MEM_TYPES		5

#define NUM_BIG_RAM_TYPES		3
#define BIG_RAM_NAME_LEN		3

#define NUM_PHY_TBUS_ADDRESSES		2048
#define PHY_DUMP_SIZE_DWORDS		(NUM_PHY_TBUS_ADDRESSES / 2)

#define SEM_FAST_MODE23_SRC_DISABLE_VAL	0x7
#define SEM_FAST_MODE4_SRC_DISABLE_VAL	0x3
#define SEM_FAST_MODE6_SRC_DISABLE_VAL	0x3f

#define SEM_SLOW_MODE1_DATA_ENABLE	0x1

#define DEBUG_BUS_CYCLE_DWORDS		8

#define VALUES_PER_BLOCK		4
#define MAX_BLOCK_VALUES_MASK		((1 << VALUES_PER_BLOCK) - 1)

#define HW_ID_BITS			3

#define NUM_CALENDAR_SLOTS		16

#define MAX_TRIGGER_STATES		3
#define TRIGGER_SETS_PER_STATE		2
#define MAX_CONSTRAINTS			4
#define MAX_FILTER_CYCLE_OFFSET		4

#define SEM_FILTER_CID_EN_MASK		0x00b
#define SEM_FILTER_EID_MASK_EN_MASK	0x013
#define SEM_FILTER_EID_RANGE_EN_MASK	0x113

#define CHUNK_SIZE_IN_DWORDS		64
#define CHUNK_SIZE_IN_BYTES		DWORDS_TO_BYTES(CHUNK_SIZE_IN_DWORDS)

#define INT_BUF_NUM_OF_LINES		192
#define INT_BUF_LINE_SIZE_IN_DWORDS	16
#define INT_BUF_SIZE_IN_DWORDS			(INT_BUF_NUM_OF_LINES * INT_BUF_LINE_SIZE_IN_DWORDS)
#define INT_BUF_SIZE_IN_CHUNKS			(INT_BUF_SIZE_IN_DWORDS / CHUNK_SIZE_IN_DWORDS)

#define PCI_BUF_LINE_SIZE_IN_DWORDS	8
#define PCI_BUF_LINE_SIZE_IN_BYTES		DWORDS_TO_BYTES(PCI_BUF_LINE_SIZE_IN_DWORDS)

#define TARGET_EN_MASK_PCI		0x3
#define TARGET_EN_MASK_NIG		0x4

#define PCI_REQ_CREDIT			1
#define PCI_PHYS_ADDR_TYPE		0

#define OPAQUE_FID(pci_func)		((pci_func << 4) | 0xff00)

#define RESET_REG_UNRESET_OFFSET	4

#define PCI_PKT_SIZE_IN_CHUNKS		1
#define PCI_PKT_SIZE_IN_BYTES			(PCI_PKT_SIZE_IN_CHUNKS * CHUNK_SIZE_IN_BYTES)

#define NIG_PKT_SIZE_IN_CHUNKS		4

#define FLUSH_DELAY_MS			500
#define STALL_DELAY_MS			500

#define SRC_MAC_ADDR_LO16		0x0a0b
#define SRC_MAC_ADDR_HI32		0x0c0d0e0f
#define ETH_TYPE			0x1000

#define STATIC_DEBUG_LINE_DWORDS	9

#define NUM_COMMON_GLOBAL_PARAMS	11

#define MAX_RECURSION_DEPTH		10

#define FW_IMG_KUKU			0
#define FW_IMG_MAIN			1
#define FW_IMG_L2B			2

#ifndef REG_FIFO_ELEMENT_DWORDS
#define REG_FIFO_ELEMENT_DWORDS		2
#endif
#define REG_FIFO_DEPTH_ELEMENTS		32
#define REG_FIFO_DEPTH_DWORDS			(REG_FIFO_ELEMENT_DWORDS * REG_FIFO_DEPTH_ELEMENTS)

#ifndef IGU_FIFO_ELEMENT_DWORDS
#define IGU_FIFO_ELEMENT_DWORDS		4
#endif
#define IGU_FIFO_DEPTH_ELEMENTS		64
#define IGU_FIFO_DEPTH_DWORDS			(IGU_FIFO_ELEMENT_DWORDS * IGU_FIFO_DEPTH_ELEMENTS)

#define SEMI_SYNC_FIFO_POLLING_DELAY_MS	5
#define SEMI_SYNC_FIFO_POLLING_COUNT	20

#ifndef PROTECTION_OVERRIDE_ELEMENT_DWORDS
#define PROTECTION_OVERRIDE_ELEMENT_DWORDS 2
#endif
#define PROTECTION_OVERRIDE_DEPTH_ELEMENTS 20
#define PROTECTION_OVERRIDE_DEPTH_DWORDS   	(PROTECTION_OVERRIDE_DEPTH_ELEMENTS 	* PROTECTION_OVERRIDE_ELEMENT_DWORDS)

#define MCP_SPAD_TRACE_OFFSIZE_ADDR		(MCP_REG_SCRATCH + 	OFFSETOF(struct static_init, sections[SPAD_SECTION_TRACE]))

#define MAX_SW_PLTAFORM_STR_SIZE	64

#define EMPTY_FW_VERSION_STR		"???_???_???_???"
#define EMPTY_FW_IMAGE_STR		"???????????????"


/***************************** Constant Arrays *******************************/

/* DBG block framing mode definitions, in descending preference order */
static struct framing_mode_defs s_framing_mode_defs[4] = {
	{ DBG_BUS_FRAME_MODE_4ST, 0x0, 0xf, DBG_BUS_SEMI_FRAME_MODE_4FAST, 10 },
	{ DBG_BUS_FRAME_MODE_4HW, 0xf, 0x0, DBG_BUS_SEMI_FRAME_MODE_4SLOW, 10 },
	{ DBG_BUS_FRAME_MODE_2ST_2HW, 0x3, 0xc, DBG_BUS_SEMI_FRAME_MODE_2FAST_2SLOW, 10 },
	{ DBG_BUS_FRAME_MODE_1ST_3HW, 0x7, 0x8, DBG_BUS_SEMI_FRAME_MODE_1FAST_3SLOW, 10 }
};

/* Chip constant definitions array */
static struct chip_defs s_chip_defs[MAX_CHIP_IDS] = {
	{ "bb", 4, DBG_BUS_NUM_FRAME_MODES, PSWRQ2_REG_ILT_MEMORY_SIZE_BB / 2, s_framing_mode_defs },
	{ "ah", 4, DBG_BUS_NUM_FRAME_MODES, PSWRQ2_REG_ILT_MEMORY_SIZE_K2 / 2, s_framing_mode_defs }
};

/* 256b framing mode definitions */
static struct framing_mode_defs s_256b_framing_mode = { DBG_BUS_FRAME_MODE_8HW, 0xf, 0x0, 0, 10 };

/* Storm constant definitions array */
static struct storm_defs s_storm_defs[] = {
	/* Tstorm */
	{	'T', BLOCK_TSEM,
		{ DBG_BUS_CLIENT_RBCT, DBG_BUS_CLIENT_RBCT }, true,
		TSEM_REG_FAST_MEMORY,
		TSEM_REG_DBG_FRAME_MODE, TSEM_REG_SLOW_DBG_ACTIVE,
		TSEM_REG_SLOW_DBG_MODE, TSEM_REG_DBG_MODE1_CFG,
		TSEM_REG_SYNC_DBG_EMPTY, TSEM_REG_DBG_GPRE_VECT,
		TCM_REG_CTX_RBC_ACCS,
		{ TCM_REG_AGG_CON_CTX, TCM_REG_SM_CON_CTX, TCM_REG_AGG_TASK_CTX, TCM_REG_SM_TASK_CTX },
		{ /* bb */ { 4, 16, 2, 4 },
		  /* k2 */ { 4, 16, 2, 4 } } },

	/* Mstorm */
	{	'M', BLOCK_MSEM,
		{ DBG_BUS_CLIENT_RBCT, DBG_BUS_CLIENT_RBCM }, false,
		MSEM_REG_FAST_MEMORY,
		MSEM_REG_DBG_FRAME_MODE, MSEM_REG_SLOW_DBG_ACTIVE,
		MSEM_REG_SLOW_DBG_MODE, MSEM_REG_DBG_MODE1_CFG,
		MSEM_REG_SYNC_DBG_EMPTY, MSEM_REG_DBG_GPRE_VECT,
		MCM_REG_CTX_RBC_ACCS,
		{ MCM_REG_AGG_CON_CTX, MCM_REG_SM_CON_CTX, MCM_REG_AGG_TASK_CTX, MCM_REG_SM_TASK_CTX },
		{ /* bb */ { 1, 10, 2, 7 },
		  /* k2 */ { 1, 10, 2, 7 } } },

	/* Ustorm */
	{	'U', BLOCK_USEM,
		{ DBG_BUS_CLIENT_RBCU, DBG_BUS_CLIENT_RBCU }, false,
		USEM_REG_FAST_MEMORY,
		USEM_REG_DBG_FRAME_MODE, USEM_REG_SLOW_DBG_ACTIVE,
		USEM_REG_SLOW_DBG_MODE, USEM_REG_DBG_MODE1_CFG,
		USEM_REG_SYNC_DBG_EMPTY, USEM_REG_DBG_GPRE_VECT,
		UCM_REG_CTX_RBC_ACCS,
		{ UCM_REG_AGG_CON_CTX, UCM_REG_SM_CON_CTX, UCM_REG_AGG_TASK_CTX, UCM_REG_SM_TASK_CTX },
		{ /* bb */ { 2, 13, 3, 3 },
		  /* k2 */ { 2, 13, 3, 3 } } },

	/* Xstorm */
	{	'X', BLOCK_XSEM,
		{ DBG_BUS_CLIENT_RBCX, DBG_BUS_CLIENT_RBCX }, false,
		XSEM_REG_FAST_MEMORY,
		XSEM_REG_DBG_FRAME_MODE, XSEM_REG_SLOW_DBG_ACTIVE,
		XSEM_REG_SLOW_DBG_MODE, XSEM_REG_DBG_MODE1_CFG,
		XSEM_REG_SYNC_DBG_EMPTY, XSEM_REG_DBG_GPRE_VECT,
		XCM_REG_CTX_RBC_ACCS,
		{ XCM_REG_AGG_CON_CTX, XCM_REG_SM_CON_CTX, 0, 0 },
		{ /* bb */ { 9, 15, 0, 0 },
		  /* k2 */ { 9, 15, 0, 0 } } },

	/* Ystorm */
	{	'Y', BLOCK_YSEM,
		{ DBG_BUS_CLIENT_RBCX, DBG_BUS_CLIENT_RBCY }, false,
		YSEM_REG_FAST_MEMORY,
		YSEM_REG_DBG_FRAME_MODE, YSEM_REG_SLOW_DBG_ACTIVE,
		YSEM_REG_SLOW_DBG_MODE, YSEM_REG_DBG_MODE1_CFG,
		YSEM_REG_SYNC_DBG_EMPTY, YSEM_REG_DBG_GPRE_VECT,
		YCM_REG_CTX_RBC_ACCS,
		{ YCM_REG_AGG_CON_CTX, YCM_REG_SM_CON_CTX, YCM_REG_AGG_TASK_CTX, YCM_REG_SM_TASK_CTX },
		{ /* bb */ { 2, 3, 2, 12 },
		  /* k2 */ { 2, 3, 2, 12 } } },

	/* Pstorm */
	{	'P', BLOCK_PSEM,
		{ DBG_BUS_CLIENT_RBCS, DBG_BUS_CLIENT_RBCS }, true,
		PSEM_REG_FAST_MEMORY,
		PSEM_REG_DBG_FRAME_MODE, PSEM_REG_SLOW_DBG_ACTIVE,
		PSEM_REG_SLOW_DBG_MODE, PSEM_REG_DBG_MODE1_CFG,
		PSEM_REG_SYNC_DBG_EMPTY, PSEM_REG_DBG_GPRE_VECT,
		PCM_REG_CTX_RBC_ACCS,
		{ 0, PCM_REG_SM_CON_CTX, 0, 0 },
		{ /* bb */ { 0, 10, 0, 0 },
		  /* k2 */ { 0, 10, 0, 0 } } },
};

/* Constraint operation types */
static struct dbg_bus_constraint_op_defs s_constraint_op_defs[] = {
	/* DBG_BUS_CONSTRAINT_OP_EQ */
	{ 0, false },

	/* DBG_BUS_CONSTRAINT_OP_NE */
	{ 5, false },

	/* DBG_BUS_CONSTRAINT_OP_LT */
	{ 1, false },

	/* DBG_BUS_CONSTRAINT_OP_LTC */
	{ 1, true },

	/* DBG_BUS_CONSTRAINT_OP_LE */
	{ 2, false },

	/* DBG_BUS_CONSTRAINT_OP_LEC */
	{ 2, true },

	/* DBG_BUS_CONSTRAINT_OP_GT */
	{ 4, false },

	/* DBG_BUS_CONSTRAINT_OP_GTC */
	{ 4, true },

	/* DBG_BUS_CONSTRAINT_OP_GE */
	{ 3, false },

	/* DBG_BUS_CONSTRAINT_OP_GEC */
	{ 3, true }
};

static const char* s_dbg_target_names[] = {
	/* DBG_BUS_TARGET_ID_INT_BUF */
	"int-buf",

	/* DBG_BUS_TARGET_ID_NIG */
	"nw",

	/* DBG_BUS_TARGET_ID_PCI */
	"pci-buf"
};

static struct storm_mode_defs s_storm_mode_defs[] = {
	/* DBG_BUS_STORM_MODE_PRINTF */
	{ "printf", true, 0, 0, 0, { true, true } },

	/* DBG_BUS_STORM_MODE_PRAM_ADDR */
	{ "pram_addr", true, 1, 0, 0, { true, true } },

	/* DBG_BUS_STORM_MODE_DRA_RW */
	{ "dra_rw", true, 2, SEM_FAST_REG_DBG_MODE23_SRC_DISABLE, 0x0, { true, true } },

	/* DBG_BUS_STORM_MODE_DRA_W */
	{ "dra_w", true, 3, SEM_FAST_REG_DBG_MODE23_SRC_DISABLE, 0x0, { true, true } },

	/* DBG_BUS_STORM_MODE_LD_ST_ADDR */
	{ "ld_st_addr", true, 4, SEM_FAST_REG_DBG_MODE4_SRC_DISABLE, 0x0, { true, true } },

	/* DBG_BUS_STORM_MODE_DRA_FSM */
	{ "dra_fsm", true, 5, 0, 0, { true, true } },

	/* DBG_BUS_STORM_MODE_FAST_DBGMUX */
	{ "fast_dbgmux", true, 8, 0, 0, { false, false } },

	/* DBG_BUS_STORM_MODE_RH */
	{ "rh", true, 6, SEM_FAST_REG_DBG_MODE6_SRC_DISABLE, 0x10, { true, true } },

	/* DBG_BUS_STORM_MODE_RH_WITH_STORE */
	{ "rh_with_store", true, 6, SEM_FAST_REG_DBG_MODE6_SRC_DISABLE, 0x0, { true, true } },

	/* DBG_BUS_STORM_MODE_FOC */
	{ "foc", false, 1, 0, 0, { true, true } },

	/* DBG_BUS_STORM_MODE_EXT_STORE */
	{ "ext_store", false, 3, 0, 0, { true, true } }
};

static struct hw_type_defs s_hw_type_defs[] = {
	/* HW_TYPE_ASIC */
	{ "asic", 1, 256, 32768 },

	/* HW_TYPE_EMUL_FULL */
	{ "emul_full", 2000, 8, 4096 },

	/* HW_TYPE_EMUL_REDUCED */
	{ "emul_reduced", 2000, 8, 4096 },

	/* HW_TYPE_FPGA */
	{ "fpga", 200, 32, 8192 },

	/* PLATFORM_CHIPSIM */
	{ "chipsim", 0, 0, 32768 }
};

static struct grc_param_defs s_grc_param_defs[] = {
	/* DBG_GRC_PARAM_DUMP_TSTORM */
	{ { 1, 1 }, 0, 1, false, false, 1, { 1, 1 } },

	/* DBG_GRC_PARAM_DUMP_MSTORM */
	{ { 1, 1 }, 0, 1, false, false, 1, { 1, 1 } },

	/* DBG_GRC_PARAM_DUMP_USTORM */
	{ { 1, 1 }, 0, 1, false, false, 1, { 1, 1 } },

	/* DBG_GRC_PARAM_DUMP_XSTORM */
	{ { 1, 1 }, 0, 1, false, false, 1, { 1, 1 } },

	/* DBG_GRC_PARAM_DUMP_YSTORM */
	{ { 1, 1 }, 0, 1, false, false, 1, { 1, 1 } },

	/* DBG_GRC_PARAM_DUMP_PSTORM */
	{ { 1, 1 }, 0, 1, false, false, 1, { 1, 1 } },

	/* DBG_GRC_PARAM_DUMP_REGS */
	{ { 1, 1 }, 0, 1, false, false, 0, { 1, 1 } },

	/* DBG_GRC_PARAM_DUMP_RAM */
	{ { 1, 1 }, 0, 1, false, false, 0, { 1, 1 } },

	/* DBG_GRC_PARAM_DUMP_PBUF */
	{ { 1, 1 }, 0, 1, false, false, 0, { 1, 1 } },

	/* DBG_GRC_PARAM_DUMP_IOR */
	{ { 0, 0 }, 0, 1, false, false, 0, { 1, 1 } },

	/* DBG_GRC_PARAM_DUMP_VFC */
	{ { 0, 0 }, 0, 1, false, false, 0, { 1, 1 } },

	/* DBG_GRC_PARAM_DUMP_CM_CTX */
	{ { 1, 1 }, 0, 1, false, false, 0, { 1, 1 } },

	/* DBG_GRC_PARAM_DUMP_ILT */
	{ { 1, 1 }, 0, 1, false, false, 0, { 1, 1 } },

	/* DBG_GRC_PARAM_DUMP_RSS */
	{ { 1, 1 }, 0, 1, false, false, 0, { 1, 1 } },

	/* DBG_GRC_PARAM_DUMP_CAU */
	{ { 1, 1 }, 0, 1, false, false, 0, { 1, 1 } },

	/* DBG_GRC_PARAM_DUMP_QM */
	{ { 1, 1 }, 0, 1, false, false, 0, { 1, 1 } },

	/* DBG_GRC_PARAM_DUMP_MCP */
	{ { 1, 1 }, 0, 1, false, false, 0, { 1, 1 } },

	/* DBG_GRC_PARAM_DUMP_DORQ */
	{ { 1, 1 }, 0, 1, false, false, 0, { 1, 1 } },

	/* DBG_GRC_PARAM_DUMP_CFC */
	{ { 1, 1 }, 0, 1, false, false, 0, { 1, 1 } },

	/* DBG_GRC_PARAM_DUMP_IGU */
	{ { 1, 1 }, 0, 1, false, false, 0, { 1, 1 } },

	/* DBG_GRC_PARAM_DUMP_BRB */
	{ { 0, 0 }, 0, 1, false, false, 0, { 1, 1 } },

	/* DBG_GRC_PARAM_DUMP_BTB */
	{ { 0, 0 }, 0, 1, false, false, 0, { 1, 1 } },

	/* DBG_GRC_PARAM_DUMP_BMB */
	{ { 0, 0 }, 0, 1, false, false, 0, { 0, 0 } },

	/* DBG_GRC_PARAM_RESERVED1 */
	{ { 0, 0 }, 0, 1, false, false, 0, { 0, 0 } },

	/* DBG_GRC_PARAM_DUMP_MULD */
	{ { 1, 1 }, 0, 1, false, false, 0, { 1, 1 } },

	/* DBG_GRC_PARAM_DUMP_PRS */
	{ { 1, 1 }, 0, 1, false, false, 0, { 1, 1 } },

	/* DBG_GRC_PARAM_DUMP_DMAE */
	{ { 1, 1 }, 0, 1, false, false, 0, { 1, 1 } },

	/* DBG_GRC_PARAM_DUMP_TM */
	{ { 1, 1 }, 0, 1, false, false, 0, { 1, 1 } },

	/* DBG_GRC_PARAM_DUMP_SDM */
	{ { 1, 1 }, 0, 1, false, false, 0, { 1, 1 } },

	/* DBG_GRC_PARAM_DUMP_DIF */
	{ { 1, 1 }, 0, 1, false, false, 0, { 1, 1 } },

	/* DBG_GRC_PARAM_DUMP_STATIC */
	{ { 1, 1 }, 0, 1, false, false, 0, { 1, 1 } },

	/* DBG_GRC_PARAM_UNSTALL */
	{ { 0, 0 }, 0, 1, false, false, 0, { 0, 0 } },

	/* DBG_GRC_PARAM_RESERVED2 */
	{ { 0, 0 }, 0, 1, false, false, 0, { 0, 0 } },

	/* DBG_GRC_PARAM_MCP_TRACE_META_SIZE */
	{ { 0, 0 }, 1, 0xffffffff, false, true, 0, { 0, 0 } },

	/* DBG_GRC_PARAM_EXCLUDE_ALL */
	{ { 0, 0 }, 0, 1, true, false, 0, { 0, 0 } },

	/* DBG_GRC_PARAM_CRASH */
	{ { 0, 0 }, 0, 1, true, false, 0, { 0, 0 } },

	/* DBG_GRC_PARAM_PARITY_SAFE */
	{ { 0, 0 }, 0, 1, false, false, 0, { 0, 0 } },

	/* DBG_GRC_PARAM_DUMP_CM */
	{ { 1, 1 }, 0, 1, false, false, 0, { 1, 1 } },

	/* DBG_GRC_PARAM_DUMP_PHY */
	{ { 0, 0 }, 0, 1, false, false, 0, { 0, 0 } },

	/* DBG_GRC_PARAM_NO_MCP */
	{ { 0, 0 }, 0, 1, false, false, 0, { 0, 0 } },

	/* DBG_GRC_PARAM_NO_FW_VER */
	{ { 0, 0 }, 0, 1, false, false, 0, { 0, 0 } },

	/* DBG_GRC_PARAM_RESERVED3 */
	{ { 0, 0 }, 0, 1, false, false, 0, { 0, 0 } },

	/* DBG_GRC_PARAM_DUMP_MCP_HW_DUMP */
	{ { 0, 1 }, 0, 1, false, false, 0, { 0, 1 } },

	/* DBG_GRC_PARAM_DUMP_ILT_CDUC */
	{ { 1, 1 }, 0, 1, false, false, 0, { 0, 0 } },

	/* DBG_GRC_PARAM_DUMP_ILT_CDUT */
	{ { 1, 1 }, 0, 1, false, false, 0, { 0, 0 } },

	/* DBG_GRC_PARAM_DUMP_CAU_EXT */
	{ { 0, 0 }, 0, 1, false, false, 0, { 1, 1 } }
};

static struct rss_mem_defs s_rss_mem_defs[] = {
	{ "rss_mem_cid", "rss_cid", 0, 32,
	{ 256, 320 } },

	{ "rss_mem_key_msb", "rss_key", 1024, 256,
	{ 128, 208 } },

	{ "rss_mem_key_lsb", "rss_key", 2048, 64,
	{ 128, 208 } },

	{ "rss_mem_info", "rss_info", 3072, 16,
	{ 128, 208 } },

	{ "rss_mem_ind", "rss_ind", 4096, 16,
	{ 16384, 26624 } }
};

static struct vfc_ram_defs s_vfc_ram_defs[] = {
	{ "vfc_ram_tt1", "vfc_ram", 0, 512 },
	{ "vfc_ram_mtt2", "vfc_ram", 512, 128 },
	{ "vfc_ram_stt2", "vfc_ram", 640, 32 },
	{ "vfc_ram_ro_vect", "vfc_ram", 672, 32 }
};

static struct big_ram_defs s_big_ram_defs[] = {
	{ "BRB", MEM_GROUP_BRB_MEM, MEM_GROUP_BRB_RAM, DBG_GRC_PARAM_DUMP_BRB,
	  BRB_REG_BIG_RAM_ADDRESS, BRB_REG_BIG_RAM_DATA, MISC_REG_BLOCK_256B_EN, { 0, 0 },
	  { 153600, 180224 } },

	{ "BTB", MEM_GROUP_BTB_MEM, MEM_GROUP_BTB_RAM, DBG_GRC_PARAM_DUMP_BTB,
	  BTB_REG_BIG_RAM_ADDRESS, BTB_REG_BIG_RAM_DATA, MISC_REG_BLOCK_256B_EN, { 0, 1 },
	  { 92160, 117760 } },

	{ "BMB", MEM_GROUP_BMB_MEM, MEM_GROUP_BMB_RAM, DBG_GRC_PARAM_DUMP_BMB,
	  BMB_REG_BIG_RAM_ADDRESS, BMB_REG_BIG_RAM_DATA, MISCS_REG_BLOCK_256B_EN, { 0, 0 },
	  { 36864, 36864 } }
};

static struct rbc_reset_defs s_rbc_reset_defs[] = {
	{ MISCS_REG_RESET_PL_HV, { 0x0, 0x400 } },
	{ MISC_REG_RESET_PL_PDA_VMAIN_1, { 0x4404040, 0x4404040 } },
	{ MISC_REG_RESET_PL_PDA_VMAIN_2, { 0x7, 0x7c00007 } },
	{ MISC_REG_RESET_PL_PDA_VAUX, { 0x2, 0x2 } },
};

static struct phy_defs s_phy_defs[] = {
	{ "nw_phy",	   NWS_REG_NWS_CMU_K2, PHY_NW_IP_REG_PHY0_TOP_TBUS_ADDR_7_0_K2, PHY_NW_IP_REG_PHY0_TOP_TBUS_ADDR_15_8_K2, PHY_NW_IP_REG_PHY0_TOP_TBUS_DATA_7_0_K2, PHY_NW_IP_REG_PHY0_TOP_TBUS_DATA_11_8_K2 },
	{ "sgmii_phy", MS_REG_MS_CMU_K2, PHY_SGMII_IP_REG_AHB_CMU_CSR_0_X132_K2, PHY_SGMII_IP_REG_AHB_CMU_CSR_0_X133_K2, PHY_SGMII_IP_REG_AHB_CMU_CSR_0_X130_K2, PHY_SGMII_IP_REG_AHB_CMU_CSR_0_X131_K2 },
	{ "pcie_phy0", PHY_PCIE_REG_PHY0_K2, PHY_PCIE_IP_REG_AHB_CMU_CSR_0_X132_K2, PHY_PCIE_IP_REG_AHB_CMU_CSR_0_X133_K2, PHY_PCIE_IP_REG_AHB_CMU_CSR_0_X130_K2, PHY_PCIE_IP_REG_AHB_CMU_CSR_0_X131_K2 },
	{ "pcie_phy1", PHY_PCIE_REG_PHY1_K2, PHY_PCIE_IP_REG_AHB_CMU_CSR_0_X132_K2, PHY_PCIE_IP_REG_AHB_CMU_CSR_0_X133_K2, PHY_PCIE_IP_REG_AHB_CMU_CSR_0_X130_K2, PHY_PCIE_IP_REG_AHB_CMU_CSR_0_X131_K2 },
};

static struct split_type_defs s_split_type_defs[] = {
	/* SPLIT_TYPE_NONE */
	{ "eng" },

	/* SPLIT_TYPE_PORT */
	{ "port" },

	/* SPLIT_TYPE_PF */
	{ "pf" },

	/* SPLIT_TYPE_PORT_PF */
	{ "port" },

	/* SPLIT_TYPE_VF */
	{"vf"}
};

/* The order of indexes that should be applied to a PCI buffer line */
static const u8 s_pci_buf_line_ind[PCI_BUF_LINE_SIZE_IN_DWORDS] = { 1, 0, 3, 2, 5, 4, 7, 6 };

/******************************** Variables **********************************/

/* The version of the calling app */
static u32 s_app_ver;

/**************************** Private Functions ******************************/

static void ecore_static_asserts(void)
{
	CHECK_ARR_SIZE(s_big_ram_defs, NUM_BIG_RAM_TYPES);
	CHECK_ARR_SIZE(s_vfc_ram_defs, NUM_VFC_RAM_TYPES);
	CHECK_ARR_SIZE(s_rss_mem_defs, NUM_RSS_MEM_TYPES);
	CHECK_ARR_SIZE(s_chip_defs, MAX_CHIP_IDS);
	CHECK_ARR_SIZE(s_hw_type_defs, MAX_HW_TYPES);
	CHECK_ARR_SIZE(s_storm_defs, MAX_DBG_STORMS);
	CHECK_ARR_SIZE(s_constraint_op_defs, MAX_DBG_BUS_CONSTRAINT_OPS);
	CHECK_ARR_SIZE(s_dbg_target_names, MAX_DBG_BUS_TARGETS);
	CHECK_ARR_SIZE(s_storm_mode_defs, MAX_DBG_BUS_STORM_MODES);
	CHECK_ARR_SIZE(s_grc_param_defs, MAX_DBG_GRC_PARAMS);
}

/* Reads and returns a single dword from the specified unaligned buffer. */
static u32 ecore_read_unaligned_dword(u8 *buf)
{
	u32 dword;

	OSAL_MEMCPY((u8 *)&dword, buf, sizeof(dword));
	return dword;
}

/* Returns the difference in bytes between the specified physical addresses.
 * Assumes that the first address is bigger then the second, and that the
 * difference is a 32-bit value.
 */
static u32 ecore_phys_addr_diff(struct dbg_bus_mem_addr *a,
				struct dbg_bus_mem_addr *b)
{
	return a->hi == b->hi ? a->lo - b->lo : b->lo - a->lo;
}

/* Returns the number of bits in the specified mask. */
static u8 ecore_count_mask_bits(u32 mask, u8 mask_bit_width)
{
	u8 i, count = 0;

	for (i = 0; i < mask_bit_width; i++)
		if (mask & (1 << i))
			count++;

	return count;
}

/* Sets the value of the specified GRC param */
static void ecore_grc_set_param(struct ecore_hwfn *p_hwfn,
				enum dbg_grc_params grc_param,
				u32 val)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;

	dev_data->grc.param_val[grc_param] = val;
}

/* Returns the value of the specified GRC param */
static u32 ecore_grc_get_param(struct ecore_hwfn *p_hwfn,
			       enum dbg_grc_params grc_param)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;

	return dev_data->grc.param_val[grc_param];
}

/* Initializes the GRC parameters */
static void ecore_dbg_grc_init_params(struct ecore_hwfn *p_hwfn)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;

	if (!dev_data->grc.params_initialized) {
		ecore_dbg_grc_set_params_default(p_hwfn);
		dev_data->grc.params_initialized = 1;
	}
}

/* Sets pointer and size for the specified binary buffer type */
static void ecore_set_dbg_bin_buf(struct ecore_hwfn *p_hwfn,
				  enum bin_dbg_buffer_type bufType,
				  const u32 *ptr,
				  u32 size)
{
	struct virt_mem_desc *buf = &p_hwfn->dbg_arrays[bufType];
	buf->ptr = (void*)ptr;
	buf->size = size;
}

/* Initializes debug data for the specified device */
static enum dbg_status ecore_dbg_dev_init(struct ecore_hwfn *p_hwfn)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	u8 num_pfs = 0, max_pfs_per_port = 0;

	if (dev_data->initialized)
		return DBG_STATUS_OK;

	if (!s_app_ver)
		return DBG_STATUS_APP_VERSION_NOT_SET;

	/* Set chip */
	if (ECORE_IS_K2(p_hwfn->p_dev)) {
		dev_data->chip_id = CHIP_K2;
		dev_data->mode_enable[MODE_K2] = 1;
		dev_data->num_vfs = MAX_NUM_VFS_K2;
		num_pfs = MAX_NUM_PFS_K2;
		max_pfs_per_port = MAX_NUM_PFS_K2 / 2;
	} else if (ECORE_IS_BB_B0(p_hwfn->p_dev)) {
		dev_data->chip_id = CHIP_BB;
		dev_data->mode_enable[MODE_BB] = 1;
		dev_data->num_vfs = MAX_NUM_VFS_BB;
		num_pfs = MAX_NUM_PFS_BB;
		max_pfs_per_port = MAX_NUM_PFS_BB;
	} else {
		return DBG_STATUS_UNKNOWN_CHIP;
	}

	/* Set HW type */
#ifdef ASIC_ONLY
	dev_data->hw_type = HW_TYPE_ASIC;
	dev_data->mode_enable[MODE_ASIC] = 1;
#else
	if (CHIP_REV_IS_ASIC(p_hwfn->p_dev)) {
		dev_data->hw_type = HW_TYPE_ASIC;
		dev_data->mode_enable[MODE_ASIC] = 1;
	} else if (CHIP_REV_IS_EMUL(p_hwfn->p_dev)) {
		if (p_hwfn->p_dev->b_is_emul_full) {
			dev_data->hw_type = HW_TYPE_EMUL_FULL;
			dev_data->mode_enable[MODE_EMUL_FULL] = 1;
		} else {
			dev_data->hw_type = HW_TYPE_EMUL_REDUCED;
			dev_data->mode_enable[MODE_EMUL_REDUCED] = 1;
		}
	} else if (CHIP_REV_IS_FPGA(p_hwfn->p_dev)) {
		dev_data->hw_type = HW_TYPE_FPGA;
		dev_data->mode_enable[MODE_FPGA] = 1;
	}
#ifdef CHIPSIM_SUPPORT
	else if (HWAL_IS_CHIPSIM(p_hwfn)) {
		dev_data->hw_type = PLATFORM_CHIPSIM;
		dev_data->mode_enable[MODE_ASIC] = 1;
	}
#endif /* CHIPSIM_SUPPORT */
	else {
		return DBG_STATUS_UNKNOWN_CHIP;
	}
#endif /* ASIC_ONLY */

	/* Set port mode */
	switch (p_hwfn->p_dev->num_ports_in_engine) {
	case 1:
		dev_data->mode_enable[MODE_PORTS_PER_ENG_1] = 1;
		break;
	case 2:
		dev_data->mode_enable[MODE_PORTS_PER_ENG_2] = 1;
		break;
	case 4:
		dev_data->mode_enable[MODE_PORTS_PER_ENG_4] = 1;
		break;
	}

	/* Set 100G mode */
	if (ECORE_IS_CMT(p_hwfn->p_dev))
		dev_data->mode_enable[MODE_100G] = 1;

	/* Set number of ports */
	if (dev_data->mode_enable[MODE_PORTS_PER_ENG_1] || dev_data->mode_enable[MODE_100G])
		dev_data->num_ports = 1;
	else if (dev_data->mode_enable[MODE_PORTS_PER_ENG_2])
		dev_data->num_ports = 2;
	else if (dev_data->mode_enable[MODE_PORTS_PER_ENG_4])
		dev_data->num_ports = 4;

	/* Set number of PFs per port */
	dev_data->num_pfs_per_port = OSAL_MIN_T(u32, num_pfs / dev_data->num_ports, max_pfs_per_port);

	/* Initializes the GRC parameters */
	ecore_dbg_grc_init_params(p_hwfn);

	dev_data->use_dmae = USE_DMAE;
	dev_data->initialized = 1;

	/* Set debug binary buffers */
#ifndef USE_DBG_BIN_FILE
	ecore_set_dbg_bin_buf(p_hwfn, BIN_BUF_DBG_MODE_TREE, (u32 *)dbg_modes_tree_buf, sizeof(dbg_modes_tree_buf));
	ecore_set_dbg_bin_buf(p_hwfn, BIN_BUF_DBG_DUMP_REG, dump_reg, sizeof(dump_reg));
	ecore_set_dbg_bin_buf(p_hwfn, BIN_BUF_DBG_DUMP_MEM, dump_mem, sizeof(dump_mem));
	ecore_set_dbg_bin_buf(p_hwfn, BIN_BUF_DBG_IDLE_CHK_REGS, idle_chk_regs, sizeof(idle_chk_regs));
	ecore_set_dbg_bin_buf(p_hwfn, BIN_BUF_DBG_IDLE_CHK_IMMS, idle_chk_imms, sizeof(idle_chk_imms));
	ecore_set_dbg_bin_buf(p_hwfn, BIN_BUF_DBG_IDLE_CHK_RULES, idle_chk_rules, sizeof(idle_chk_rules));
	ecore_set_dbg_bin_buf(p_hwfn, BIN_BUF_DBG_ATTN_BLOCKS, attn_block, sizeof(attn_block));
	ecore_set_dbg_bin_buf(p_hwfn, BIN_BUF_DBG_ATTN_REGS, attn_reg, sizeof(attn_reg));
	ecore_set_dbg_bin_buf(p_hwfn, BIN_BUF_DBG_BLOCKS, dbg_blocks, sizeof(dbg_blocks));
	ecore_set_dbg_bin_buf(p_hwfn, BIN_BUF_DBG_BLOCKS_CHIP_DATA, dbg_blocks_per_chip, sizeof(dbg_blocks_per_chip));
	ecore_set_dbg_bin_buf(p_hwfn, BIN_BUF_DBG_BUS_LINES, (u32 *)dbg_bus_lines, sizeof(dbg_bus_lines));
	ecore_set_dbg_bin_buf(p_hwfn, BIN_BUF_DBG_RESET_REGS, dbg_reset_regs, sizeof(dbg_reset_regs));
#endif

	return DBG_STATUS_OK;
}

static const struct dbg_block *get_dbg_block(struct ecore_hwfn *p_hwfn,
					     enum block_id block_id)
{
	return (const struct dbg_block *)p_hwfn->dbg_arrays[BIN_BUF_DBG_BLOCKS].ptr + block_id;
}

static const struct dbg_block_chip *ecore_get_dbg_block_per_chip(struct ecore_hwfn *p_hwfn,
								 enum block_id block_id)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;

	return (const struct dbg_block_chip *)p_hwfn->dbg_arrays[BIN_BUF_DBG_BLOCKS_CHIP_DATA].ptr + block_id * MAX_CHIP_IDS + dev_data->chip_id;
}

/* Returns OSAL_NULL for signature line, latency line and non-existing lines */
static const struct dbg_bus_line *get_dbg_bus_line(struct ecore_hwfn *p_hwfn,
						   enum block_id block_id,
						   u8 line_num)
{
	const struct dbg_block_chip *block;
	bool has_latency_events;

	block = ecore_get_dbg_block_per_chip(p_hwfn, block_id);
	has_latency_events = GET_FIELD(block->flags, DBG_BLOCK_CHIP_HAS_LATENCY_EVENTS);

	if (!line_num || (line_num == 1 && has_latency_events) || line_num >= NUM_DBG_LINES(block))
		return OSAL_NULL;

	return (const struct dbg_bus_line *)p_hwfn->dbg_arrays[BIN_BUF_DBG_BUS_LINES].ptr + block->dbg_bus_lines_offset + line_num - NUM_EXTRA_DBG_LINES(block);
}

static const struct dbg_reset_reg *ecore_get_dbg_reset_reg(struct ecore_hwfn *p_hwfn,
							   u8 reset_reg_id)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;

	return (const struct dbg_reset_reg *)p_hwfn->dbg_arrays[BIN_BUF_DBG_RESET_REGS].ptr + reset_reg_id * MAX_CHIP_IDS + dev_data->chip_id;
}

/* Reads the FW info structure for the specified Storm from the chip,
 * and writes it to the specified fw_info pointer.
 */
static void ecore_read_storm_fw_info(struct ecore_hwfn *p_hwfn,
				     struct ecore_ptt *p_ptt,
				     u8 storm_id,
				     struct fw_info *fw_info)
{
	struct storm_defs *storm = &s_storm_defs[storm_id];
	struct fw_info_location fw_info_location;
	u32 addr, i, *dest;

	OSAL_MEMSET(&fw_info_location, 0, sizeof(fw_info_location));
	OSAL_MEMSET(fw_info, 0, sizeof(*fw_info));

	/* Read first the address that points to fw_info location.
	 * The address is located in the last line of the Storm RAM.
	 */
	addr = storm->sem_fast_mem_addr + SEM_FAST_REG_INT_RAM + DWORDS_TO_BYTES(SEM_FAST_REG_INT_RAM_SIZE) - sizeof(fw_info_location);

	dest = (u32 *)&fw_info_location;

	for (i = 0; i < BYTES_TO_DWORDS(sizeof(fw_info_location)); i++, addr += BYTES_IN_DWORD)
		dest[i] = ecore_rd(p_hwfn, p_ptt, addr);

	/* Read FW version info from Storm RAM */
	if (fw_info_location.size > 0 && fw_info_location.size <= sizeof(*fw_info)) {
		addr = fw_info_location.grc_addr;
		dest = (u32 *)fw_info;
		for (i = 0; i < BYTES_TO_DWORDS(fw_info_location.size); i++, addr += BYTES_IN_DWORD)
			dest[i] = ecore_rd(p_hwfn, p_ptt, addr);
	}
}

/* Dumps the specified string to the specified buffer.
 * Returns the dumped size in bytes.
 */
static u32 ecore_dump_str(char *dump_buf,
			  bool dump,
			  const char *str)
{
	if (dump)
		OSAL_STRCPY(dump_buf, str);

	return (u32)OSAL_STRLEN(str) + 1;
}

/* Dumps zeros to align the specified buffer to dwords.
 * Returns the dumped size in bytes.
 */
static u32 ecore_dump_align(char *dump_buf,
			    bool dump,
			    u32 byte_offset)
{
	u8 offset_in_dword, align_size;

	offset_in_dword = (u8)(byte_offset & 0x3);
	align_size = offset_in_dword ? BYTES_IN_DWORD - offset_in_dword : 0;

	if (dump && align_size)
		OSAL_MEMSET(dump_buf, 0, align_size);

	return align_size;
}

/* Writes the specified string param to the specified buffer.
 * Returns the dumped size in dwords.
 */
static u32 ecore_dump_str_param(u32 *dump_buf,
				bool dump,
				const char *param_name,
				const char *param_val)
{
	char *char_buf = (char *)dump_buf;
	u32 offset = 0;

	/* Dump param name */
	offset += ecore_dump_str(char_buf + offset, dump, param_name);

	/* Indicate a string param value */
	if (dump)
		*(char_buf + offset) = 1;
	offset++;

	/* Dump param value */
	offset += ecore_dump_str(char_buf + offset, dump, param_val);

	/* Align buffer to next dword */
	offset += ecore_dump_align(char_buf + offset, dump, offset);

	return BYTES_TO_DWORDS(offset);
}

/* Writes the specified numeric param to the specified buffer.
 * Returns the dumped size in dwords.
 */
static u32 ecore_dump_num_param(u32 *dump_buf,
				bool dump,
				const char *param_name,
				u32 param_val)
{
	char *char_buf = (char *)dump_buf;
	u32 offset = 0;

	/* Dump param name */
	offset += ecore_dump_str(char_buf + offset, dump, param_name);

	/* Indicate a numeric param value */
	if (dump)
		*(char_buf + offset) = 0;
	offset++;

	/* Align buffer to next dword */
	offset += ecore_dump_align(char_buf + offset, dump, offset);

	/* Dump param value (and change offset from bytes to dwords) */
	offset = BYTES_TO_DWORDS(offset);
	if (dump)
		*(dump_buf + offset) = param_val;
	offset++;

	return offset;
}

/* Reads the FW version and writes it as a param to the specified buffer.
 * Returns the dumped size in dwords.
 */
static u32 ecore_dump_fw_ver_param(struct ecore_hwfn *p_hwfn,
				   struct ecore_ptt *p_ptt,
				   u32 *dump_buf,
				   bool dump)
{
	char fw_ver_str[16] = EMPTY_FW_VERSION_STR;
	char fw_img_str[16] = EMPTY_FW_IMAGE_STR;
	struct fw_info fw_info;
	u32 offset = 0;

	OSAL_MEMSET(&fw_info, 0, sizeof(struct fw_info));

	if (dump && !ecore_grc_get_param(p_hwfn, DBG_GRC_PARAM_NO_FW_VER)) {
		/* Read FW info from chip */
		ecore_read_fw_info(p_hwfn, p_ptt, &fw_info);

		/* Create FW version/image strings */
		if (OSAL_SNPRINTF(fw_ver_str, sizeof(fw_ver_str), "%d_%d_%d_%d", fw_info.ver.num.major, fw_info.ver.num.minor, fw_info.ver.num.rev, fw_info.ver.num.eng) < 0)
			DP_NOTICE(p_hwfn, true, "Unexpected debug error: invalid FW version string\n");
		switch (fw_info.ver.image_id) {
		case FW_IMG_KUKU: OSAL_STRCPY(fw_img_str, "kuku"); break;
		case FW_IMG_MAIN: OSAL_STRCPY(fw_img_str, "main"); break;
		case FW_IMG_L2B: OSAL_STRCPY(fw_img_str, "l2b"); break;
		default: OSAL_STRCPY(fw_img_str, "unknown"); break;
		}
	}

	/* Dump FW version, image and timestamp */
	offset += ecore_dump_str_param(dump_buf + offset, dump, "fw-version", fw_ver_str);
	offset += ecore_dump_str_param(dump_buf + offset, dump, "fw-image", fw_img_str);
	offset += ecore_dump_num_param(dump_buf + offset, dump, "fw-timestamp", fw_info.ver.timestamp);

	return offset;
}

/* Reads the MFW version and writes it as a param to the specified buffer.
 * Returns the dumped size in dwords.
 */
static u32 ecore_dump_mfw_ver_param(struct ecore_hwfn *p_hwfn,
				    struct ecore_ptt *p_ptt,
				    u32 *dump_buf,
				    bool dump)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	char mfw_ver_str[16] = EMPTY_FW_VERSION_STR;

	if (dump && dev_data->hw_type == HW_TYPE_ASIC && !ecore_grc_get_param(p_hwfn, DBG_GRC_PARAM_NO_FW_VER)) {
		u32 public_data_addr, global_section_offsize_addr, global_section_offsize, global_section_addr, mfw_ver;

		/* Find MCP public data GRC address. Needs to be ORed with
		 * MCP_REG_SCRATCH due to a HW bug.
		 */
		public_data_addr = ecore_rd(p_hwfn, p_ptt, MISC_REG_SHARED_MEM_ADDR) | MCP_REG_SCRATCH;

		/* Find MCP public global section offset */
		global_section_offsize_addr = public_data_addr + OFFSETOF(struct mcp_public_data, sections) + sizeof(offsize_t) * PUBLIC_GLOBAL;
		global_section_offsize = ecore_rd(p_hwfn, p_ptt, global_section_offsize_addr);
		global_section_addr = MCP_REG_SCRATCH + (global_section_offsize & OFFSIZE_OFFSET_MASK) * 4;

		/* Read MFW version from MCP public global section */
		mfw_ver = ecore_rd(p_hwfn, p_ptt, global_section_addr + OFFSETOF(struct public_global, mfw_ver));

		/* Dump MFW version param */
		if (OSAL_SNPRINTF(mfw_ver_str, sizeof(mfw_ver_str), "%d_%d_%d_%d", (u8)(mfw_ver >> 24), (u8)(mfw_ver >> 16), (u8)(mfw_ver >> 8), (u8)mfw_ver) < 0)
			DP_NOTICE(p_hwfn, true, "Unexpected debug error: invalid MFW version string\n");
	}

	return ecore_dump_str_param(dump_buf, dump, "mfw-version", mfw_ver_str);
}

/* Reads the chip revision from the chip and writes it as a param to the
 * specified buffer. Returns the dumped size in dwords.
 */
static u32 ecore_dump_chip_revision_param(struct ecore_hwfn *p_hwfn,
					  struct ecore_ptt *p_ptt,
					  u32 *dump_buf,
					  bool dump)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	char param_str[3] = "??";

	if (dev_data->hw_type == HW_TYPE_ASIC) {
		u32 chip_rev, chip_metal;

		chip_rev = ecore_rd(p_hwfn, p_ptt, MISCS_REG_CHIP_REV);
		chip_metal = ecore_rd(p_hwfn, p_ptt, MISCS_REG_CHIP_METAL);

		param_str[0] = 'a' + (u8)chip_rev;
		param_str[1] = '0' + (u8)chip_metal;
	}

	return ecore_dump_str_param(dump_buf, dump, "chip-revision", param_str);
}

/* Writes a section header to the specified buffer.
 * Returns the dumped size in dwords.
 */
static u32 ecore_dump_section_hdr(u32 *dump_buf,
				  bool dump,
				  const char *name,
				  u32 num_params)
{
	return ecore_dump_num_param(dump_buf, dump, name, num_params);
}

/* Writes the common global params to the specified buffer.
 * Returns the dumped size in dwords.
 */
static u32 ecore_dump_common_global_params(struct ecore_hwfn *p_hwfn,
					   struct ecore_ptt *p_ptt,
					   u32 *dump_buf,
					   bool dump,
					   u8 num_specific_global_params)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	char sw_platform_str[MAX_SW_PLTAFORM_STR_SIZE];
	u32 offset = 0;
	u8 num_params;

	/* Fill platform string */
	ecore_set_platform_str(p_hwfn, sw_platform_str, MAX_SW_PLTAFORM_STR_SIZE);

	/* Dump global params section header */
	num_params = NUM_COMMON_GLOBAL_PARAMS + num_specific_global_params + (dev_data->chip_id == CHIP_BB ? 1 : 0);
	offset += ecore_dump_section_hdr(dump_buf + offset, dump, "global_params", num_params);

	/* Store params */
	offset += ecore_dump_fw_ver_param(p_hwfn, p_ptt, dump_buf + offset, dump);
	offset += ecore_dump_mfw_ver_param(p_hwfn, p_ptt, dump_buf + offset, dump);
	offset += ecore_dump_chip_revision_param(p_hwfn, p_ptt, dump_buf + offset, dump);
	offset += ecore_dump_num_param(dump_buf + offset, dump, "tools-version", TOOLS_VERSION);
	offset += ecore_dump_str_param(dump_buf + offset, dump, "chip", s_chip_defs[dev_data->chip_id].name);
	offset += ecore_dump_str_param(dump_buf + offset, dump, "platform", s_hw_type_defs[dev_data->hw_type].name);
	offset += ecore_dump_str_param(dump_buf + offset, dump, "sw-platform", sw_platform_str);
	offset += ecore_dump_num_param(dump_buf + offset, dump, "pci-func", p_hwfn->abs_pf_id);
	offset += ecore_dump_num_param(dump_buf + offset, dump, "epoch", OSAL_GET_EPOCH(p_hwfn));
	if (dev_data->chip_id == CHIP_BB)
		offset += ecore_dump_num_param(dump_buf + offset, dump, "path", ECORE_PATH_ID(p_hwfn));

	return offset;
}

/* Writes the "last" section (including CRC) to the specified buffer at the
 * given offset. Returns the dumped size in dwords.
 */
static u32 ecore_dump_last_section(u32 *dump_buf,
				   u32 offset,
				   bool dump)
{
	u32 start_offset = offset;

	/* Dump CRC section header */
	offset += ecore_dump_section_hdr(dump_buf + offset, dump, "last", 0);

	/* Calculate CRC32 and add it to the dword after the "last" section */
	if (dump)
		*(dump_buf + offset) = ~OSAL_CRC32(0xffffffff, (u8 *)dump_buf, DWORDS_TO_BYTES(offset));

	offset++;

	return offset - start_offset;
}

/* Update blocks reset state  */
static void ecore_update_blocks_reset_state(struct ecore_hwfn *p_hwfn,
					    struct ecore_ptt *p_ptt)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	u32 reg_val[NUM_DBG_RESET_REGS] = { 0 };
	u8 reset_reg_id;
	u32 block_id;

	if (dev_data->hw_type == PLATFORM_CHIPSIM)
		return;

	/* Read reset registers */
	for (reset_reg_id = 0; reset_reg_id < NUM_DBG_RESET_REGS; reset_reg_id++) {
		const struct dbg_reset_reg *reset_reg;
		bool reset_reg_removed;
		u32 reset_reg_addr;

		reset_reg = ecore_get_dbg_reset_reg(p_hwfn, reset_reg_id);
		reset_reg_removed = GET_FIELD(reset_reg->data, DBG_RESET_REG_IS_REMOVED);
		reset_reg_addr = DWORDS_TO_BYTES(GET_FIELD(reset_reg->data, DBG_RESET_REG_ADDR));

		if (!reset_reg_removed)
			reg_val[reset_reg_id] = ecore_rd(p_hwfn, p_ptt, reset_reg_addr);
	}

	/* Check if blocks are in reset */
	for (block_id = 0; block_id < NUM_PHYS_BLOCKS; block_id++) {
		const struct dbg_block_chip *block;
		bool has_reset_reg;
		bool is_removed;

		block = ecore_get_dbg_block_per_chip(p_hwfn, (enum block_id)block_id);
		is_removed = GET_FIELD(block->flags, DBG_BLOCK_CHIP_IS_REMOVED);
		has_reset_reg = GET_FIELD(block->flags, DBG_BLOCK_CHIP_HAS_RESET_REG);

		if (!is_removed && has_reset_reg)
			dev_data->block_in_reset[block_id] = !(reg_val[block->reset_reg_id] & (1 << block->reset_reg_bit_offset));
	}
}

/* is_mode_match recursive function */
static bool ecore_is_mode_match_rec(struct ecore_hwfn *p_hwfn,
				    u16 *modes_buf_offset,
				    u8 rec_depth)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	bool arg1, arg2;
	u8 tree_val;

	if (rec_depth > MAX_RECURSION_DEPTH) {
		DP_NOTICE(p_hwfn, true, "Unexpected error: is_mode_match_rec exceeded the max recursion depth. This is probably due to a corrupt init/debug buffer.\n");
		return false;
	}

	/* Get next element from modes tree buffer */
	tree_val = ((u8 *)p_hwfn->dbg_arrays[BIN_BUF_DBG_MODE_TREE].ptr)[(*modes_buf_offset)++];

	switch (tree_val) {
	case INIT_MODE_OP_NOT:
		return !ecore_is_mode_match_rec(p_hwfn, modes_buf_offset, rec_depth + 1);
	case INIT_MODE_OP_OR:
	case INIT_MODE_OP_AND:
		arg1 = ecore_is_mode_match_rec(p_hwfn, modes_buf_offset, rec_depth + 1);
		arg2 = ecore_is_mode_match_rec(p_hwfn, modes_buf_offset, rec_depth + 1);
		return (tree_val == INIT_MODE_OP_OR) ? (arg1 || arg2) : (arg1 && arg2);
	default: return dev_data->mode_enable[tree_val - MAX_INIT_MODE_OPS] > 0;
	}
}

/* Returns true if the mode (specified using modes_buf_offset) is enabled */
static bool ecore_is_mode_match(struct ecore_hwfn *p_hwfn,
				u16 *modes_buf_offset)
{
	return ecore_is_mode_match_rec(p_hwfn, modes_buf_offset, 0);
}

/* Try to empty the SEMI sync fifo. Must be done after messages output
 * were disabled in all Storms.
 * Returns true if the SEMI sync FIFO was emptied in all Storms, false otherwise.
 */
static bool ecore_empty_sem_sync_fifo(struct ecore_hwfn *p_hwfn,
				      struct ecore_ptt *p_ptt)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	u8 storm_id, num_fifos_to_empty = MAX_DBG_STORMS;
	bool is_fifo_empty[MAX_DBG_STORMS] = { false };
	u32 polling_ms, polling_count = 0;

	polling_ms = SEMI_SYNC_FIFO_POLLING_DELAY_MS * s_hw_type_defs[dev_data->hw_type].delay_factor;

	while (num_fifos_to_empty && polling_count < SEMI_SYNC_FIFO_POLLING_COUNT) {
		for (storm_id = 0; storm_id < MAX_DBG_STORMS; storm_id++) {
			struct storm_defs *storm = &s_storm_defs[storm_id];

			if (is_fifo_empty[storm_id])
				continue;

			/* Check if sync fifo got empty */
			if (dev_data->block_in_reset[storm->sem_block_id] || ecore_rd(p_hwfn, p_ptt, storm->sem_sync_dbg_empty_addr)) {
				is_fifo_empty[storm_id] = true;
				num_fifos_to_empty--;
			}
		}

		/* Check if need to continue polling */
		if (!num_fifos_to_empty)
			break;

		OSAL_MSLEEP(polling_ms);
		polling_count++;
	}

	return !num_fifos_to_empty;
}

/* Enable / disable the Debug block */
static void ecore_bus_enable_dbg_block(struct ecore_hwfn *p_hwfn,
				       struct ecore_ptt *p_ptt,
				       bool enable)
{
	ecore_wr(p_hwfn, p_ptt, DBG_REG_DBG_BLOCK_ON, enable ? 1 : 0);
}

/* Resets the Debug block */
static void ecore_bus_reset_dbg_block(struct ecore_hwfn *p_hwfn,
				      struct ecore_ptt *p_ptt)
{
	u32 reset_reg_addr, old_reset_reg_val, new_reset_reg_val;
	const struct dbg_reset_reg *reset_reg;
	const struct dbg_block_chip *block;

	block = ecore_get_dbg_block_per_chip(p_hwfn, BLOCK_DBG);
	reset_reg = ecore_get_dbg_reset_reg(p_hwfn, block->reset_reg_id);
	reset_reg_addr = DWORDS_TO_BYTES(GET_FIELD(reset_reg->data, DBG_RESET_REG_ADDR));

	old_reset_reg_val = ecore_rd(p_hwfn, p_ptt, reset_reg_addr);
	new_reset_reg_val = old_reset_reg_val & ~(1 << block->reset_reg_bit_offset);

	ecore_wr(p_hwfn, p_ptt, reset_reg_addr, new_reset_reg_val);
	ecore_wr(p_hwfn, p_ptt, reset_reg_addr, old_reset_reg_val);
}

/* Returns the dword mask for the specified block parameters. A dword mask
 * is an 8-bit value containing a bit for each dword in the debug bus cycle,
 * indicating if this dword is recorded (1) or not (0).
 */
static u8 ecore_bus_get_dword_mask(u8 enable_mask,
				   u8 right_shift)
{
	return ((enable_mask | (enable_mask << VALUES_PER_BLOCK)) >> right_shift) & MAX_BLOCK_VALUES_MASK;
}

/* Enable / disable Debug Bus clients according to the specified mask
 * (1 = enable, 0 = disable).
 */
static void ecore_bus_enable_clients(struct ecore_hwfn *p_hwfn,
				     struct ecore_ptt *p_ptt,
				     u32 client_mask)
{
	ecore_wr(p_hwfn, p_ptt, DBG_REG_CLIENT_ENABLE, client_mask);
}

/* Enables the specified Storm for Debug Bus. Assumes a valid Storm ID. */
static void ecore_bus_add_storm_input(struct ecore_hwfn *p_hwfn,
				      struct ecore_ptt *p_ptt,
				      enum dbg_storms storm_id,
				      u8 semi_framing_mode_id)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	u32 base_addr, sem_filter_params = 0;
	struct dbg_bus_storm_data *storm_bus;
	struct storm_mode_defs *storm_mode;
	struct storm_defs *storm;

	storm = &s_storm_defs[storm_id];
	storm_bus = &dev_data->bus.storms[storm_id];
	storm_mode = &s_storm_mode_defs[storm_bus->mode];
	base_addr = storm->sem_fast_mem_addr;

	ecore_wr(p_hwfn, p_ptt, storm->sem_frame_mode_addr, semi_framing_mode_id);

	/* Config SEM */
	if (storm_mode->is_fast_dbg) {
		/* Enable fast debug */
		ecore_wr(p_hwfn, p_ptt, base_addr + SEM_FAST_REG_DEBUG_MODE, storm_mode->id_in_hw);
		ecore_wr(p_hwfn, p_ptt, base_addr + SEM_FAST_REG_DEBUG_ACTIVE, 1);
	} else {
		/* Enable slow debug */
		ecore_wr(p_hwfn, p_ptt, storm->sem_slow_enable_addr, 1);
		ecore_wr(p_hwfn, p_ptt, storm->sem_slow_mode_addr, storm_mode->id_in_hw);
		ecore_wr(p_hwfn, p_ptt, storm->sem_slow_mode1_conf_addr, SEM_SLOW_MODE1_DATA_ENABLE);
	}

	/* Enable messages. Must be done after enabling SEM_FAST_REG_DEBUG_ACTIVE
	 * otherwise messages will be dropped after the SEMI sync fifo is filled.
	 */
	if (storm_mode->src_disable_reg_addr)
		ecore_wr(p_hwfn, p_ptt, base_addr + storm_mode->src_disable_reg_addr, storm_mode->src_enable_val);

	/* Config SEM cid filter */
	if (storm_bus->cid_filter_en) {
		ecore_wr(p_hwfn, p_ptt, base_addr + SEM_FAST_REG_FILTER_CID, storm_bus->cid);
		sem_filter_params |= SEM_FILTER_CID_EN_MASK;
	}

	/* Config SEM eid filter */
	if (storm_bus->eid_filter_en) {
		const union dbg_bus_storm_eid_params *eid_filter = &storm_bus->eid_filter_params;

		if (storm_bus->eid_range_not_mask) {
			ecore_wr(p_hwfn, p_ptt, base_addr + SEM_FAST_REG_EVENT_ID_RANGE_STRT, eid_filter->range.min);
			ecore_wr(p_hwfn, p_ptt, base_addr + SEM_FAST_REG_EVENT_ID_RANGE_END, eid_filter->range.max);
			sem_filter_params |= SEM_FILTER_EID_RANGE_EN_MASK;
		} else {
			ecore_wr(p_hwfn, p_ptt, base_addr + SEM_FAST_REG_FILTER_EVENT_ID, eid_filter->mask.val);
			ecore_wr(p_hwfn, p_ptt, base_addr + SEM_FAST_REG_EVENT_ID_MASK, ~eid_filter->mask.mask);
			sem_filter_params |= SEM_FILTER_EID_MASK_EN_MASK;
		}
	}

	/* Config accumulaed SEM filter parameters (if any) */
	if (sem_filter_params)
		ecore_wr(p_hwfn, p_ptt, base_addr + SEM_FAST_REG_RECORD_FILTER_ENABLE, sem_filter_params);
}

/* Configure the DBG block client mask */
static void ecore_bus_config_client_mask(struct ecore_hwfn *p_hwfn,
					 struct ecore_ptt *p_ptt)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	struct dbg_bus_data *bus = &dev_data->bus;
	u32 block_id, client_mask = 0;
	u8 storm_id;

	/* Update client mask for Storm inputs */
	if (bus->num_enabled_storms)
		for (storm_id = 0; storm_id < MAX_DBG_STORMS; storm_id++) {
			struct storm_defs *storm = &s_storm_defs[storm_id];

			if (bus->storms[storm_id].enabled)
				client_mask |= (1 << storm->dbg_client_id[dev_data->chip_id]);
		}

	/* Update client mask for block inputs */
	if (bus->num_enabled_blocks) {
		for (block_id = 0; block_id < MAX_BLOCK_ID; block_id++) {
			struct dbg_bus_block_data *block_bus;
			const struct dbg_block_chip *block;

			block = ecore_get_dbg_block_per_chip(p_hwfn, (enum block_id)block_id);
			block_bus = &bus->blocks[block_id];

			if (block_bus->enable_mask && block_id != BLOCK_DBG)
				client_mask |= (1 << block->dbg_client_id);
		}
	}

	/* Update client mask for GRC input */
	if (bus->grc_input_en)
		client_mask |= (1 << DBG_BUS_CLIENT_CPU);

	/* Update client mask for timestamp input */
	if (bus->timestamp_input_en)
		client_mask |= (1 << DBG_BUS_CLIENT_TIMESTAMP);

	ecore_bus_enable_clients(p_hwfn, p_ptt, client_mask);
}

/* Chooses and returns a matching DBG block framing mode.
 * If no matching framing mode found, return null.
 */
static struct framing_mode_defs *ecore_bus_get_framing_mode(struct ecore_hwfn *p_hwfn)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	struct dbg_bus_data *bus = &dev_data->bus;
	struct framing_mode_defs *framing_modes;
	u8 num_framing_modes, i;

	if (bus->mode_256b_en)
		return &s_256b_framing_mode;

	num_framing_modes = s_chip_defs[dev_data->chip_id].num_framing_modes;
	framing_modes = s_chip_defs[dev_data->chip_id].framing_modes;

	for (i = 0; i < num_framing_modes; i++) {
		struct framing_mode_defs *framing_mode = &framing_modes[i];

		/* Check if the requested block dwords fit into the framing mode
		 * block dwords. If Storm dwords are requested, check that the
		 * framing mode contains Storm dwords as well.
		 */
		if (((bus->blocks_dword_mask & framing_mode->blocks_dword_mask) == bus->blocks_dword_mask) && (!bus->num_enabled_storms || framing_mode->storms_dword_mask))
			return framing_mode;
	}

	return OSAL_NULL;
}

/* Configure the DBG block Storm data */
static enum dbg_status ecore_bus_config_storm_inputs(struct ecore_hwfn *p_hwfn,
						     struct ecore_ptt *p_ptt)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	struct dbg_bus_data *bus = &dev_data->bus;
	u8 storm_id, i, next_storm_id = 0;
	u32 storm_id_mask = 0;

	/* Check if SEMI sync FIFO is empty */
	for (storm_id = 0; storm_id < MAX_DBG_STORMS; storm_id++) {
		struct dbg_bus_storm_data *storm_bus = &bus->storms[storm_id];
		struct storm_defs *storm = &s_storm_defs[storm_id];

		if (storm_bus->enabled && !ecore_rd(p_hwfn, p_ptt, storm->sem_sync_dbg_empty_addr))
			return DBG_STATUS_SEMI_FIFO_NOT_EMPTY;
	}

	for (storm_id = 0; storm_id < MAX_DBG_STORMS; storm_id++) {
		struct dbg_bus_storm_data *storm_bus = &bus->storms[storm_id];

		if (storm_bus->enabled)
			storm_id_mask |= (storm_bus->hw_id << (storm_id * HW_ID_BITS));
	}

	ecore_wr(p_hwfn, p_ptt, DBG_REG_STORM_ID_NUM, storm_id_mask);

	/* Disable storm stall if recording to internal buffer in one-shot */
	ecore_wr(p_hwfn, p_ptt, DBG_REG_NO_GRANT_ON_FULL, (dev_data->bus.target == DBG_BUS_TARGET_ID_INT_BUF && bus->one_shot_en) ? 0 : 1);

	/* Configure calendar */
	for (i = 0; i < NUM_CALENDAR_SLOTS; i++, next_storm_id = (next_storm_id + 1) % MAX_DBG_STORMS) {
		/* Find next enabled Storm */
		for (; !dev_data->bus.storms[next_storm_id].enabled; next_storm_id = (next_storm_id + 1) % MAX_DBG_STORMS);

		/* Configure calendar slot */
		ecore_wr(p_hwfn, p_ptt, DBG_REG_CALENDAR_SLOT0 + DWORDS_TO_BYTES(i), next_storm_id);
	}

	return DBG_STATUS_OK;
}

/* Assigns block HW IDs to each dword in the debug bus cycles. The HW IDs are
 * written to the block_hw_ids array (expected to be of size
 * DEBUG_BUS_CYCLE_DWORDS).
 */
static enum dbg_status ecore_bus_assign_block_hw_ids(struct ecore_hwfn *p_hwfn,
						     u8 *block_hw_ids)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	struct dbg_bus_data *bus = &dev_data->bus;
	u8 num_block_dwords, max_blocks_hw_ids;
	u8 dword_id, block_id, next_hw_id;
	enum hw_id_alloc_mode hw_id_mode;

	/* Default mode is HW ID per dword */
	hw_id_mode = HW_ID_MODE_PER_DWORD;

	if (bus->blocks_dword_overlap) {
		/* Some of the dwords contain data from multiple blocks -
		 * use a single HW ID.
		 */
		hw_id_mode = HW_ID_MODE_SINGLE;
	} else if (bus->filter_en || bus->trigger_en) {
		if (bus->filter_en) {
			/* Filtering on a single HW ID - use a single HW ID */
			hw_id_mode = HW_ID_MODE_SINGLE;
		} else if (bus->trigger_en) {
			bool multi_constraint_dwords = false, multi_constraint_blocks = false;
			u8 state_id;

			/* In single HW ID filter mode, we must use a single
			 * HW ID for all constraints. Therefore, for each
			 * trigger state, check if there are constraints on
			 * multiple dwords. If there are, do the following:
			 * - If the constraints cover a single block, use HW ID
			 *   per block.
			 * - If the constraints cover multiple blocks - use
			 *   a single HW ID.
			 */
			for (state_id = 0; state_id < bus->next_trigger_state && !multi_constraint_blocks; state_id++) {
				struct dbg_bus_trigger_state_data *trigger_state = &bus->trigger_states[state_id];
				u8 num_constraint_dwords = 0, num_constraint_blocks = 0;

				num_constraint_dwords = ecore_count_mask_bits(trigger_state->constraint_dword_mask, DEBUG_BUS_CYCLE_DWORDS);
				if (num_constraint_dwords <= 1)
					continue;

				multi_constraint_dwords = true;

				/* This state has multiple dwords per constraint -
				 * check if these dwords contain multiple blocks
				 */
				for (block_id = 0; block_id < MAX_BLOCK_ID && !multi_constraint_blocks; block_id++) {
					struct dbg_bus_block_data *block_bus = &bus->blocks[block_id];

					if (trigger_state->constraint_dword_mask & block_bus->dword_mask)
						multi_constraint_blocks = (++num_constraint_blocks > 1);
				}
			}

			if (multi_constraint_dwords)
				hw_id_mode = multi_constraint_blocks ? HW_ID_MODE_SINGLE : HW_ID_MODE_PER_BLOCK;
		}
	}

	/* Find number of remaining HW IDs for blocks, and update the HW ID
	 * allocation mode if the remaining IDs are not sufficient.
	 */
	num_block_dwords = ecore_count_mask_bits(bus->blocks_dword_mask, DEBUG_BUS_CYCLE_DWORDS);
	max_blocks_hw_ids = DEBUG_BUS_CYCLE_DWORDS - bus->num_enabled_storms;

	if (hw_id_mode == HW_ID_MODE_PER_DWORD && num_block_dwords > max_blocks_hw_ids)
		hw_id_mode = HW_ID_MODE_PER_BLOCK;
	if (hw_id_mode == HW_ID_MODE_PER_BLOCK && bus->num_enabled_blocks > max_blocks_hw_ids)
		hw_id_mode = HW_ID_MODE_SINGLE;
	if (hw_id_mode == HW_ID_MODE_SINGLE && !max_blocks_hw_ids)
		return DBG_STATUS_INSUFFICIENT_HW_IDS;

	/* Assign block HW IDs */
	OSAL_MEMSET(block_hw_ids, 0, DEBUG_BUS_CYCLE_DWORDS);
	next_hw_id = bus->num_enabled_storms;
	if (hw_id_mode == HW_ID_MODE_PER_DWORD) {
		for (dword_id = 0; dword_id < DEBUG_BUS_CYCLE_DWORDS; dword_id++)
			if (bus->blocks_dword_mask & (1 << dword_id))
				block_hw_ids[dword_id] = next_hw_id++;
	} else if (hw_id_mode == HW_ID_MODE_PER_BLOCK) {
		for (block_id = 0; block_id < MAX_BLOCK_ID; block_id++) {
			struct dbg_bus_block_data *block_bus = &bus->blocks[block_id];
			if (!block_bus->dword_mask)
				continue;
			for (dword_id = 0; dword_id < DEBUG_BUS_CYCLE_DWORDS; dword_id++)
				if (block_bus->dword_mask & (1 << dword_id))
					block_hw_ids[dword_id] = next_hw_id;
			next_hw_id++;
		}
	} else
		OSAL_MEMSET(block_hw_ids, next_hw_id, DEBUG_BUS_CYCLE_DWORDS);

	return DBG_STATUS_OK;
}

/* Configure the DBG block HW blocks data.
 * The assignment of block HW IDs to each dword in the debug bus cycle is
 * written to the block_hw_ids array (expected to be of size
 * DEBUG_BUS_CYCLE_DWORDS).
 */
static enum dbg_status ecore_bus_config_block_inputs(struct ecore_hwfn *p_hwfn,
						     struct ecore_ptt *p_ptt,
						     u8 *block_hw_ids)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	struct dbg_bus_data *bus = &dev_data->bus;
	u8 dword_id;

	/* Configure HW ID mask */
	bus->hw_id_mask = 0;
	for (dword_id = 0; dword_id < DEBUG_BUS_CYCLE_DWORDS; dword_id++)
		bus->hw_id_mask |= (block_hw_ids[dword_id] << (dword_id * HW_ID_BITS));
	ecore_wr(p_hwfn, p_ptt, DBG_REG_HW_ID_NUM, bus->hw_id_mask);

	/* Configure additional K2 PCIE registers */
	if (dev_data->chip_id == CHIP_K2 && (bus->blocks[BLOCK_PCIE].enable_mask || bus->blocks[BLOCK_PHY_PCIE].enable_mask)) {
		ecore_wr(p_hwfn, p_ptt, PCIE_REG_DBG_REPEAT_THRESHOLD_COUNT_K2, 1);
		ecore_wr(p_hwfn, p_ptt, PCIE_REG_DBG_FW_TRIGGER_ENABLE_K2, 1);
	}

	return DBG_STATUS_OK;
}

/* Configure the DBG block filters and triggers.
 * The assignment of block HW IDs to each dword in the debug bus cycle is
 * written to the block_hw_ids array (expected to be of size
 * DEBUG_BUS_CYCLE_DWORDS).
 */
static enum dbg_status ecore_bus_config_filters_and_triggers(struct ecore_hwfn *p_hwfn,
							     struct ecore_ptt *p_ptt,
							     struct framing_mode_defs *framing_mode,
							     u8 *block_hw_ids)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	struct dbg_bus_data *bus = &dev_data->bus;
	enum dbg_bus_filter_types filter_type;

	/* Configure filter type */
	if (bus->filter_en) {
		if (bus->trigger_en) {
			if (bus->filter_pre_trigger)
				filter_type = bus->filter_post_trigger ? DBG_BUS_FILTER_TYPE_ON : DBG_BUS_FILTER_TYPE_PRE;
			else
				filter_type = bus->filter_post_trigger ? DBG_BUS_FILTER_TYPE_POST : DBG_BUS_FILTER_TYPE_OFF;
		} else {
			filter_type = DBG_BUS_FILTER_TYPE_ON;
		}
	} else {
		filter_type = DBG_BUS_FILTER_TYPE_OFF;
	}

	ecore_wr(p_hwfn, p_ptt, DBG_REG_FILTER_ENABLE, filter_type);

	/* Configure the filter HW ID - in Single HW ID mode, there can be
	 * either (1) a single Storm input with no block inputs, or
	 * (2) multiple block inputs with no Storm inputs.
	 */
	if (bus->filter_en) {
		if (!(bus->filter_constraint_dword_mask & bus->blocks_dword_mask) && !(bus->filter_constraint_dword_mask & framing_mode->storms_dword_mask))
			return DBG_STATUS_INVALID_FILTER_TRIGGER_DWORDS;
		if (bus->num_enabled_storms > 1 || (bus->num_enabled_storms == 1 && bus->num_enabled_blocks))
			return DBG_STATUS_FILTER_SINGLE_HW_ID;
		ecore_wr(p_hwfn, p_ptt, DBG_REG_FILTER_ID_NUM, 0);
	}

	/* Configure the HW ID for each trigger state */
	if (bus->trigger_en) {
		u8 state_id;

		for (state_id = 0; state_id < bus->next_trigger_state; state_id++) {
			struct dbg_bus_trigger_state_data *state = &bus->trigger_states[state_id];
			u8 dword_id, hw_id = 0;

			if (state->constraint_dword_mask & framing_mode->storms_dword_mask) {
				/* Triggering on Storm data - use HW ID of
				 * specified Storm.
				 */
				if (bus->num_enabled_storms > 1 || (state->constraint_dword_mask & bus->blocks_dword_mask))
					return DBG_STATUS_TRIGGER_SINGLE_HW_ID;
				if (state->storm_id == MAX_DBG_STORMS)
					return DBG_STATUS_MISSING_TRIGGER_STATE_STORM;
				hw_id = bus->storms[state->storm_id].hw_id;
			} else if (state->constraint_dword_mask & bus->blocks_dword_mask) {
				/* Triggering on block data - all state's
				 * constraints have same HW ID. Use the first.
				 */
				for (dword_id = 0; dword_id < DEBUG_BUS_CYCLE_DWORDS; dword_id++) {
					if (!(state->constraint_dword_mask & (1 << dword_id)))
						continue;
					hw_id = block_hw_ids[dword_id];
					break;
				}
			} else
				return DBG_STATUS_INVALID_FILTER_TRIGGER_DWORDS;

			ecore_wr(p_hwfn, p_ptt, DBG_REG_TRIGGER_STATE_ID_0 + state_id * BYTES_IN_DWORD, hw_id);
		}
	}

	return DBG_STATUS_OK;
}

static void ecore_bus_config_dbg_line(struct ecore_hwfn *p_hwfn,
				      struct ecore_ptt *p_ptt,
				      enum block_id block_id,
				      u8 line_id,
				      u8 enable_mask,
				      u8 right_shift,
				      u8 force_valid_mask,
				      u8 force_frame_mask)
{
	const struct dbg_block_chip *block = ecore_get_dbg_block_per_chip(p_hwfn, block_id);

	ecore_wr(p_hwfn, p_ptt, DWORDS_TO_BYTES(block->dbg_select_reg_addr), line_id);
	ecore_wr(p_hwfn, p_ptt, DWORDS_TO_BYTES(block->dbg_dword_enable_reg_addr), enable_mask);
	ecore_wr(p_hwfn, p_ptt, DWORDS_TO_BYTES(block->dbg_shift_reg_addr), right_shift);
	ecore_wr(p_hwfn, p_ptt, DWORDS_TO_BYTES(block->dbg_force_valid_reg_addr), force_valid_mask);
	ecore_wr(p_hwfn, p_ptt, DWORDS_TO_BYTES(block->dbg_force_frame_reg_addr), force_frame_mask);
}

/* Disable debug bus in all blocks */
static void ecore_bus_disable_blocks(struct ecore_hwfn *p_hwfn,
				     struct ecore_ptt *p_ptt)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	u32 block_id;

	/* Disable all blocks */
	for (block_id = 0; block_id < MAX_BLOCK_ID; block_id++) {
		const struct dbg_block_chip *block_per_chip = ecore_get_dbg_block_per_chip(p_hwfn, (enum block_id)block_id);

		if (GET_FIELD(block_per_chip->flags, DBG_BLOCK_CHIP_IS_REMOVED) || dev_data->block_in_reset[block_id])
			continue;

		/* Disable debug bus */
		if (GET_FIELD(block_per_chip->flags, DBG_BLOCK_CHIP_HAS_DBG_BUS)) {
			u16 modes_buf_offset = GET_FIELD(block_per_chip->dbg_bus_mode.data, DBG_MODE_HDR_MODES_BUF_OFFSET);
			bool eval_mode = GET_FIELD(block_per_chip->dbg_bus_mode.data, DBG_MODE_HDR_EVAL_MODE) > 0;

			if (!eval_mode || ecore_is_mode_match(p_hwfn, &modes_buf_offset))
				ecore_wr(p_hwfn, p_ptt, DWORDS_TO_BYTES(block_per_chip->dbg_dword_enable_reg_addr), 0);
		}
	}
}

/* Disables Debug Bus block inputs */
static enum dbg_status ecore_bus_disable_inputs(struct ecore_hwfn *p_hwfn,
						struct ecore_ptt *p_ptt,
						bool empty_semi_fifos)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	u8 storm_id;

	/* Disable messages output in all Storms */
	for (storm_id = 0; storm_id < MAX_DBG_STORMS; storm_id++) {
		struct storm_defs *storm = &s_storm_defs[storm_id];

		if (dev_data->block_in_reset[storm->sem_block_id])
			continue;
		
		ecore_wr(p_hwfn, p_ptt, storm->sem_fast_mem_addr + SEM_FAST_REG_DBG_MODE23_SRC_DISABLE, SEM_FAST_MODE23_SRC_DISABLE_VAL);
		ecore_wr(p_hwfn, p_ptt, storm->sem_fast_mem_addr + SEM_FAST_REG_DBG_MODE4_SRC_DISABLE, SEM_FAST_MODE4_SRC_DISABLE_VAL);
		ecore_wr(p_hwfn, p_ptt, storm->sem_fast_mem_addr + SEM_FAST_REG_DBG_MODE6_SRC_DISABLE, SEM_FAST_MODE6_SRC_DISABLE_VAL);
	}

	/* Empty SEMI sync FIFO. If we fail to empty the FIFO when recording to
	 * PCI/NW, change recording target to DBG block internal buffer and try
	 * emptying the FIFO again.
	 */
	if (empty_semi_fifos) {
		bool emptied = ecore_empty_sem_sync_fifo(p_hwfn, p_ptt);

		if (!emptied && dev_data->bus.state == DBG_BUS_STATE_RECORDING && dev_data->bus.target != DBG_BUS_TARGET_ID_INT_BUF) {
			ecore_wr(p_hwfn, p_ptt, DBG_REG_DEBUG_TARGET, DBG_BUS_TARGET_ID_INT_BUF);
			emptied = ecore_empty_sem_sync_fifo(p_hwfn, p_ptt);
			if (!emptied)
				DP_NOTICE(p_hwfn, false, "Warning: failed to empty the SEMI sync FIFO. It means that the last few messages from the SEMI could not be sent to the DBG block. This can happen when the DBG block is blocked (e.g. due to a PCI problem).\n");
		}
	}

	/* Disable debug in all Storms */
	for (storm_id = 0; storm_id < MAX_DBG_STORMS; storm_id++) {
		struct storm_defs *storm = &s_storm_defs[storm_id];
		u32 base_addr = storm->sem_fast_mem_addr;

		if (dev_data->block_in_reset[storm->sem_block_id])
			continue;

		ecore_wr(p_hwfn, p_ptt, base_addr + SEM_FAST_REG_DEBUG_ACTIVE, 0);
		ecore_wr(p_hwfn, p_ptt, base_addr + SEM_FAST_REG_RECORD_FILTER_ENABLE, DBG_BUS_FILTER_TYPE_OFF);
		ecore_wr(p_hwfn, p_ptt, storm->sem_frame_mode_addr, DBG_BUS_SEMI_FRAME_MODE_4SLOW);
		ecore_wr(p_hwfn, p_ptt, storm->sem_slow_enable_addr, 0);
	}

	/* Disable all clients */
	ecore_bus_enable_clients(p_hwfn, p_ptt, 0);

	/* Disable debug bus in all blocks */
	ecore_bus_disable_blocks(p_hwfn, p_ptt);

	/* Disable timestamp */
	ecore_wr(p_hwfn, p_ptt, DBG_REG_TIMESTAMP_VALID_EN, 0);

	/* Disable filters and triggers */
	ecore_wr(p_hwfn, p_ptt, DBG_REG_FILTER_ENABLE, DBG_BUS_FILTER_TYPE_OFF);
	ecore_wr(p_hwfn, p_ptt, DBG_REG_TRIGGER_ENABLE, 0);

	return DBG_STATUS_OK;
}

/* Sets a Debug Bus trigger/filter constraint */
static void ecore_bus_set_constraint(struct ecore_hwfn *p_hwfn,
				     struct ecore_ptt *p_ptt,
				     bool is_filter,
				     u8 constraint_id,
				     u8 hw_op_val,
				     u32 data_val,
				     u32 data_mask,
				     u8 frame_bit,
				     u8 frame_mask,
				     u16 dword_offset,
				     u16 range,
				     u8 cyclic_bit,
				     u8 must_bit)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	u32 reg_offset = constraint_id * BYTES_IN_DWORD;
	u8 curr_trigger_state;

	/* For trigger only - set register offset according to state */
	if (!is_filter) {
		curr_trigger_state = dev_data->bus.next_trigger_state - 1;
		reg_offset += curr_trigger_state * TRIGGER_SETS_PER_STATE * MAX_CONSTRAINTS * BYTES_IN_DWORD;
	}

	ecore_wr(p_hwfn, p_ptt, (is_filter ? DBG_REG_FILTER_CNSTR_OPRTN_0 : DBG_REG_TRIGGER_STATE_SET_CNSTR_OPRTN_0) + reg_offset, hw_op_val);
	ecore_wr(p_hwfn, p_ptt, (is_filter ? DBG_REG_FILTER_CNSTR_DATA_0 : DBG_REG_TRIGGER_STATE_SET_CNSTR_DATA_0) + reg_offset, data_val);
	ecore_wr(p_hwfn, p_ptt, (is_filter ? DBG_REG_FILTER_CNSTR_DATA_MASK_0 : DBG_REG_TRIGGER_STATE_SET_CNSTR_DATA_MASK_0) + reg_offset, data_mask);
	ecore_wr(p_hwfn, p_ptt, (is_filter ? DBG_REG_FILTER_CNSTR_FRAME_0 : DBG_REG_TRIGGER_STATE_SET_CNSTR_FRAME_0) + reg_offset, frame_bit);
	ecore_wr(p_hwfn, p_ptt, (is_filter ? DBG_REG_FILTER_CNSTR_FRAME_MASK_0 : DBG_REG_TRIGGER_STATE_SET_CNSTR_FRAME_MASK_0) + reg_offset, frame_mask);
	ecore_wr(p_hwfn, p_ptt, (is_filter ? DBG_REG_FILTER_CNSTR_OFFSET_0 : DBG_REG_TRIGGER_STATE_SET_CNSTR_OFFSET_0) + reg_offset, dword_offset);
	ecore_wr(p_hwfn, p_ptt, (is_filter ? DBG_REG_FILTER_CNSTR_RANGE_0 : DBG_REG_TRIGGER_STATE_SET_CNSTR_RANGE_0) + reg_offset, range);
	ecore_wr(p_hwfn, p_ptt, (is_filter ? DBG_REG_FILTER_CNSTR_CYCLIC_0 : DBG_REG_TRIGGER_STATE_SET_CNSTR_CYCLIC_0) + reg_offset, cyclic_bit);
	ecore_wr(p_hwfn, p_ptt, (is_filter ? DBG_REG_FILTER_CNSTR_MUST_0 : DBG_REG_TRIGGER_STATE_SET_CNSTR_MUST_0) + reg_offset, must_bit);
}

static enum dbg_status ecore_bus_add_block_input(struct ecore_hwfn *p_hwfn,
						 enum block_id block_id,
						 u8 line_num,
						 u8 enable_mask,
						 u8 right_shift,
						 u8 force_valid_mask,
						 u8 force_frame_mask,
						 u8 is_256b_line)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	struct dbg_bus_data *bus = &dev_data->bus;
	struct dbg_bus_block_data *block_bus;

	block_bus = &bus->blocks[block_id];

	/* Set block data */
	block_bus->line_num = line_num;
	block_bus->enable_mask = enable_mask;
	block_bus->right_shift = right_shift;
	block_bus->force_valid_mask = force_valid_mask;
	block_bus->force_frame_mask = force_frame_mask;
	SET_FIELD(block_bus->flags, DBG_BUS_BLOCK_DATA_IS_256B_LINE, is_256b_line);

	/* Set dword mask */
	block_bus->dword_mask = ecore_bus_get_dword_mask(enable_mask, right_shift);
	if ((bus->blocks_dword_mask & block_bus->dword_mask))
		bus->blocks_dword_overlap = 1;
	bus->blocks_dword_mask |= block_bus->dword_mask;

	/* Set 256b mode */
	if (is_256b_line)
		bus->mode_256b_en = 1;

	bus->num_enabled_blocks++;

	return DBG_STATUS_OK;
}

/* Reads the specified DBG Bus internal buffer range and copy it to the
 * specified buffer. Returns the dumped size in dwords.
 */
static u32 ecore_bus_dump_int_buf_range(struct ecore_hwfn *p_hwfn,
					struct ecore_ptt *p_ptt,
					u32 *dump_buf,
					bool dump,
					u32 start_line,
					u32 end_line)
{
	u32 line, reg_addr, i, offset = 0;

	if (!dump)
		return (end_line - start_line + 1) * INT_BUF_LINE_SIZE_IN_DWORDS;

	for (line = start_line, reg_addr = DBG_REG_INTR_BUFFER + DWORDS_TO_BYTES(start_line * INT_BUF_LINE_SIZE_IN_DWORDS);
		line <= end_line;
		line++, offset += INT_BUF_LINE_SIZE_IN_DWORDS)
		for (i = 0; i < INT_BUF_LINE_SIZE_IN_DWORDS; i++, reg_addr += BYTES_IN_DWORD)
			dump_buf[offset + INT_BUF_LINE_SIZE_IN_DWORDS - 1 - i] = ecore_rd(p_hwfn, p_ptt, reg_addr);

	return offset;
}

/* Reads the DBG Bus internal buffer and copy its contents to a buffer.
 * Returns the dumped size in dwords.
 */
static u32 ecore_bus_dump_int_buf(struct ecore_hwfn *p_hwfn,
				  struct ecore_ptt *p_ptt,
				  u32 *dump_buf,
				  bool dump)
{
	u32 last_written_line, offset = 0;

	last_written_line = ecore_rd(p_hwfn, p_ptt, DBG_REG_INTR_BUFFER_WR_PTR);

	if (ecore_rd(p_hwfn, p_ptt, DBG_REG_WRAP_ON_INT_BUFFER)) {
		/* Internal buffer was wrapped: first dump from write pointer
		 * to buffer end, then dump from buffer start to write pointer.
		 */
		if (last_written_line < INT_BUF_NUM_OF_LINES - 1)
			offset += ecore_bus_dump_int_buf_range(p_hwfn, p_ptt, dump_buf + offset, dump, last_written_line + 1, INT_BUF_NUM_OF_LINES - 1);
		offset += ecore_bus_dump_int_buf_range(p_hwfn, p_ptt, dump_buf + offset, dump, 0, last_written_line);
	} else if (last_written_line) {
		/* Internal buffer wasn't wrapped: dump from buffer start until
		 *  write pointer.
		 */
		if (!ecore_rd(p_hwfn, p_ptt, DBG_REG_INTR_BUFFER_RD_PTR))
			offset += ecore_bus_dump_int_buf_range(p_hwfn, p_ptt, dump_buf + offset, dump, 0, last_written_line);
		else
			DP_NOTICE(p_hwfn, true, "Unexpected Debug Bus error: internal buffer read pointer is not zero\n");
	}

	return offset;
}

/* Reads the specified DBG Bus PCI buffer range and copy it to the specified
 * buffer. Returns the dumped size in dwords.
 */
static u32 ecore_bus_dump_pci_buf_range(struct ecore_hwfn *p_hwfn,
					u32 *dump_buf,
					bool dump,
					u32 start_line,
					u32 end_line)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	u32 offset = 0;

	/* Extract PCI buffer pointer from virtual address */
	void *virt_addr_lo = &dev_data->bus.pci_buf.virt_addr.lo;
	u32 *pci_buf_start = (u32 *)(osal_uintptr_t)*((u64 *)virt_addr_lo);
	u32 *pci_buf, line, i;

	if (!dump)
		return (end_line - start_line + 1) * PCI_BUF_LINE_SIZE_IN_DWORDS;

	for (line = start_line, pci_buf = pci_buf_start + start_line * PCI_BUF_LINE_SIZE_IN_DWORDS;
	     line <= end_line;
	     line++, offset += PCI_BUF_LINE_SIZE_IN_DWORDS)
		for (i = 0; i < PCI_BUF_LINE_SIZE_IN_DWORDS; i++, pci_buf++)
			dump_buf[offset + s_pci_buf_line_ind[i]] = *pci_buf;

	return offset;
}

/* Copies the DBG Bus PCI buffer to the specified buffer.
 * Returns the dumped size in dwords.
 */
static u32 ecore_bus_dump_pci_buf(struct ecore_hwfn *p_hwfn,
				  struct ecore_ptt *p_ptt,
				  u32 *dump_buf,
				  bool dump)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	u32 next_wr_byte_offset, next_wr_line_offset;
	struct dbg_bus_mem_addr next_wr_phys_addr;
	u32 pci_buf_size_in_lines, offset = 0;

	pci_buf_size_in_lines = dev_data->bus.pci_buf.size / PCI_BUF_LINE_SIZE_IN_BYTES;

	/* Extract write pointer (physical address) */
	next_wr_phys_addr.lo = ecore_rd(p_hwfn, p_ptt, DBG_REG_EXT_BUFFER_WR_PTR);
	next_wr_phys_addr.hi = ecore_rd(p_hwfn, p_ptt, DBG_REG_EXT_BUFFER_WR_PTR + BYTES_IN_DWORD);

	/* Convert write pointer to offset */
	next_wr_byte_offset = ecore_phys_addr_diff(&next_wr_phys_addr, &dev_data->bus.pci_buf.phys_addr);
	if ((next_wr_byte_offset % PCI_BUF_LINE_SIZE_IN_BYTES) || next_wr_byte_offset > dev_data->bus.pci_buf.size)
		return 0;
	next_wr_line_offset = next_wr_byte_offset / PCI_BUF_LINE_SIZE_IN_BYTES;

	/* PCI buffer wrapped: first dump from write pointer to buffer end. */
	if (ecore_rd(p_hwfn, p_ptt, DBG_REG_WRAP_ON_EXT_BUFFER))
		offset += ecore_bus_dump_pci_buf_range(p_hwfn, dump_buf + offset, dump, next_wr_line_offset, pci_buf_size_in_lines - 1);

	/* Dump from buffer start until write pointer */
	if (next_wr_line_offset)
		offset += ecore_bus_dump_pci_buf_range(p_hwfn, dump_buf + offset, dump, 0, next_wr_line_offset - 1);

	return offset;
}

/* Copies the DBG Bus recorded data to the specified buffer.
 * Returns the dumped size in dwords.
 */
static u32 ecore_bus_dump_data(struct ecore_hwfn *p_hwfn,
			       struct ecore_ptt *p_ptt,
			       u32 *dump_buf,
			       bool dump)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;

	switch (dev_data->bus.target) {
	case DBG_BUS_TARGET_ID_INT_BUF:
		return ecore_bus_dump_int_buf(p_hwfn, p_ptt, dump_buf, dump);
	case DBG_BUS_TARGET_ID_PCI:
		return ecore_bus_dump_pci_buf(p_hwfn, p_ptt, dump_buf, dump);
	default:
		break;
	}

	return 0;
}

/* Frees the Debug Bus PCI buffer */
static void ecore_bus_free_pci_buf(struct ecore_hwfn *p_hwfn)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	dma_addr_t pci_buf_phys_addr;
	void *virt_addr_lo;
	u32 *pci_buf;

	/* Extract PCI buffer pointer from virtual address */
	virt_addr_lo = &dev_data->bus.pci_buf.virt_addr.lo;
	pci_buf = (u32 *)(osal_uintptr_t)*((u64 *)virt_addr_lo);

	if (!dev_data->bus.pci_buf.size)
		return;

	OSAL_MEMCPY(&pci_buf_phys_addr, &dev_data->bus.pci_buf.phys_addr, sizeof(pci_buf_phys_addr));

	OSAL_DMA_FREE_COHERENT(p_hwfn->p_dev, pci_buf, pci_buf_phys_addr, dev_data->bus.pci_buf.size);

	dev_data->bus.pci_buf.size = 0;
}

/* Dumps the list of DBG Bus inputs (blocks/Storms) to the specified buffer.
 * Returns the dumped size in dwords.
 */
static u32 ecore_bus_dump_inputs(struct ecore_hwfn *p_hwfn,
				 struct ecore_ptt *p_ptt,
				 u32 *dump_buf,
				 bool dump)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	char storm_name[8] = "?storm";
	u32 block_id, offset = 0;
	u8 storm_id;

	/* Store storms */
	for (storm_id = 0; storm_id < MAX_DBG_STORMS; storm_id++) {
		struct dbg_bus_storm_data *storm_bus = &dev_data->bus.storms[storm_id];
		struct storm_defs *storm = &s_storm_defs[storm_id];
		u8 num_input_params = 3;
		u32 gpre_vect = 0;

		if (!dev_data->bus.storms[storm_id].enabled)
			continue;

		if (dump) {
			/* Read GPRE vector */
			gpre_vect = ecore_rd(p_hwfn, p_ptt, storm->sem_gpre_vect_addr);
			if (gpre_vect)
				num_input_params++;
		}

		/* Dump section header */
		storm_name[0] = storm->letter;
		offset += ecore_dump_section_hdr(dump_buf + offset, dump, "bus_input", num_input_params);
		offset += ecore_dump_str_param(dump_buf + offset, dump, "name", storm_name);
		offset += ecore_dump_num_param(dump_buf + offset, dump, "id", storm_bus->hw_id);
		offset += ecore_dump_str_param(dump_buf + offset, dump, "mode", s_storm_mode_defs[storm_bus->mode].name);
		if (gpre_vect)
			offset += ecore_dump_num_param(dump_buf + offset, dump, "gpre_vect", gpre_vect);
	}

	/* Store blocks */
	for (block_id = 0; block_id < MAX_BLOCK_ID; block_id++) {
		struct dbg_bus_block_data *block_bus = &dev_data->bus.blocks[block_id];
		const struct dbg_block *block = get_dbg_block(p_hwfn, (enum block_id)block_id);
		u8 is_256b_line;

		if (!block_bus->enable_mask)
			continue;

		is_256b_line = GET_FIELD(block_bus->flags, DBG_BUS_BLOCK_DATA_IS_256B_LINE);

		/* Dump section header */
		offset += ecore_dump_section_hdr(dump_buf + offset, dump, "bus_input", 5);
		offset += ecore_dump_str_param(dump_buf + offset, dump, "name", (const char *)block->name);
		offset += ecore_dump_num_param(dump_buf + offset, dump, "line", block_bus->line_num);
		offset += ecore_dump_num_param(dump_buf + offset, dump, "en", block_bus->enable_mask);
		offset += ecore_dump_num_param(dump_buf + offset, dump, "shr", block_bus->right_shift);
		offset += ecore_dump_num_param(dump_buf + offset, dump, "256b", is_256b_line);
	}

	return offset;
}

/* Dumps the Debug Bus header (params, inputs, data header) to the specified
 * buffer. Returns the dumped size in dwords.
 */
static u32 ecore_bus_dump_hdr(struct ecore_hwfn *p_hwfn,
			      struct ecore_ptt *p_ptt,
			      u32 *dump_buf,
			      bool dump)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	char hw_id_mask_str[16];
	u8 num_specific_params;
	u32 offset = 0;

	if (OSAL_SNPRINTF(hw_id_mask_str, sizeof(hw_id_mask_str), "0x%x", dev_data->bus.hw_id_mask) < 0)
		DP_NOTICE(p_hwfn, true, "Unexpected debug error : invalid HW ID mask\n");

	num_specific_params = 4 + /* 4 must match below amount of params */
		(dev_data->bus.mode_256b_en ? 1 : 0) +
		(dev_data->bus.target != DBG_BUS_TARGET_ID_NIG ? 1 : 0);

	/* Dump global params */
	offset += ecore_dump_common_global_params(p_hwfn, p_ptt, dump_buf + offset, dump, num_specific_params);
	offset += /* 1 */ ecore_dump_str_param(dump_buf + offset, dump, "dump-type", "debug-bus");
	offset += /* 2 */ ecore_dump_str_param(dump_buf + offset, dump, "wrap-mode", dev_data->bus.one_shot_en ? "one-shot" : "wrap-around");
	offset += /* 3 */ ecore_dump_str_param(dump_buf + offset, dump, "hw-id-mask", hw_id_mask_str);
	offset += /* 4 */ ecore_dump_str_param(dump_buf + offset, dump, "target", s_dbg_target_names[dev_data->bus.target]);
	/* Additional/Less parameters require matching of number in call to dump_common_global_params() */

	if (dev_data->bus.target != DBG_BUS_TARGET_ID_NIG) {
		u32 buf_size_bytes = (dev_data->bus.target == DBG_BUS_TARGET_ID_PCI ? dev_data->bus.pci_buf.size : DWORDS_TO_BYTES(INT_BUF_SIZE_IN_DWORDS));

		offset += ecore_dump_num_param(dump_buf + offset, dump, "target-size-kb", buf_size_bytes / 1024);
	}
	if (dev_data->bus.mode_256b_en)
		offset += ecore_dump_num_param(dump_buf + offset, dump, "hw-dwords", 8);

	offset += ecore_bus_dump_inputs(p_hwfn, p_ptt, dump_buf + offset, dump);

	if (dev_data->bus.target != DBG_BUS_TARGET_ID_NIG) {
		/* Start data section */
		u32 recorded_dwords = 0;
		
		if (dump)
			recorded_dwords = ecore_bus_dump_data(p_hwfn, p_ptt, OSAL_NULL, false);

		offset += ecore_dump_section_hdr(dump_buf + offset, dump, "bus_data", 1);
		offset += ecore_dump_num_param(dump_buf + offset, dump, "size", recorded_dwords);
	}

	return offset;
}

/* Returns true if the specified entity (indicated by GRC param) should be
 * included in the dump, false otherwise.
 */
static bool ecore_grc_is_included(struct ecore_hwfn *p_hwfn,
				  enum dbg_grc_params grc_param)
{
	return ecore_grc_get_param(p_hwfn, grc_param) > 0;
}

/* Returns the storm_id that matches the specified Storm letter,
 * or MAX_DBG_STORMS if invalid storm letter.
 */
static enum dbg_storms ecore_get_storm_id_from_letter(char storm_letter)
{
	u8 storm_id;

	for (storm_id = 0; storm_id < MAX_DBG_STORMS; storm_id++)
		if (s_storm_defs[storm_id].letter == storm_letter)
			return (enum dbg_storms)storm_id;

	return MAX_DBG_STORMS;
}

/* Returns true of the specified Storm should be included in the dump, false
 * otherwise.
 */
static bool ecore_grc_is_storm_included(struct ecore_hwfn *p_hwfn,
					enum dbg_storms storm)
{
	return ecore_grc_get_param(p_hwfn, (enum dbg_grc_params)storm) > 0;
}

/* Returns true if the specified memory should be included in the dump, false
 * otherwise.
 */
static bool ecore_grc_is_mem_included(struct ecore_hwfn *p_hwfn,
				      enum block_id block_id,
				      u8 mem_group_id)
{
	const struct dbg_block *block;
	u8 i;

	block = get_dbg_block(p_hwfn, block_id);

	/* If the block is associated with a Storm, check Storm match */
	if (block->associated_storm_letter) {
		enum dbg_storms associated_storm_id = ecore_get_storm_id_from_letter(block->associated_storm_letter);

		if (associated_storm_id == MAX_DBG_STORMS || !ecore_grc_is_storm_included(p_hwfn, associated_storm_id))
			return false;
	}

	for (i = 0; i < NUM_BIG_RAM_TYPES; i++) {
		struct big_ram_defs *big_ram = &s_big_ram_defs[i];

		if (mem_group_id == big_ram->mem_group_id || mem_group_id == big_ram->ram_mem_group_id)
			return ecore_grc_is_included(p_hwfn, big_ram->grc_param);
	}

	switch (mem_group_id) {
	case MEM_GROUP_PXP_ILT:
	case MEM_GROUP_PXP_MEM:
		return ecore_grc_is_included(p_hwfn, DBG_GRC_PARAM_DUMP_PXP);
	case MEM_GROUP_RAM:
		return ecore_grc_is_included(p_hwfn, DBG_GRC_PARAM_DUMP_RAM);
	case MEM_GROUP_PBUF:
		return ecore_grc_is_included(p_hwfn, DBG_GRC_PARAM_DUMP_PBUF);
	case MEM_GROUP_CAU_MEM:
	case MEM_GROUP_CAU_SB:
	case MEM_GROUP_CAU_PI:
		return ecore_grc_is_included(p_hwfn, DBG_GRC_PARAM_DUMP_CAU);
	case MEM_GROUP_CAU_MEM_EXT:
		return ecore_grc_is_included(p_hwfn, DBG_GRC_PARAM_DUMP_CAU_EXT);
	case MEM_GROUP_QM_MEM:
		return ecore_grc_is_included(p_hwfn, DBG_GRC_PARAM_DUMP_QM);
	case MEM_GROUP_CFC_MEM:
	case MEM_GROUP_CONN_CFC_MEM:
	case MEM_GROUP_TASK_CFC_MEM:
		return ecore_grc_is_included(p_hwfn, DBG_GRC_PARAM_DUMP_CFC) || ecore_grc_is_included(p_hwfn, DBG_GRC_PARAM_DUMP_CM_CTX);
	case MEM_GROUP_DORQ_MEM:
		return ecore_grc_is_included(p_hwfn, DBG_GRC_PARAM_DUMP_DORQ);
	case MEM_GROUP_IGU_MEM:
	case MEM_GROUP_IGU_MSIX:
		return ecore_grc_is_included(p_hwfn, DBG_GRC_PARAM_DUMP_IGU);
	case MEM_GROUP_MULD_MEM:
		return ecore_grc_is_included(p_hwfn, DBG_GRC_PARAM_DUMP_MULD);
	case MEM_GROUP_PRS_MEM:
		return ecore_grc_is_included(p_hwfn, DBG_GRC_PARAM_DUMP_PRS);
	case MEM_GROUP_DMAE_MEM:
		return ecore_grc_is_included(p_hwfn, DBG_GRC_PARAM_DUMP_DMAE);
	case MEM_GROUP_TM_MEM:
		return ecore_grc_is_included(p_hwfn, DBG_GRC_PARAM_DUMP_TM);
	case MEM_GROUP_SDM_MEM:
		return ecore_grc_is_included(p_hwfn, DBG_GRC_PARAM_DUMP_SDM);
	case MEM_GROUP_TDIF_CTX:
	case MEM_GROUP_RDIF_CTX:
		return ecore_grc_is_included(p_hwfn, DBG_GRC_PARAM_DUMP_DIF);
	case MEM_GROUP_CM_MEM:
		return ecore_grc_is_included(p_hwfn, DBG_GRC_PARAM_DUMP_CM);
	case MEM_GROUP_IOR:
		return ecore_grc_is_included(p_hwfn, DBG_GRC_PARAM_DUMP_IOR);
	default:
		return true;
	}
}

/* Stalls all Storms */
static void ecore_grc_stall_storms(struct ecore_hwfn *p_hwfn,
				   struct ecore_ptt *p_ptt,
				   bool stall)
{
	u32 reg_addr;
	u8 storm_id;

	for (storm_id = 0; storm_id < MAX_DBG_STORMS; storm_id++) {
		if (!ecore_grc_is_storm_included(p_hwfn, (enum dbg_storms)storm_id))
			continue;

		reg_addr = s_storm_defs[storm_id].sem_fast_mem_addr + SEM_FAST_REG_STALL_0;
		ecore_wr(p_hwfn, p_ptt, reg_addr, stall ? 1 : 0);
	}

	OSAL_MSLEEP(STALL_DELAY_MS);
}

/* Takes all blocks out of reset. If rbc_only is true, only RBC clients are
 * taken out of reset.
 */
static void ecore_grc_unreset_blocks(struct ecore_hwfn *p_hwfn,
				     struct ecore_ptt *p_ptt,
				     bool rbc_only)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	u32 i;

	/* Take RBCs out of reset */
	for (i = 0; i < OSAL_ARRAY_SIZE(s_rbc_reset_defs); i++)
		if (s_rbc_reset_defs[i].reset_val[dev_data->chip_id])
			ecore_wr(p_hwfn, p_ptt, s_rbc_reset_defs[i].reset_reg_addr + RESET_REG_UNRESET_OFFSET, s_rbc_reset_defs[i].reset_val[dev_data->chip_id]);

	if (!rbc_only) {
		u32 reg_val[NUM_DBG_RESET_REGS] = { 0 };
		u8 reset_reg_id;
		u32 block_id;

		/* Fill reset regs values */
		for (block_id = 0; block_id < NUM_PHYS_BLOCKS; block_id++) {
			bool is_removed, has_reset_reg, unreset_before_dump;
			const struct dbg_block_chip *block;

			block = ecore_get_dbg_block_per_chip(p_hwfn, (enum block_id)block_id);
			is_removed = GET_FIELD(block->flags, DBG_BLOCK_CHIP_IS_REMOVED);
			has_reset_reg = GET_FIELD(block->flags, DBG_BLOCK_CHIP_HAS_RESET_REG);
			unreset_before_dump = GET_FIELD(block->flags, DBG_BLOCK_CHIP_UNRESET_BEFORE_DUMP);

			if (!is_removed && has_reset_reg && unreset_before_dump)
				reg_val[block->reset_reg_id] |= (1 << block->reset_reg_bit_offset);
		}

		/* Write reset registers */
		for (reset_reg_id = 0; reset_reg_id < NUM_DBG_RESET_REGS; reset_reg_id++) {
			const struct dbg_reset_reg *reset_reg;
			u32 reset_reg_addr;

			reset_reg = ecore_get_dbg_reset_reg(p_hwfn, reset_reg_id);

			if (GET_FIELD(reset_reg->data, DBG_RESET_REG_IS_REMOVED))
				continue;

			if (reg_val[reset_reg_id]) {
				reset_reg_addr = DWORDS_TO_BYTES(GET_FIELD(reset_reg->data, DBG_RESET_REG_ADDR));
				ecore_wr(p_hwfn, p_ptt, reset_reg_addr + RESET_REG_UNRESET_OFFSET, reg_val[reset_reg_id]);
			}
		}
	}
}

/* Returns the attention block data of the specified block */
static const struct dbg_attn_block_type_data *ecore_get_block_attn_data(struct ecore_hwfn *p_hwfn,
									enum block_id block_id,
									enum dbg_attn_type attn_type)
{
	const struct dbg_attn_block *base_attn_block_arr = (const struct dbg_attn_block *)p_hwfn->dbg_arrays[BIN_BUF_DBG_ATTN_BLOCKS].ptr;

	return &base_attn_block_arr[block_id].per_type_data[attn_type];
}

/* Returns the attention registers of the specified block */
static const struct dbg_attn_reg *ecore_get_block_attn_regs(struct ecore_hwfn *p_hwfn,
							    enum block_id block_id,
							    enum dbg_attn_type attn_type,
							    u8 *num_attn_regs)
{
	const struct dbg_attn_block_type_data *block_type_data = ecore_get_block_attn_data(p_hwfn, block_id, attn_type);

	*num_attn_regs = block_type_data->num_regs;

	return (const struct dbg_attn_reg *)p_hwfn->dbg_arrays[BIN_BUF_DBG_ATTN_REGS].ptr + block_type_data->regs_offset;
}

/* For each block, clear the status of all parities */
static void ecore_grc_clear_all_prty(struct ecore_hwfn *p_hwfn,
				     struct ecore_ptt *p_ptt)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	const struct dbg_attn_reg *attn_reg_arr;
	u8 reg_idx, num_attn_regs;
	u32 block_id;

	for (block_id = 0; block_id < NUM_PHYS_BLOCKS; block_id++) {
		if (dev_data->block_in_reset[block_id])
			continue;

		attn_reg_arr = ecore_get_block_attn_regs(p_hwfn, (enum block_id)block_id, ATTN_TYPE_PARITY, &num_attn_regs);

		for (reg_idx = 0; reg_idx < num_attn_regs; reg_idx++) {
			const struct dbg_attn_reg *reg_data = &attn_reg_arr[reg_idx];
			u16 modes_buf_offset;
			bool eval_mode;

			/* Check mode */
			eval_mode = GET_FIELD(reg_data->mode.data, DBG_MODE_HDR_EVAL_MODE) > 0;
			modes_buf_offset = GET_FIELD(reg_data->mode.data, DBG_MODE_HDR_MODES_BUF_OFFSET);

			/* If Mode match: clear parity status */
			if (!eval_mode || ecore_is_mode_match(p_hwfn, &modes_buf_offset))
				ecore_rd(p_hwfn, p_ptt, DWORDS_TO_BYTES(reg_data->sts_clr_address));
		}
	}
}

/* Finds the meta data image in NVRAM */
static enum dbg_status ecore_find_nvram_image(struct ecore_hwfn *p_hwfn,
					      struct ecore_ptt *p_ptt,
					      u32 image_type,
					      u32 *nvram_offset_bytes,
					      u32 *nvram_size_bytes)
{
	u32 ret_mcp_resp, ret_mcp_param, ret_txn_size;
	struct mcp_file_att file_att;
	int nvm_result;

	/* Call NVRAM get file command */
	nvm_result = ecore_mcp_nvm_rd_cmd(p_hwfn, p_ptt, DRV_MSG_CODE_NVM_GET_FILE_ATT, image_type, &ret_mcp_resp, &ret_mcp_param, &ret_txn_size, (u32 *)&file_att, false);

	/* Check response */
	if (nvm_result || (ret_mcp_resp & FW_MSG_CODE_MASK) != FW_MSG_CODE_NVM_OK)
		return DBG_STATUS_NVRAM_GET_IMAGE_FAILED;

	/* Update return values */
	*nvram_offset_bytes = file_att.nvm_start_addr;
	*nvram_size_bytes = file_att.len;

	DP_VERBOSE(p_hwfn, ECORE_MSG_DEBUG, "find_nvram_image: found NVRAM image of type %d in NVRAM offset %d bytes with size %d bytes\n", image_type, *nvram_offset_bytes, *nvram_size_bytes);

	/* Check alignment */
	if (*nvram_size_bytes & 0x3)
		return DBG_STATUS_NON_ALIGNED_NVRAM_IMAGE;

	return DBG_STATUS_OK;
}

/* Reads data from NVRAM */
static enum dbg_status ecore_nvram_read(struct ecore_hwfn *p_hwfn,
					struct ecore_ptt *p_ptt,
					u32 nvram_offset_bytes,
					u32 nvram_size_bytes,
					u32 *ret_buf)
{
	u32 ret_mcp_resp, ret_mcp_param, ret_read_size, bytes_to_copy;
	s32 bytes_left = nvram_size_bytes;
	u32 read_offset = 0, param = 0;

	DP_VERBOSE(p_hwfn, ECORE_MSG_DEBUG, "nvram_read: reading image of size %d bytes from NVRAM\n", nvram_size_bytes);

	do {
		bytes_to_copy = (bytes_left > MCP_DRV_NVM_BUF_LEN) ? MCP_DRV_NVM_BUF_LEN : bytes_left;

		/* Call NVRAM read command */
		SET_MFW_FIELD(param, DRV_MB_PARAM_NVM_OFFSET, nvram_offset_bytes + read_offset);
		SET_MFW_FIELD(param, DRV_MB_PARAM_NVM_LEN, bytes_to_copy);
		if (ecore_mcp_nvm_rd_cmd(p_hwfn, p_ptt, DRV_MSG_CODE_NVM_READ_NVRAM, param, &ret_mcp_resp, &ret_mcp_param, &ret_read_size, (u32 *)((u8 *)ret_buf + read_offset), false))
			return DBG_STATUS_NVRAM_READ_FAILED;

		/* Check response */
		if ((ret_mcp_resp  & FW_MSG_CODE_MASK) != FW_MSG_CODE_NVM_OK)
			return DBG_STATUS_NVRAM_READ_FAILED;

		/* Update read offset */
		read_offset += ret_read_size;
		bytes_left -= ret_read_size;
	} while (bytes_left > 0);

	return DBG_STATUS_OK;
}

/* Dumps GRC registers section header. Returns the dumped size in dwords.
 * the following parameters are dumped:
 * - count: no. of dumped entries
 * - split_type: split type
 * - split_id: split ID (dumped only if split_id != SPLIT_TYPE_NONE)
 * - reg_type_name: register type name (dumped only if reg_type_name != OSAL_NULL)
 */
static u32 ecore_grc_dump_regs_hdr(u32 *dump_buf,
				   bool dump,
				   u32 num_reg_entries,
				   enum init_split_types split_type,
				   u8 split_id,
				   const char *reg_type_name)
{
	u8 num_params = 2 + (split_type != SPLIT_TYPE_NONE ? 1 : 0) + (reg_type_name ? 1 : 0);
	u32 offset = 0;

	offset += ecore_dump_section_hdr(dump_buf + offset, dump, "grc_regs", num_params);
	offset += ecore_dump_num_param(dump_buf + offset, dump, "count", num_reg_entries);
	offset += ecore_dump_str_param(dump_buf + offset, dump, "split", s_split_type_defs[split_type].name);
	if (split_type != SPLIT_TYPE_NONE)
		offset += ecore_dump_num_param(dump_buf + offset, dump, "id", split_id);
	if (reg_type_name)
		offset += ecore_dump_str_param(dump_buf + offset, dump, "type", reg_type_name);

	return offset;
}

/* Dumps the GRC registers in the specified address range.
 * Returns the dumped size in dwords.
 * The addr and len arguments are specified in dwords.
 */
static u32 ecore_grc_dump_addr_range(struct ecore_hwfn *p_hwfn,
				     struct ecore_ptt *p_ptt,
				     u32 *dump_buf,
				     bool dump,
				     u32 addr,
				     u32 len,
				     bool wide_bus,
				     enum init_split_types split_type,
				     u8 split_id)

{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	u8 port_id = 0, pf_id = 0, vf_id = 0;
	bool read_using_dmae = false;
	u32 thresh;

	if (!dump)
		return len;

	switch (split_type) {
	case SPLIT_TYPE_PORT:
		port_id = split_id;
		break;
	case SPLIT_TYPE_PF:
		pf_id = split_id;
		break;
	case SPLIT_TYPE_PORT_PF:
		port_id = split_id / dev_data->num_pfs_per_port;
		pf_id = port_id + dev_data->num_ports * (split_id % dev_data->num_pfs_per_port);
		break;
	case SPLIT_TYPE_VF:
		vf_id = split_id;
		break;
	default:
		break;
	}

	/* Try reading using DMAE */
	if (dev_data->use_dmae && split_type != SPLIT_TYPE_VF && (len >= s_hw_type_defs[dev_data->hw_type].dmae_thresh || (PROTECT_WIDE_BUS && wide_bus))) {
		struct dmae_params dmae_params;

		/* Set DMAE params */
		OSAL_MEMSET(&dmae_params, 0, sizeof(dmae_params));
		SET_FIELD(dmae_params.flags, DMAE_PARAMS_COMPLETION_DST, 1);
		switch (split_type) {
		case SPLIT_TYPE_PORT:
			SET_FIELD(dmae_params.flags, DMAE_PARAMS_PORT_VALID, 1);
			dmae_params.port_id = port_id;
			break;
		case SPLIT_TYPE_PF:
			SET_FIELD(dmae_params.flags, DMAE_PARAMS_SRC_PF_VALID, 1);
			dmae_params.src_pf_id = pf_id;
			break;
		case SPLIT_TYPE_PORT_PF:
			SET_FIELD(dmae_params.flags, DMAE_PARAMS_PORT_VALID, 1);
			SET_FIELD(dmae_params.flags, DMAE_PARAMS_SRC_PF_VALID, 1);
			dmae_params.port_id = port_id;
			dmae_params.src_pf_id = pf_id;
			break;
		default:
			break;
		}

		/* Execute DMAE command */
		read_using_dmae = !ecore_dmae_grc2host(p_hwfn, p_ptt, DWORDS_TO_BYTES(addr), (u64)(osal_uintptr_t)(dump_buf), len, &dmae_params);
		if (!read_using_dmae) {
			dev_data->use_dmae = 0;
			DP_VERBOSE(p_hwfn, ECORE_MSG_DEBUG, "Failed reading from chip using DMAE, using GRC instead\n");
		}
	}

	/* If not read using DMAE, read using GRC */
	if (!read_using_dmae) {

		/* Set pretend */
		if (split_type != dev_data->pretend.split_type || split_id != dev_data->pretend.split_id) {
			switch (split_type) {
			case SPLIT_TYPE_PORT:
				ecore_port_pretend(p_hwfn, p_ptt, port_id);
				break;
			case SPLIT_TYPE_PF:
				ecore_fid_pretend(p_hwfn, p_ptt, FIELD_VALUE(PXP_PRETEND_CONCRETE_FID_PFID, pf_id));
				break;
			case SPLIT_TYPE_PORT_PF:
				ecore_port_fid_pretend(p_hwfn, p_ptt, port_id, FIELD_VALUE(PXP_PRETEND_CONCRETE_FID_PFID, pf_id));
				break;
			case SPLIT_TYPE_VF:
				ecore_fid_pretend(p_hwfn, p_ptt, FIELD_VALUE(PXP_PRETEND_CONCRETE_FID_VFVALID, 1) | FIELD_VALUE(PXP_PRETEND_CONCRETE_FID_VFID, vf_id));
				break;
			default:
				break;
			}

			dev_data->pretend.split_type = (u8)split_type;
			dev_data->pretend.split_id = split_id;
		}

		/* Read registers using GRC */
		ecore_read_regs(p_hwfn, p_ptt, dump_buf, addr, len);
	}

	/* Print log */
	dev_data->num_regs_read += len;
	thresh = s_hw_type_defs[dev_data->hw_type].log_thresh;
	if ((dev_data->num_regs_read / thresh) > ((dev_data->num_regs_read - len) / thresh))
		DP_VERBOSE(p_hwfn, ECORE_MSG_DEBUG, "Dumped %d registers...\n", dev_data->num_regs_read);

	return len;
}

/* Dumps GRC registers sequence header. Returns the dumped size in dwords.
 * The addr and len arguments are specified in dwords.
 */
static u32 ecore_grc_dump_reg_entry_hdr(u32 *dump_buf,
					bool dump,
					u32 addr,
					u32 len)
{
	if (dump)
		*dump_buf = addr | (len << REG_DUMP_LEN_SHIFT);

	return 1;
}

/* Dumps GRC registers sequence. Returns the dumped size in dwords.
 * The addr and len arguments are specified in dwords.
 */
static u32 ecore_grc_dump_reg_entry(struct ecore_hwfn *p_hwfn,
				    struct ecore_ptt *p_ptt,
				    u32 *dump_buf,
				    bool dump,
				    u32 addr,
				    u32 len,
				    bool wide_bus,
				    enum init_split_types split_type,
				    u8 split_id)
{
	u32 offset = 0;

	offset += ecore_grc_dump_reg_entry_hdr(dump_buf, dump, addr, len);
	offset += ecore_grc_dump_addr_range(p_hwfn, p_ptt, dump_buf + offset, dump, addr, len, wide_bus, split_type, split_id);

	return offset;
}

/* Dumps GRC registers sequence with skip cycle.
 * Returns the dumped size in dwords.
 * - addr:	start GRC address in dwords
 * - total_len:	total no. of dwords to dump
 * - read_len:	no. consecutive dwords to read
 * - skip_len:	no. of dwords to skip (and fill with zeros)
 */
static u32 ecore_grc_dump_reg_entry_skip(struct ecore_hwfn *p_hwfn,
					 struct ecore_ptt *p_ptt,
					 u32 *dump_buf,
					 bool dump,
					 u32 addr,
					 u32 total_len,
					 u32 read_len,
					 u32 skip_len)
{
	u32 offset = 0, reg_offset = 0;

	offset += ecore_grc_dump_reg_entry_hdr(dump_buf, dump, addr, total_len);

	if (!dump)
		return offset + total_len;

	while (reg_offset < total_len) {
		u32 curr_len = OSAL_MIN_T(u32, read_len, total_len - reg_offset);

		offset += ecore_grc_dump_addr_range(p_hwfn, p_ptt, dump_buf + offset, dump, addr, curr_len, false, SPLIT_TYPE_NONE, 0);
		reg_offset += curr_len;
		addr += curr_len;

		if (reg_offset < total_len) {
			curr_len = OSAL_MIN_T(u32, skip_len, total_len - skip_len);
			OSAL_MEMSET(dump_buf + offset, 0, DWORDS_TO_BYTES(curr_len));
			offset += curr_len;
			reg_offset += curr_len;
			addr += curr_len;
		}
	}

	return offset;
}

/* Dumps GRC registers entries. Returns the dumped size in dwords. */
static u32 ecore_grc_dump_regs_entries(struct ecore_hwfn *p_hwfn,
				       struct ecore_ptt *p_ptt,
				       struct virt_mem_desc input_regs_arr,
				       u32 *dump_buf,
				       bool dump,
				       enum init_split_types split_type,
				       u8 split_id,
				       bool block_enable[MAX_BLOCK_ID],
				       u32 *num_dumped_reg_entries)
{
	u32 i, offset = 0, input_offset = 0;
	bool mode_match = true;

	*num_dumped_reg_entries = 0;

	while (input_offset < BYTES_TO_DWORDS(input_regs_arr.size)) {
		const struct dbg_dump_cond_hdr *cond_hdr = (const struct dbg_dump_cond_hdr *)input_regs_arr.ptr + input_offset++;
		u16 modes_buf_offset;
		bool eval_mode;

		/* Check mode/block */
		eval_mode = GET_FIELD(cond_hdr->mode.data, DBG_MODE_HDR_EVAL_MODE) > 0;
		if (eval_mode) {
			modes_buf_offset = GET_FIELD(cond_hdr->mode.data, DBG_MODE_HDR_MODES_BUF_OFFSET);
			mode_match = ecore_is_mode_match(p_hwfn, &modes_buf_offset);
		}

		if (!mode_match || !block_enable[cond_hdr->block_id]) {
			input_offset += cond_hdr->data_size;
			continue;
		}

		for (i = 0; i < cond_hdr->data_size; i++, input_offset++) {
			const struct dbg_dump_reg *reg = (const struct dbg_dump_reg *)input_regs_arr.ptr + input_offset;

			offset += ecore_grc_dump_reg_entry(p_hwfn, p_ptt, dump_buf + offset, dump,
				GET_FIELD(reg->data, DBG_DUMP_REG_ADDRESS),
				GET_FIELD(reg->data, DBG_DUMP_REG_LENGTH),
				GET_FIELD(reg->data, DBG_DUMP_REG_WIDE_BUS),
				split_type, split_id);
			(*num_dumped_reg_entries)++;
		}
	}

	return offset;
}

/* Dumps GRC registers entries. Returns the dumped size in dwords. */
static u32 ecore_grc_dump_split_data(struct ecore_hwfn *p_hwfn,
				     struct ecore_ptt *p_ptt,
				     struct virt_mem_desc input_regs_arr,
				     u32 *dump_buf,
				     bool dump,
				     bool block_enable[MAX_BLOCK_ID],
				     enum init_split_types split_type,
				     u8 split_id,
				     const char *reg_type_name)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	enum init_split_types hdr_split_type = split_type;
	u32 num_dumped_reg_entries, offset;
	u8 hdr_split_id = split_id;

	/* In PORT_PF split type, print a port split header */
	if (split_type == SPLIT_TYPE_PORT_PF) {
		hdr_split_type = SPLIT_TYPE_PORT;
		hdr_split_id = split_id / dev_data->num_pfs_per_port;
	}

	/* Calculate register dump header size (and skip it for now) */
	offset = ecore_grc_dump_regs_hdr(dump_buf, false, 0, hdr_split_type, hdr_split_id, reg_type_name);

	/* Dump registers */
	offset += ecore_grc_dump_regs_entries(p_hwfn, p_ptt, input_regs_arr, dump_buf + offset, dump, split_type, split_id, block_enable, &num_dumped_reg_entries);

	/* Write register dump header */
	if (dump && num_dumped_reg_entries > 0)
		ecore_grc_dump_regs_hdr(dump_buf, dump, num_dumped_reg_entries, hdr_split_type, hdr_split_id, reg_type_name);

	return num_dumped_reg_entries > 0 ? offset : 0;
}

/* Dumps registers according to the input registers array. Returns the dumped
 * size in dwords.
 */
static u32 ecore_grc_dump_registers(struct ecore_hwfn *p_hwfn,
				    struct ecore_ptt *p_ptt,
				    u32 *dump_buf,
				    bool dump,
				    bool block_enable[MAX_BLOCK_ID],
				    const char *reg_type_name)
{
	struct virt_mem_desc *dbg_buf = &p_hwfn->dbg_arrays[BIN_BUF_DBG_DUMP_REG];
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	u32 offset = 0, input_offset = 0;

	while (input_offset < BYTES_TO_DWORDS(dbg_buf->size)) {
		const struct dbg_dump_split_hdr *split_hdr;
		struct virt_mem_desc curr_input_regs_arr;
		enum init_split_types split_type;
		u16 split_count = 0;
		u32 split_data_size;
		u8 split_id;

		split_hdr = (const struct dbg_dump_split_hdr *)dbg_buf->ptr + input_offset++;
		split_type = (enum init_split_types)GET_FIELD(split_hdr->hdr, DBG_DUMP_SPLIT_HDR_SPLIT_TYPE_ID);
		split_data_size = GET_FIELD(split_hdr->hdr, DBG_DUMP_SPLIT_HDR_DATA_SIZE);
		curr_input_regs_arr.ptr = (u32 *)p_hwfn->dbg_arrays[BIN_BUF_DBG_DUMP_REG].ptr + input_offset;
		curr_input_regs_arr.size = DWORDS_TO_BYTES(split_data_size);

		switch (split_type) {
		case SPLIT_TYPE_NONE:
			split_count = 1;
			break;
		case SPLIT_TYPE_PORT:
			split_count = dev_data->num_ports;
			break;
		case SPLIT_TYPE_PF:
		case SPLIT_TYPE_PORT_PF:
			split_count = dev_data->num_ports * dev_data->num_pfs_per_port;
			break;
		case SPLIT_TYPE_VF:
			split_count = dev_data->num_vfs;
			break;
		default: return 0;
		};

		for (split_id = 0; split_id < split_count; split_id++)
			offset += ecore_grc_dump_split_data(p_hwfn, p_ptt, curr_input_regs_arr, dump_buf + offset, dump, block_enable, split_type, split_id, reg_type_name);

		input_offset += split_data_size;
	}

	/* Cancel pretends (pretend to original PF) */
	if (dump) {
		ecore_fid_pretend(p_hwfn, p_ptt, FIELD_VALUE(PXP_PRETEND_CONCRETE_FID_PFID, p_hwfn->rel_pf_id));
		dev_data->pretend.split_type = SPLIT_TYPE_NONE;
		dev_data->pretend.split_id = 0;
	}

	return offset;
}

/* Dump reset registers. Returns the dumped size in dwords. */
static u32 ecore_grc_dump_reset_regs(struct ecore_hwfn *p_hwfn,
				     struct ecore_ptt *p_ptt,
				     u32 *dump_buf,
				     bool dump)
{
	u32 offset = 0, num_regs = 0;
	u8 reset_reg_id;

	/* Calculate header size */
	offset += ecore_grc_dump_regs_hdr(dump_buf, false, 0, SPLIT_TYPE_NONE, 0, "RESET_REGS");

	/* Write reset registers */
	for (reset_reg_id = 0; reset_reg_id < NUM_DBG_RESET_REGS; reset_reg_id++) {
		const struct dbg_reset_reg *reset_reg;
		u32 reset_reg_addr;

		reset_reg = ecore_get_dbg_reset_reg(p_hwfn, reset_reg_id);

		if (GET_FIELD(reset_reg->data, DBG_RESET_REG_IS_REMOVED))
			continue;

		reset_reg_addr = GET_FIELD(reset_reg->data, DBG_RESET_REG_ADDR);
		offset += ecore_grc_dump_reg_entry(p_hwfn, p_ptt, dump_buf + offset, dump, reset_reg_addr, 1, false, SPLIT_TYPE_NONE, 0);
		num_regs++;
	}

	/* Write header */
	if (dump)
		ecore_grc_dump_regs_hdr(dump_buf, true, num_regs, SPLIT_TYPE_NONE, 0, "RESET_REGS");

	return offset;
}

/* Dump registers that are modified during GRC Dump and therefore must be
 * dumped first. Returns the dumped size in dwords.
 */
static u32 ecore_grc_dump_modified_regs(struct ecore_hwfn *p_hwfn,
					struct ecore_ptt *p_ptt,
					u32 *dump_buf,
					bool dump)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	u32 block_id, offset = 0, stall_regs_offset;
	const struct dbg_attn_reg *attn_reg_arr;
	u8 storm_id, reg_idx, num_attn_regs;
	u32 num_reg_entries = 0;

	/* Write empty header for attention registers */
	offset += ecore_grc_dump_regs_hdr(dump_buf, false, 0, SPLIT_TYPE_NONE, 0, "ATTN_REGS");

	/* Write parity registers */
	for (block_id = 0; block_id < NUM_PHYS_BLOCKS; block_id++) {
		if (dev_data->block_in_reset[block_id] && dump)
			continue;

		attn_reg_arr = ecore_get_block_attn_regs(p_hwfn, (enum block_id)block_id, ATTN_TYPE_PARITY, &num_attn_regs);

		for (reg_idx = 0; reg_idx < num_attn_regs; reg_idx++) {
			const struct dbg_attn_reg *reg_data = &attn_reg_arr[reg_idx];
			u16 modes_buf_offset;
			bool eval_mode;

			/* Check mode */
			eval_mode = GET_FIELD(reg_data->mode.data, DBG_MODE_HDR_EVAL_MODE) > 0;
			modes_buf_offset = GET_FIELD(reg_data->mode.data, DBG_MODE_HDR_MODES_BUF_OFFSET);
			if (eval_mode && !ecore_is_mode_match(p_hwfn, &modes_buf_offset))
				continue;

			/* Mode match: read & dump registers */
			offset += ecore_grc_dump_reg_entry(p_hwfn, p_ptt, dump_buf + offset, dump, reg_data->mask_address, 1, false, SPLIT_TYPE_NONE, 0);
			offset += ecore_grc_dump_reg_entry(p_hwfn, p_ptt, dump_buf + offset, dump, GET_FIELD(reg_data->data, DBG_ATTN_REG_STS_ADDRESS), 1, false, SPLIT_TYPE_NONE, 0);
			num_reg_entries += 2;
		}
	}

	/* Overwrite header for attention registers */
	if (dump)
		ecore_grc_dump_regs_hdr(dump_buf, true, num_reg_entries, SPLIT_TYPE_NONE, 0, "ATTN_REGS");

	/* Write empty header for stall registers */
	stall_regs_offset = offset;
	offset += ecore_grc_dump_regs_hdr(dump_buf, false, 0, SPLIT_TYPE_NONE, 0, "REGS");

	/* Write Storm stall status registers */
	for (storm_id = 0, num_reg_entries = 0; storm_id < MAX_DBG_STORMS; storm_id++) {
		struct storm_defs *storm = &s_storm_defs[storm_id];

		if (dev_data->block_in_reset[storm->sem_block_id] && dump)
			continue;

		offset += ecore_grc_dump_reg_entry(p_hwfn, p_ptt, dump_buf + offset, dump, BYTES_TO_DWORDS(storm->sem_fast_mem_addr + SEM_FAST_REG_STALLED), 1, false, SPLIT_TYPE_NONE, 0);
		num_reg_entries++;
	}

	/* Overwrite header for stall registers */
	if (dump)
		ecore_grc_dump_regs_hdr(dump_buf + stall_regs_offset, true, num_reg_entries, SPLIT_TYPE_NONE, 0, "REGS");

	return offset;
}

/* Dumps registers that can't be represented in the debug arrays */
static u32 ecore_grc_dump_special_regs(struct ecore_hwfn *p_hwfn,
				       struct ecore_ptt *p_ptt,
				       u32 *dump_buf,
				       bool dump)
{
	u32 offset = 0;

	offset += ecore_grc_dump_regs_hdr(dump_buf, dump, 2, SPLIT_TYPE_NONE, 0, "REGS");

	/* Dump R/TDIF_REG_DEBUG_ERROR_INFO_SIZE (every 8'th register should be
	 * skipped).
	 */
	offset += ecore_grc_dump_reg_entry_skip(p_hwfn, p_ptt, dump_buf + offset, dump, BYTES_TO_DWORDS(RDIF_REG_DEBUG_ERROR_INFO), RDIF_REG_DEBUG_ERROR_INFO_SIZE, 7, 1);
	offset += ecore_grc_dump_reg_entry_skip(p_hwfn, p_ptt, dump_buf + offset, dump, BYTES_TO_DWORDS(TDIF_REG_DEBUG_ERROR_INFO), TDIF_REG_DEBUG_ERROR_INFO_SIZE, 7, 1);

	return offset;
}

/* Dumps a GRC memory header (section and params). Returns the dumped size in
 * dwords. The following parameters are dumped:
 * - name:	   dumped only if it's not OSAL_NULL.
 * - addr:	   in dwords, dumped only if name is OSAL_NULL.
 * - len:	   in dwords, always dumped.
 * - width:	   dumped if it's not zero.
 * - packed:	   dumped only if it's not false.
 * - mem_group:	   always dumped.
 * - is_storm:	   true only if the memory is related to a Storm.
 * - storm_letter: valid only if is_storm is true.
 */
static u32 ecore_grc_dump_mem_hdr(struct ecore_hwfn *p_hwfn,
				  u32 *dump_buf,
				  bool dump,
				  const char *name,
				  u32 addr,
				  u32 len,
				  u32 bit_width,
				  bool packed,
				  const char *mem_group,
				  char storm_letter)
{
	u8 num_params = 3;
	u32 offset = 0;
	char buf[64];

	if (!len)
		DP_NOTICE(p_hwfn, true, "Unexpected GRC Dump error: dumped memory size must be non-zero\n");

	if (bit_width)
		num_params++;
	if (packed)
		num_params++;

	/* Dump section header */
	offset += ecore_dump_section_hdr(dump_buf + offset, dump, "grc_mem", num_params);

	if (name) {
		/* Dump name */
		if (storm_letter) {
			OSAL_STRCPY(buf, "?STORM_");
			buf[0] = storm_letter;
			OSAL_STRCPY(buf + OSAL_STRLEN(buf), name);
		} else {
			OSAL_STRCPY(buf, name);
		}

		offset += ecore_dump_str_param(dump_buf + offset, dump, "name", buf);
	} else {
		/* Dump address */
		u32 addr_in_bytes = DWORDS_TO_BYTES(addr);

		offset += ecore_dump_num_param(dump_buf + offset, dump, "addr", addr_in_bytes);
	}

	/* Dump len */
	offset += ecore_dump_num_param(dump_buf + offset, dump, "len", len);

	/* Dump bit width */
	if (bit_width)
		offset += ecore_dump_num_param(dump_buf + offset, dump, "width", bit_width);

	/* Dump packed */
	if (packed)
		offset += ecore_dump_num_param(dump_buf + offset, dump, "packed", 1);

	/* Dump reg type */
	if (storm_letter) {
		OSAL_STRCPY(buf, "?STORM_");
		buf[0] = storm_letter;
		OSAL_STRCPY(buf + OSAL_STRLEN(buf), mem_group);
	} else {
		OSAL_STRCPY(buf, mem_group);
	}

	offset += ecore_dump_str_param(dump_buf + offset, dump, "type", buf);

	return offset;
}

/* Dumps a single GRC memory. If name is OSAL_NULL, the memory is stored by address.
 * Returns the dumped size in dwords.
 * The addr and len arguments are specified in dwords.
 */
static u32 ecore_grc_dump_mem(struct ecore_hwfn *p_hwfn,
			      struct ecore_ptt *p_ptt,
			      u32 *dump_buf,
			      bool dump,
			      const char *name,
			      u32 addr,
			      u32 len,
			      bool wide_bus,
			      u32 bit_width,
			      bool packed,
			      const char *mem_group,
			      char storm_letter)
{
	u32 offset = 0;

	offset += ecore_grc_dump_mem_hdr(p_hwfn, dump_buf + offset, dump, name, addr, len, bit_width, packed, mem_group, storm_letter);
	offset += ecore_grc_dump_addr_range(p_hwfn, p_ptt, dump_buf + offset, dump, addr, len, wide_bus, SPLIT_TYPE_NONE, 0);

	return offset;
}

/* Dumps GRC memories entries. Returns the dumped size in dwords. */
static u32 ecore_grc_dump_mem_entries(struct ecore_hwfn *p_hwfn,
				      struct ecore_ptt *p_ptt,
				      struct virt_mem_desc input_mems_arr,
				      u32 *dump_buf,
				      bool dump)
{
	u32 i, offset = 0, input_offset = 0;
	bool mode_match = true;

	while (input_offset < BYTES_TO_DWORDS(input_mems_arr.size)) {
		const struct dbg_dump_cond_hdr *cond_hdr;
		u16 modes_buf_offset;
		u32 num_entries;
		bool eval_mode;

		cond_hdr = (const struct dbg_dump_cond_hdr *)input_mems_arr.ptr + input_offset++;
		num_entries = cond_hdr->data_size / MEM_DUMP_ENTRY_SIZE_DWORDS;

		/* Check required mode */
		eval_mode = GET_FIELD(cond_hdr->mode.data, DBG_MODE_HDR_EVAL_MODE) > 0;
		if (eval_mode) {
			modes_buf_offset = GET_FIELD(cond_hdr->mode.data, DBG_MODE_HDR_MODES_BUF_OFFSET);
			mode_match = ecore_is_mode_match(p_hwfn, &modes_buf_offset);
		}

		if (!mode_match) {
			input_offset += cond_hdr->data_size;
			continue;
		}

		for (i = 0; i < num_entries; i++, input_offset += MEM_DUMP_ENTRY_SIZE_DWORDS) {
			const struct dbg_dump_mem *mem = (const struct dbg_dump_mem *)((u32 *)input_mems_arr.ptr + input_offset);
			const struct dbg_block *block;
			char storm_letter = 0;
			u32 mem_addr, mem_len;
			bool mem_wide_bus;
			u8 mem_group_id;

			mem_group_id = GET_FIELD(mem->dword0, DBG_DUMP_MEM_MEM_GROUP_ID);
			if (mem_group_id >= MEM_GROUPS_NUM) {
				DP_NOTICE(p_hwfn, true, "Invalid mem_group_id\n");
				return 0;
			}

			if (!ecore_grc_is_mem_included(p_hwfn, (enum block_id)cond_hdr->block_id, mem_group_id))
				continue;

			mem_addr = GET_FIELD(mem->dword0, DBG_DUMP_MEM_ADDRESS);
			mem_len = GET_FIELD(mem->dword1, DBG_DUMP_MEM_LENGTH);
			mem_wide_bus = GET_FIELD(mem->dword1, DBG_DUMP_MEM_WIDE_BUS);

			block = get_dbg_block(p_hwfn, (enum block_id)cond_hdr->block_id);

			/* If memory is associated with Storm, udpate Storm details */
			if (block->associated_storm_letter)
				storm_letter = block->associated_storm_letter;

			/* Dump memory */
			offset += ecore_grc_dump_mem(p_hwfn, p_ptt, dump_buf + offset, dump, OSAL_NULL, mem_addr, mem_len, mem_wide_bus, 0, false, s_mem_group_names[mem_group_id], storm_letter);
		}
	}

	return offset;
}

/* Dumps GRC memories according to the input array dump_mem.
 * Returns the dumped size in dwords.
 */
static u32 ecore_grc_dump_memories(struct ecore_hwfn *p_hwfn,
				   struct ecore_ptt *p_ptt,
				   u32 *dump_buf,
				   bool dump)
{
	struct virt_mem_desc *dbg_buf = &p_hwfn->dbg_arrays[BIN_BUF_DBG_DUMP_MEM];
	u32 offset = 0, input_offset = 0;

	while (input_offset < BYTES_TO_DWORDS(dbg_buf->size)) {
		const struct dbg_dump_split_hdr *split_hdr;
		struct virt_mem_desc curr_input_mems_arr;
		enum init_split_types split_type;
		u32 split_data_size;

		split_hdr = (const struct dbg_dump_split_hdr *)dbg_buf->ptr + input_offset++;
		split_type = (enum init_split_types)GET_FIELD(split_hdr->hdr, DBG_DUMP_SPLIT_HDR_SPLIT_TYPE_ID);
		split_data_size = GET_FIELD(split_hdr->hdr, DBG_DUMP_SPLIT_HDR_DATA_SIZE);
		curr_input_mems_arr.ptr = (u32 *)dbg_buf->ptr + input_offset;
		curr_input_mems_arr.size = DWORDS_TO_BYTES(split_data_size);

		if (split_type == SPLIT_TYPE_NONE)
			offset += ecore_grc_dump_mem_entries(p_hwfn, p_ptt, curr_input_mems_arr, dump_buf + offset, dump);
		else
			DP_NOTICE(p_hwfn, true, "Dumping split memories is currently not supported\n");

		input_offset += split_data_size;
	}

	return offset;
}

/* Dumps GRC context data for the specified Storm.
 * Returns the dumped size in dwords.
 * The lid_size argument is specified in quad-regs.
 */
static u32 ecore_grc_dump_ctx_data(struct ecore_hwfn *p_hwfn,
				   struct ecore_ptt *p_ptt,
				   u32 *dump_buf,
				   bool dump,
				   const char *name,
				   u32 num_lids,
				   enum cm_ctx_types ctx_type,
				   u8 storm_id)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	struct storm_defs *storm = &s_storm_defs[storm_id];
	u32 i, lid, lid_size, total_size;
	u32 rd_reg_addr, offset = 0;

	/* Convert quad-regs to dwords */
	lid_size = storm->cm_ctx_lid_sizes[dev_data->chip_id][ctx_type] * 4;

	if (!lid_size)
		return 0;

	total_size = num_lids * lid_size;

	offset += ecore_grc_dump_mem_hdr(p_hwfn, dump_buf + offset, dump, name, 0, total_size, lid_size * 32, false, name, storm->letter);

	if (!dump)
		return offset + total_size;

	rd_reg_addr = BYTES_TO_DWORDS(storm->cm_ctx_rd_addr[ctx_type]);

	/* Dump context data */
	for (lid = 0; lid < num_lids; lid++) {
		for (i = 0; i < lid_size; i++) {
			ecore_wr(p_hwfn, p_ptt, storm->cm_ctx_wr_addr, (i << 9) | lid);
			offset += ecore_grc_dump_addr_range(p_hwfn, p_ptt, dump_buf + offset, dump, rd_reg_addr, 1, false, SPLIT_TYPE_NONE, 0);
		}
	}

	return offset;
}

/* Dumps GRC contexts. Returns the dumped size in dwords. */
static u32 ecore_grc_dump_ctx(struct ecore_hwfn *p_hwfn,
			      struct ecore_ptt *p_ptt,
			      u32 *dump_buf,
			      bool dump)
{
	u32 offset = 0;
	u8 storm_id;

	for (storm_id = 0; storm_id < MAX_DBG_STORMS; storm_id++) {
		if (!ecore_grc_is_storm_included(p_hwfn, (enum dbg_storms)storm_id))
			continue;

		/* Dump Conn AG context size */
		offset += ecore_grc_dump_ctx_data(p_hwfn, p_ptt, dump_buf + offset, dump, "CONN_AG_CTX", NUM_OF_LCIDS, CM_CTX_CONN_AG, storm_id);

		/* Dump Conn ST context size */
		offset += ecore_grc_dump_ctx_data(p_hwfn, p_ptt, dump_buf + offset, dump, "CONN_ST_CTX", NUM_OF_LCIDS, CM_CTX_CONN_ST, storm_id);

		/* Dump Task AG context size */
		offset += ecore_grc_dump_ctx_data(p_hwfn, p_ptt, dump_buf + offset, dump, "TASK_AG_CTX", NUM_OF_LTIDS, CM_CTX_TASK_AG, storm_id);

		/* Dump Task ST context size */
		offset += ecore_grc_dump_ctx_data(p_hwfn, p_ptt, dump_buf + offset, dump, "TASK_ST_CTX", NUM_OF_LTIDS, CM_CTX_TASK_ST, storm_id);
	}

	return offset;
}

#define VFC_STATUS_RESP_READY_BIT	0
#define VFC_STATUS_BUSY_BIT		1
#define VFC_STATUS_SENDING_CMD_BIT	2

#define VFC_POLLING_DELAY_MS	1
#define VFC_POLLING_COUNT		20

/* Reads data from VFC. Returns the number of dwords read (0 on error).
 * Sizes are specified in dwords.
 */
static u32 ecore_grc_dump_read_from_vfc(struct ecore_hwfn *p_hwfn,
					struct ecore_ptt *p_ptt,
					struct storm_defs *storm,
					u32 *cmd_data,
					u32 cmd_size,
					u32 *addr_data,
					u32 addr_size,
					u32 resp_size,
					u32 *dump_buf)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	u32 vfc_status, polling_ms, polling_count = 0, i;
	bool is_ready = false;

	polling_ms = VFC_POLLING_DELAY_MS * s_hw_type_defs[dev_data->hw_type].delay_factor;

	/* Write VFC command */
	ARR_REG_WR(p_hwfn, p_ptt, storm->sem_fast_mem_addr + SEM_FAST_REG_VFC_DATA_WR, cmd_data, cmd_size);

	/* Write VFC address */
	ARR_REG_WR(p_hwfn, p_ptt, storm->sem_fast_mem_addr + SEM_FAST_REG_VFC_ADDR, addr_data, addr_size);

	/* Read response */
	for (i = 0; i < resp_size; i++) {
		/* Poll until ready */
		do {
			ecore_grc_dump_addr_range(p_hwfn, p_ptt, &vfc_status, true, BYTES_TO_DWORDS(storm->sem_fast_mem_addr + SEM_FAST_REG_VFC_STATUS), 1, false, SPLIT_TYPE_NONE, 0);
			is_ready = vfc_status & (1 << VFC_STATUS_RESP_READY_BIT);

			if (!is_ready) {
				if (polling_count++ == VFC_POLLING_COUNT)
					return 0;

				OSAL_MSLEEP(polling_ms);
			}
		} while (!is_ready);

		ecore_grc_dump_addr_range(p_hwfn, p_ptt, dump_buf + i, true, BYTES_TO_DWORDS(storm->sem_fast_mem_addr + SEM_FAST_REG_VFC_DATA_RD), 1, false, SPLIT_TYPE_NONE, 0);
	}

	return resp_size;
}

/* Dump VFC CAM. Returns the dumped size in dwords. */
static u32 ecore_grc_dump_vfc_cam(struct ecore_hwfn *p_hwfn,
				  struct ecore_ptt *p_ptt,
				  u32 *dump_buf,
				  bool dump,
				  u8 storm_id)
{
	u32 total_size = VFC_CAM_NUM_ROWS * VFC_CAM_RESP_DWORDS;
	struct storm_defs *storm = &s_storm_defs[storm_id];
	u32 cam_addr[VFC_CAM_ADDR_DWORDS] = { 0 };
	u32 cam_cmd[VFC_CAM_CMD_DWORDS] = { 0 };
	u32 row, offset = 0;

	offset += ecore_grc_dump_mem_hdr(p_hwfn, dump_buf + offset, dump, "vfc_cam", 0, total_size, 256, false, "vfc_cam", storm->letter);

	if (!dump)
		return offset + total_size;

	/* Prepare CAM address */
	SET_VAR_FIELD(cam_addr, VFC_CAM_ADDR, OP, VFC_OPCODE_CAM_RD);

	/* Read VFC CAM data */
	for (row = 0; row < VFC_CAM_NUM_ROWS; row++) {
		SET_VAR_FIELD(cam_cmd, VFC_CAM_CMD, ROW, row);
		offset += ecore_grc_dump_read_from_vfc(p_hwfn, p_ptt, storm, cam_cmd, VFC_CAM_CMD_DWORDS, cam_addr, VFC_CAM_ADDR_DWORDS, VFC_CAM_RESP_DWORDS, dump_buf + offset);
	}

	return offset;
}

/* Dump VFC RAM. Returns the dumped size in dwords. */
static u32 ecore_grc_dump_vfc_ram(struct ecore_hwfn *p_hwfn,
				  struct ecore_ptt *p_ptt,
				  u32 *dump_buf,
				  bool dump,
				  u8 storm_id,
				  struct vfc_ram_defs *ram_defs)
{
	u32 total_size = ram_defs->num_rows * VFC_RAM_RESP_DWORDS;
	struct storm_defs *storm = &s_storm_defs[storm_id];
	u32 ram_addr[VFC_RAM_ADDR_DWORDS] = { 0 };
	u32 ram_cmd[VFC_RAM_CMD_DWORDS] = { 0 };
	u32 row, offset = 0;

	offset += ecore_grc_dump_mem_hdr(p_hwfn, dump_buf + offset, dump, ram_defs->mem_name, 0, total_size, 256, false, ram_defs->type_name, storm->letter);

	if (!dump)
		return offset + total_size;

	/* Prepare RAM address */
	SET_VAR_FIELD(ram_addr, VFC_RAM_ADDR, OP, VFC_OPCODE_RAM_RD);

	/* Read VFC RAM data */
	for (row = ram_defs->base_row; row < ram_defs->base_row + ram_defs->num_rows; row++) {
		SET_VAR_FIELD(ram_addr, VFC_RAM_ADDR, ROW, row);
		offset += ecore_grc_dump_read_from_vfc(p_hwfn, p_ptt, storm, ram_cmd, VFC_RAM_CMD_DWORDS, ram_addr, VFC_RAM_ADDR_DWORDS, VFC_RAM_RESP_DWORDS, dump_buf + offset);
	}

	return offset;
}

/* Dumps GRC VFC data. Returns the dumped size in dwords. */
static u32 ecore_grc_dump_vfc(struct ecore_hwfn *p_hwfn,
			      struct ecore_ptt *p_ptt,
			      u32 *dump_buf,
			      bool dump)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	u8 storm_id, i;
	u32 offset = 0;

	for (storm_id = 0; storm_id < MAX_DBG_STORMS; storm_id++) {
		if (!ecore_grc_is_storm_included(p_hwfn, (enum dbg_storms)storm_id) ||
		    !s_storm_defs[storm_id].has_vfc ||
		    (storm_id == DBG_PSTORM_ID && (dev_data->hw_type == HW_TYPE_EMUL_REDUCED || dev_data->hw_type == HW_TYPE_FPGA)))
			continue;

		/* Read CAM */
		offset += ecore_grc_dump_vfc_cam(p_hwfn, p_ptt, dump_buf + offset, dump, storm_id);

		/* Read RAM */
		for (i = 0; i < NUM_VFC_RAM_TYPES; i++)
			offset += ecore_grc_dump_vfc_ram(p_hwfn, p_ptt, dump_buf + offset, dump, storm_id, &s_vfc_ram_defs[i]);
	}

	return offset;
}

/* Dumps GRC RSS data. Returns the dumped size in dwords. */
static u32 ecore_grc_dump_rss(struct ecore_hwfn *p_hwfn,
			      struct ecore_ptt *p_ptt,
			      u32 *dump_buf,
			      bool dump)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	u32 offset = 0;
	u8 rss_mem_id;

	for (rss_mem_id = 0; rss_mem_id < NUM_RSS_MEM_TYPES; rss_mem_id++) {
		u32 rss_addr, num_entries, total_dwords;
		struct rss_mem_defs *rss_defs;
		bool packed;

		rss_defs = &s_rss_mem_defs[rss_mem_id];
		rss_addr = rss_defs->addr;
		num_entries = rss_defs->num_entries[dev_data->chip_id];
		total_dwords = (num_entries * rss_defs->entry_width) / 32;
		packed = (rss_defs->entry_width == 16);

		offset += ecore_grc_dump_mem_hdr(p_hwfn, dump_buf + offset, dump, rss_defs->mem_name, 0, total_dwords,	rss_defs->entry_width, packed, rss_defs->type_name, 0);

		/* Dump RSS data */
		if (!dump) {
			offset += total_dwords;
			continue;
		}

		while (total_dwords) {
			u32 num_dwords_to_read = OSAL_MIN_T(u32, RSS_REG_RSS_RAM_DATA_SIZE, total_dwords);
			ecore_wr(p_hwfn, p_ptt, RSS_REG_RSS_RAM_ADDR, rss_addr);
			offset += ecore_grc_dump_addr_range(p_hwfn, p_ptt, dump_buf + offset, dump, BYTES_TO_DWORDS(RSS_REG_RSS_RAM_DATA), num_dwords_to_read, false, SPLIT_TYPE_NONE, 0);
			total_dwords -= num_dwords_to_read;
			rss_addr++;
		}
	}

	return offset;
}

/* Dumps GRC Big RAM. Returns the dumped size in dwords. */
static u32 ecore_grc_dump_big_ram(struct ecore_hwfn *p_hwfn,
				  struct ecore_ptt *p_ptt,
				  u32 *dump_buf,
				  bool dump,
				  u8 big_ram_id)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	u32 block_size, ram_size, offset = 0, reg_val, i;
	char mem_name[12] = "???_BIG_RAM";
	char type_name[8] = "???_RAM";
	struct big_ram_defs *big_ram;

	big_ram = &s_big_ram_defs[big_ram_id];
	ram_size = big_ram->ram_size[dev_data->chip_id];

	reg_val = ecore_rd(p_hwfn, p_ptt, big_ram->is_256b_reg_addr);
	block_size = reg_val & (1 << big_ram->is_256b_bit_offset[dev_data->chip_id]) ? 256 : 128;

	OSAL_STRNCPY(type_name, big_ram->instance_name, BIG_RAM_NAME_LEN);
	OSAL_STRNCPY(mem_name, big_ram->instance_name, BIG_RAM_NAME_LEN);

	/* Dump memory header */
	offset += ecore_grc_dump_mem_hdr(p_hwfn, dump_buf + offset, dump, mem_name, 0, ram_size, block_size * 8, false, type_name, 0);

	/* Read and dump Big RAM data */
	if (!dump)
		return offset + ram_size;

	/* Dump Big RAM */
	for (i = 0; i < DIV_ROUND_UP(ram_size, BRB_REG_BIG_RAM_DATA_SIZE); i++) {
		ecore_wr(p_hwfn, p_ptt, big_ram->addr_reg_addr, i);
		offset += ecore_grc_dump_addr_range(p_hwfn, p_ptt, dump_buf + offset, dump, BYTES_TO_DWORDS(big_ram->data_reg_addr), BRB_REG_BIG_RAM_DATA_SIZE, false, SPLIT_TYPE_NONE, 0);
	}

	return offset;
}

/* Dumps MCP scratchpad. Returns the dumped size in dwords. */
static u32 ecore_grc_dump_mcp(struct ecore_hwfn *p_hwfn,
			      struct ecore_ptt *p_ptt,
			      u32 *dump_buf,
			      bool dump)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	bool block_enable[MAX_BLOCK_ID] = { 0 };
	bool halted = false;
	u32 offset = 0;

	/* Halt MCP */
	if (dump && dev_data->hw_type == HW_TYPE_ASIC && !ecore_grc_get_param(p_hwfn, DBG_GRC_PARAM_NO_MCP)) {
		halted = !ecore_mcp_halt(p_hwfn, p_ptt);
		if (!halted)
			DP_NOTICE(p_hwfn, false, "MCP halt failed!\n");
	}

	/* Dump MCP scratchpad */
	offset += ecore_grc_dump_mem(p_hwfn, p_ptt, dump_buf + offset, dump, OSAL_NULL, BYTES_TO_DWORDS(MCP_REG_SCRATCH),
		MCP_REG_SCRATCH_SIZE, false, 0, false, "MCP", 0);

	/* Dump MCP cpu_reg_file */
	offset += ecore_grc_dump_mem(p_hwfn, p_ptt, dump_buf + offset, dump, OSAL_NULL, BYTES_TO_DWORDS(MCP_REG_CPU_REG_FILE),
		MCP_REG_CPU_REG_FILE_SIZE, false, 0, false, "MCP", 0);

	/* Dump MCP registers */
	block_enable[BLOCK_MCP] = true;
	offset += ecore_grc_dump_registers(p_hwfn, p_ptt, dump_buf + offset, dump, block_enable, "MCP");

	/* Dump required non-MCP registers */
	offset += ecore_grc_dump_regs_hdr(dump_buf + offset, dump, 1, SPLIT_TYPE_NONE, 0, "MCP");
	offset += ecore_grc_dump_reg_entry(p_hwfn, p_ptt, dump_buf + offset, dump, BYTES_TO_DWORDS(MISC_REG_SHARED_MEM_ADDR), 1, false, SPLIT_TYPE_NONE, 0);

	/* Release MCP */
	if (halted && ecore_mcp_resume(p_hwfn, p_ptt))
		DP_NOTICE(p_hwfn, false, "Failed to resume MCP after halt!\n");

	return offset;
}

/* Dumps the tbus indirect memory for all PHYs.
 * Returns the dumped size in dwords.
 */
static u32 ecore_grc_dump_phy(struct ecore_hwfn *p_hwfn,
			      struct ecore_ptt *p_ptt,
			      u32 *dump_buf,
			      bool dump)
{
	u32 offset = 0, tbus_lo_offset, tbus_hi_offset;
	char mem_name[32];
	u8 phy_id;

	for (phy_id = 0; phy_id < OSAL_ARRAY_SIZE(s_phy_defs); phy_id++) {
		u32 addr_lo_addr, addr_hi_addr, data_lo_addr, data_hi_addr;
		struct phy_defs *phy_defs;
		u8 *bytes_buf;

		phy_defs = &s_phy_defs[phy_id];
		addr_lo_addr = phy_defs->base_addr + phy_defs->tbus_addr_lo_addr;
		addr_hi_addr = phy_defs->base_addr + phy_defs->tbus_addr_hi_addr;
		data_lo_addr = phy_defs->base_addr + phy_defs->tbus_data_lo_addr;
		data_hi_addr = phy_defs->base_addr + phy_defs->tbus_data_hi_addr;

		if (OSAL_SNPRINTF(mem_name, sizeof(mem_name), "tbus_%s", phy_defs->phy_name) < 0)
			DP_NOTICE(p_hwfn, true, "Unexpected debug error: invalid PHY memory name\n");

		offset += ecore_grc_dump_mem_hdr(p_hwfn, dump_buf + offset, dump, mem_name, 0, PHY_DUMP_SIZE_DWORDS, 16, true, mem_name, 0);

		if (!dump) {
			offset += PHY_DUMP_SIZE_DWORDS;
			continue;
		}

		bytes_buf = (u8 *)(dump_buf + offset);
		for (tbus_hi_offset = 0; tbus_hi_offset < (NUM_PHY_TBUS_ADDRESSES >> 8); tbus_hi_offset++) {
			ecore_wr(p_hwfn, p_ptt, addr_hi_addr, tbus_hi_offset);
			for (tbus_lo_offset = 0; tbus_lo_offset < 256; tbus_lo_offset++) {
				ecore_wr(p_hwfn, p_ptt, addr_lo_addr, tbus_lo_offset);
				*(bytes_buf++) = (u8)ecore_rd(p_hwfn, p_ptt, data_lo_addr);
				*(bytes_buf++) = (u8)ecore_rd(p_hwfn, p_ptt, data_hi_addr);
			}
		}

		offset += PHY_DUMP_SIZE_DWORDS;
	}

	return offset;
}

/* Dumps the MCP HW dump from NVRAM. Returns the dumped size in dwords. */
static u32 ecore_grc_dump_mcp_hw_dump(struct ecore_hwfn *p_hwfn,
				      struct ecore_ptt *p_ptt,
				      u32 *dump_buf,
				      bool dump)
{
	u32 hw_dump_offset_bytes = 0, hw_dump_size_bytes = 0;
	u32 hw_dump_size_dwords = 0, offset = 0;
	enum dbg_status status;

	/* Read HW dump image from NVRAM */
	status = ecore_find_nvram_image(p_hwfn, p_ptt, NVM_TYPE_HW_DUMP_OUT, &hw_dump_offset_bytes, &hw_dump_size_bytes);
	if (status != DBG_STATUS_OK)
		return 0;

	hw_dump_size_dwords = BYTES_TO_DWORDS(hw_dump_size_bytes);

	/* Dump HW dump image section */
	offset += ecore_dump_section_hdr(dump_buf + offset, dump, "mcp_hw_dump", 1);
	offset += ecore_dump_num_param(dump_buf + offset, dump, "size", hw_dump_size_dwords);

	/* Read MCP HW dump image into dump buffer */
	if (dump && hw_dump_size_dwords) {
		status = ecore_nvram_read(p_hwfn, p_ptt, hw_dump_offset_bytes, hw_dump_size_bytes, dump_buf + offset);
		if (status != DBG_STATUS_OK) {
			DP_NOTICE(p_hwfn, false, "Failed to read MCP HW Dump image from NVRAM\n");
			return 0;
		}
	}
	offset += hw_dump_size_dwords;

	return offset;
}

/* Dumps Static Debug data. Returns the dumped size in dwords. */
static u32 ecore_grc_dump_static_debug(struct ecore_hwfn *p_hwfn,
				       struct ecore_ptt *p_ptt,
				       u32 *dump_buf,
				       bool dump)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	u32 block_id, line_id, offset = 0;

	/* don't dump static debug if a debug bus recording is in progress */
	if (dump && ecore_rd(p_hwfn, p_ptt, DBG_REG_DBG_BLOCK_ON))
		return 0;

	if (dump) {
		/* Disable debug bus in all blocks */
		ecore_bus_disable_blocks(p_hwfn, p_ptt);

		ecore_bus_reset_dbg_block(p_hwfn, p_ptt);
		ecore_wr(p_hwfn, p_ptt, DBG_REG_FRAMING_MODE, DBG_BUS_FRAME_MODE_8HW);
		ecore_wr(p_hwfn, p_ptt, DBG_REG_DEBUG_TARGET, DBG_BUS_TARGET_ID_INT_BUF);
		ecore_wr(p_hwfn, p_ptt, DBG_REG_FULL_MODE, 1);
		ecore_bus_enable_dbg_block(p_hwfn, p_ptt, true);
	}

	/* Dump all static debug lines for each relevant block */
	for (block_id = 0; block_id < MAX_BLOCK_ID; block_id++) {
		const struct dbg_block_chip *block_per_chip;
		const struct dbg_block *block;
		bool is_removed, has_dbg_bus;
		u16 modes_buf_offset;
		u32 block_dwords;

		block_per_chip = ecore_get_dbg_block_per_chip(p_hwfn, (enum block_id)block_id);
		is_removed = GET_FIELD(block_per_chip->flags, DBG_BLOCK_CHIP_IS_REMOVED);
		has_dbg_bus = GET_FIELD(block_per_chip->flags, DBG_BLOCK_CHIP_HAS_DBG_BUS);

        /* Temporary workaround NWS parity (CQ106101) - Since read+clear for NWS parity is not working, skip NWS block: */
        if (block_id == BLOCK_NWS)
            continue;

		if (!is_removed && has_dbg_bus && GET_FIELD(block_per_chip->dbg_bus_mode.data, DBG_MODE_HDR_EVAL_MODE) > 0) {
			modes_buf_offset = GET_FIELD(block_per_chip->dbg_bus_mode.data, DBG_MODE_HDR_MODES_BUF_OFFSET);
			if (!ecore_is_mode_match(p_hwfn, &modes_buf_offset))
				has_dbg_bus = false;
		}

		if (is_removed || !has_dbg_bus)
			continue;

		block_dwords = NUM_DBG_LINES(block_per_chip) * STATIC_DEBUG_LINE_DWORDS;

		/* Dump static section params */
		block = get_dbg_block(p_hwfn, (enum block_id)block_id);
		offset += ecore_grc_dump_mem_hdr(p_hwfn, dump_buf + offset, dump, (const char *)block->name, 0, block_dwords, 32, false, "STATIC", 0);

		if (!dump) {
			offset += block_dwords;
			continue;
		}

		/* If all lines are invalid - dump zeros */
		if (dev_data->block_in_reset[block_id]) {
			OSAL_MEMSET(dump_buf + offset, 0, DWORDS_TO_BYTES(block_dwords));
			offset += block_dwords;
			continue;
		}

		/* Enable block's client */
		ecore_bus_enable_clients(p_hwfn, p_ptt, 1 << block_per_chip->dbg_client_id);
		for (line_id = 0; line_id < (u32)NUM_DBG_LINES(block_per_chip); line_id++) {
			/* Configure debug line ID */
			ecore_bus_config_dbg_line(p_hwfn, p_ptt, (enum block_id)block_id, (u8)line_id, 0xf, 0, 0, 0);

			/* Read debug line info */
			offset += ecore_grc_dump_addr_range(p_hwfn, p_ptt, dump_buf + offset, dump, BYTES_TO_DWORDS(DBG_REG_CALENDAR_OUT_DATA), STATIC_DEBUG_LINE_DWORDS, true, SPLIT_TYPE_NONE, 0);
		}

		/* Disable block's client and debug output */
		ecore_bus_enable_clients(p_hwfn, p_ptt, 0);
		ecore_bus_config_dbg_line(p_hwfn, p_ptt, (enum block_id)block_id, 0, 0, 0, 0, 0);
	}

	if (dump) {
		ecore_bus_enable_dbg_block(p_hwfn, p_ptt, false);
		ecore_bus_enable_clients(p_hwfn, p_ptt, 0);
	}

	return offset;
}

/* Performs GRC Dump to the specified buffer.
 * Returns the dumped size in dwords.
 */
static enum dbg_status ecore_grc_dump(struct ecore_hwfn *p_hwfn,
				      struct ecore_ptt *p_ptt,
				      u32 *dump_buf,
				      bool dump,
				      u32 *num_dumped_dwords)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	u32 dwords_read, offset = 0;
	bool is_asic, parities_masked = false;
	u8 i;

	is_asic = dev_data->hw_type == HW_TYPE_ASIC;
	*num_dumped_dwords = 0;
	dev_data->num_regs_read = 0;

	/* Update reset state */
	if (dump)
		ecore_update_blocks_reset_state(p_hwfn, p_ptt);

	/* Dump global params */
	offset += ecore_dump_common_global_params(p_hwfn, p_ptt, dump_buf + offset, dump, 4); /* 4 must match below amount of params */
	offset += /* 1 */ ecore_dump_str_param(dump_buf + offset, dump, "dump-type", "grc-dump");
	offset += /* 2 */ ecore_dump_num_param(dump_buf + offset, dump, "num-lcids", NUM_OF_LCIDS);
	offset += /* 3 */ ecore_dump_num_param(dump_buf + offset, dump, "num-ltids", NUM_OF_LTIDS);
	offset += /* 4 */ ecore_dump_num_param(dump_buf + offset, dump, "num-ports", dev_data->num_ports);
	/* Additional/Less parameters require matching of number in call to dump_common_global_params() */

	/* Dump reset registers (dumped before taking blocks out of reset ) */
	if (ecore_grc_is_included(p_hwfn, DBG_GRC_PARAM_DUMP_REGS))
		offset += ecore_grc_dump_reset_regs(p_hwfn, p_ptt, dump_buf + offset, dump);

	/* Take all blocks out of reset (using reset registers) */
	if (dump) {
		ecore_grc_unreset_blocks(p_hwfn, p_ptt, false);
		ecore_update_blocks_reset_state(p_hwfn, p_ptt);
	}

	/* Disable all parities */
	if (dump) {
		if (!is_asic) {
			/* Disable parities using GRC (no MCP) */
			ecore_mask_parities_no_mcp(p_hwfn, p_ptt, true);
			parities_masked = true;
		} else if (!ecore_grc_get_param(p_hwfn, DBG_GRC_PARAM_NO_MCP)) {
			/* Disable parities using MFW command */
			parities_masked = !ecore_mcp_mask_parities(p_hwfn, p_ptt, 1);
			if (!parities_masked) {
				DP_NOTICE(p_hwfn, false, "Failed to mask parities using MFW\n");
				if (ecore_grc_get_param(p_hwfn, DBG_GRC_PARAM_PARITY_SAFE))
					return DBG_STATUS_MCP_COULD_NOT_MASK_PRTY;
			}
		}
	}

	/* Dump modified registers (dumped before modifying them) */
	if (ecore_grc_is_included(p_hwfn, DBG_GRC_PARAM_DUMP_REGS))
		offset += ecore_grc_dump_modified_regs(p_hwfn, p_ptt, dump_buf + offset, dump);

	/* Stall storms */
	if (dump && (ecore_grc_is_included(p_hwfn, DBG_GRC_PARAM_DUMP_IOR) || ecore_grc_is_included(p_hwfn, DBG_GRC_PARAM_DUMP_VFC)))
		ecore_grc_stall_storms(p_hwfn, p_ptt, true);

	/* Dump all regs  */
	if (ecore_grc_is_included(p_hwfn, DBG_GRC_PARAM_DUMP_REGS)) {
		bool block_enable[MAX_BLOCK_ID];

		/* Dump all blocks except MCP */
		for (i = 0; i < MAX_BLOCK_ID; i++)
			block_enable[i] = true;
		block_enable[BLOCK_MCP] = false;
		offset += ecore_grc_dump_registers(p_hwfn, p_ptt, dump_buf + offset, dump, block_enable, OSAL_NULL);

		/* Dump special registers */
		offset += ecore_grc_dump_special_regs(p_hwfn, p_ptt, dump_buf + offset, dump);
	}

	/* Dump memories */
	offset += ecore_grc_dump_memories(p_hwfn, p_ptt, dump_buf + offset, dump);

	/* Dump MCP */
	if (ecore_grc_is_included(p_hwfn, DBG_GRC_PARAM_DUMP_MCP))
		offset += ecore_grc_dump_mcp(p_hwfn, p_ptt, dump_buf + offset, dump);

	/* Dump context */
	if (ecore_grc_is_included(p_hwfn, DBG_GRC_PARAM_DUMP_CM_CTX))
		offset += ecore_grc_dump_ctx(p_hwfn, p_ptt, dump_buf + offset, dump);

	/* Dump RSS memories */
	if (ecore_grc_is_included(p_hwfn, DBG_GRC_PARAM_DUMP_RSS))
		offset += ecore_grc_dump_rss(p_hwfn, p_ptt, dump_buf + offset, dump);

	/* Dump Big RAM */
	for (i = 0; i < NUM_BIG_RAM_TYPES; i++)
		if (ecore_grc_is_included(p_hwfn, s_big_ram_defs[i].grc_param))
			offset += ecore_grc_dump_big_ram(p_hwfn, p_ptt, dump_buf + offset, dump, i);

	/* Dump VFC */
	if (ecore_grc_is_included(p_hwfn, DBG_GRC_PARAM_DUMP_VFC)) {
		dwords_read = ecore_grc_dump_vfc(p_hwfn, p_ptt, dump_buf + offset, dump);
		offset += dwords_read;
		if (!dwords_read)
			return DBG_STATUS_VFC_READ_ERROR;
	}

	/* Dump PHY tbus */
	if (ecore_grc_is_included(p_hwfn, DBG_GRC_PARAM_DUMP_PHY) && dev_data->chip_id == CHIP_K2 && dev_data->hw_type == HW_TYPE_ASIC)
		offset += ecore_grc_dump_phy(p_hwfn, p_ptt, dump_buf + offset, dump);

	/* Dump MCP HW Dump */
	if (ecore_grc_is_included(p_hwfn, DBG_GRC_PARAM_DUMP_MCP_HW_DUMP) &&
	    dev_data->hw_type == HW_TYPE_ASIC &&
	    !ecore_grc_get_param(p_hwfn, DBG_GRC_PARAM_NO_MCP) &&
	    OSAL_NVM_IS_ACCESS_ENABLED(p_hwfn))
		offset += ecore_grc_dump_mcp_hw_dump(p_hwfn, p_ptt, dump_buf + offset, dump);

	/* Dump static debug data (only if not during debug bus recording) */
	if (ecore_grc_is_included(p_hwfn, DBG_GRC_PARAM_DUMP_STATIC) && (!dump || dev_data->bus.state == DBG_BUS_STATE_IDLE))
		offset += ecore_grc_dump_static_debug(p_hwfn, p_ptt, dump_buf + offset, dump);

	/* Dump last section */
	offset += ecore_dump_last_section(dump_buf, offset, dump);

	if (dump) {
		/* Unstall storms */
		if (ecore_grc_get_param(p_hwfn, DBG_GRC_PARAM_UNSTALL))
			ecore_grc_stall_storms(p_hwfn, p_ptt, false);

		/* Clear parity status */
		ecore_grc_clear_all_prty(p_hwfn, p_ptt);

		/* Enable parities */
		if (parities_masked) {
			if (!is_asic)
				ecore_mask_parities_no_mcp(p_hwfn, p_ptt, false);
			else
				ecore_mcp_mask_parities(p_hwfn, p_ptt, 0);
		}
	}

	*num_dumped_dwords = offset;

	return DBG_STATUS_OK;
}

/* Writes the specified failing Idle Check rule to the specified buffer.
 * Returns the dumped size in dwords.
 */
static u32 ecore_idle_chk_dump_failure(struct ecore_hwfn *p_hwfn,
				       struct ecore_ptt *p_ptt,
				       u32 *dump_buf,
				       bool dump,
				       u16 rule_id,
				       const struct dbg_idle_chk_rule *rule,
				       u16 fail_entry_id,
				       u32 *cond_reg_values)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	const struct dbg_idle_chk_cond_reg *cond_regs;
	const struct dbg_idle_chk_info_reg *info_regs;
	u32 i, next_reg_offset = 0, offset = 0;
	struct dbg_idle_chk_result_hdr *hdr;
	const union dbg_idle_chk_reg *regs;
	u8 reg_id;

	hdr = (struct dbg_idle_chk_result_hdr *)dump_buf;
	regs = (const union dbg_idle_chk_reg *)p_hwfn->dbg_arrays[BIN_BUF_DBG_IDLE_CHK_REGS].ptr + rule->reg_offset;
	cond_regs = &regs[0].cond_reg;
	info_regs = &regs[rule->num_cond_regs].info_reg;

	/* Dump rule data */
	if (dump) {
		OSAL_MEMSET(hdr, 0, sizeof(*hdr));
		hdr->rule_id = rule_id;
		hdr->mem_entry_id = fail_entry_id;
		hdr->severity = rule->severity;
		hdr->num_dumped_cond_regs = rule->num_cond_regs;
	}

	offset += IDLE_CHK_RESULT_HDR_DWORDS;

	/* Dump condition register values */
	for (reg_id = 0; reg_id < rule->num_cond_regs; reg_id++) {
		const struct dbg_idle_chk_cond_reg *reg = &cond_regs[reg_id];
		struct dbg_idle_chk_result_reg_hdr *reg_hdr;

		reg_hdr = (struct dbg_idle_chk_result_reg_hdr *)(dump_buf + offset);

		/* Write register header */
		if (!dump) {
			offset += IDLE_CHK_RESULT_REG_HDR_DWORDS + reg->entry_size;
			continue;
		}

		offset += IDLE_CHK_RESULT_REG_HDR_DWORDS;
		OSAL_MEMSET(reg_hdr, 0, sizeof(*reg_hdr));
		reg_hdr->start_entry = reg->start_entry;
		reg_hdr->size = reg->entry_size;
		SET_FIELD(reg_hdr->data, DBG_IDLE_CHK_RESULT_REG_HDR_IS_MEM, reg->num_entries > 1 || reg->start_entry > 0 ? 1 : 0);
		SET_FIELD(reg_hdr->data, DBG_IDLE_CHK_RESULT_REG_HDR_REG_ID, reg_id);

		/* Write register values */
		for (i = 0; i < reg_hdr->size; i++, next_reg_offset++, offset++)
			dump_buf[offset] = cond_reg_values[next_reg_offset];
	}

	/* Dump info register values */
	for (reg_id = 0; reg_id < rule->num_info_regs; reg_id++) {
		const struct dbg_idle_chk_info_reg *reg = &info_regs[reg_id];
		u32 block_id;

		/* Check if register's block is in reset */
		if (!dump) {
			offset += IDLE_CHK_RESULT_REG_HDR_DWORDS + reg->size;
			continue;
		}

		block_id = GET_FIELD(reg->data, DBG_IDLE_CHK_INFO_REG_BLOCK_ID);
		if (block_id >= MAX_BLOCK_ID) {
			DP_NOTICE(p_hwfn, true, "Invalid block_id\n");
			return 0;
		}

		if (!dev_data->block_in_reset[block_id]) {
			struct dbg_idle_chk_result_reg_hdr *reg_hdr;
			bool wide_bus, eval_mode, mode_match = true;
			u16 modes_buf_offset;
			u32 addr;

			reg_hdr = (struct dbg_idle_chk_result_reg_hdr *)(dump_buf + offset);

			/* Check mode */
			eval_mode = GET_FIELD(reg->mode.data, DBG_MODE_HDR_EVAL_MODE) > 0;
			if (eval_mode) {
				modes_buf_offset = GET_FIELD(reg->mode.data, DBG_MODE_HDR_MODES_BUF_OFFSET);
				mode_match = ecore_is_mode_match(p_hwfn, &modes_buf_offset);
			}

			if (!mode_match)
				continue;

			addr = GET_FIELD(reg->data, DBG_IDLE_CHK_INFO_REG_ADDRESS);
			wide_bus = GET_FIELD(reg->data, DBG_IDLE_CHK_INFO_REG_WIDE_BUS);

			/* Write register header */
			offset += IDLE_CHK_RESULT_REG_HDR_DWORDS;
			hdr->num_dumped_info_regs++;
			OSAL_MEMSET(reg_hdr, 0, sizeof(*reg_hdr));
			reg_hdr->size = reg->size;
			SET_FIELD(reg_hdr->data, DBG_IDLE_CHK_RESULT_REG_HDR_REG_ID, rule->num_cond_regs + reg_id);

			/* Write register values */
			offset += ecore_grc_dump_addr_range(p_hwfn, p_ptt, dump_buf + offset, dump, addr, reg->size, wide_bus, SPLIT_TYPE_NONE, 0);
		}
	}

	return offset;
}

/* Dumps idle check rule entries. Returns the dumped size in dwords. */
static u32 ecore_idle_chk_dump_rule_entries(struct ecore_hwfn *p_hwfn,
					    struct ecore_ptt *p_ptt,
					    u32 *dump_buf,
					    bool dump,
					    const struct dbg_idle_chk_rule *input_rules,
					    u32 num_input_rules,
					    u32 *num_failing_rules)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	u32 cond_reg_values[IDLE_CHK_MAX_ENTRIES_SIZE];
	u32 i, offset = 0;
	u16 entry_id;
	u8 reg_id;

	*num_failing_rules = 0;

	for (i = 0; i < num_input_rules; i++) {
		const struct dbg_idle_chk_cond_reg *cond_regs;
		const struct dbg_idle_chk_rule *rule;
		const union dbg_idle_chk_reg *regs;
		u16 num_reg_entries = 1;
		bool check_rule = true;
		const u32 *imm_values;

		rule = &input_rules[i];
		regs = (const union dbg_idle_chk_reg *)p_hwfn->dbg_arrays[BIN_BUF_DBG_IDLE_CHK_REGS].ptr + rule->reg_offset;
		cond_regs = &regs[0].cond_reg;
		imm_values = (u32 *)p_hwfn->dbg_arrays[BIN_BUF_DBG_IDLE_CHK_IMMS].ptr + rule->imm_offset;

		/* Check if all condition register blocks are out of reset, and
		 * find maximal number of entries (all condition registers that
		 * are memories must have the same size, which is > 1).
		 */
		for (reg_id = 0; reg_id < rule->num_cond_regs && check_rule; reg_id++) {
			u32 block_id = GET_FIELD(cond_regs[reg_id].data, DBG_IDLE_CHK_COND_REG_BLOCK_ID);

			if (block_id >= MAX_BLOCK_ID) {
				DP_NOTICE(p_hwfn, true, "Invalid block_id\n");
				return 0;
			}

			check_rule = !dev_data->block_in_reset[block_id];
			if (cond_regs[reg_id].num_entries > num_reg_entries)
				num_reg_entries = cond_regs[reg_id].num_entries;
		}

		if (!check_rule && dump)
			continue;

		if (!dump) {
			u32 entry_dump_size = ecore_idle_chk_dump_failure(p_hwfn, p_ptt, dump_buf + offset, false, rule->rule_id, rule, 0, OSAL_NULL);

			offset += num_reg_entries * entry_dump_size;
			(*num_failing_rules) += num_reg_entries;
			continue;
		}

		/* Go over all register entries (number of entries is the same for all
		 * condition registers).
		 */
		for (entry_id = 0; entry_id < num_reg_entries; entry_id++) {
			u32 next_reg_offset = 0;

			/* Read current entry of all condition registers */
			for (reg_id = 0; reg_id < rule->num_cond_regs; reg_id++) {
				const struct dbg_idle_chk_cond_reg *reg = &cond_regs[reg_id];
				u32 padded_entry_size, addr;
				bool wide_bus;

				/* Find GRC address (if it's a memory, the address of the
				 * specific entry is calculated).
				 */
				addr = GET_FIELD(reg->data, DBG_IDLE_CHK_COND_REG_ADDRESS);
				wide_bus = GET_FIELD(reg->data, DBG_IDLE_CHK_COND_REG_WIDE_BUS);
				if (reg->num_entries > 1 || reg->start_entry > 0) {
					padded_entry_size = reg->entry_size > 1 ? OSAL_ROUNDUP_POW_OF_TWO(reg->entry_size) : 1;
					addr += (reg->start_entry + entry_id) * padded_entry_size;
				}

				/* Read registers */
				if (next_reg_offset + reg->entry_size >= IDLE_CHK_MAX_ENTRIES_SIZE) {
					DP_NOTICE(p_hwfn, true, "idle check registers entry is too large\n");
					return 0;
				}

				next_reg_offset += ecore_grc_dump_addr_range(p_hwfn, p_ptt, cond_reg_values + next_reg_offset, dump, addr, reg->entry_size, wide_bus, SPLIT_TYPE_NONE, 0);
			}

			/* Call rule condition function. if returns true, it's a failure.*/
			if ((*cond_arr[rule->cond_id])(cond_reg_values, imm_values)) {
				offset += ecore_idle_chk_dump_failure(p_hwfn, p_ptt, dump_buf + offset, dump, rule->rule_id, rule, entry_id, cond_reg_values);
				(*num_failing_rules)++;
			}
		}
	}

	return offset;
}

/* Performs Idle Check Dump to the specified buffer.
 * Returns the dumped size in dwords.
 */
static u32 ecore_idle_chk_dump(struct ecore_hwfn *p_hwfn,
			       struct ecore_ptt *p_ptt,
			       u32 *dump_buf,
			       bool dump)
{
	struct virt_mem_desc *dbg_buf = &p_hwfn->dbg_arrays[BIN_BUF_DBG_IDLE_CHK_RULES];
	u32 num_failing_rules_offset, offset = 0, input_offset = 0, num_failing_rules = 0;

	/* Dump global params */
	offset += ecore_dump_common_global_params(p_hwfn, p_ptt, dump_buf + offset, dump, 1); /* 1 must match below amount of params */
	offset += /* 1 */ ecore_dump_str_param(dump_buf + offset, dump, "dump-type", "idle-chk");
	/* Additional/Less parameters require matching of number in call to dump_common_global_params() */

	/* Dump idle check section header with a single parameter */
	offset += ecore_dump_section_hdr(dump_buf + offset, dump, "idle_chk", 1);
	num_failing_rules_offset = offset;
	offset += ecore_dump_num_param(dump_buf + offset, dump, "num_rules", 0);

	while (input_offset < BYTES_TO_DWORDS(dbg_buf->size)) {
		const struct dbg_idle_chk_cond_hdr *cond_hdr = (const struct dbg_idle_chk_cond_hdr *)dbg_buf->ptr + input_offset++;
		bool eval_mode, mode_match = true;
		u32 curr_failing_rules;
		u16 modes_buf_offset;

		/* Check mode */
		eval_mode = GET_FIELD(cond_hdr->mode.data, DBG_MODE_HDR_EVAL_MODE) > 0;
		if (eval_mode) {
			modes_buf_offset = GET_FIELD(cond_hdr->mode.data, DBG_MODE_HDR_MODES_BUF_OFFSET);
			mode_match = ecore_is_mode_match(p_hwfn, &modes_buf_offset);
		}

		if (mode_match) {
			const struct dbg_idle_chk_rule *rule = (const struct dbg_idle_chk_rule *)((u32 *)dbg_buf->ptr + input_offset);

			offset += ecore_idle_chk_dump_rule_entries(p_hwfn, p_ptt, dump_buf + offset, dump, rule, cond_hdr->data_size / IDLE_CHK_RULE_SIZE_DWORDS, &curr_failing_rules);
			num_failing_rules += curr_failing_rules;
		}

		input_offset += cond_hdr->data_size;
	}

	/* Overwrite num_rules parameter */
	if (dump)
		ecore_dump_num_param(dump_buf + num_failing_rules_offset, dump, "num_rules", num_failing_rules);

	/* Dump last section */
	offset += ecore_dump_last_section(dump_buf, offset, dump);

	return offset;
}

/* Get info on the MCP Trace data in the scratchpad:
 * - trace_data_grc_addr (OUT): trace data GRC address in bytes
 * - trace_data_size (OUT): trace data size in bytes (without the header)
 */
static enum dbg_status ecore_mcp_trace_get_data_info(struct ecore_hwfn *p_hwfn,
						     struct ecore_ptt *p_ptt,
						     u32 *trace_data_grc_addr,
						     u32 *trace_data_size)
{
	u32 spad_trace_offsize, signature;

	/* Read trace section offsize structure from MCP scratchpad */
	spad_trace_offsize = ecore_rd(p_hwfn, p_ptt, MCP_SPAD_TRACE_OFFSIZE_ADDR);

	/* Extract trace section address from offsize (in scratchpad) */
	*trace_data_grc_addr = MCP_REG_SCRATCH + SECTION_OFFSET(spad_trace_offsize);

	/* Read signature from MCP trace section */
	signature = ecore_rd(p_hwfn, p_ptt, *trace_data_grc_addr + OFFSETOF(struct mcp_trace, signature));

	if (signature != MFW_TRACE_SIGNATURE)
		return DBG_STATUS_INVALID_TRACE_SIGNATURE;

	/* Read trace size from MCP trace section */
	*trace_data_size = ecore_rd(p_hwfn, p_ptt, *trace_data_grc_addr + OFFSETOF(struct mcp_trace, size));

	return DBG_STATUS_OK;
}

/* Reads MCP trace meta data image from NVRAM
 * - running_bundle_id (OUT): running bundle ID (invalid when loaded from file)
 * - trace_meta_offset (OUT): trace meta offset in NVRAM in bytes (invalid when
 *			      loaded from file).
 * - trace_meta_size (OUT):   size in bytes of the trace meta data.
 */
static enum dbg_status ecore_mcp_trace_get_meta_info(struct ecore_hwfn *p_hwfn,
						     struct ecore_ptt *p_ptt,
						     u32 trace_data_size_bytes,
						     u32 *running_bundle_id,
						     u32 *trace_meta_offset,
						     u32 *trace_meta_size)
{
	u32 spad_trace_offsize, nvram_image_type, running_mfw_addr;

	/* Read MCP trace section offsize structure from MCP scratchpad */
	spad_trace_offsize = ecore_rd(p_hwfn, p_ptt, MCP_SPAD_TRACE_OFFSIZE_ADDR);

	/* Find running bundle ID */
	running_mfw_addr = MCP_REG_SCRATCH + SECTION_OFFSET(spad_trace_offsize) + SECTION_SIZE(spad_trace_offsize) + trace_data_size_bytes;
	*running_bundle_id = ecore_rd(p_hwfn, p_ptt, running_mfw_addr);
	if (*running_bundle_id > 1)
		return DBG_STATUS_INVALID_NVRAM_BUNDLE;

	/* Find image in NVRAM */
	nvram_image_type = (*running_bundle_id == DIR_ID_1) ? NVM_TYPE_MFW_TRACE1 : NVM_TYPE_MFW_TRACE2;
	return ecore_find_nvram_image(p_hwfn, p_ptt, nvram_image_type, trace_meta_offset, trace_meta_size);
}

/* Reads the MCP Trace meta data from NVRAM into the specified buffer */
static enum dbg_status ecore_mcp_trace_read_meta(struct ecore_hwfn *p_hwfn,
						 struct ecore_ptt *p_ptt,
						 u32 nvram_offset_in_bytes,
						 u32 size_in_bytes,
						 u32 *buf)
{
	u8 modules_num, module_len, i, *byte_buf = (u8 *)buf;
	enum dbg_status status;
	u32 signature;

	/* Read meta data from NVRAM */
	status = ecore_nvram_read(p_hwfn, p_ptt, nvram_offset_in_bytes, size_in_bytes, buf);
	if (status != DBG_STATUS_OK)
		return status;

	/* Extract and check first signature */
	signature = ecore_read_unaligned_dword(byte_buf);
	byte_buf += sizeof(signature);
	if (signature != NVM_MAGIC_VALUE)
		return DBG_STATUS_INVALID_TRACE_SIGNATURE;

	/* Extract number of modules */
	modules_num = *(byte_buf++);

	/* Skip all modules */
	for (i = 0; i < modules_num; i++) {
		module_len = *(byte_buf++);
		byte_buf += module_len;
	}

	/* Extract and check second signature */
	signature = ecore_read_unaligned_dword(byte_buf);
	byte_buf += sizeof(signature);
	if (signature != NVM_MAGIC_VALUE)
		return DBG_STATUS_INVALID_TRACE_SIGNATURE;

	return DBG_STATUS_OK;
}

/* Dump MCP Trace */
static enum dbg_status ecore_mcp_trace_dump(struct ecore_hwfn *p_hwfn,
					    struct ecore_ptt *p_ptt,
					    u32 *dump_buf,
					    bool dump,
					    u32 *num_dumped_dwords)
{
	u32 trace_meta_offset_bytes = 0, trace_meta_size_bytes = 0, trace_meta_size_dwords = 0;
	u32 trace_data_grc_addr, trace_data_size_bytes, trace_data_size_dwords;
	u32 running_bundle_id, offset = 0;
	enum dbg_status status;
	int halted = 0;
	bool use_mfw;

	*num_dumped_dwords = 0;

	use_mfw = !ecore_grc_get_param(p_hwfn, DBG_GRC_PARAM_NO_MCP);

	/* Get trace data info */
	status = ecore_mcp_trace_get_data_info(p_hwfn, p_ptt, &trace_data_grc_addr, &trace_data_size_bytes);
	if (status != DBG_STATUS_OK)
		return status;

	/* Dump global params */
	offset += ecore_dump_common_global_params(p_hwfn, p_ptt, dump_buf + offset, dump, 1); /* 1 must match below amount of params */
	offset += /* 1 */ ecore_dump_str_param(dump_buf + offset, dump, "dump-type", "mcp-trace");
	/* Additional/Less parameters require matching of number in call to dump_common_global_params() */

	/* Halt MCP while reading from scratchpad so the read data will be
	 * consistent. if halt fails, MCP trace is taken anyway, with a small
	 * risk that it may be corrupt.
	 */
	if (dump && use_mfw) {
		halted = !ecore_mcp_halt(p_hwfn, p_ptt);
		if (!halted)
			DP_NOTICE(p_hwfn, false, "MCP halt failed!\n");
	}

	/* Find trace data size */
	trace_data_size_dwords = DIV_ROUND_UP(trace_data_size_bytes + sizeof(struct mcp_trace), BYTES_IN_DWORD);

	/* Dump trace data section header and param */
	offset += ecore_dump_section_hdr(dump_buf + offset, dump, "mcp_trace_data", 1);
	offset += ecore_dump_num_param(dump_buf + offset, dump, "size", trace_data_size_dwords);

	/* Read trace data from scratchpad into dump buffer */
	offset += ecore_grc_dump_addr_range(p_hwfn, p_ptt, dump_buf + offset, dump, BYTES_TO_DWORDS(trace_data_grc_addr), trace_data_size_dwords, false, SPLIT_TYPE_NONE, 0);

	/* Resume MCP (only if halt succeeded) */
	if (halted && ecore_mcp_resume(p_hwfn, p_ptt))
		DP_NOTICE(p_hwfn, false, "Failed to resume MCP after halt!\n");

	/* Dump trace meta section header */
	offset += ecore_dump_section_hdr(dump_buf + offset, dump, "mcp_trace_meta", 1);

	/* If MCP Trace meta size parameter was set, use it.
	 * Otherwise, read trace meta (if NVRAM access is enabled).
	 * trace_meta_size_bytes is dword-aligned.
	 */
	trace_meta_size_bytes = ecore_grc_get_param(p_hwfn, DBG_GRC_PARAM_MCP_TRACE_META_SIZE);
	if ((!trace_meta_size_bytes || dump) && OSAL_NVM_IS_ACCESS_ENABLED(p_hwfn) && use_mfw)
		status = ecore_mcp_trace_get_meta_info(p_hwfn, p_ptt, trace_data_size_bytes, &running_bundle_id, &trace_meta_offset_bytes, &trace_meta_size_bytes);
	if (status == DBG_STATUS_OK)
		trace_meta_size_dwords = BYTES_TO_DWORDS(trace_meta_size_bytes);

	/* Dump trace meta size param */
	offset += ecore_dump_num_param(dump_buf + offset, dump, "size", trace_meta_size_dwords);

	/* Read trace meta image into dump buffer */
	if (dump && trace_meta_size_dwords)
		status = ecore_mcp_trace_read_meta(p_hwfn, p_ptt, trace_meta_offset_bytes, trace_meta_size_bytes, dump_buf + offset);
	if (status == DBG_STATUS_OK)
		offset += trace_meta_size_dwords;

	/* Dump last section */
	offset += ecore_dump_last_section(dump_buf, offset, dump);

	*num_dumped_dwords = offset;

	/* If MFW can't be used, indicate that the dump doesn't contain the
	 * meta data from NVRAM.
	 */
	return use_mfw ? status : DBG_STATUS_NVRAM_GET_IMAGE_FAILED;
}

/* Dump GRC FIFO */
static enum dbg_status ecore_reg_fifo_dump(struct ecore_hwfn *p_hwfn,
					   struct ecore_ptt *p_ptt,
					   u32 *dump_buf,
					   bool dump,
					   u32 *num_dumped_dwords)
{
	u32 dwords_read, size_param_offset, offset = 0;
	bool fifo_has_data;

	*num_dumped_dwords = 0;

	/* Dump global params */
	offset += ecore_dump_common_global_params(p_hwfn, p_ptt, dump_buf + offset, dump, 1); /* 1 must match below amount of params */
	offset += /* 1 */ ecore_dump_str_param(dump_buf + offset, dump, "dump-type", "reg-fifo");
	/* Additional/Less parameters require matching of number in call to dump_common_global_params() */

	/* Dump fifo data section header and param. The size param is 0 for
	 * now, and is overwritten after reading the FIFO.
	 */
	offset += ecore_dump_section_hdr(dump_buf + offset, dump, "reg_fifo_data", 1);
	size_param_offset = offset;
	offset += ecore_dump_num_param(dump_buf + offset, dump, "size", 0);

	if (dump) {
		fifo_has_data = ecore_rd(p_hwfn, p_ptt, GRC_REG_TRACE_FIFO_VALID_DATA) > 0;

		/* Pull available data from fifo. Use DMAE since this is
		 * widebus memory and must be accessed atomically. Test for
		 * dwords_read not passing buffer size since more entries could
		 * be added to the buffer as we
		 * are emptying it.
		 */
		for (dwords_read = 0; fifo_has_data && dwords_read < REG_FIFO_DEPTH_DWORDS; dwords_read += REG_FIFO_ELEMENT_DWORDS) {
			offset += ecore_grc_dump_addr_range(p_hwfn, p_ptt, dump_buf + offset, true, BYTES_TO_DWORDS(GRC_REG_TRACE_FIFO), REG_FIFO_ELEMENT_DWORDS, true, SPLIT_TYPE_NONE, 0);
			fifo_has_data = ecore_rd(p_hwfn, p_ptt, GRC_REG_TRACE_FIFO_VALID_DATA) > 0;
		}

		ecore_dump_num_param(dump_buf + size_param_offset, dump, "size", dwords_read);
	} else {
		/* FIFO max size is REG_FIFO_DEPTH_DWORDS. There is no way to
		 * test how much data is available, except for reading it.
		 */
		offset += REG_FIFO_DEPTH_DWORDS;
	}

	/* Dump last section */
	offset += ecore_dump_last_section(dump_buf, offset, dump);

	*num_dumped_dwords = offset;

	return DBG_STATUS_OK;
}

/* Dump IGU FIFO */
static enum dbg_status ecore_igu_fifo_dump(struct ecore_hwfn *p_hwfn,
					   struct ecore_ptt *p_ptt,
					   u32 *dump_buf,
					   bool dump,
					   u32 *num_dumped_dwords)
{
	u32 dwords_read, size_param_offset, offset = 0;
	bool fifo_has_data;

	*num_dumped_dwords = 0;

	/* Dump global params */
	offset += ecore_dump_common_global_params(p_hwfn, p_ptt, dump_buf + offset, dump, 1); /* 1 must match below amount of params */
	offset += /* 1 */ ecore_dump_str_param(dump_buf + offset, dump, "dump-type", "igu-fifo");
	/* Additional/Less parameters require matching of number in call to dump_common_global_params() */

	/* Dump fifo data section header and param. The size param is 0 for
	 * now, and is overwritten after reading the FIFO.
	 */
	offset += ecore_dump_section_hdr(dump_buf + offset, dump, "igu_fifo_data", 1);
	size_param_offset = offset;
	offset += ecore_dump_num_param(dump_buf + offset, dump, "size", 0);

	if (dump) {
		fifo_has_data = ecore_rd(p_hwfn, p_ptt, IGU_REG_ERROR_HANDLING_DATA_VALID) > 0;

		/* Pull available data from fifo. Use DMAE since this is
		 * widebus memory and must be accessed atomically. Test for
		 * dwords_read not passing buffer size since more entries could
		 * be added to the buffer as we are emptying it.
		 */
		for (dwords_read = 0; fifo_has_data && dwords_read < IGU_FIFO_DEPTH_DWORDS; dwords_read += IGU_FIFO_ELEMENT_DWORDS) {
			offset += ecore_grc_dump_addr_range(p_hwfn, p_ptt, dump_buf + offset, true, BYTES_TO_DWORDS(IGU_REG_ERROR_HANDLING_MEMORY), IGU_FIFO_ELEMENT_DWORDS, true, SPLIT_TYPE_NONE, 0);
			fifo_has_data = ecore_rd(p_hwfn, p_ptt, IGU_REG_ERROR_HANDLING_DATA_VALID) > 0;
		}

		ecore_dump_num_param(dump_buf + size_param_offset, dump, "size", dwords_read);
	} else {
		/* FIFO max size is IGU_FIFO_DEPTH_DWORDS. There is no way to
		 * test how much data is available, except for reading it.
		 */
		offset += IGU_FIFO_DEPTH_DWORDS;
	}

	/* Dump last section */
	offset += ecore_dump_last_section(dump_buf, offset, dump);

	*num_dumped_dwords = offset;

	return DBG_STATUS_OK;
}

/* Protection Override dump */
static enum dbg_status ecore_protection_override_dump(struct ecore_hwfn *p_hwfn,
						      struct ecore_ptt *p_ptt,
						      u32 *dump_buf,
						      bool dump,
						      u32 *num_dumped_dwords)
{
	u32 size_param_offset, override_window_dwords, offset = 0;

	*num_dumped_dwords = 0;

	/* Dump global params */
	offset += ecore_dump_common_global_params(p_hwfn, p_ptt, dump_buf + offset, dump, 1); /* 1 must match below amount of params */
	offset += /* 1 */ ecore_dump_str_param(dump_buf + offset, dump, "dump-type", "protection-override");
	/* Additional/Less parameters require matching of number in call to dump_common_global_params() */

	/* Dump data section header and param. The size param is 0 for now,
	 * and is overwritten after reading the data.
	 */
	offset += ecore_dump_section_hdr(dump_buf + offset, dump, "protection_override_data", 1);
	size_param_offset = offset;
	offset += ecore_dump_num_param(dump_buf + offset, dump, "size", 0);

	if (dump) {
		/* Add override window info to buffer */
		override_window_dwords = ecore_rd(p_hwfn, p_ptt, GRC_REG_NUMBER_VALID_OVERRIDE_WINDOW) * PROTECTION_OVERRIDE_ELEMENT_DWORDS;
		if (override_window_dwords) {
			offset += ecore_grc_dump_addr_range(p_hwfn, p_ptt, dump_buf + offset, true, BYTES_TO_DWORDS(GRC_REG_PROTECTION_OVERRIDE_WINDOW), override_window_dwords, true, SPLIT_TYPE_NONE, 0);
			ecore_dump_num_param(dump_buf + size_param_offset, dump, "size", override_window_dwords);
		}
	} else {
		offset += PROTECTION_OVERRIDE_DEPTH_DWORDS;
	}

	/* Dump last section */
	offset += ecore_dump_last_section(dump_buf, offset, dump);

	*num_dumped_dwords = offset;

	return DBG_STATUS_OK;
}

/* Performs FW Asserts Dump to the specified buffer.
 * Returns the dumped size in dwords.
 */
static u32 ecore_fw_asserts_dump(struct ecore_hwfn *p_hwfn,
				 struct ecore_ptt *p_ptt,
				 u32 *dump_buf,
				 bool dump)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	struct fw_asserts_ram_section *asserts;
	char storm_letter_str[2] = "?";
	struct fw_info fw_info;
	u32 offset = 0;
	u8 storm_id;

	/* Dump global params */
	offset += ecore_dump_common_global_params(p_hwfn, p_ptt, dump_buf + offset, dump, 1); /* 1 must match below amount of params */
	offset += /* 1 */ ecore_dump_str_param(dump_buf + offset, dump, "dump-type", "fw-asserts");
	/* Additional/Less parameters require matching of number in call to dump_common_global_params() */

	/* Find Storm dump size */
	for (storm_id = 0; storm_id < MAX_DBG_STORMS; storm_id++) {
		u32 fw_asserts_section_addr, next_list_idx_addr, next_list_idx, last_list_idx, addr;
		struct storm_defs *storm = &s_storm_defs[storm_id];

		if (dev_data->block_in_reset[storm->sem_block_id])
			continue;

		/* Read FW info for the current Storm  */
		ecore_read_storm_fw_info(p_hwfn, p_ptt, storm_id, &fw_info);

		asserts = &fw_info.fw_asserts_section;

		/* Dump FW Asserts section header and params */
		storm_letter_str[0] = storm->letter;
		offset += ecore_dump_section_hdr(dump_buf + offset, dump, "fw_asserts", 2);
		offset += ecore_dump_str_param(dump_buf + offset, dump, "storm", storm_letter_str);
		offset += ecore_dump_num_param(dump_buf + offset, dump, "size", asserts->list_element_dword_size);

		/* Read and dump FW Asserts data */
		if (!dump) {
			offset += asserts->list_element_dword_size;
			continue;
		}

		fw_asserts_section_addr = storm->sem_fast_mem_addr + SEM_FAST_REG_INT_RAM +
			RAM_LINES_TO_BYTES(asserts->section_ram_line_offset);
		next_list_idx_addr = fw_asserts_section_addr + DWORDS_TO_BYTES(asserts->list_next_index_dword_offset);
		next_list_idx = ecore_rd(p_hwfn, p_ptt, next_list_idx_addr);
		last_list_idx = (next_list_idx > 0 ? next_list_idx : asserts->list_num_elements) - 1;
		addr = BYTES_TO_DWORDS(fw_asserts_section_addr) + asserts->list_dword_offset +
		       last_list_idx * asserts->list_element_dword_size;
		offset += ecore_grc_dump_addr_range(p_hwfn, p_ptt, dump_buf + offset, dump, addr, asserts->list_element_dword_size, false, SPLIT_TYPE_NONE, 0);
	}

	/* Dump last section */
	offset += ecore_dump_last_section(dump_buf, offset, dump);

	return offset;
}

/* Dumps the specified ILT pages to the specified buffer.
 * Returns the dumped size in dwords.
 */
static u32 ecore_ilt_dump_pages_range(u32 *dump_buf,
				      bool dump,
				      u32 start_page_id,
				      u32 num_pages,
				      struct phys_mem_desc *ilt_pages,
				      bool dump_page_ids)
{
	u32 page_id, end_page_id, offset = 0;

	if (num_pages == 0)
		return offset;

	end_page_id = start_page_id + num_pages - 1;

	for (page_id = start_page_id; page_id <= end_page_id; page_id++) {
		struct phys_mem_desc *mem_desc = &ilt_pages[page_id];

        /**
         *
         * if (page_id >= ->p_cxt_mngr->ilt_shadow_size)
         *     break;
         */

		if (!ilt_pages[page_id].virt_addr)
			continue;

		if (dump_page_ids) {
			/* Copy page ID to dump buffer */
			if (dump)
				*(dump_buf + offset) = page_id;
			offset++;
		} else {
			/* Copy page memory to dump buffer */
			if (dump)
				OSAL_MEMCPY(dump_buf + offset, mem_desc->virt_addr, mem_desc->size);
			offset += BYTES_TO_DWORDS(mem_desc->size);
		}
	}

	return offset;
}

/* Dumps a section containing the dumped ILT pages.
 * Returns the dumped size in dwords.
 */
static u32 ecore_ilt_dump_pages_section(struct ecore_hwfn *p_hwfn,
					u32 *dump_buf,
					bool dump,
					u32 valid_conn_pf_pages,
					u32 valid_conn_vf_pages,
					struct phys_mem_desc *ilt_pages,
					bool dump_page_ids)
{
#if ((!defined VMWARE) && (!defined UEFI))
	struct ecore_ilt_client_cfg *clients = p_hwfn->p_cxt_mngr->clients;
#endif
	u32 pf_start_line, start_page_id, offset = 0;
	u32 cdut_pf_init_pages, cdut_vf_init_pages;
	u32 cdut_pf_work_pages, cdut_vf_work_pages;
	u32 base_data_offset, size_param_offset;
	u32 cdut_pf_pages, cdut_vf_pages;
	const char *section_name;
	u8 i;

	section_name = dump_page_ids ? "ilt_page_ids" : "ilt_page_mem";
	cdut_pf_init_pages = ecore_get_cdut_num_pf_init_pages(p_hwfn);
	cdut_vf_init_pages = ecore_get_cdut_num_vf_init_pages(p_hwfn);
	cdut_pf_work_pages = ecore_get_cdut_num_pf_work_pages(p_hwfn);
	cdut_vf_work_pages = ecore_get_cdut_num_vf_work_pages(p_hwfn);
	cdut_pf_pages = cdut_pf_init_pages + cdut_pf_work_pages;
	cdut_vf_pages = cdut_vf_init_pages + cdut_vf_work_pages;
	pf_start_line = p_hwfn->p_cxt_mngr->pf_start_line; 

	offset += ecore_dump_section_hdr(dump_buf + offset, dump, section_name, 1);

	/* Dump size parameter (0 for now, overwritten with real size later) */
	size_param_offset = offset;
	offset += ecore_dump_num_param(dump_buf + offset, dump, "size", 0);
	base_data_offset = offset;

	/* CDUC pages are ordered as follows:
	* - PF pages - valid section (included in PF connection type mapping)
	* - PF pages - invalid section (not dumped)
	* - For each VF in the PF:
	*   - VF pages - valid section (included in VF connection type mapping)
	*   - VF pages - invalid section (not dumped)
	*/
	if (ecore_grc_get_param(p_hwfn, DBG_GRC_PARAM_DUMP_ILT_CDUC)) {

		/* Dump connection PF pages */
		start_page_id = clients[ILT_CLI_CDUC].first.val - pf_start_line;
		offset += ecore_ilt_dump_pages_range(dump_buf + offset, dump, start_page_id, valid_conn_pf_pages, ilt_pages, dump_page_ids);

		/* Dump connection VF pages */
		start_page_id += clients[ILT_CLI_CDUC].pf_total_lines;
		for (i = 0; i < p_hwfn->p_cxt_mngr->vf_count; i++, start_page_id += clients[ILT_CLI_CDUC].vf_total_lines)
			offset += ecore_ilt_dump_pages_range(dump_buf + offset, dump, start_page_id, valid_conn_vf_pages, ilt_pages, dump_page_ids);
	}

	/* CDUT pages are ordered as follows:
	* - PF init pages (not dumped)
	* - PF work pages
	* - For each VF in the PF:
	*   - VF init pages (not dumped)
	*   - VF work pages
	*/
	if (ecore_grc_get_param(p_hwfn, DBG_GRC_PARAM_DUMP_ILT_CDUT)) {

		/* Dump task PF pages */
		start_page_id = clients[ILT_CLI_CDUT].first.val + cdut_pf_init_pages - pf_start_line;
		offset += ecore_ilt_dump_pages_range(dump_buf + offset, dump, start_page_id, cdut_pf_work_pages, ilt_pages, dump_page_ids);

		/* Dump task VF pages */
		start_page_id = clients[ILT_CLI_CDUT].first.val + cdut_pf_pages + cdut_vf_init_pages - pf_start_line;
		for (i = 0; i < p_hwfn->p_cxt_mngr->vf_count; i++, start_page_id += cdut_vf_pages)
			offset += ecore_ilt_dump_pages_range(dump_buf + offset, dump, start_page_id, cdut_vf_work_pages, ilt_pages, dump_page_ids);
	}

	/* Overwrite size param */
	if (dump)
		ecore_dump_num_param(dump_buf + size_param_offset, dump, "size", offset - base_data_offset);

	return offset;
}

/* Performs ILT Dump to the specified buffer.
 * Returns the dumped size in dwords.
 */
static u32 ecore_ilt_dump(struct ecore_hwfn *p_hwfn,
			  struct ecore_ptt *p_ptt,
			  u32 *dump_buf,
			  bool dump)
{
#if ((!defined VMWARE) && (!defined UEFI))
	struct ecore_ilt_client_cfg *clients = p_hwfn->p_cxt_mngr->clients;
#endif
	u32 valid_conn_vf_cids, valid_conn_vf_pages, offset = 0;
	u32 valid_conn_pf_cids, valid_conn_pf_pages, num_pages;
	u32 num_cids_per_page, conn_ctx_size;
	u32 cduc_page_size, cdut_page_size;
	struct phys_mem_desc *ilt_pages;
	u8 conn_type;

	cduc_page_size = 1 << (clients[ILT_CLI_CDUC].p_size.val + PXP_ILT_PAGE_SIZE_NUM_BITS_MIN);
	cdut_page_size = 1 << (clients[ILT_CLI_CDUT].p_size.val + PXP_ILT_PAGE_SIZE_NUM_BITS_MIN);
	conn_ctx_size = p_hwfn->p_cxt_mngr->conn_ctx_size;
	num_cids_per_page = (int)(cduc_page_size / conn_ctx_size);
	ilt_pages = p_hwfn->p_cxt_mngr->ilt_shadow;

	/* Dump global params */
	offset += ecore_dump_common_global_params(p_hwfn, p_ptt, dump_buf + offset, dump, 22); /* 22 must match below amount of params */
	offset += /*  1 */ ecore_dump_str_param(dump_buf + offset, dump, "dump-type", "ilt-dump");
	offset += /*  2 */ ecore_dump_num_param(dump_buf + offset, dump, "cduc-page-size", cduc_page_size);
	offset += /*  3 */ ecore_dump_num_param(dump_buf + offset, dump, "cduc-first-page-id", clients[ILT_CLI_CDUC].first.val);
	offset += /*  4 */ ecore_dump_num_param(dump_buf + offset, dump, "cduc-last-page-id", clients[ILT_CLI_CDUC].last.val);
	offset += /*  5 */ ecore_dump_num_param(dump_buf + offset, dump, "cduc-num-pf-pages", clients[ILT_CLI_CDUC].pf_total_lines);
	offset += /*  6 */ ecore_dump_num_param(dump_buf + offset, dump, "cduc-num-vf-pages", clients[ILT_CLI_CDUC].vf_total_lines);
	offset += /*  7 */ ecore_dump_num_param(dump_buf + offset, dump, "max-conn-ctx-size", conn_ctx_size);
	offset += /*  8 */ ecore_dump_num_param(dump_buf + offset, dump, "cdut-page-size", cdut_page_size);
	offset += /*  9 */ ecore_dump_num_param(dump_buf + offset, dump, "cdut-first-page-id", clients[ILT_CLI_CDUT].first.val);
	offset += /* 10 */ ecore_dump_num_param(dump_buf + offset, dump, "cdut-last-page-id", clients[ILT_CLI_CDUT].last.val);
	offset += /* 11 */ ecore_dump_num_param(dump_buf + offset, dump, "cdut-num-pf-init-pages", ecore_get_cdut_num_pf_init_pages(p_hwfn));
	offset += /* 12 */ ecore_dump_num_param(dump_buf + offset, dump, "cdut-num-vf-init-pages", ecore_get_cdut_num_vf_init_pages(p_hwfn));
	offset += /* 13 */ ecore_dump_num_param(dump_buf + offset, dump, "cdut-num-pf-work-pages", ecore_get_cdut_num_pf_work_pages(p_hwfn));
	offset += /* 14 */ ecore_dump_num_param(dump_buf + offset, dump, "cdut-num-vf-work-pages", ecore_get_cdut_num_vf_work_pages(p_hwfn));
	offset += /* 15 */ ecore_dump_num_param(dump_buf + offset, dump, "max-task-ctx-size", p_hwfn->p_cxt_mngr->task_ctx_size);
	offset += /* 16 */ ecore_dump_num_param(dump_buf + offset, dump, "task-type-id", p_hwfn->p_cxt_mngr->task_type_id);
	offset += /* 17 */ ecore_dump_num_param(dump_buf + offset, dump, "first-vf-id-in-pf", p_hwfn->p_cxt_mngr->first_vf_in_pf);
	offset += /* 18 */ ecore_dump_num_param(dump_buf + offset, dump, "num-vfs-in-pf", p_hwfn->p_cxt_mngr->vf_count);
	offset += /* 19 */ ecore_dump_num_param(dump_buf + offset, dump, "ptr-size-bytes", sizeof(void*));
	offset += /* 20 */ ecore_dump_num_param(dump_buf + offset, dump, "pf-start-line", p_hwfn->p_cxt_mngr->pf_start_line);
	offset += /* 21 */ ecore_dump_num_param(dump_buf + offset, dump, "page-mem-desc-size-dwords", PAGE_MEM_DESC_SIZE_DWORDS);
    offset += /* 22 */ ecore_dump_num_param(dump_buf + offset, dump, "ilt-shadow-size", p_hwfn->p_cxt_mngr->ilt_shadow_size);
	/* Additional/Less parameters require matching of number in call to dump_common_global_params() */

	/* Dump section containing number of PF CIDs per connection type */
	offset += ecore_dump_section_hdr(dump_buf + offset, dump, "num_pf_cids_per_conn_type", 1);
	offset += ecore_dump_num_param(dump_buf + offset, dump, "size", NUM_OF_CONNECTION_TYPES);
	for (conn_type = 0, valid_conn_pf_cids = 0; conn_type < NUM_OF_CONNECTION_TYPES; conn_type++, offset++) {
		u32 num_pf_cids = p_hwfn->p_cxt_mngr->conn_cfg[conn_type].cid_count;

		if (dump)
			*(dump_buf + offset) = num_pf_cids;
		valid_conn_pf_cids += num_pf_cids;
	}

	/* Dump section containing number of VF CIDs per connection type */
	offset += ecore_dump_section_hdr(dump_buf + offset, dump, "num_vf_cids_per_conn_type", 1);
	offset += ecore_dump_num_param(dump_buf + offset, dump, "size", NUM_OF_CONNECTION_TYPES);
	for (conn_type = 0, valid_conn_vf_cids = 0; conn_type < NUM_OF_CONNECTION_TYPES; conn_type++, offset++) {
		u32 num_vf_cids = p_hwfn->p_cxt_mngr->conn_cfg[conn_type].cids_per_vf;

		if (dump)
			*(dump_buf + offset) = num_vf_cids;
		valid_conn_vf_cids += num_vf_cids;
	}

	/* Dump section containing physical memory descriptors for each ILT page */
    num_pages = p_hwfn->p_cxt_mngr->ilt_shadow_size; /* struct dbg_tools_data *dev_data = &p_hwfn->dbg_info; s_chip_defs[dev_data->chip_id].num_ilt_pages; */
	offset += ecore_dump_section_hdr(dump_buf + offset, dump, "ilt_page_desc", 1);
	offset += ecore_dump_num_param(dump_buf + offset, dump, "size", num_pages * PAGE_MEM_DESC_SIZE_DWORDS);

	/* Copy memory descriptors to dump buffer */
	if (dump) {
		u32 page_id;

		for (page_id = 0; page_id < num_pages; page_id++, offset += PAGE_MEM_DESC_SIZE_DWORDS)
			OSAL_MEMCPY(dump_buf + offset, &ilt_pages[page_id], DWORDS_TO_BYTES(PAGE_MEM_DESC_SIZE_DWORDS));
	}
	else {
		offset += num_pages * PAGE_MEM_DESC_SIZE_DWORDS;
	}

	valid_conn_pf_pages = DIV_ROUND_UP(valid_conn_pf_cids, num_cids_per_page);
	valid_conn_vf_pages = DIV_ROUND_UP(valid_conn_vf_cids, num_cids_per_page);

	/* Dump ILT pages IDs */
	offset += ecore_ilt_dump_pages_section(p_hwfn, dump_buf + offset, dump, valid_conn_pf_pages, valid_conn_vf_pages, ilt_pages, true);

	/* Dump ILT pages memory */
	offset += ecore_ilt_dump_pages_section(p_hwfn, dump_buf + offset, dump, valid_conn_pf_pages, valid_conn_vf_pages, ilt_pages, false);

	/* Dump last section */
	offset += ecore_dump_last_section(dump_buf, offset, dump);

	return offset;
}

/* Set/clear bits in the specified register */
static void ecore_update_reg_bits(struct ecore_hwfn *p_hwfn,
				  struct ecore_ptt *p_ptt,
				  u32 reg_addr,
				  bool set,
				  u32 mask)
{
	u32 val;

	val = ecore_rd(p_hwfn, p_ptt, reg_addr);
	ecore_wr(p_hwfn, p_ptt, reg_addr, set ? (val | mask) : (val & ~mask));
}

/* Set/clear parity AEU bits */
static void ecore_update_parity_aeu_bits(struct ecore_hwfn *p_hwfn,
					 struct ecore_ptt *p_ptt,
					 int aeu_id,
					 bool set,
					 u32 parity_mask,
					 u32 sys_kill_mask)
{
	u32 aeu_reg_offset;

	if (aeu_id < 1 || aeu_id > 9)
		return;

	aeu_reg_offset = (aeu_id - 1) * 4;

	/* Set all parity bits to trigger Close-The-Gate NIG */
	ecore_update_reg_bits(p_hwfn, p_ptt, MISC_REG_AEU_ENABLE1_IGU_OUT_0 + aeu_reg_offset, set, parity_mask);
	ecore_update_reg_bits(p_hwfn, p_ptt, MISC_REG_AEU_ENABLE1_IGU_OUT_1 + aeu_reg_offset, set, parity_mask);
	ecore_update_reg_bits(p_hwfn, p_ptt, MISC_REG_AEU_ENABLE1_IGU_OUT_2 + aeu_reg_offset, set, parity_mask);
	ecore_update_reg_bits(p_hwfn, p_ptt, MISC_REG_AEU_ENABLE1_IGU_OUT_3 + aeu_reg_offset, set, parity_mask);
	ecore_update_reg_bits(p_hwfn, p_ptt, MISC_REG_AEU_ENABLE1_IGU_OUT_4 + aeu_reg_offset, set, parity_mask);
	ecore_update_reg_bits(p_hwfn, p_ptt, MISC_REG_AEU_ENABLE1_IGU_OUT_5 + aeu_reg_offset, set, parity_mask);
	ecore_update_reg_bits(p_hwfn, p_ptt, MISC_REG_AEU_ENABLE1_IGU_OUT_6 + aeu_reg_offset, set, parity_mask);
	ecore_update_reg_bits(p_hwfn, p_ptt, MISC_REG_AEU_ENABLE1_IGU_OUT_7 + aeu_reg_offset, set, parity_mask);

	/* Set all parity bits to trigger Close-The-Gate NIG */
	ecore_update_reg_bits(p_hwfn, p_ptt, MISC_REG_AEU_ENABLE1_NIG + aeu_reg_offset, set, parity_mask);

	/* Set all parity bits to trigger Close-The-Gate PXP */
	ecore_update_reg_bits(p_hwfn, p_ptt, MISC_REG_AEU_ENABLE1_PXP + aeu_reg_offset, set, parity_mask);

	/* Route the parity bits to MCP output 0 in the AEU. When triggered,
	 * the MCPF_ATTENTIONS_BITS_PARITY_ERROR bit will be set.
	 */
	ecore_update_reg_bits(p_hwfn, p_ptt, MISC_REG_AEU_ENABLE1_MCP_OUT_0 + aeu_reg_offset, set, parity_mask);

	/* Route parity to all loaded driver.
	 * Configure specific blocks to uncorrectable errors per definition.
	 */
	ecore_update_reg_bits(p_hwfn, p_ptt, MISC_REG_AEU_ENABLE1_GLB_UNC_ERR + aeu_reg_offset, set, sys_kill_mask);

	/* Configure specific blocks to generate system kill per definition */
	ecore_update_reg_bits(p_hwfn, p_ptt, MISC_REG_AEU_ENABLE1_SYS_KILL + aeu_reg_offset, set, sys_kill_mask);
}

/***************************** Public Functions *******************************/

enum dbg_status ecore_dbg_set_bin_ptr(struct ecore_hwfn *p_hwfn,
				      const u8 *const bin_ptr)
{
	struct bin_buffer_hdr *buf_hdrs = (struct bin_buffer_hdr *)bin_ptr;
	u8 buf_id;

	/* Convert binary data to debug arrays */
	for (buf_id = 0; buf_id < MAX_BIN_DBG_BUFFER_TYPE; buf_id++)
		ecore_set_dbg_bin_buf(p_hwfn, (enum bin_dbg_buffer_type)buf_id, (u32 *)(bin_ptr + buf_hdrs[buf_id].offset), buf_hdrs[buf_id].length);

	return DBG_STATUS_OK;
}

enum dbg_status ecore_dbg_set_app_ver(u32 ver)
{
	if (ver < TOOLS_VERSION)
		return DBG_STATUS_UNSUPPORTED_APP_VERSION;

	s_app_ver = ver;

	return DBG_STATUS_OK;
}

u32 ecore_dbg_get_fw_func_ver(void)
{
	return TOOLS_VERSION;
}

enum chip_ids ecore_dbg_get_chip_id(struct ecore_hwfn *p_hwfn)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;

	return (enum chip_ids)dev_data->chip_id;
}

void ecore_read_regs(struct ecore_hwfn *p_hwfn,
		     struct ecore_ptt *p_ptt,
		     u32 *buf,
		     u32 addr,
		     u32 len)
{
	u32 i;

	for (i = 0; i < len; i++)
		buf[i] = ecore_rd(p_hwfn, p_ptt, DWORDS_TO_BYTES(addr + i));
}

bool ecore_read_fw_info(struct ecore_hwfn *p_hwfn,
			struct ecore_ptt *p_ptt,
			struct fw_info *fw_info)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	u8 storm_id;

	for (storm_id = 0; storm_id < MAX_DBG_STORMS; storm_id++) {
		struct storm_defs *storm = &s_storm_defs[storm_id];

		/* Skip Storm if it's in reset */
		if (dev_data->block_in_reset[storm->sem_block_id])
			continue;

		/* Read FW info for the current Storm */
		ecore_read_storm_fw_info(p_hwfn, p_ptt, storm_id, fw_info);

		return true;
	}

	return false;
}

enum dbg_status ecore_dbg_bus_reset(struct ecore_hwfn *p_hwfn,
				    struct ecore_ptt *p_ptt,
				    bool one_shot_en,
				    OSAL_UNUSED u8 hw_dwords,
				    OSAL_UNUSED bool unify_inputs,
				    bool grc_input_en)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	enum dbg_status status;

	status = ecore_dbg_dev_init(p_hwfn);
	if (status != DBG_STATUS_OK)
		return status;

	DP_VERBOSE(p_hwfn, ECORE_MSG_DEBUG, "dbg_bus_reset: one_shot_en = %d, grc_input_en = %d\n", one_shot_en, grc_input_en);

	if (ecore_rd(p_hwfn, p_ptt, DBG_REG_DBG_BLOCK_ON))
		return DBG_STATUS_DBG_BUS_IN_USE;

	/* Update reset state of all blocks */
	ecore_update_blocks_reset_state(p_hwfn, p_ptt);

	/* Disable all debug inputs */
	status = ecore_bus_disable_inputs(p_hwfn, p_ptt, false);
	if (status != DBG_STATUS_OK)
		return status;

	/* Reset DBG block */
	ecore_bus_reset_dbg_block(p_hwfn, p_ptt);

	/* Set one-shot / wrap-around */
	ecore_wr(p_hwfn, p_ptt, DBG_REG_FULL_MODE, one_shot_en ? 0 : 1);

	/* Init state params */
	OSAL_MEMSET(&dev_data->bus, 0, sizeof(dev_data->bus));
	dev_data->bus.target = DBG_BUS_TARGET_ID_INT_BUF;
	dev_data->bus.state = DBG_BUS_STATE_READY;
	dev_data->bus.one_shot_en = one_shot_en;
	dev_data->bus.grc_input_en = grc_input_en;

	/* Init special DBG block */
	if (grc_input_en)
		ecore_bus_add_block_input(p_hwfn, BLOCK_DBG, 0, 0x1, 0, 0, 0, 0);

	return DBG_STATUS_OK;
}

enum dbg_status ecore_dbg_bus_set_pci_output(struct ecore_hwfn *p_hwfn,
					     struct ecore_ptt *p_ptt,
					     u16 buf_size_kb)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	dma_addr_t pci_buf_phys_addr;
	void *pci_buf;

	DP_VERBOSE(p_hwfn, ECORE_MSG_DEBUG, "dbg_bus_set_pci_output: buf_size_kb = %d\n", buf_size_kb);

	if (dev_data->bus.target != DBG_BUS_TARGET_ID_INT_BUF)
		return DBG_STATUS_OUTPUT_ALREADY_SET;
	if (dev_data->bus.state != DBG_BUS_STATE_READY || dev_data->bus.pci_buf.size > 0)
		return DBG_STATUS_DBG_BLOCK_NOT_RESET;

	dev_data->bus.target = DBG_BUS_TARGET_ID_PCI;
	dev_data->bus.pci_buf.size = buf_size_kb * 1024;
	if (dev_data->bus.pci_buf.size % PCI_PKT_SIZE_IN_BYTES)
		return DBG_STATUS_INVALID_ARGS;

	pci_buf = OSAL_DMA_ALLOC_COHERENT(p_hwfn->p_dev, &pci_buf_phys_addr, dev_data->bus.pci_buf.size);
	if (!pci_buf)
		return DBG_STATUS_PCI_BUF_ALLOC_FAILED;

	OSAL_MEMCPY(&dev_data->bus.pci_buf.phys_addr, &pci_buf_phys_addr, sizeof(pci_buf_phys_addr));

	dev_data->bus.pci_buf.virt_addr.lo = (u32)((u64)(osal_uintptr_t)pci_buf);
	dev_data->bus.pci_buf.virt_addr.hi = (u32)((u64)(osal_uintptr_t)pci_buf >> 32);

	ecore_wr(p_hwfn, p_ptt, DBG_REG_PCI_EXT_BUFFER_STRT_ADDR_LSB, dev_data->bus.pci_buf.phys_addr.lo);
	ecore_wr(p_hwfn, p_ptt, DBG_REG_PCI_EXT_BUFFER_STRT_ADDR_MSB, dev_data->bus.pci_buf.phys_addr.hi);
	ecore_wr(p_hwfn, p_ptt, DBG_REG_TARGET_PACKET_SIZE, PCI_PKT_SIZE_IN_CHUNKS);
	ecore_wr(p_hwfn, p_ptt, DBG_REG_PCI_EXT_BUFFER_SIZE, dev_data->bus.pci_buf.size / PCI_PKT_SIZE_IN_BYTES);
	ecore_wr(p_hwfn, p_ptt, DBG_REG_PCI_FUNC_NUM, OPAQUE_FID(p_hwfn->rel_pf_id));
	ecore_wr(p_hwfn, p_ptt, DBG_REG_PCI_LOGIC_ADDR, PCI_PHYS_ADDR_TYPE);
	ecore_wr(p_hwfn, p_ptt, DBG_REG_PCI_REQ_CREDIT, PCI_REQ_CREDIT);
	ecore_wr(p_hwfn, p_ptt, DBG_REG_DEBUG_TARGET, DBG_BUS_TARGET_ID_PCI);
	ecore_wr(p_hwfn, p_ptt, DBG_REG_OUTPUT_ENABLE, TARGET_EN_MASK_PCI);

	return DBG_STATUS_OK;
}

enum dbg_status ecore_dbg_bus_set_nw_output(struct ecore_hwfn *p_hwfn,
					    struct ecore_ptt *p_ptt,
					    u8 port_id,
					    u32 dest_addr_lo32,
					    u16 dest_addr_hi16,
					    u16 data_limit_size_kb,
					    bool send_to_other_engine,
					    bool rcv_from_other_engine)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;

	DP_VERBOSE(p_hwfn, ECORE_MSG_DEBUG, "dbg_bus_set_nw_output: port_id = %d, dest_addr_lo32 = 0x%x, dest_addr_hi16 = 0x%x, data_limit_size_kb = %d, send_to_other_engine = %d, rcv_from_other_engine = %d\n", port_id, dest_addr_lo32, dest_addr_hi16, data_limit_size_kb, send_to_other_engine, rcv_from_other_engine);

	if (dev_data->bus.target != DBG_BUS_TARGET_ID_INT_BUF)
		return DBG_STATUS_OUTPUT_ALREADY_SET;
	if (dev_data->bus.state != DBG_BUS_STATE_READY)
		return DBG_STATUS_DBG_BLOCK_NOT_RESET;
	if ((send_to_other_engine || rcv_from_other_engine) && dev_data->chip_id != CHIP_BB)
		return DBG_STATUS_OTHER_ENGINE_BB_ONLY;
	if (port_id >= dev_data->num_ports || (send_to_other_engine && rcv_from_other_engine))
		return DBG_STATUS_INVALID_ARGS;

	dev_data->bus.target = DBG_BUS_TARGET_ID_NIG;
	dev_data->bus.rcv_from_other_engine = rcv_from_other_engine;

	ecore_wr(p_hwfn, p_ptt, DBG_REG_OUTPUT_ENABLE, TARGET_EN_MASK_NIG);
	ecore_wr(p_hwfn, p_ptt, DBG_REG_DEBUG_TARGET, DBG_BUS_TARGET_ID_NIG);

	if (send_to_other_engine)
		ecore_wr(p_hwfn, p_ptt, DBG_REG_OTHER_ENGINE_MODE, DBG_BUS_OTHER_ENGINE_MODE_CROSS_ENGINE_TX);
	else
		ecore_wr(p_hwfn, p_ptt, NIG_REG_DEBUG_PORT, port_id);

	if (rcv_from_other_engine) {
		ecore_wr(p_hwfn, p_ptt, DBG_REG_OTHER_ENGINE_MODE, DBG_BUS_OTHER_ENGINE_MODE_CROSS_ENGINE_RX);
	} else {
		/* Configure ethernet header of 14 bytes */
		ecore_wr(p_hwfn, p_ptt, DBG_REG_ETHERNET_HDR_WIDTH, 0);
		ecore_wr(p_hwfn, p_ptt, DBG_REG_ETHERNET_HDR_7, dest_addr_lo32);
		ecore_wr(p_hwfn, p_ptt, DBG_REG_ETHERNET_HDR_6, (u32)SRC_MAC_ADDR_LO16 | ((u32)dest_addr_hi16 << 16));
		ecore_wr(p_hwfn, p_ptt, DBG_REG_ETHERNET_HDR_5, SRC_MAC_ADDR_HI32);
		ecore_wr(p_hwfn, p_ptt, DBG_REG_ETHERNET_HDR_4, (u32)ETH_TYPE << 16);
		ecore_wr(p_hwfn, p_ptt, DBG_REG_TARGET_PACKET_SIZE, NIG_PKT_SIZE_IN_CHUNKS);
		if (data_limit_size_kb)
			ecore_wr(p_hwfn, p_ptt, DBG_REG_NIG_DATA_LIMIT_SIZE, (data_limit_size_kb * 1024) / CHUNK_SIZE_IN_BYTES);
	}

	return DBG_STATUS_OK;
}

enum dbg_status ecore_dbg_bus_enable_block(struct ecore_hwfn *p_hwfn,
					   enum block_id block_id,
					   u8 line_num,
					   u8 enable_mask,
					   u8 right_shift,
					   u8 force_valid_mask,
					   u8 force_frame_mask)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	struct dbg_bus_data *bus = &dev_data->bus;
	struct dbg_bus_block_data *block_bus;
	const struct dbg_bus_line *dbg_line;
	const struct dbg_block_chip *block;
	bool is_removed, has_dbg_bus;
	u16 modes_buf_offset;
	u8 is_256b_line;

	block = ecore_get_dbg_block_per_chip(p_hwfn, block_id);
	is_removed = GET_FIELD(block->flags, DBG_BLOCK_CHIP_IS_REMOVED);
	has_dbg_bus = GET_FIELD(block->flags, DBG_BLOCK_CHIP_HAS_DBG_BUS);

	/* Check debug bus mode */
	if (!is_removed && has_dbg_bus && GET_FIELD(block->dbg_bus_mode.data, DBG_MODE_HDR_EVAL_MODE) > 0) {
		modes_buf_offset = GET_FIELD(block->dbg_bus_mode.data, DBG_MODE_HDR_MODES_BUF_OFFSET);
		if (!ecore_is_mode_match(p_hwfn, &modes_buf_offset))
			has_dbg_bus = false;
	}

	block_bus = &bus->blocks[block_id];
	dbg_line = get_dbg_bus_line(p_hwfn, block_id, line_num);
	is_256b_line = dbg_line && GET_FIELD(dbg_line->data, DBG_BUS_LINE_IS_256B) ? 1 : 0;

	DP_VERBOSE(p_hwfn, ECORE_MSG_DEBUG, "dbg_bus_enable_block: block = %d, line_num = %d, enable_mask = 0x%x, right_shift = %d, force_valid_mask = 0x%x, force_frame_mask = 0x%x\n", block_id, line_num, enable_mask, right_shift, force_valid_mask, force_frame_mask);

	if (bus->state != DBG_BUS_STATE_READY)
		return DBG_STATUS_DBG_BLOCK_NOT_RESET;
	if (block_id >= MAX_BLOCK_ID)
		return DBG_STATUS_INVALID_ARGS;
	if (block_bus->enable_mask)
		return DBG_STATUS_BLOCK_ALREADY_ENABLED;
	if (is_removed ||
	    !has_dbg_bus ||
	    line_num >= NUM_DBG_LINES(block) ||
	    !enable_mask ||
	    enable_mask > MAX_BLOCK_VALUES_MASK ||
	    force_valid_mask > MAX_BLOCK_VALUES_MASK ||
	    force_frame_mask > MAX_BLOCK_VALUES_MASK)
		return DBG_STATUS_INVALID_ARGS;
	if (dev_data->block_in_reset[block_id])
		return DBG_STATUS_BLOCK_IN_RESET;
	if (is_256b_line && (bus->trigger_en || bus->filter_en))
		return DBG_STATUS_NO_FILTER_TRIGGER_256B;

	/* Verify the following in all enabled blocks:
	 * - configured debug lines have the same width (128/256 bit).
	 */
	if (bus->num_enabled_blocks) {
		u8 curr_block_id;

		for (curr_block_id = 0; curr_block_id < MAX_BLOCK_ID; curr_block_id++) {
			struct dbg_bus_block_data *curr_block_bus = &bus->blocks[curr_block_id];

			if (curr_block_id == block_id)
				continue;

			if (curr_block_bus->enable_mask) {
				u8 curr_is_256b_line = GET_FIELD(block_bus->flags, DBG_BUS_BLOCK_DATA_IS_256B_LINE);

				if (is_256b_line != curr_is_256b_line)
					return DBG_STATUS_NON_MATCHING_LINES;
			}
		}
	}

	ecore_bus_add_block_input(p_hwfn, block_id, line_num, enable_mask, right_shift, force_valid_mask, force_frame_mask, is_256b_line);

	return DBG_STATUS_OK;
}

enum dbg_status ecore_dbg_bus_enable_storm(struct ecore_hwfn *p_hwfn,
					   enum dbg_storms storm_id,
					   enum dbg_bus_storm_modes storm_mode)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	struct dbg_bus_data *bus = &dev_data->bus;
	struct dbg_bus_storm_data *storm_bus;
	struct storm_defs *storm;

	DP_VERBOSE(p_hwfn, ECORE_MSG_DEBUG, "dbg_bus_enable_storm: storm = %d, storm_mode = %d\n", storm_id, storm_mode);

	if (bus->state != DBG_BUS_STATE_READY)
		return DBG_STATUS_DBG_BLOCK_NOT_RESET;
	if (storm_id >= MAX_DBG_STORMS)
		return DBG_STATUS_INVALID_ARGS;
	if (storm_mode >= MAX_DBG_BUS_STORM_MODES)
		return DBG_STATUS_INVALID_ARGS;
	if (!s_storm_mode_defs[storm_mode].exists[dev_data->chip_id])
		return DBG_STATUS_INVALID_STORM_DBG_MODE;
	if (bus->storms[storm_id].enabled)
		return DBG_STATUS_STORM_ALREADY_ENABLED;

	storm = &s_storm_defs[storm_id];
	storm_bus = &bus->storms[storm_id];

	if (dev_data->block_in_reset[storm->sem_block_id])
		return DBG_STATUS_BLOCK_IN_RESET;

	storm_bus->enabled = true;
	storm_bus->mode = (u8)storm_mode;
	storm_bus->hw_id = bus->num_enabled_storms++;

	return DBG_STATUS_OK;
}

enum dbg_status ecore_dbg_bus_enable_timestamp(struct ecore_hwfn *p_hwfn,
					       struct ecore_ptt *p_ptt,
					       u8 valid_mask,
					       u8 frame_mask,
					       u32 tick_len)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	struct dbg_bus_data *bus = &dev_data->bus;

	u8 max_en_mask = (1 << (s_chip_defs[dev_data->chip_id].dwords_per_cycle - 1)) - 1;

	DP_VERBOSE(p_hwfn, ECORE_MSG_DEBUG, "dbg_bus_enable_timestamp: valid_mask = 0x%x, frame_mask = 0x%x, tick_len = %d\n", valid_mask, frame_mask, tick_len);

	if (bus->state != DBG_BUS_STATE_READY)
		return DBG_STATUS_DBG_BLOCK_NOT_RESET;
	if (valid_mask > max_en_mask || frame_mask > max_en_mask)
		return DBG_STATUS_INVALID_ARGS;

	bus->timestamp_input_en = true;

	ecore_wr(p_hwfn, p_ptt, DBG_REG_TIMESTAMP_VALID_EN, valid_mask);
	ecore_wr(p_hwfn, p_ptt, DBG_REG_TIMESTAMP_FRAME_EN, frame_mask);
	ecore_wr(p_hwfn, p_ptt, DBG_REG_TIMESTAMP_TICK, tick_len);

	ecore_bus_add_block_input(p_hwfn, BLOCK_DBG, 0, 0x1, 0, false, false, false);

	return DBG_STATUS_OK;
}

enum dbg_status ecore_dbg_bus_add_eid_range_sem_filter(struct ecore_hwfn *p_hwfn,
						       enum dbg_storms storm_id,
						       u8 min_eid,
						       u8 max_eid)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	struct dbg_bus_storm_data *storm_bus;

	storm_bus = &dev_data->bus.storms[storm_id];

	DP_VERBOSE(p_hwfn, ECORE_MSG_DEBUG, "dbg_bus_add_eid_range_sem_filter: storm = %d, min_eid = 0x%x, max_eid = 0x%x\n", storm_id, min_eid, max_eid);

	if (storm_id >= MAX_DBG_STORMS)
		return DBG_STATUS_INVALID_ARGS;
	if (min_eid > max_eid)
		return DBG_STATUS_INVALID_ARGS;
	if (!storm_bus->enabled)
		return DBG_STATUS_STORM_NOT_ENABLED;

	storm_bus->eid_filter_en = 1;
	storm_bus->eid_range_not_mask = 1;
	storm_bus->eid_filter_params.range.min = min_eid;
	storm_bus->eid_filter_params.range.max = max_eid;

	return DBG_STATUS_OK;
}

enum dbg_status ecore_dbg_bus_add_eid_mask_sem_filter(struct ecore_hwfn *p_hwfn,
						      enum dbg_storms storm_id,
						      u8 eid_val,
						      u8 eid_mask)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	struct dbg_bus_storm_data *storm_bus;

	storm_bus = &dev_data->bus.storms[storm_id];

	DP_VERBOSE(p_hwfn, ECORE_MSG_DEBUG, "dbg_bus_add_eid_mask_sem_filter: storm = %d, eid_val = 0x%x, eid_mask = 0x%x\n", storm_id, eid_val, eid_mask);

	if (storm_id >= MAX_DBG_STORMS)
		return DBG_STATUS_INVALID_ARGS;
	if (!storm_bus->enabled)
		return DBG_STATUS_STORM_NOT_ENABLED;

	storm_bus->eid_filter_en = 1;
	storm_bus->eid_range_not_mask = 0;
	storm_bus->eid_filter_params.mask.val = eid_val;
	storm_bus->eid_filter_params.mask.mask = eid_mask;

	return DBG_STATUS_OK;
}

enum dbg_status ecore_dbg_bus_add_cid_sem_filter(struct ecore_hwfn *p_hwfn,
						 enum dbg_storms storm_id,
						 u32 cid)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	struct dbg_bus_storm_data *storm_bus;

	storm_bus = &dev_data->bus.storms[storm_id];

	DP_VERBOSE(p_hwfn, ECORE_MSG_DEBUG, "dbg_bus_add_cid_sem_filter: storm = %d, cid = 0x%x\n", storm_id, cid);

	if (storm_id >= MAX_DBG_STORMS)
		return DBG_STATUS_INVALID_ARGS;
	if (!storm_bus->enabled)
		return DBG_STATUS_STORM_NOT_ENABLED;

	storm_bus->cid_filter_en = 1;
	storm_bus->cid = cid;

	return DBG_STATUS_OK;
}

enum dbg_status ecore_dbg_bus_enable_filter(struct ecore_hwfn *p_hwfn,
					    struct ecore_ptt *p_ptt,
					    OSAL_UNUSED enum block_id block_id,
					    u8 msg_len)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	struct dbg_bus_data *bus = &dev_data->bus;

	DP_VERBOSE(p_hwfn, ECORE_MSG_DEBUG, "dbg_bus_enable_filter: msg_len = %d\n", msg_len);

	if (bus->state != DBG_BUS_STATE_READY)
		return DBG_STATUS_DBG_BLOCK_NOT_RESET;
	if (bus->filter_en)
		return DBG_STATUS_FILTER_ALREADY_ENABLED;
	if (bus->mode_256b_en)
		return DBG_STATUS_NO_FILTER_TRIGGER_256B;

	ecore_wr(p_hwfn, p_ptt, DBG_REG_FILTER_MSG_LENGTH_ENABLE, msg_len > 0 ? 1 : 0);
	if (msg_len > 0)
		ecore_wr(p_hwfn, p_ptt, DBG_REG_FILTER_MSG_LENGTH, msg_len - 1);

	bus->filter_msg_len = msg_len;
	bus->filter_en = true;
	bus->next_constraint_id = 0;
	bus->adding_filter = true;

	return DBG_STATUS_OK;
}

enum dbg_status ecore_dbg_bus_enable_trigger(struct ecore_hwfn *p_hwfn,
					     struct ecore_ptt *p_ptt,
					     bool rec_pre_trigger,
					     u8 pre_chunks,
					     bool rec_post_trigger,
					     u32 post_cycles,
					     bool filter_pre_trigger,
					     bool filter_post_trigger)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	enum dbg_bus_post_trigger_types post_trigger_type;
	enum dbg_bus_pre_trigger_types pre_trigger_type;
	struct dbg_bus_data *bus = &dev_data->bus;

	DP_VERBOSE(p_hwfn, ECORE_MSG_DEBUG, "dbg_bus_enable_trigger: rec_pre_trigger = %d, pre_chunks = %d, rec_post_trigger = %d, post_cycles = %d, filter_pre_trigger = %d, filter_post_trigger = %d\n", rec_pre_trigger, pre_chunks, rec_post_trigger, post_cycles, filter_pre_trigger, filter_post_trigger);

	if (bus->state != DBG_BUS_STATE_READY)
		return DBG_STATUS_DBG_BLOCK_NOT_RESET;
	if (bus->trigger_en)
		return DBG_STATUS_TRIGGER_ALREADY_ENABLED;
	if (rec_pre_trigger && pre_chunks >= INT_BUF_SIZE_IN_CHUNKS)
		return DBG_STATUS_INVALID_ARGS;
	if (bus->mode_256b_en)
		return DBG_STATUS_NO_FILTER_TRIGGER_256B;

	bus->trigger_en = true;
	bus->filter_pre_trigger = filter_pre_trigger;
	bus->filter_post_trigger = filter_post_trigger;

	if (rec_pre_trigger) {
		pre_trigger_type = pre_chunks ? DBG_BUS_PRE_TRIGGER_NUM_CHUNKS : DBG_BUS_PRE_TRIGGER_FROM_ZERO;
		ecore_wr(p_hwfn, p_ptt, DBG_REG_RCRD_ON_WINDOW_PRE_NUM_CHUNKS, pre_chunks);
	} else {
		pre_trigger_type = DBG_BUS_PRE_TRIGGER_DROP;
	}

	if (rec_post_trigger) {
		post_trigger_type = DBG_BUS_POST_TRIGGER_RECORD;
		ecore_wr(p_hwfn, p_ptt, DBG_REG_RCRD_ON_WINDOW_POST_NUM_CYCLES, post_cycles ? post_cycles : 0xffffffff);
	} else {
		post_trigger_type = DBG_BUS_POST_TRIGGER_DROP;
	}

	ecore_wr(p_hwfn, p_ptt, DBG_REG_RCRD_ON_WINDOW_PRE_TRGR_EVNT_MODE, pre_trigger_type);
	ecore_wr(p_hwfn, p_ptt, DBG_REG_RCRD_ON_WINDOW_POST_TRGR_EVNT_MODE, post_trigger_type);
	ecore_wr(p_hwfn, p_ptt, DBG_REG_TRIGGER_ENABLE, 1);

	return DBG_STATUS_OK;
}

enum dbg_status ecore_dbg_bus_add_trigger_state(struct ecore_hwfn *p_hwfn,
						struct ecore_ptt *p_ptt,
						enum block_id block_id,
						u8 msg_len,
						u16 count_to_next)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	struct dbg_bus_trigger_state_data *trigger_state;
	struct dbg_bus_data *bus = &dev_data->bus;
	u8 reg_offset;

	DP_VERBOSE(p_hwfn, ECORE_MSG_DEBUG, "dbg_bus_add_trigger_state: block_id = %d, msg_len = %d, count_to_next = %d\n", block_id, msg_len, count_to_next);

	if (!bus->trigger_en)
		return DBG_STATUS_TRIGGER_NOT_ENABLED;
	if (bus->next_trigger_state == MAX_TRIGGER_STATES)
		return DBG_STATUS_TOO_MANY_TRIGGER_STATES;
	if (!count_to_next)
		return DBG_STATUS_INVALID_ARGS;

	trigger_state = &bus->trigger_states[bus->next_trigger_state];
	trigger_state->msg_len = msg_len;

	reg_offset = bus->next_trigger_state * BYTES_IN_DWORD;

	if (block_id != MAX_BLOCK_ID) {
		const struct dbg_block *block;
		enum dbg_storms associated_storm_id;

		block = get_dbg_block(p_hwfn, block_id);

		if (!block->associated_storm_letter)
			return DBG_STATUS_INVALID_ARGS;

		associated_storm_id = ecore_get_storm_id_from_letter(block->associated_storm_letter);
		if (associated_storm_id == MAX_DBG_STORMS || !bus->storms[associated_storm_id].enabled)
			return DBG_STATUS_STORM_NOT_ENABLED;

		trigger_state->storm_id = (u8)associated_storm_id;
	} else
		trigger_state->storm_id = MAX_DBG_STORMS;

	bus->next_constraint_id = 0;
	bus->adding_filter = false;

	/* Set trigger state registers */
	ecore_wr(p_hwfn, p_ptt, DBG_REG_TRIGGER_STATE_MSG_LENGTH_ENABLE_0 + reg_offset, msg_len > 0 ? 1 : 0);
	if (msg_len > 0)
		ecore_wr(p_hwfn, p_ptt, DBG_REG_TRIGGER_STATE_MSG_LENGTH_0 + reg_offset, msg_len - 1);

	/* Set trigger set registers */
	reg_offset = bus->next_trigger_state * TRIGGER_SETS_PER_STATE * BYTES_IN_DWORD;
	ecore_wr(p_hwfn, p_ptt, DBG_REG_TRIGGER_STATE_SET_COUNT_0 + reg_offset, count_to_next);

	/* Set next state to final state, and overwrite previous next state
	 * (if any).
	 */
	ecore_wr(p_hwfn, p_ptt, DBG_REG_TRIGGER_STATE_SET_NXT_STATE_0 + reg_offset, MAX_TRIGGER_STATES);
	if (bus->next_trigger_state > 0) {
		reg_offset = (bus->next_trigger_state - 1) * TRIGGER_SETS_PER_STATE * BYTES_IN_DWORD;
		ecore_wr(p_hwfn, p_ptt, DBG_REG_TRIGGER_STATE_SET_NXT_STATE_0 + reg_offset, bus->next_trigger_state);
	}

	bus->next_trigger_state++;

	return DBG_STATUS_OK;
}

enum dbg_status ecore_dbg_bus_add_constraint(struct ecore_hwfn *p_hwfn,
					     struct ecore_ptt *p_ptt,
					     enum dbg_bus_constraint_ops constraint_op,
					     u32 data_val,
					     u32 data_mask,
					     bool compare_frame,
					     u8 frame_bit,
					     u8 cycle_offset,
					     u8 dword_offset_in_cycle,
					     bool is_mandatory)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	struct dbg_bus_data *bus = &dev_data->bus;
	u16 dword_offset, range = 0;

	DP_VERBOSE(p_hwfn, ECORE_MSG_DEBUG, "dbg_bus_add_constraint: op = %d, data_val = 0x%x, data_mask = 0x%x, compare_frame = %d, frame_bit = %d, cycle_offset = %d, dword_offset_in_cycle = %d, is_mandatory = %d\n", constraint_op, data_val, data_mask, compare_frame, frame_bit, cycle_offset, dword_offset_in_cycle, is_mandatory);

	if (!bus->filter_en && !dev_data->bus.trigger_en)
		return DBG_STATUS_CANT_ADD_CONSTRAINT;
	if (bus->trigger_en && !bus->adding_filter && !bus->next_trigger_state)
		return DBG_STATUS_CANT_ADD_CONSTRAINT;
	if (bus->next_constraint_id >= MAX_CONSTRAINTS)
		return DBG_STATUS_TOO_MANY_CONSTRAINTS;
	if (constraint_op >= MAX_DBG_BUS_CONSTRAINT_OPS || frame_bit > 1 ||
	    dword_offset_in_cycle >= s_chip_defs[dev_data->chip_id].dwords_per_cycle ||
	    (bus->adding_filter && cycle_offset >= MAX_FILTER_CYCLE_OFFSET))
		return DBG_STATUS_INVALID_ARGS;
	if (compare_frame &&
	    constraint_op != DBG_BUS_CONSTRAINT_OP_EQ &&
	    constraint_op != DBG_BUS_CONSTRAINT_OP_NE)
		return DBG_STATUS_INVALID_ARGS;

	dword_offset = cycle_offset * s_chip_defs[dev_data->chip_id].dwords_per_cycle + dword_offset_in_cycle;

	/* Add selected dword to constraints dword mask */
	if (bus->adding_filter) {
		bus->filter_constraint_dword_mask |= (u8)(1 << dword_offset_in_cycle);
	} else {
		u8 curr_trigger_state_id = bus->next_trigger_state - 1;
		struct dbg_bus_trigger_state_data *trigger_state;

		trigger_state = &bus->trigger_states[curr_trigger_state_id];
		trigger_state->constraint_dword_mask |= (u8)(1 << dword_offset_in_cycle);
	}

	/* Prepare data mask and range */
	if (constraint_op == DBG_BUS_CONSTRAINT_OP_EQ ||
		constraint_op == DBG_BUS_CONSTRAINT_OP_NE) {
		data_mask = ~data_mask;
	} else {
		u8 lsb, width;

		/* Extract lsb and width from mask */
		if (!data_mask)
			return DBG_STATUS_INVALID_ARGS;

		for (lsb = 0; lsb < 32 && !(data_mask & 1); lsb++, data_mask >>= 1);
		for (width = 0; width < 32 - lsb && (data_mask & 1); width++, data_mask >>= 1);
		if (data_mask)
			return DBG_STATUS_INVALID_ARGS;
		range = (lsb << 5) | (width - 1);
	}

	/* Add constraint */
	ecore_bus_set_constraint(p_hwfn, p_ptt, dev_data->bus.adding_filter ? 1 : 0,
		dev_data->bus.next_constraint_id,
		s_constraint_op_defs[constraint_op].hw_op_val,
		data_val, data_mask, frame_bit,
		compare_frame ? 0 : 1, dword_offset, range,
		s_constraint_op_defs[constraint_op].is_cyclic ? 1 : 0,
		is_mandatory ? 1 : 0);

	/* If first constraint, fill other 3 constraints with dummy constraints
	 * that always match (using the same offset).
	 */
	if (!dev_data->bus.next_constraint_id) {
		u8 i;

		for (i = 1; i < MAX_CONSTRAINTS; i++)
			ecore_bus_set_constraint(p_hwfn, p_ptt, bus->adding_filter ? 1 : 0,
				i, DBG_BUS_CONSTRAINT_OP_EQ, 0, 0xffffffff,
				0, 1, dword_offset, 0, 0, 1);
	}

	bus->next_constraint_id++;

	return DBG_STATUS_OK;
}

enum dbg_status ecore_dbg_bus_start(struct ecore_hwfn *p_hwfn,
				    struct ecore_ptt *p_ptt)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	struct dbg_bus_data *bus = &dev_data->bus;
	u8 block_hw_ids[DEBUG_BUS_CYCLE_DWORDS];
	struct framing_mode_defs *framing_mode;
	enum dbg_status status;
	u32 block_id;
	u8 storm_id;

	DP_VERBOSE(p_hwfn, ECORE_MSG_DEBUG, "dbg_bus_start\n");

	if (bus->state != DBG_BUS_STATE_READY)
		return DBG_STATUS_DBG_BLOCK_NOT_RESET;

	/* Check if any input was enabled */
	if (!bus->num_enabled_storms &&
	    !bus->num_enabled_blocks &&
	    !bus->rcv_from_other_engine)
		return DBG_STATUS_NO_INPUT_ENABLED;

	/* Configure framing mode */
	framing_mode = ecore_bus_get_framing_mode(p_hwfn);
	if (!framing_mode)
		return DBG_STATUS_NO_MATCHING_FRAMING_MODE;

	ecore_wr(p_hwfn, p_ptt, DBG_REG_FRAMING_MODE, framing_mode->id);
	ecore_wr(p_hwfn, p_ptt, DBG_REG_FULL_BUFFER_THR, framing_mode->full_buf_thr);

	/* Configure DBG block for Storm inputs */
	if (bus->num_enabled_storms) {
		status = ecore_bus_config_storm_inputs(p_hwfn, p_ptt);
		if (status != DBG_STATUS_OK)
			return status;
	}

	/* Assigns block HW IDs */
	status = ecore_bus_assign_block_hw_ids(p_hwfn, block_hw_ids);
	if (status != DBG_STATUS_OK)
		return status;

	/* Configure DBG block for block inputs */
	if (bus->num_enabled_blocks) {
		status = ecore_bus_config_block_inputs(p_hwfn, p_ptt, block_hw_ids);
		if (status != DBG_STATUS_OK)
			return status;
	}

	/* Configure DBG block filters and triggers */
	if (bus->filter_en || bus->trigger_en) {
		status = ecore_bus_config_filters_and_triggers(p_hwfn, p_ptt, framing_mode, block_hw_ids);
		if (status != DBG_STATUS_OK)
			return status;
	}

	/* Restart timestamp */
	ecore_wr(p_hwfn, p_ptt, DBG_REG_TIMESTAMP, 0);

	/* Enable debug block */
	ecore_bus_enable_dbg_block(p_hwfn, p_ptt, 1);

	/* Configure enabled blocks - must be done before the DBG block is
	 * enabled.
	 */
	if (dev_data->bus.num_enabled_blocks) {
		/* Config blocks */
		for (block_id = 0; block_id < MAX_BLOCK_ID; block_id++) {
			struct dbg_bus_block_data *block_bus = &bus->blocks[block_id];

			if (!block_bus->enable_mask || block_id == BLOCK_DBG)
				continue;

			ecore_bus_config_dbg_line(p_hwfn, p_ptt, (enum block_id)block_id,
				block_bus->line_num,
				block_bus->enable_mask,
				block_bus->right_shift,
				block_bus->force_valid_mask,
				block_bus->force_frame_mask);
		}
	}

	/* Configure client mask */
	ecore_bus_config_client_mask(p_hwfn, p_ptt);

	/* Configure enabled Storms - must be done after the DBG block is
	 * enabled.
	 */
	if (dev_data->bus.num_enabled_storms) {
		for (storm_id = 0; storm_id < MAX_DBG_STORMS; storm_id++) {
			if (!dev_data->bus.storms[storm_id].enabled)
				continue;
			ecore_bus_add_storm_input(p_hwfn, p_ptt, (enum dbg_storms)storm_id, framing_mode->semi_framing_mode_id);
		}
	}

	dev_data->bus.state = DBG_BUS_STATE_RECORDING;

	return DBG_STATUS_OK;
}

enum dbg_status ecore_dbg_bus_stop(struct ecore_hwfn *p_hwfn,
				   struct ecore_ptt *p_ptt)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	struct dbg_bus_data *bus = &dev_data->bus;
	enum dbg_status status = DBG_STATUS_OK;

	DP_VERBOSE(p_hwfn, ECORE_MSG_DEBUG, "dbg_bus_stop\n");

	if (bus->state != DBG_BUS_STATE_RECORDING)
		return DBG_STATUS_RECORDING_NOT_STARTED;

	status = ecore_bus_disable_inputs(p_hwfn, p_ptt, true);
	if (status != DBG_STATUS_OK)
		return status;

	ecore_wr(p_hwfn, p_ptt, DBG_REG_CPU_TIMEOUT, 1);

	OSAL_MSLEEP(FLUSH_DELAY_MS);

	ecore_bus_enable_dbg_block(p_hwfn, p_ptt, false);

	/* Check if trigger worked */
	if (bus->trigger_en) {
		u32 trigger_state = ecore_rd(p_hwfn, p_ptt, DBG_REG_TRIGGER_STATUS_CUR_STATE);

		if (trigger_state != MAX_TRIGGER_STATES)
			return DBG_STATUS_DATA_DIDNT_TRIGGER;
	}

	bus->state = DBG_BUS_STATE_STOPPED;

	return status;
}

enum dbg_status ecore_dbg_bus_get_dump_buf_size(struct ecore_hwfn *p_hwfn,
						struct ecore_ptt *p_ptt,
						u32 *buf_size)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	struct dbg_bus_data *bus = &dev_data->bus;
	enum dbg_status status;

	status = ecore_dbg_dev_init(p_hwfn);

	*buf_size = 0;

	if (status != DBG_STATUS_OK)
		return status;

	/* Add dump header */
	*buf_size = (u32)ecore_bus_dump_hdr(p_hwfn, p_ptt, OSAL_NULL, false);

	switch (bus->target) {
	case DBG_BUS_TARGET_ID_INT_BUF:
		*buf_size += INT_BUF_SIZE_IN_DWORDS; break;
	case DBG_BUS_TARGET_ID_PCI:
		*buf_size += BYTES_TO_DWORDS(bus->pci_buf.size); break;
	default:
		break;
	}

	/* Dump last section */
	*buf_size += ecore_dump_last_section(OSAL_NULL, 0, false);

	return DBG_STATUS_OK;
}

enum dbg_status ecore_dbg_bus_dump(struct ecore_hwfn *p_hwfn,
				   struct ecore_ptt *p_ptt,
				   u32 *dump_buf,
				   u32 buf_size_in_dwords,
				   u32 *num_dumped_dwords)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	u32 min_buf_size_in_dwords, block_id, offset = 0;
	struct dbg_bus_data *bus = &dev_data->bus;
	enum dbg_status status;
	u8 storm_id;

	*num_dumped_dwords = 0;

	status = ecore_dbg_bus_get_dump_buf_size(p_hwfn, p_ptt, &min_buf_size_in_dwords);
	if (status != DBG_STATUS_OK)
		return status;

	DP_VERBOSE(p_hwfn, ECORE_MSG_DEBUG, "dbg_bus_dump: dump_buf = 0x%p, buf_size_in_dwords = %d\n", dump_buf, buf_size_in_dwords);

	if (bus->state != DBG_BUS_STATE_RECORDING && bus->state != DBG_BUS_STATE_STOPPED)
		return DBG_STATUS_RECORDING_NOT_STARTED;

	if (bus->state == DBG_BUS_STATE_RECORDING) {
		enum dbg_status stop_state = ecore_dbg_bus_stop(p_hwfn, p_ptt);
		if (stop_state != DBG_STATUS_OK)
			return stop_state;
	}

	if (buf_size_in_dwords < min_buf_size_in_dwords)
		return DBG_STATUS_DUMP_BUF_TOO_SMALL;

	if (bus->target == DBG_BUS_TARGET_ID_PCI && !bus->pci_buf.size)
		return DBG_STATUS_PCI_BUF_NOT_ALLOCATED;

	/* Dump header */
	offset += ecore_bus_dump_hdr(p_hwfn, p_ptt, dump_buf + offset, true);

	/* Dump recorded data */
	if (bus->target != DBG_BUS_TARGET_ID_NIG) {
		u32 recorded_dwords = ecore_bus_dump_data(p_hwfn, p_ptt, dump_buf + offset, true);

		if (!recorded_dwords)
			return DBG_STATUS_NO_DATA_RECORDED;
		if (recorded_dwords % CHUNK_SIZE_IN_DWORDS)
			return DBG_STATUS_DUMP_NOT_CHUNK_ALIGNED;
		offset += recorded_dwords;
	}

	/* Dump last section */
	offset += ecore_dump_last_section(dump_buf, offset, true);

	/* If recorded to PCI buffer - free the buffer */
	ecore_bus_free_pci_buf(p_hwfn);

	/* Clear debug bus parameters */
	bus->state = DBG_BUS_STATE_IDLE;
	bus->num_enabled_blocks = 0;
	bus->num_enabled_storms = 0;
	bus->filter_en = bus->trigger_en = 0;

	/* Clear block data */
	for (block_id = 0; block_id < MAX_BLOCK_ID; block_id++)
		bus->blocks[block_id].enable_mask = 0;

	/* Clear Storm data */
	for (storm_id = 0; storm_id < MAX_DBG_STORMS; storm_id++) {
		struct dbg_bus_storm_data *storm_bus = &bus->storms[storm_id];

		storm_bus->enabled = false;
		storm_bus->eid_filter_en = storm_bus->cid_filter_en = 0;
	}

	*num_dumped_dwords = offset;

	return DBG_STATUS_OK;
}

enum dbg_status ecore_dbg_grc_config(struct ecore_hwfn *p_hwfn,
				     enum dbg_grc_params grc_param,
				     u32 val)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	enum dbg_status status;
	int i;

	DP_VERBOSE(p_hwfn, ECORE_MSG_DEBUG, "dbg_grc_config: paramId = %d, val = %d\n", grc_param, val);

	status = ecore_dbg_dev_init(p_hwfn);
	if (status != DBG_STATUS_OK)
		return status;

	/* Initializes the GRC parameters (if not initialized). Needed in order
	 * to set the default parameter values for the first time.
	 */
	ecore_dbg_grc_init_params(p_hwfn);

	if (grc_param >= MAX_DBG_GRC_PARAMS)
		return DBG_STATUS_INVALID_ARGS;
	if (val < s_grc_param_defs[grc_param].min ||
		val > s_grc_param_defs[grc_param].max)
		return DBG_STATUS_INVALID_ARGS;

	if (s_grc_param_defs[grc_param].is_preset) {
		/* Preset param */

		/* Disabling a preset is not allowed. Call
		 * dbg_grc_set_params_default instead.
		 */
		if (!val)
			return DBG_STATUS_INVALID_ARGS;

		/* Update all params with the preset values */
		for (i = 0; i < MAX_DBG_GRC_PARAMS; i++) {
			u32 preset_val;

			/* Skip persistent params */
			if (s_grc_param_defs[i].is_persistent)
				continue;

			/* Find preset value */
			if (grc_param == DBG_GRC_PARAM_EXCLUDE_ALL)
				preset_val = s_grc_param_defs[i].exclude_all_preset_val;
			else if (grc_param == DBG_GRC_PARAM_CRASH)
				preset_val = s_grc_param_defs[i].crash_preset_val[dev_data->chip_id];
			else
				return DBG_STATUS_INVALID_ARGS;

			ecore_grc_set_param(p_hwfn, (enum dbg_grc_params)i, preset_val);
		}
	} else {
		/* Regular param - set its value */
		ecore_grc_set_param(p_hwfn, grc_param, val);
	}

	return DBG_STATUS_OK;
}

/* Assign default GRC param values */
void ecore_dbg_grc_set_params_default(struct ecore_hwfn *p_hwfn)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	u32 i;

	for (i = 0; i < MAX_DBG_GRC_PARAMS; i++)
		if (!s_grc_param_defs[i].is_persistent)
			dev_data->grc.param_val[i] = s_grc_param_defs[i].default_val[dev_data->chip_id];
}

enum dbg_status ecore_dbg_grc_get_dump_buf_size(struct ecore_hwfn *p_hwfn,
						struct ecore_ptt *p_ptt,
						u32 *buf_size)
{
	enum dbg_status status = ecore_dbg_dev_init(p_hwfn);

	*buf_size = 0;

	if (status != DBG_STATUS_OK)
		return status;

	if (!p_hwfn->dbg_arrays[BIN_BUF_DBG_MODE_TREE].ptr ||
	    !p_hwfn->dbg_arrays[BIN_BUF_DBG_DUMP_REG].ptr || !p_hwfn->dbg_arrays[BIN_BUF_DBG_DUMP_MEM].ptr ||
	    !p_hwfn->dbg_arrays[BIN_BUF_DBG_ATTN_BLOCKS].ptr || !p_hwfn->dbg_arrays[BIN_BUF_DBG_ATTN_REGS].ptr)
		return DBG_STATUS_DBG_ARRAY_NOT_SET;

	return ecore_grc_dump(p_hwfn, p_ptt, OSAL_NULL, false, buf_size);
}

enum dbg_status ecore_dbg_grc_dump(struct ecore_hwfn *p_hwfn,
				   struct ecore_ptt *p_ptt,
				   u32 *dump_buf,
				   u32 buf_size_in_dwords,
				   u32 *num_dumped_dwords)
{
	u32 needed_buf_size_in_dwords;
	enum dbg_status status;

	*num_dumped_dwords = 0;

	status = ecore_dbg_grc_get_dump_buf_size(p_hwfn, p_ptt, &needed_buf_size_in_dwords);
	if (status != DBG_STATUS_OK)
		return status;

	if (buf_size_in_dwords < needed_buf_size_in_dwords)
		return DBG_STATUS_DUMP_BUF_TOO_SMALL;

	/* Doesn't do anything, needed for compile time asserts */
	ecore_static_asserts();

	/* GRC Dump */
	status = ecore_grc_dump(p_hwfn, p_ptt, dump_buf, true, num_dumped_dwords);

	/* Revert GRC params to their default */
	ecore_dbg_grc_set_params_default(p_hwfn);

	return status;
}

enum dbg_status ecore_dbg_idle_chk_get_dump_buf_size(struct ecore_hwfn *p_hwfn,
						     struct ecore_ptt *p_ptt,
						     u32 *buf_size)
{
	struct dbg_tools_data *dev_data = &p_hwfn->dbg_info;
	struct idle_chk_data *idle_chk = &dev_data->idle_chk;
	enum dbg_status status;

	*buf_size = 0;

	status = ecore_dbg_dev_init(p_hwfn);
	if (status != DBG_STATUS_OK)
		return status;

	if (!p_hwfn->dbg_arrays[BIN_BUF_DBG_MODE_TREE].ptr || !p_hwfn->dbg_arrays[BIN_BUF_DBG_IDLE_CHK_REGS].ptr ||
	    !p_hwfn->dbg_arrays[BIN_BUF_DBG_IDLE_CHK_IMMS].ptr || !p_hwfn->dbg_arrays[BIN_BUF_DBG_IDLE_CHK_RULES].ptr)
		return DBG_STATUS_DBG_ARRAY_NOT_SET;

	if (!idle_chk->buf_size_set) {
		idle_chk->buf_size = ecore_idle_chk_dump(p_hwfn, p_ptt, OSAL_NULL, false);
		idle_chk->buf_size_set = true;
	}

	*buf_size = idle_chk->buf_size;

	return DBG_STATUS_OK;
}

enum dbg_status ecore_dbg_idle_chk_dump(struct ecore_hwfn *p_hwfn,
					struct ecore_ptt *p_ptt,
					u32 *dump_buf,
					u32 buf_size_in_dwords,
					u32 *num_dumped_dwords)
{
	u32 needed_buf_size_in_dwords;
	enum dbg_status status;

	*num_dumped_dwords = 0;

	status = ecore_dbg_idle_chk_get_dump_buf_size(p_hwfn, p_ptt, &needed_buf_size_in_dwords);
	if (status != DBG_STATUS_OK)
		return status;

	if (buf_size_in_dwords < needed_buf_size_in_dwords)
		return DBG_STATUS_DUMP_BUF_TOO_SMALL;

	/* Update reset state */
	ecore_grc_unreset_blocks(p_hwfn, p_ptt, true);
	ecore_update_blocks_reset_state(p_hwfn, p_ptt);

	/* Idle Check Dump */
	*num_dumped_dwords = ecore_idle_chk_dump(p_hwfn, p_ptt, dump_buf, true);

	/* Revert GRC params to their default */
	ecore_dbg_grc_set_params_default(p_hwfn);

	return DBG_STATUS_OK;
}

enum dbg_status ecore_dbg_mcp_trace_get_dump_buf_size(struct ecore_hwfn *p_hwfn,
						      struct ecore_ptt *p_ptt,
						      u32 *buf_size)
{
	enum dbg_status status = ecore_dbg_dev_init(p_hwfn);

	*buf_size = 0;

	if (status != DBG_STATUS_OK)
		return status;

	return ecore_mcp_trace_dump(p_hwfn, p_ptt, OSAL_NULL, false, buf_size);
}

enum dbg_status ecore_dbg_mcp_trace_dump(struct ecore_hwfn *p_hwfn,
					 struct ecore_ptt *p_ptt,
					 u32 *dump_buf,
					 u32 buf_size_in_dwords,
					 u32 *num_dumped_dwords)
{
	u32 needed_buf_size_in_dwords;
	enum dbg_status status;

	status = ecore_dbg_mcp_trace_get_dump_buf_size(p_hwfn, p_ptt, &needed_buf_size_in_dwords);
	if (status != DBG_STATUS_OK && status != DBG_STATUS_NVRAM_GET_IMAGE_FAILED)
		return status;

	if (buf_size_in_dwords < needed_buf_size_in_dwords)
		return DBG_STATUS_DUMP_BUF_TOO_SMALL;

	/* Update reset state */
	ecore_update_blocks_reset_state(p_hwfn, p_ptt);

	/* Perform dump */
	status = ecore_mcp_trace_dump(p_hwfn, p_ptt, dump_buf, true, num_dumped_dwords);

	/* Revert GRC params to their default */
	ecore_dbg_grc_set_params_default(p_hwfn);

	return status;
}

enum dbg_status ecore_dbg_reg_fifo_get_dump_buf_size(struct ecore_hwfn *p_hwfn,
						     struct ecore_ptt *p_ptt,
						     u32 *buf_size)
{
	enum dbg_status status = ecore_dbg_dev_init(p_hwfn);

	*buf_size = 0;

	if (status != DBG_STATUS_OK)
		return status;

	return ecore_reg_fifo_dump(p_hwfn, p_ptt, OSAL_NULL, false, buf_size);
}

enum dbg_status ecore_dbg_reg_fifo_dump(struct ecore_hwfn *p_hwfn,
					struct ecore_ptt *p_ptt,
					u32 *dump_buf,
					u32 buf_size_in_dwords,
					u32 *num_dumped_dwords)
{
	u32 needed_buf_size_in_dwords;
	enum dbg_status status;

	*num_dumped_dwords = 0;

	status = ecore_dbg_reg_fifo_get_dump_buf_size(p_hwfn, p_ptt, &needed_buf_size_in_dwords);
	if (status != DBG_STATUS_OK)
		return status;

	if (buf_size_in_dwords < needed_buf_size_in_dwords)
		return DBG_STATUS_DUMP_BUF_TOO_SMALL;

	/* Update reset state */
	ecore_update_blocks_reset_state(p_hwfn, p_ptt);

	status = ecore_reg_fifo_dump(p_hwfn, p_ptt, dump_buf, true, num_dumped_dwords);

	/* Revert GRC params to their default */
	ecore_dbg_grc_set_params_default(p_hwfn);

	return status;
}

enum dbg_status ecore_dbg_igu_fifo_get_dump_buf_size(struct ecore_hwfn *p_hwfn,
						     struct ecore_ptt *p_ptt,
						     u32 *buf_size)
{
	enum dbg_status status = ecore_dbg_dev_init(p_hwfn);

	*buf_size = 0;

	if (status != DBG_STATUS_OK)
		return status;

	return ecore_igu_fifo_dump(p_hwfn, p_ptt, OSAL_NULL, false, buf_size);
}

enum dbg_status ecore_dbg_igu_fifo_dump(struct ecore_hwfn *p_hwfn,
					struct ecore_ptt *p_ptt,
					u32 *dump_buf,
					u32 buf_size_in_dwords,
					u32 *num_dumped_dwords)
{
	u32 needed_buf_size_in_dwords;
	enum dbg_status status;

	*num_dumped_dwords = 0;

	status = ecore_dbg_igu_fifo_get_dump_buf_size(p_hwfn, p_ptt, &needed_buf_size_in_dwords);
	if (status != DBG_STATUS_OK)
		return status;

	if (buf_size_in_dwords < needed_buf_size_in_dwords)
		return DBG_STATUS_DUMP_BUF_TOO_SMALL;

	/* Update reset state */
	ecore_update_blocks_reset_state(p_hwfn, p_ptt);

	status = ecore_igu_fifo_dump(p_hwfn, p_ptt, dump_buf, true, num_dumped_dwords);

	/* Revert GRC params to their default */
	ecore_dbg_grc_set_params_default(p_hwfn);

	return status;
}

enum dbg_status ecore_dbg_protection_override_get_dump_buf_size(struct ecore_hwfn *p_hwfn,
								struct ecore_ptt *p_ptt,
								u32 *buf_size)
{
	enum dbg_status status = ecore_dbg_dev_init(p_hwfn);

	*buf_size = 0;

	if (status != DBG_STATUS_OK)
		return status;

	return ecore_protection_override_dump(p_hwfn, p_ptt, OSAL_NULL, false, buf_size);
}

enum dbg_status ecore_dbg_protection_override_dump(struct ecore_hwfn *p_hwfn,
						   struct ecore_ptt *p_ptt,
						   u32 *dump_buf,
						   u32 buf_size_in_dwords,
						   u32 *num_dumped_dwords)
{
	u32 needed_buf_size_in_dwords;
	enum dbg_status status;

	*num_dumped_dwords = 0;

	status = ecore_dbg_protection_override_get_dump_buf_size(p_hwfn, p_ptt, &needed_buf_size_in_dwords);
	if (status != DBG_STATUS_OK)
		return status;

	if (buf_size_in_dwords < needed_buf_size_in_dwords)
		return DBG_STATUS_DUMP_BUF_TOO_SMALL;

	/* Update reset state */
	ecore_update_blocks_reset_state(p_hwfn, p_ptt);

	status = ecore_protection_override_dump(p_hwfn, p_ptt, dump_buf, true, num_dumped_dwords);

	/* Revert GRC params to their default */
	ecore_dbg_grc_set_params_default(p_hwfn);

	return status;
}

enum dbg_status ecore_dbg_fw_asserts_get_dump_buf_size(struct ecore_hwfn *p_hwfn,
						       struct ecore_ptt *p_ptt,
						       u32 *buf_size)
{
	enum dbg_status status = ecore_dbg_dev_init(p_hwfn);

	*buf_size = 0;

	if (status != DBG_STATUS_OK)
		return status;

	/* Update reset state */
	ecore_update_blocks_reset_state(p_hwfn, p_ptt);

	*buf_size = ecore_fw_asserts_dump(p_hwfn, p_ptt, OSAL_NULL, false);

	return DBG_STATUS_OK;
}

enum dbg_status ecore_dbg_fw_asserts_dump(struct ecore_hwfn *p_hwfn,
					  struct ecore_ptt *p_ptt,
					  u32 *dump_buf,
					  u32 buf_size_in_dwords,
					  u32 *num_dumped_dwords)
{
	u32 needed_buf_size_in_dwords;
	enum dbg_status status;

	*num_dumped_dwords = 0;

	status = ecore_dbg_fw_asserts_get_dump_buf_size(p_hwfn, p_ptt, &needed_buf_size_in_dwords);
	if (status != DBG_STATUS_OK)
		return status;

	if (buf_size_in_dwords < needed_buf_size_in_dwords)
		return DBG_STATUS_DUMP_BUF_TOO_SMALL;

	*num_dumped_dwords = ecore_fw_asserts_dump(p_hwfn, p_ptt, dump_buf, true);

	/* Revert GRC params to their default */
	ecore_dbg_grc_set_params_default(p_hwfn);

	return DBG_STATUS_OK;
}

enum dbg_status ecore_dbg_ilt_get_dump_buf_size(struct ecore_hwfn *p_hwfn,
						struct ecore_ptt *p_ptt,
						u32 *buf_size)
{
	enum dbg_status status = ecore_dbg_dev_init(p_hwfn);

	*buf_size = 0;

	if (status != DBG_STATUS_OK)
		return status;

	*buf_size = ecore_ilt_dump(p_hwfn, p_ptt, OSAL_NULL, false);

	return DBG_STATUS_OK;
}

enum dbg_status ecore_dbg_ilt_dump(struct ecore_hwfn *p_hwfn,
				   struct ecore_ptt *p_ptt,
				   u32 *dump_buf,
				   u32 buf_size_in_dwords,
				   u32 *num_dumped_dwords)
{
	u32 needed_buf_size_in_dwords;
	enum dbg_status status;

	*num_dumped_dwords = 0;

	status = ecore_dbg_ilt_get_dump_buf_size(p_hwfn, p_ptt, &needed_buf_size_in_dwords);
	if (status != DBG_STATUS_OK)
		return status;

	if (buf_size_in_dwords < needed_buf_size_in_dwords)
		return DBG_STATUS_DUMP_BUF_TOO_SMALL;

	*num_dumped_dwords = ecore_ilt_dump(p_hwfn, p_ptt, dump_buf, true);

	/* Reveret GRC params to their default */
	ecore_dbg_grc_set_params_default(p_hwfn);

	return DBG_STATUS_OK;
}

enum dbg_status ecore_dbg_read_attn(struct ecore_hwfn *p_hwfn,
				    struct ecore_ptt *p_ptt,
				    enum block_id block_id,
				    enum dbg_attn_type attn_type,
				    bool clear_status,
				    struct dbg_attn_block_result *results)
{
	enum dbg_status status = ecore_dbg_dev_init(p_hwfn);
	u8 reg_idx, num_attn_regs, num_result_regs = 0;
	const struct dbg_attn_reg *attn_reg_arr;

	if (status != DBG_STATUS_OK)
		return status;

	if (!p_hwfn->dbg_arrays[BIN_BUF_DBG_MODE_TREE].ptr || !p_hwfn->dbg_arrays[BIN_BUF_DBG_ATTN_BLOCKS].ptr || !p_hwfn->dbg_arrays[BIN_BUF_DBG_ATTN_REGS].ptr)
		return DBG_STATUS_DBG_ARRAY_NOT_SET;

	attn_reg_arr = ecore_get_block_attn_regs(p_hwfn, block_id, attn_type, &num_attn_regs);

	for (reg_idx = 0; reg_idx < num_attn_regs; reg_idx++) {
		const struct dbg_attn_reg *reg_data = &attn_reg_arr[reg_idx];
		struct dbg_attn_reg_result *reg_result;
		u32 sts_addr, sts_val;
		u16 modes_buf_offset;
		bool eval_mode;

		/* Check mode */
		eval_mode = GET_FIELD(reg_data->mode.data, DBG_MODE_HDR_EVAL_MODE) > 0;
		modes_buf_offset = GET_FIELD(reg_data->mode.data, DBG_MODE_HDR_MODES_BUF_OFFSET);
		if (eval_mode && !ecore_is_mode_match(p_hwfn, &modes_buf_offset))
			continue;

		/* Mode match - read attention status register */
		sts_addr = clear_status ? reg_data->sts_clr_address : GET_FIELD(reg_data->data, DBG_ATTN_REG_STS_ADDRESS);
		sts_val = ecore_rd(p_hwfn, p_ptt, DWORDS_TO_BYTES(sts_addr));
		if (!sts_val)
			continue;

		/* Non-zero attention status - add to results */
		reg_result = &results->reg_results[num_result_regs];
		SET_FIELD(reg_result->data, DBG_ATTN_REG_RESULT_STS_ADDRESS, sts_addr);
		SET_FIELD(reg_result->data, DBG_ATTN_REG_RESULT_NUM_REG_ATTN, GET_FIELD(reg_data->data, DBG_ATTN_REG_NUM_REG_ATTN));
		reg_result->block_attn_offset = reg_data->block_attn_offset;
		reg_result->sts_val = sts_val;
		reg_result->mask_val = ecore_rd(p_hwfn, p_ptt, DWORDS_TO_BYTES(reg_data->mask_address));
		num_result_regs++;
	}

	results->block_id = (u8)block_id;
	results->names_offset = ecore_get_block_attn_data(p_hwfn, block_id, attn_type)->names_offset;
	SET_FIELD(results->data, DBG_ATTN_BLOCK_RESULT_ATTN_TYPE, attn_type);
	SET_FIELD(results->data, DBG_ATTN_BLOCK_RESULT_NUM_REGS, num_result_regs);

	return DBG_STATUS_OK;
}

enum dbg_status ecore_dbg_print_attn(struct ecore_hwfn *p_hwfn,
				     struct dbg_attn_block_result *results)
{
	enum dbg_attn_type attn_type;
	u8 num_regs, i;

	num_regs = GET_FIELD(results->data, DBG_ATTN_BLOCK_RESULT_NUM_REGS);
	attn_type = (enum dbg_attn_type)GET_FIELD(results->data, DBG_ATTN_BLOCK_RESULT_ATTN_TYPE);

	for (i = 0; i < num_regs; i++) {
		struct dbg_attn_reg_result *reg_result;
		const char *attn_type_str;
		u32 sts_addr;

		reg_result = &results->reg_results[i];
		attn_type_str = (attn_type == ATTN_TYPE_INTERRUPT ? "interrupt" : "parity");
		sts_addr = GET_FIELD(reg_result->data, DBG_ATTN_REG_RESULT_STS_ADDRESS);
		DP_NOTICE(p_hwfn, false, "%s: address 0x%08x, status 0x%08x, mask 0x%08x\n", attn_type_str, (u32)DWORDS_TO_BYTES(sts_addr), reg_result->sts_val, reg_result->mask_val);
	}

	return DBG_STATUS_OK;
}

bool ecore_is_block_in_reset(struct ecore_hwfn *p_hwfn,
			     struct ecore_ptt *p_ptt,
			     enum block_id block_id)
{
	const struct dbg_reset_reg *reset_reg;
	const struct dbg_block_chip *block;
	u32 reset_reg_addr;

	block = ecore_get_dbg_block_per_chip(p_hwfn, block_id);

	if (!GET_FIELD(block->flags, DBG_BLOCK_CHIP_HAS_RESET_REG))
		return false;

	reset_reg = ecore_get_dbg_reset_reg(p_hwfn, block->reset_reg_id);

	if (GET_FIELD(reset_reg->data, DBG_RESET_REG_IS_REMOVED))
		return false;

	reset_reg_addr = GET_FIELD(reset_reg->data, DBG_RESET_REG_ADDR);
	return (ecore_rd(p_hwfn, p_ptt, reset_reg_addr) & (1 << block->reset_reg_bit_offset)) > 0;
}

void ecore_mask_parities_no_mcp(struct ecore_hwfn *p_hwfn,
				struct ecore_ptt *p_ptt,
				bool mask)
{
	/* Clear General Attn 35 which is used for driver initiated parity */
	ecore_wr(p_hwfn, p_ptt, MISC_REG_AEU_GENERAL_ATTN_35, 0);

	ecore_update_parity_aeu_bits(p_hwfn, p_ptt, 2, !mask, AEU_INPUT2_BITS_PARITY_ERROR, AEU_INPUT2_BITS_GENERATE_SYSTEM_KILL);
	ecore_update_parity_aeu_bits(p_hwfn, p_ptt, 4, !mask,
		ECORE_IS_BB_B0(p_hwfn->p_dev) ? AEU_INPUT4_BITS_PARITY_ERROR_BB : AEU_INPUT4_BITS_PARITY_ERROR_K2,
		ECORE_IS_BB_B0(p_hwfn->p_dev) ? AEU_INPUT4_BITS_GENERATE_SYSTEM_KILL_BB : AEU_INPUT4_BITS_GENERATE_SYSTEM_KILL_K2);
	ecore_update_parity_aeu_bits(p_hwfn, p_ptt, 5, !mask, AEU_INPUT5_BITS_PARITY_ERROR, AEU_INPUT5_BITS_GENERATE_SYSTEM_KILL);
	ecore_update_parity_aeu_bits(p_hwfn, p_ptt, 6, !mask, AEU_INPUT6_BITS_PARITY_ERROR, AEU_INPUT6_BITS_GENERATE_SYSTEM_KILL);
	ecore_update_parity_aeu_bits(p_hwfn, p_ptt, 7, !mask, AEU_INPUT7_BITS_PARITY_ERROR, AEU_INPUT7_BITS_GENERATE_SYSTEM_KILL);
	ecore_update_parity_aeu_bits(p_hwfn, p_ptt, 8, !mask,
		ECORE_IS_BB_B0(p_hwfn->p_dev) ? AEU_INPUT8_BITS_PARITY_ERROR_BB : AEU_INPUT8_BITS_PARITY_ERROR_K2,
		ECORE_IS_BB_B0(p_hwfn->p_dev) ? AEU_INPUT8_BITS_GENERATE_SYSTEM_KILL_BB : AEU_INPUT8_BITS_GENERATE_SYSTEM_KILL_K2);
	ecore_update_parity_aeu_bits(p_hwfn, p_ptt, 9, !mask, AEU_INPUT9_BITS_PARITY_ERROR, AEU_INPUT9_BITS_GENERATE_SYSTEM_KILL);

	ecore_update_reg_bits(p_hwfn, p_ptt, MISC_REG_AEU_MASK_ATTN_MCP, !mask, 1);

	ecore_wr(p_hwfn, p_ptt, MISC_REG_AEU_SYS_KILL_BEHAVIOR, 0);
	ecore_wr(p_hwfn, p_ptt, MISC_REG_AEU_GENERAL_MASK, 0);
}

